(*
Copyright (c) 1997,1998 The Regents of the University of California.
All rights reserved.

Permission to use, copy, modify, and distribute this software for any
purpose, without fee, and without written agreement is hereby granted,
provided that the above copyright notice and the following two
paragraphs appear in all copies of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*)

signature SYMTAB =
    sig
	type identifier
	type info
	type symbolTable
	type scope
	type scopeInfo

	(**** Symbol Tables ****)

	(* Make a new symbol table *)
	val mkSymTab : unit->symbolTable

	(* Reset a symbol table *)
	val resetSymTab : symbolTable->unit


	(**** Scopes ****)

	(* Enter a new scope *)
	val enterScope : symbolTable->scopeInfo->unit

	(* Leave the current scope *)
	val exitScope : symbolTable->unit

	(* Get the current scope *)
	val getScope : symbolTable->scope

	(* Get the current scope, but make all levels except the root
	   non-updatable. *)
	val getCScope : symbolTable->scope

	(* Set the current scope *)
	val setScope : symbolTable->scope->unit

	(* Get the information associated with a scope *)
	val getScopeInfo : scope->scopeInfo

	(* Given a scope, return its parent *)
	val parentScope : scope->scope option

	(* Apply a function to all the information in a scope *)
	val mapScope : (info->info)->scope->scope

	(* Given a scope, return the number of scopes above it
	   in the symbol table *)
	val scopeDepth : scope->int

	(* Return true iff the scope is the outermost (global) scope *)
	val isGlobalScope : scope->bool

	(**** Symbols ****)

	(* Insert a symbol *)
	val insert : symbolTable->identifier->info->unit

	(* Lookup a symbol in a particular scope without changing
	 the current scope. *)
	val findInScope : scope->identifier->{info : info,
					      scope : scope} option

	(* Lookup a symbol *)
	val find : symbolTable->identifier->{info : info,
					     scope : scope} option
    end
