(* Chris Harrelson - August 6, 1999 *)

signature typeCheck = sig

   structure AST : AST

   (* exceptions *)

   exception Unsupported of AST.region * string
   exception ParseError of AST.region * string

   (* end exceptions *)

   (* datatypes *)

   type program_function_type = (SourceMap.sourcemap * AST.program -> (AST.program option)) option
   type statement_function_type = (SourceMap.sourcemap * AST.statement -> (AST.statement option)) option
   type expression_function_type = (SourceMap.sourcemap * AST.expression -> (AST.expression option)) option
   type declaration_function_type = (SourceMap.sourcemap * AST.declaration -> (AST.declaration option)) option
      
   type ast_type = {sourceMap : SourceMap.sourcemap, tree : typeCheckCParser.GenParser.result}
   type error_report = AST.region * string
      
   datatype callback_type =
      Callbacks of
      {(* the whole program *)
       program : program_function_type,
       (* statement types *)
       return : statement_function_type,
       expression : statement_function_type,
       ifThen : statement_function_type,
       ifThenElse : statement_function_type,
       whileDo : statement_function_type,
       doWhile : statement_function_type,
       for : statement_function_type,
       goTo : statement_function_type,
       switch : statement_function_type,
       caseStat : statement_function_type, (* case is a reserved word, so we can't use it *)
       labeled : statement_function_type,
       break : statement_function_type,
       continue : statement_function_type,
       compound : statement_function_type,
       (* expression types *)
       variable : expression_function_type,
       constant : expression_function_type,
       conditional : expression_function_type,
       application : expression_function_type,
       operation : expression_function_type,
       assignment : expression_function_type,
       comma : expression_function_type,
       cast : expression_function_type,
       sizeOfExpr : expression_function_type,
       sizeOfTypeExpr : expression_function_type,
       fieldDirect : expression_function_type,
       fieldIndirect : expression_function_type,
       postfix : expression_function_type,
       index : expression_function_type,
       extension : expression_function_type,
       (* declaration types *)
       ellipsisDecl : declaration_function_type,
       varDecl : declaration_function_type,
       typeDef : declaration_function_type,
       funDef : declaration_function_type,
       funDecl : declaration_function_type}

   (* end datatypes *)

   val check_types : ast_type -> callback_type -> (ast_type * error_report list)

end


(* you can cut-and-paste this into your file to make things quicker to type:

      {(* the whole program *)
       program = ,
       (* statements *)
       return = ,
       expression = ,
       ifThen = ,
       ifThenElse = ,
       whileDo = ,
       doWhile = ,
       for = ,
       goTo = ,
       switch = ,
       caseStat = ,
       labeled = ,
       break = ,
       continue = ,
       compound = ,
       (* expressions *)
       variable = ,
       constant = ,
       conditional = ,
       application = ,
       operation = ,
       assignment = ,
       comma = ,
       cast = ,
       sizeOfExpr = ,
       sizeOfTypeExpr = ,
       fieldDirect = ,
       fieldIndirect = ,
       postfix = ,
       index = ,
       extension = ,
       (* declarations *)
       ellipsisDecl = ,
       varDecl = ,
       typeDef = ,
       funDef = ,
       funDecl = }

*)