(*
Copyright (c) 1997,1998 The Regents of the University of California.
All rights reserved.

Permission to use, copy, modify, and distribute this software for any
purpose, without fee, and without written agreement is hereby granted,
provided that the above copyright notice and the following two
paragraphs appear in all copies of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*)

functor UnparseFn(structure AST : AST) =
    struct

	structure AST = AST

	fun identifier_opt (SOME i) = AST.idToString i
	  | identifier_opt NONE = ""

	fun constant (AST.Integer n) = Int.toString n
	  | constant (AST.BigInteger iden) = AST.idToString iden
	  | constant (AST.String iden) = AST.idToString iden
	  | constant (AST.Character iden) = AST.idToString iden
	  | constant (AST.Real iden) = AST.idToString iden

	fun operator AST.Plus = "+"
	  | operator AST.Minus = "-"
	  | operator AST.Mul = "*"
	  | operator AST.Div = "/"
	  | operator AST.Mod = "%"
	  | operator AST.Incr = "++"
	  | operator AST.Decr = "--"
	  | operator AST.And = "&&"
	  | operator AST.Or = "||"
	  | operator AST.Not = "!"
	  | operator AST.BitAnd = "&"
	  | operator AST.BitOr = "|"
	  | operator AST.BitNot = "~"
	  | operator AST.BitXor = "^"
	  | operator AST.Rshift = ">>"
	  | operator AST.Lshift = "<<"
	  | operator AST.Lt = "<"
	  | operator AST.Le = "<="
	  | operator AST.Gt = ">"
	  | operator AST.Ge = ">="
	  | operator AST.Eq = "=="
	  | operator AST.Ne = "!="
	  | operator AST.Address = "&"
	  | operator AST.Deref = "*"
	  | operator AST.UnaryMinus = "-"
	  | operator AST.Flow = "???"
	and operator_opt (SOME oper) = operator oper
	  | operator_opt NONE = ""

	fun expression (AST.Variable (iden,_)) = AST.idToString iden
	  | expression (AST.Constant (const, _)) = constant const
	  | expression (AST.Conditional(g, i, e, _)) =
	    (expression g) ^ " ? " ^ (expression i) ^ " : " ^
	    (expression e)
	  | expression (AST.Application(e, es, _, _)) =
	    (expression e) ^ "(" ^ (expression_list es) ^ ")"
	  | expression (AST.Operation(oper, [], _)) =
	    (operator oper) ^ "???"
	  | expression (AST.Operation(oper, [e], _)) =
	    (operator oper) ^ (expression e)
	  | expression (AST.Operation(oper, e1::es, _)) =
	    (expression e1) ^ (operator oper) ^ (expression_list es)
	  | expression (AST.Assignment(e1, opopt, e2, _)) =
	    (expression e1) ^ " " ^ (operator_opt opopt) ^ "= " ^
	    (expression e2)
	  | expression (AST.Comma(e1, e2, _)) =
	    (expression e1) ^ ", " ^ (expression e2)
	  | expression (AST.Cast(stype, e, _)) =
	    "(" ^ (syntacticType "" stype) ^ ") " ^ (expression e)
	  | expression (AST.SizeOfExpr (e, _)) =
	    "sizeof(" ^ (expression e) ^ ")"
	  | expression (AST.SizeOfTypeExpr (stype, _)) =
	    "sizeof(" ^ (syntacticType "" stype) ^ ")"
	  | expression (AST.FieldDirect(e, id, _)) =
	    (expression e) ^ "." ^ (AST.idToString id)
	  | expression (AST.FieldIndirect(e, id, _)) =
	    (expression e) ^ "->" ^ (AST.idToString id)
	  | expression (AST.Postfix(oper, e, _)) =
	    (expression e) ^ (operator oper)
	  | expression (AST.Index(e1, e2, _)) =
	    (expression e1) ^ "[" ^ (expression e2) ^ "]"
	  | expression (AST.Extension(stmt, _)) =
	    "(" ^ (statement stmt) ^ ")"

	and expression_list [] = ""
	  | expression_list [e] = (expression e)
	  | expression_list (e::es) = (expression e) ^ ", " ^
	    (expression_list es)

	and expression_opt (SOME e) = (expression e)
	  | expression_opt NONE = ""

	and syntacticType name (AST.SyntacticType(sc, ts, tql, dl)) =
	    (storageClass_opt sc) ^
	    (typeQualifier_list tql) ^
	    (typeSpecifier ts) ^ " " ^
	    (declarator_list name dl)

	and aggregate AST.Struct = "struct"
	  | aggregate AST.Union = "union"

	and storageClass AST.Register = "register"
	  | storageClass AST.Auto = "auto"
	  | storageClass AST.Static = "static"
	  | storageClass AST.Extern = "extern"
	  | storageClass AST.Typedef = "typedef"

	and storageClass_opt (SOME sc) = (storageClass sc ^ " ")
	  | storageClass_opt NONE = ""

	and typeQualifier AST.Const = "const"
	  | typeQualifier AST.Volatile = "volatile"
	  | typeQualifier AST.Inline = "__inline"
	  | typeQualifier AST.NonConst = "nonconst"
	  | typeQualifier (AST.Qualifier id) = AST.idToString id

	and typeQualifier_list [] = ""
	  | typeQualifier_list [d] = (typeQualifier d) ^ " "
	  | typeQualifier_list (d::ds) = (typeQualifier d) ^ (typeQualifier_list ds)
(*
	and declarator_list _ [] = ""
	  | declarator_list n (d::rest) =
	    case d of
		AST.Pointer tql => "[" ^ typeQualifier_list tql ^ "ptr(" ^
		                   (declarator_list n rest) ^ ")]"
	      | AST.Array e => "array(" ^ (declarator_list n rest) ^ ")"
	      | AST.BitField e => "bitfield(" ^ (declarator_list n rest) ^ ")"
	      | AST.Function dl => (declarator_list n rest) ^
				   "f(" ^ parameter_list dl ^ ")"
	      | AST.OldFunction _ => (declarator_list n rest )^ "f()"
	    
*)
	and declarator_list name dl =
	    let fun wrap b acc = if b then "("^acc^")" else acc

		fun decl (acc,np) [] = acc
		  | decl (acc,np) (d::rest) =
		      (case d of
			   AST.Pointer tql =>
			       decl ("*" ^ (typeQualifier_list tql) ^ acc,true) rest
			 | AST.Array e =>
			       decl (wrap np acc ^ "[" ^
				     (expression_opt e) ^ "]", false) rest
			 | AST.BitField e =>
			       decl (wrap np acc ^ ":" ^ (expression e), false) rest
			 | AST.Function dl =>
			       decl (wrap np acc ^ "f(" ^ parameter_list dl ^ ")",false) rest
			 | AST.OldFunction _ => 
			       decl (wrap np acc ^ "f()", false) rest
		      )
	    in
		decl (name,false) dl
	    end

	and typeSpecifier AST.Void = "void"
	  | typeSpecifier AST.Char = "char"
	  | typeSpecifier AST.UnsignedChar = "unsigned char"
	  | typeSpecifier AST.ShortInt = "short"
	  | typeSpecifier AST.UnsignedShortInt = "unsigned short"
	  | typeSpecifier AST.Int = "int"
	  | typeSpecifier AST.UnsignedInt = "unsigned"
	  | typeSpecifier AST.LongInt = "long"
	  | typeSpecifier AST.LongLong = "long long"
	  | typeSpecifier AST.UnsignedLongInt = "unsigned long"
	  | typeSpecifier AST.UnsignedLongLong = "unsigned long long"
	  | typeSpecifier AST.Float = "float"
	  | typeSpecifier AST.Double = "double"
	  | typeSpecifier AST.LongDouble = "long double"
	  | typeSpecifier (AST.NamedType iden) = AST.idToString iden
	  | typeSpecifier (AST.Enum(iden_opt, ieopt_list)) =
	    "enum" ^ (identifier_opt iden_opt) ^ "{" ^
	    (enumTypeSpecifier_list ieopt_list) ^ "}"
	  | typeSpecifier (AST.Aggregate(ag, iden_opt, dl)) =
	    (aggregate ag) ^ " " ^ (identifier_opt iden_opt) ^ "{"
	    ^ (declaration_list dl) ^ "}"
	  | typeSpecifier (AST.NamedAggregate(ag, id)) =
	    (aggregate ag) ^ " " ^ (AST.idToString id)
	  | typeSpecifier (AST.TypeOfExpr(e, _)) =   (* johno added this *)
	    "typeof(" ^ (expression e) ^ ")"
	  | typeSpecifier (AST.TypeOfTypeExpr(stype, _)) =  (* johno *)
	    "typeof(" ^ (syntacticType "" stype) ^")"

	and enumTypeSpecifier_list [] = ""
	  | enumTypeSpecifier_list [(id, SOME e)] =
	    (AST.idToString id) ^ " = " ^ (expression e)
	  | enumTypeSpecifier_list [(id, NONE)] =
	    (AST.idToString id)
	  | enumTypeSpecifier_list ((id, SOME e)::ieopts) =
	    (AST.idToString id) ^ " = " ^ (expression e) ^ ", " ^
	    (enumTypeSpecifier_list ieopts)
	  | enumTypeSpecifier_list ((id, NONE)::ieopts) =
	    (AST.idToString id) ^ ", " ^
	    (enumTypeSpecifier_list ieopts)


	and typeSpecifier_opt (SOME ts) = typeSpecifier ts
	  | typeSpecifier_opt NONE = ""

	and declaration (AST.EllipsisDecl{...}) = "..."
	  | declaration (AST.VarDecl{name, typ, init, region}) =
	    (syntacticType (AST.idToString name) typ) ^
	    (initializer_opt init)
	  | declaration (AST.TypeDef{name, typ,...}) =
	    (syntacticType (AST.idToString name) typ)
	  | declaration (AST.FunDef{name, typ, return, params, decls, body, region, ...}) =
	    (syntacticType (AST.idToString name) return) ^ "(" ^
	    (parameter_list params) ^ ") {\n" ^
	    (declaration_list decls) ^ "\n" ^ (statement_list body) ^ "\n}\n"
	  | declaration (AST.FunDecl{name, typ, return, params=SOME params, region}) =
	    (syntacticType (AST.idToString name) return) ^ "(" ^
	    (parameter_list params) ^ ")"
	  | declaration (AST.FunDecl{name, typ, return, params=NONE, region}) =
	    (syntacticType (AST.idToString name) return) ^ "()" 


	and terminated_declaration d =
	    case d of
		AST.FunDef _ => declaration d
	      | _ => declaration d ^ ";\n"

	and declaration_list [] = ""
	  | declaration_list (d::ds) = (terminated_declaration d) ^
	    (declaration_list ds)

	and parameter_list [] = "void"
	  | parameter_list pl =
	    let fun p [] = ""
		  | p [d] = (declaration d)
		  | p (d::ds) = (declaration d) ^ ", " ^ p ds
	    in
		p pl
	    end

	and initializer (AST.Initial e) = " = " ^ (expression e)
	  | initializer (AST.InitialList il) = 
	           "{" ^ (initializer_list il) ^ "}"

	and initializer_list [] = ""
	  | initializer_list [i] = initializer i
	  | initializer_list (i::is) = (initializer i) ^ ", " ^
	                               (initializer_list is)

	and initializer_opt (SOME i) = initializer i
	  | initializer_opt NONE = ""

	and statement (AST.Return(SOME e, idopt, _)) =
	    "return " ^ (expression e) ^ "; /* " ^
	    (identifier_opt idopt) ^ " */\n"
	  | statement (AST.Return(NONE, idopt, _)) =
	    "return; /* " ^ (identifier_opt idopt) ^ " */\n"
	  | statement (AST.Expression (eopt, _)) =
	    (expression_opt eopt) ^ ";\n"
	  | statement (AST.IfThen(e, s, _)) =
	    "if (" ^ (expression e) ^ ") " ^ (statement s) ^ "\n"
	  | statement (AST.IfThenElse(e, s1, s2, _)) =
	    "if (" ^ (expression e) ^") " ^ (statement s1) ^ "\nelse " ^
	    (statement s2) ^ "\n"
	  | statement (AST.WhileDo(e, s, _)) =
	    "while (" ^ (expression e) ^ ") " ^ (statement s)
	  | statement (AST.DoWhile(s, e, _)) =
	    "do " ^ (statement s) ^ " while (" ^ (expression e) ^ ");\n"
	  | statement (AST.For(eopt1, eopt2, eopt3, s, _)) =
	    "for (" ^ (expression_opt eopt1) ^ "; " ^
	    (expression_opt eopt2) ^ "; " ^
	    (expression_opt eopt3) ^ ") " ^ (statement s)
	  | statement (AST.GoTo (iden, _)) = "goto " ^ (AST.idToString iden)
	  | statement (AST.Switch(e, s, _)) =
	    "switch (" ^ (expression e) ^ ")\n" ^ (statement s)
	  | statement (AST.Case(SOME e, SOME s, _)) =
	    "case " ^ (expression e) ^ ": " ^ (statement s)
	  | statement (AST.Case(NONE, SOME s, _)) =
	    "default: " ^ (statement s)
	  | statement (AST.Case(NONE, NONE, _)) =
	    "default: "
	  | statement (AST.Labeled(iden, SOME s, _)) =
	    (AST.idToString iden) ^ ": " ^ (statement s)
	  | statement (AST.Labeled(iden, NONE, _)) =
	    (AST.idToString iden) ^ ": "
	  | statement (AST.Break _) = "break;\n"
	  | statement (AST.Continue _) = "continue;\n"
	  | statement (AST.Compound{decls, body, ...}) =
	    "{\n" ^ (declaration_list decls) ^ "\n\n" ^
	    (statement_list body) ^ "}\n"

	and statement_list [] = ""
	  | statement_list [s] = (statement s)
	  | statement_list (s::ss) = (statement s) ^ (statement_list ss)

	fun program (AST.Program {decls,...}) = declaration_list decls

    end
