(*
Copyright (c) 1997,1998 The Regents of the University of California.
All rights reserved.

Permission to use, copy, modify, and distribute this software for any
purpose, without fee, and without written agreement is hereby granted,
provided that the above copyright notice and the following two
paragraphs appear in all copies of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*)

signature GRAPH =
    sig

	type node
	type graph
	type 'a jlist = 'a Util.jlist

	val debug : bool ref

	(* The empty graph *)
	val empty : unit->graph

	(* Reset a graph *)
	val reset : graph->unit

	(* Add a node if it doesn't already exist *)
	val addNode:graph->node->unit

	(* Add a directed edge *)
	val addEdge : graph->{from:node,to:node}->unit

	(* Return the precessors/successors of a node. *)
	val predecessors : graph->node->node jlist
	val successors : graph->node->node jlist

	(* Make two nodes equivalent *)
	val mkEquiv : graph->node*node->unit

	(* Return a list of root nodes in the graph, i.e.
	   nodes with no predecessors *)
	val findRoots : graph->node list

	(* Compute the strongly connected components and collapse them *)
	(*val collapseStronglyConnectedComponents : graph->unit*)

	(* Apply a function to each node of the graph in
	   final depth first order, starting at the root;
	   the function is applied after all the children have
	   been visited. *)
	val DFSApply : graph->(node jlist->unit)->unit

	(* Apply a function to each strongly connected component of the
	   graph (computes connected components).
	   The function is applied final depth first order, starting at the roots;
	   the function is applied after all the children have
	   been visited. *)
	val DFSCApply : graph->(node jlist->unit)->unit

	(* Output the graph in a format parseable by dot to
	 a file *)
	val dotPrint : graph->string->unit
    end
