structure Sockets:
    sig
	type addr
	type active
	type passive
	type 'a socket
		    
	val server_init: int (* port *) -> passive socket * addr
	val server_connect: passive socket -> active socket * int
	val server_close: 'a socket * 'b socket -> unit
	    
	val client_init: string (* server *) *
	                 int (* server port *) *
			 int (* client port *)
		      -> active socket
	val client_init': int (* server port *) -> active socket
	val client_close: 'a socket -> unit

	val get_port: 'a socket * addr -> int
	    
	val send_to: active socket * string -> unit
	val recv_from: active socket -> int * string
    end =
struct

    type addr = INetSock.sock_addr
    type 'a socket = 'a INetSock.stream_sock
    type passive = Socket.passive
    type active = Socket.active

    fun server_init port =
	let
	    val sock = INetSock.TCP.socket () 
	    val addr = Socket.Ctl.getSockName sock
	    val ouraddr = INetSock.any port
	    val _ = Socket.Ctl.setREUSEADDR(sock,false)
	    val _ = Socket.bind (sock, ouraddr)

	    val _ = Socket.listen (sock, 10)
	in
	    (sock, ouraddr)
	end


    fun server_connect sock =
	let
	    val (sock', addr') = Socket.accept sock
	    val port' = #2(INetSock.fromAddr addr')
	in
	    (sock', port')
	end
    
    fun get_port (sock, addr) = #2(INetSock.fromAddr addr)
	
    fun client_init(server, sport, cport) =
	let
	    val sock = INetSock.TCP.socket () 
	    val ouraddr = INetSock.any cport
	    val _ = Socket.bind (sock, ouraddr)
	    val server_addr = 
		case server of
		    "" => INetSock.any sport
		|   srv => (case NetHostDB.getByName srv of
				SOME host => INetSock.toAddr (NetHostDB.addr host, sport)
			     |  NONE => INetSock.any sport)
	    val _ = Socket.connect (sock, server_addr)
	in
	    sock
	end

    (* client_init': just choose some random port on the host we are running *)
    fun client_init' sport = client_init("", sport, 0)

    fun server_close (server_sock, client_sock) =
	(Socket.close client_sock; Socket.close server_sock)

    fun client_close sock = Socket.close sock

    fun send_to (sock, s) =
	let
	    fun sendVec (sock, vec) = 
		let val len = Word8Vector.length vec
		    fun send i = Socket.sendVec (sock, {buf=vec, i=i, sz=NONE})
		    fun put i = if (i < len) then put(i + send i) else ()
		in
		    put 0
		end
	    val vec = Word8Vector.fromList (map Byte.charToByte (explode s))
	in
	    sendVec(sock, vec)
	end
		
    fun recv_from (sock) =
	let
	    val max_recv = 1024
	    val v = Socket.recvVec(sock,max_recv)
	    val res = Byte.bytesToString v
	in
	    (max_recv, res)
	end
    
end (* structure Sockets *)
