signature CONSTRAINT =
  sig
    val reset : unit -> unit

    (* A set of values of some integer quantity *)
    type term

    (* Create a new term (possibly associated with an identifier) *)
    val freshTerm  : string option -> term
    (* Special case for constants: Create a new term for an integer constant *)
    val constantTerm : int -> term
    (* The top of the lattice; i.e. the set of all integers *)
    val top : unit -> term

    (* Expressions over terms *)
    val op + : term * term -> term
    val op - : term * term -> term
    val op * : term * term -> term
    val op / : term * term -> term
    val negate : term -> term
    val union  : term * term -> term
    val intersect  : term * term -> term
    val min  : term * term -> term 

    (* Force this to be a singleton set. *)
    val makeSingleton : term -> unit
    
    (* Constraints over terms *)
    val op =   : term * term -> unit (* unification *)
    val subset : term * term -> unit

    (* A constraint for when term t must be wholly >= than term u,
       i.e. the condition `x >= y for all x in t and y in u' *)
    val op >=  : term * term -> unit

    datatype endpoint = Int of int | Infinity | NegativeInfinity

    (* (lo,hi) = the interval lo..hi (inclusive) *)
    type range = endpoint * endpoint

    (* A variant form of the `>=' constraint:
       Query whether the listed `>=' constraint is compatible
       with the rest of the constraints.
       The result of the query is not available until the system is
       solved, so you must supply a function to call if the
       `query constraint' is inconsistent with the system of
       `non-query constraints'.
       In the case of an incompatible `query constraint', the
       supplied function will be called later with the possible
       range of values for the first term and the possible range
       of values for the second term and a string, and you can
       deal with the error condition however you like.
       The string contains a list of identifiers that this term
       depends upon, according to the output of the solver; this
       is very loosely defined, and is intended really only for
       debugging. *)
    val queryge  : term * term * (range * range * string -> unit) -> unit

    (* For printouts and debugging. *)
    val toString : term -> string

    (* Solve the constraint system *)
    val solve    : unit -> unit
  end
