structure Walk =

    struct

        (* open Bane.Pretty *)

        val doDebug = ref(false) : bool ref
        val allDebugs = ref([]) : string list ref
        val allWarnings = ref([]) : string list ref
        val all0Holes = ref([]) : string list ref
        val all1Holes = ref([]) : string list ref
        val all2Holes = ref([]) : string list ref
        (*
          fun debug x = print ("Debug: "^x^"\n")
          fun warn x  = print ("Warning: "^x^"\n")
        *)
        fun debug x = if !doDebug
                      then allDebugs := (!allDebugs) @ [x^"\n"]
                      else ()
        fun warn x = allWarnings := (!allWarnings) @ [x^"\n"]
        fun showProgress x  = (print ("--- "^x^"\n"); debug ("--- "^x))
        fun hole0 x = all0Holes := (!all0Holes) @
              ["Almost certainly a buffer overflow in "^x^"\n"]
        fun hole1 x = all1Holes := (!all1Holes) @
              ["Possibly a buffer overflow in "^x^"\n"]
        fun hole2 x = all2Holes := (!all2Holes) @
              ["Slight chance of a buffer overflow in "^x^"\n"]
        fun flushOutput () =
              (print ("\nPOSSIBLE VULNERABILITIES:\n"^(concat (!all0Holes))
                      ^(concat (!all1Holes))^(concat (!all2Holes)));
               print ("\nCAVEATS:\n"^(concat (!allWarnings)));
               if !doDebug
               then print ("\nDEBUGGING:\n"^(concat (!allDebugs)))
               else ())

        fun idAppend id str = AST.mkIdentifier ((AST.idToString id) ^ str)

        open Context

        fun applyOpt f NONE = ()
          | applyOpt f (SOME x) = f x
	
        val ge'     = Constraint.>=;     infix ge'
        val eq'     = Constraint.=;      infix eq'
        val plus'   = Constraint.+;      infix plus'
        val minus'  = Constraint.-;      infix minus'
        val mul'    = Constraint.*;      infix mul'
        val div'    = Constraint./;      infix div'
        val subset' = Constraint.subset; infix subset'
        val union'  = Constraint.union;  infix union'
        val const'  = Constraint.constantTerm
        val min'    = Constraint.min;
        val queryge = Constraint.queryge;
        fun le' (x,y) = y ge' x;         infix le'
        val top'    = Constraint.top;
        (* Because string lengths are constrained to be >= 1. *)
        fun forceLenTop t =
              ((const' 1) subset' t; (t plus' (const' 1)) subset' t)

        (* `assign (dstCtx,srcCtx)' treats the effect of `dst = src;' *)
        fun assign (None,_) = ()
          | assign (_,None) = ()
          | assign (Int di, Int si) = si subset' di
          | assign (Str {siz=ds,len=dl}, Str {siz=ss,len=sl}) =
              (* FIX: this should do a points-to analysis. *)
              ((* debug "string assignment."; *) sl subset' dl; ss subset' ds)
          | assign (Str {siz=ds,len=dl}, Int _) =
              (* Treat this as an implicit cast to the `char *' type. *)
              (debug "Assigning int to string."; ())
(*
          | assign (Int _,Str _) = (debug "BUG? assign(int,str)."; raise Fail "assign(int,str)")
*)
          | assign (_,_) = (warn "BUG: unexpected assignment types."; ())
        fun assign' (Str {len=dl,...}, Str {len=sl,...}) =
              (debug "(initializer) string assignment."; sl subset' dl)
          | assign' (x,y) = assign (x,y)

        (* `advance (Str x, Int i)' returns a `Str y' context which models
           the effect of `y = x+i;' when x,y are strings. *)
        (* FIX: I'm not sure if this code is correct (!) -- check it. *)
        fun advance' (Str {siz,len}, Int iset) =
              (* XXX: The reason why I chose this approach is subtle.
                 (One useful property: len <= siz implies len-i <= siz-i.)
                 An older, worse approach:
                       (val new as Str {siz=ns,len=nl} = freshStr NONE
                        (const' 0) subset' ns; siz subset' ns;
                        (const' 1) subset' nl; len subset' nl; new) *)
              let val ival = Constraint.freshTerm NONE
                   (* ^-- represents the value of i at some single instance,
                          rather than representing the set of possible values
                          for i across all instances. *)
              in
                 Constraint.makeSingleton ival;
                 ival subset' iset;
                 Str{siz=(siz minus' ival), len=(len minus' ival)}
              end
          | advance' (_,_) =
              (print "Error: unexpected ptr advance types.\n";
               raise (Fail "advance"))
        fun advance (Str {siz,len}, Int i) =
              Str {siz=(siz minus' i), len=(len minus' i)}
          | advance (_,_) =
              (print "Error: unexpected ptr advance types.\n";
               raise (Fail "advance"))


        fun freshStr x =
              let open Constraint
                  val (siz,len) =
                         case x of
                             NONE => (freshTerm NONE, freshTerm NONE)
                           | SOME xx => (freshTerm (SOME("siz("^xx^")")),
                                         freshTerm (SOME("len("^xx^")")))
                  val name = case x of SOME(n) => n | NONE => "(unnamed)"
                  fun endptToStr Infinity = "+Infinity"
                    | endptToStr NegativeInfinity = "-Infinity"
                    | endptToStr (Int(i)) = (Int.toString i)
                  fun rangeToStr (x,y) = (endptToStr x)^".."^(endptToStr y)
                  fun elt (NegativeInfinity, (Int(_)|Infinity)) = true
                    | elt (Int _,Infinity) = true
                    | elt (Int i,Int j) = i<j
                    | elt _ = false
                  fun chance ((sl,sh),(ll,lh)) =
                        if elt (sh,ll) then hole0
                        else if elt (sh,lh) then hole1
                        else hole2
                  fun overflow (sr,lr,d) =
                        (chance (sr,lr)) 
                              ("`"^name^"':\n  " ^
                               (rangeToStr sr)^" bytes allocated, " ^
                               (rangeToStr lr)^" bytes used.\n"^d)
                  fun unnatural (sr,lr) =
                        print ("Warning: unnatural string len for "^name^": " ^
                               (rangeToStr sr)^" bytes allocated, " ^
                               (rangeToStr lr)^" bytes used.\n")
              in
                  queryge (siz, len, overflow);
                  (* Now they are queries (don't create infeasible systems). *)
                  (*
                  queryge (siz, (const' 0), unnatural);
                  queryge (len, (const' 1), unnatural);
                              (* ^^^^^^^^ Even the empty str has a '\0' byte. *)
                  *)
                  Str({siz=siz,len=len})
              end

        datatype typKind = StrKind | IntKind | NoneKind;
        fun tkind (AST.SyntacticType(_, (AST.Char|AST.UnsignedChar), _,
                         ([AST.Array _]|[AST.Pointer _]))) = StrKind
          | tkind (AST.SyntacticType(_, (AST.Int|AST.UnsignedInt
	           |AST.ShortInt|AST.UnsignedShortInt|AST.LongInt
                   |AST.Char|AST.UnsignedChar|(AST.Enum _)
		   |AST.UnsignedLongInt|AST.LongLong|AST.UnsignedLongLong),
                         _, [])) = IntKind
          | tkind (AST.SyntacticType(_, (AST.Void|AST.Float|AST.Double
                         |AST.LongDouble), _, [])) = NoneKind
          | tkind (AST.SyntacticType(_, (AST.Aggregate _|AST.NamedAggregate _),
                         _, _)) = NoneKind
          | tkind (AST.SyntacticType(_, _, _, _::_)) = NoneKind
          | tkind t = (debug ("type = "^(Unparse.syntacticType "XXX" t));
                       raise (Fail "unknown type"))

        val killEllipsis = List.filter
                              (fn AST.EllipsisDecl _ => false | _ => true)
        (*
        fun consteq (AST.Constant(c,_)) (AST.Constant(c',_)) = (c = c')
          | consteq _ _ = raise (Fail "consteq expected consts")
        and ideq i i' = ((AST.idToString i) = (AST.idToString i'))
        *)
        fun ideq (i,i') = (AST.IdKey.compare (i,i') = EQUAL)
        fun dneq (l,l') (AST.VarDecl {typ=t,...}, AST.VarDecl {typ=t',...}) =
              typeq' ((l,t), (l',t'))
          | dneq (l,l') (AST.FunDecl {typ=t,...}, AST.FunDecl {typ=t',...}) =
              typeq' ((l,t), (l',t'))
          | dneq _ _ = raise (Fail "dneq got confused")
        and dnseq z (xs, xs') =
              (case (killEllipsis xs, killEllipsis xs') of
                   (([],_)|(_,[])) => true
                        (* XXX: Old-style fun decls often omit params. *)
                 | (ds,ds')        => ListPair.all (dneq z) (ds,ds'))
        and dclreq _ ((AST.Pointer(_)|AST.Array(_)),
                      (AST.Pointer(_)|AST.Array(_))) = true
          | dclreq _ (AST.BitField _, AST.BitField _) = true
          | dclreq z (AST.Function ds, AST.Function ds') = dnseq z (ds, ds')
          | dclreq z (AST.OldFunction(xs,_), AST.OldFunction(xs',_)) =
              (case (xs,xs') of
                   (([],_)|(_,[])) => true
                 | _    => (debug "dclreq oldfun with non-empty lists"; false))
          | dclreq _ _ = raise (Fail "dclreq got confused")
        and dclrseq z (xs,ys) = ListPair.all (dclreq z) (xs,ys)
        and speceq _ (AST.Void, AST.Void) = true
          | speceq _ ((AST.Float|AST.Double|AST.LongDouble),
                      (AST.Float|AST.Double|AST.LongDouble)) = true
          | speceq _ ((AST.Char|AST.UnsignedChar),
                      (AST.Char|AST.UnsignedChar)) = true
          | speceq _ ((AST.Int|AST.UnsignedInt|AST.ShortInt|AST.LongInt
                      |AST.UnsignedShortInt|AST.UnsignedLongInt
                      |AST.LongLong|AST.UnsignedLongLong|(AST.Enum _)),
                     (AST.Int|AST.UnsignedInt|AST.ShortInt|AST.LongInt
                     |AST.UnsignedShortInt|AST.UnsignedLongInt
                     |AST.LongLong|AST.UnsignedLongLong|(AST.Enum _))) = true
          | speceq _ (AST.NamedType i, AST.NamedType i') = ideq (i,i')
          | speceq _ (AST.NamedAggregate (a,i), AST.NamedAggregate (a',i')) =
               (a = a') andalso ideq (i,i')
          | speceq _ (AST.Aggregate (a,SOME(i),_), AST.NamedAggregate (a',i')) =
               (a = a') andalso ideq (i,i')
          | speceq _ (AST.NamedAggregate (a,i), AST.Aggregate (a',SOME(i'),_)) =
               (a = a') andalso ideq (i,i')
          | speceq z (AST.Aggregate (a,i,ds), AST.Aggregate (a',i',ds')) =
               (a = a') andalso (dnseq z (ds,ds') orelse
                   (case (i,i') of
                        (SOME x, SOME x') => ideq (x,x')
                      | _ => false))
          | speceq _ _ = false
        and typeq' ((l,AST.SyntacticType(_,s,_,ds)),
                       (l',AST.SyntacticType(_,s',_,ds'))) =
              speceq (l,l') (s, s') andalso dclrseq (l,l') (ds,ds')
        fun typeq (ct, ct') =
              typeq' (CType.resolveType ct, CType.resolveType ct')
              


        (* A symbol table, keyed on variable name, which stores
           information about the Context associated with that variable. *)
        datatype varInfo = VarInfo of {varCtx : Context, visible : bool}

        fun varCtx (VarInfo {varCtx,...}) = varCtx

        val varMap = ref(PTUtil.IdMap.empty) : varInfo PTUtil.IdMap.map ref

        fun getVarInfo id = PTUtil.IdMap.find(!varMap, id)

        (*
        fun makeTemporary id =
              let val SOME(VarInfo {varCtx,visible}) = getVarInfo id
                  val newInfo = VarInfo {varCtx=varCtx,visible=false}
              in
                  varMap := PTUtil.IdMap.insert(!varMap, id, newInfo)
              end
        *)

        fun getFret funId =
	      AST.mkIdentifier ("@" ^ (AST.idToString funId) ^ "_return")

        (* Maps function names to the list of terms representing
           the formal parameters. *)
        val funMap = ref(PTUtil.IdMap.empty) : Context list PTUtil.IdMap.map ref

        fun unify (Str({siz=xs,len=xl}), Str({siz=ys,len=yl})) =
               (xs eq' ys; xl eq' yl; ())
          | unify (Int(x), Int(y)) = (x eq' y; ())
          | unify ((Int _, Str _)|(Str _, Int _)) =
               (print "Err: Bad unification.\n"; raise (Fail "unify"))
          | unify _ = ()
        fun unifyl ([],[] )= ()
          | unifyl (([],_)|(_,[])) =
              (print "Err: Inconsistent fun params lists.\n";
               raise (Fail "FunParams"))
          | unifyl (x::xs,y::ys) = (unify (x,y); unifyl (xs,ys))


        val allocID = ref 0
        fun freshAllocStr () =
              let val i = !allocID
                  val v = AST.mkIdentifier ("alloc@" ^ Int.toString(i))
              in  allocID := i+1;
                  freshStr(SOME(AST.idToString v))
              end

        (* Poor man's map: (syntacticType * identifier) -> Context. *)
        val structMap = ref(PTUtil.IdMap.empty) :
                      (CType.extType * Context) list PTUtil.IdMap.map ref
           

        fun reset () = (
               varMap := PTUtil.IdMap.empty;
               funMap := PTUtil.IdMap.empty;
               structMap := PTUtil.IdMap.empty;
               allocID := 0;
               Constraint.reset ()
              )

        val warnedStrcat = ref false
        fun warnStrcat () =
               if not (!warnedStrcat)
               then (warn "str[n]cat() calls are not checked...";
                      warnedStrcat := true)
               else ()

        
        fun getStr id =
              (* Need to strip @0 and quotes before running `size'. *)
              let val all = AST.idToString id
                  (* val noAt = PTUtil.beforeLast #"@" all *)
                  val noFinalQuote = PTUtil.beforeLast #"\"" all
                  val noQuotes = PTUtil.afterFirst #"\"" noFinalQuote
              in
                  noQuotes
              end

        fun promoteToPtr (t,ct) =
              if CType.isArrayType ct
              then (t, CType.arrayToPtr ct)
              else (t,ct)


        (* Helps extract width and precision fields from printf formats. *)
        fun getPrec (xs,pl) =
              if "" = Substring.string xs
              then (NONE,xs,pl)
              else
              if "." = Substring.string (Substring.slice (xs,0,SOME(1)))
              then
              let val isInt = Char.contains "0123456789*"
                  val _ = debug ("saw \""^(Substring.string xs)
                                 ^"\", checking prec...")
                  val noDot = Substring.slice (xs,1,NONE)
                  val (precS,rhs) = Substring.splitl isInt noDot
                  val precS' = Substring.string precS
                  val (precOpt,pl') =
                    if precS'="" then (print "Bad fmt precision.\n";
                                       raise (Fail "printf"))
                    else if precS'="*" then
                      case pl of
                          (Int i)::ps => (SOME(i),ps)
                        | (None::ps) => (NONE,ps)
                        | _ => (print "Bad fmt param list.\n";
                                raise (Fail "printf"))
                    else
                       case Int.fromString precS' of
                           NONE => (print "Bad fmt int.\n";
                                    raise (Fail "printf"))
                         | SOME(n) => (SOME(const' n),pl)
                  val _ = debug ("prec = \""^precS'^"\" = "^
                            (case precOpt of NONE => "(none)"
                               | SOME(w) => Constraint.toString w))
              in  (precOpt,rhs,pl')
              end
              else (NONE,xs,pl)
        fun getWidthPrec (xs,pl) =
              let val isInt = Char.contains "0123456789*"
                  val (widthS,rhs) = Substring.splitl isInt xs
                  val widthS' = Substring.string widthS
                  val (widthOpt,pl') =
                    if widthS'="" then (NONE,pl)
                    else if widthS'="*" then
                      case pl of
                          (Int i)::ps => (SOME(i),ps)
                        | (None::ps) => (NONE,ps)
                        | _ => (print "Bad fmt param list (width prec).\n";
                                raise (Fail "printf"))
                    else
                       case Int.fromString widthS' of
                           NONE => (print "Bad fmt int (width prec).\n";
                                    raise (Fail "printf"))
                         | SOME(n) => (SOME(const' n),pl)
              in  (widthOpt, getPrec (rhs,pl'))
              end
        val SOME percentSign = Char.fromString "%"

        (* Model a printf() style invocation.
           `f' is the format string, `pl' is the list of Context's
           of the parameters following the format string.
           Returns a Context corresponding to the result. *)
        fun parseFmt f pl =
              let open Substring
              in
              case first f of
                  NONE => const' 1
                | SOME x =>
                    let val f' = slice (f,1,NONE)
                    in
                        if x=percentSign
                        then
                        case first f' of
                             NONE => const' 2
                           | SOME y =>
                               if y=percentSign then
                               (const' 1) plus' (parseFmt (slice(f',1,NONE)) pl)
                               else expandPercent f' pl
                        else (const' 1) plus' (parseFmt f' pl)
                    end
              end
        and expandPercent s pl =
              let val ignoreIt = Char.contains "#0-+hlL"
                  val (_,s') = Substring.splitl ignoreIt s
                  val (widthOpt, (precOpt, s'', pl')) = getWidthPrec (s',pl)
              in
                  case Substring.first s'' of
                     NONE => (print "Bad fmt str.\n"; raise (Fail "printf"))
                   | SOME(c) => expand' c widthOpt precOpt
                                   (Substring.slice(s'',1,NONE)) pl'
              end
        and expand' c widthOpt precOpt xs pl =
              if SOME(c)=(Char.fromString "s")
              then
              let val (n,pl') =
                           case pl of
                              (Str {siz,len})::ps =>
                                    (* Don't double-count the '\0'. *)
                                    (len minus' (const' 1), ps)
                            | (None::ps) =>
                                    let val t = Constraint.freshTerm NONE
                                    in  forceLenTop t; (t,ps)
                                    end
                            | _ => (print "Bad `%s' param.\n";
                                    raise (Fail "printf"))
                  val n' = case precOpt of
                              NONE => n
                            | SOME(m) => min' (m,n)
                  val n'' = case widthOpt of
                              NONE => n'
                            | SOME(m) => m union' n
                              (* XXX: max' would be a little better *)
              in  n'' plus' (parseFmt xs pl')
              end
              (* XXX: ignore precision, width fields for now, for
                 the rest of the expansions *)
              else if Char.contains "dixXuU" c
              then case pl of
                       ((Int(_)|None)::ps) => 
                           (* For a 64-bit long, it can take up to 19 digits
                              to express 2^63 base 10, plus one for the minus
                              sign. *)
                           let val t = Constraint.freshTerm NONE
                           in  (const' 1) subset' t; (const' 20) subset' t;
                               t plus' (parseFmt xs ps)
                           end
                      | _ => (print "Bad %[dixXuU] param.\n";
                              raise (Fail "printf"))
              else if Char.contains "o" c
              then case pl of
                       ((Int(_)|None)::ps) => 
                           (* For a 64-bit long, it can take up to 21 digits
                              to express 2^63 in octal, plus one for the minus
                              sign. *)
                           let val t = Constraint.freshTerm NONE
                           in  (const' 1) subset' t; (const' 22) subset' t;
                               t plus' (parseFmt xs ps)
                           end
                      | _ => (print "Bad %o param.\n"; raise (Fail "printf"))
              else if SOME(c)=(Char.fromString "c")
              then case pl of
                       ((Int(_)|None)::ps) => 
                          (const' 1) plus' (parseFmt xs ps)
                     | _ => (print "Bad %c param.\n"; raise (Fail "printf"))
              else if Char.contains "eEfgG" c
              then case pl of
                       ((Int(_)|None)::ps) => 
                           (* BUG: I didn't bother to calculate how many digits
                              long it can be, but 32 should be safe. *)
                           let val t = Constraint.freshTerm NONE
                           in  (const' 1) subset' t; (const' 32) subset' t;
                               t plus' (parseFmt xs ps)
                           end
                      | _ => (print "Bad %f param.\n"; raise (Fail "printf"))
              else (print "Unknown %-cmd in fmt.\n"; raise (Fail "printf"))

        (* Check a sprintf() call.  Called with list of (unprocessed) args
           and list of (processed) arg-contexts. *)
        fun modelSprintf (_::AST.Constant(AST.String fmtId,_)::_)
                         ((dst as Str {siz=ds,len=dl})::_::rest) =
              let val fmt = getStr fmtId
                  val _ = debug ("sprintf(.,\""^fmt^"\",...).")
                  val parsedLen = parseFmt (Substring.all fmt) rest
              in
                  parsedLen subset' dl;
                  dst
              end
          | modelSprintf _ (None::_) =
              (warn "sprintf(unknown,...)"; None)
          | modelSprintf _ ((dst as Str {siz=ds,len=dl})::_) =
              (warn "sprintf(.,unknown,...)"; forceLenTop dl; dst)
          | modelSprintf _ _ =
              (warn "Weird sprintf() call."; None)
        fun modelSnprintf (_::_::AST.Constant(AST.String fmtId,_)::_)
                         ((dst as Str {siz=ds,len=dl})::nt::_::rest) =
              let val fmt = getStr fmtId
                  val _ = debug ("snprintf(.,"
                                 ^(case nt of Int(n) => Constraint.toString n
                                              | _ => "(??)")
                                 ^",\""^fmt^"\",...).")
                  val parsedLen = parseFmt (Substring.all fmt) rest
              in
                  case nt of
                      Int(n) => (min' (parsedLen,n)) subset' dl
                    | None   => (warn "ignoring 2nd arg...";
                                 parsedLen subset' dl)
                    | _      => (print "Bad 2nd arg in snprintf().\n";
                                 raise (Fail "modelSnprintf"));
                  dst
              end
          | modelSnprintf _ (None::_) =
              (warn "snprintf(unknown,...)"; None)
          | modelSnprintf _ _ =
              (warn "Weird snprintf() call.\n"; None)

        (* For modelling argv[].  (HACK! FIX) *)
        val starArgvId = AST.mkIdentifier "argv[0]@main()"
              

        (* Have to do it in two passes:
           One = remember global decls (for typing),
           Two = generate constraints. *)
        datatype Pass = One | Two;

        (* These are here because ML doesn't allow forward references, argh!
           All this just so freshVarCtx can call expression. *)
        (*
          freshVarCtx (AST.SyntacticType(_, AST.Char, _,
	           [AST.Array(SOME(AST.Constant(AST.Integer(i),_)))] )) v =
         *)
        fun freshVarCtx (lookup, AST.SyntacticType(_, AST.Char, _,
	           [AST.Array(SOME(e))] )) v =
              (case #1 (expr lookup e) of
                   Int i =>
	             let val s as Str {siz,len} = freshStr(SOME(v))
	             in
	                 (* siz eq' i; *)
                     i subset' siz;
                     s
	             end
                 | _ => freshStr(SOME(v)))
          | freshVarCtx (lookup,typ) v =
              case tkind (#2 (CType.resolveType (lookup,typ))) of
                  StrKind  => freshStr(SOME(v))
                | IntKind  => Int(Constraint.freshTerm (SOME v))
                | NoneKind => None
        and createVar lookup id visible typ =
            let val vname = AST.idToString id
                val varCtx = freshVarCtx (lookup,typ) vname
                val isVisible = visible andalso (not (PTUtil.varIsFretName id))
                      (* ^-- Don't print function return info. *)
                val vInfo = VarInfo {varCtx=varCtx, visible=isVisible}
            in
                (debug ("New var " ^ vname ^ " ["
		        ^ (Context.toString varCtx) ^ "]"));
                varMap := PTUtil.IdMap.insert(!varMap, id, vInfo);
                vInfo
            end
        and newVar lookup id visible typ =
            let val vOpt = getVarInfo id
            in  case vOpt of
                    SOME v => v
                            (*
                              (print ("Debug: var "^(AST.idToString id)^
                                      " already exists, not creating.\n"); v)
                            *)
                  | NONE   => createVar lookup id visible typ
            end
        and newUnknownVar lookup id warn =
              (* Generate fresh varinfo's even when we can't
                 find the variable, because it might be a previously-
                 unmentioned symbol that isn't bound until link time. *)
              let val fallback = AST.SyntacticType(NONE, AST.Void, [], [])
                  val t = case lookup AST.VarNames id of
                              (SOME(AST.SymInfo {typ,...})) => typ
                            | NONE => (if warn then
                                       debug ("Couldn't find typ for "
                                         ^ (AST.idToString id)
                                         ^ ", assuming void.")
                                       else (); fallback)
              in
                if warn then
                  debug ("Couldn't find variable info for "
                         ^ (AST.idToString id) ^ "; trying to make "
                         ^ "something up.")
                else ();
                varCtx (newVar lookup id true t)
              end
        and applyFun lookup id pl =
              (case checkStrOp lookup (AST.idToString id) pl of
                  SOME(x) => x
                | NONE => (doFormals id pl; doRet lookup id pl))
        and doFormals id pl =
              let fun mapAssign (x::xs) (y::ys) =
                        (assign (x,y); mapAssign xs ys)
                    | mapAssign [] [] = ()
                    | mapAssign _ _ =
                        (warn ("Wrong number of fun args calling `"
                              ^(AST.idToString id)^"'?");
                         (* raise (Fail "FunParams"); *) ())
                  val formals = PTUtil.IdMap.find(!funMap, id)
              in
                case formals of
                  NONE => warn ("undeclared function "
                                 ^ (AST.idToString id) ^ "().")
                          (* FIX: should really do newUnknownVar's? *)
                | SOME(formalCtxs) => mapAssign formalCtxs pl
              end
        and doRet lookup id pl =
              let val fret = getFret id
              in
                  case getVarInfo fret of
                       NONE => (warn ("no return info for "
                                      ^(AST.idToString id) ^ "().");
                                newUnknownVar lookup fret false)
                     | SOME vInfo => varCtx vInfo
              end
        and rememberFunParams lookup funName params ignoreFormalNames =
              let fun doFormal NONE (AST.VarDecl{name,init=NONE,typ,...}) =
                        if ignoreFormalNames
                        then freshVarCtx (lookup,typ) (AST.idToString name)
                        else varCtx (newVar lookup name true typ)
                    | doFormal (SOME ctx)
                                   (AST.VarDecl{name,init=NONE,typ,...}) =
                        let val vInfo = VarInfo {varCtx=ctx,visible=true}
                        in
                            if ignoreFormalNames then () else
                            varMap := PTUtil.IdMap.insert(!varMap, name, vInfo);
                            (* ^^ No checking that the FunDef and FunDecl
                                  are consistent; leave that to gcc -Wall. *)
                            ctx
                        end
                    | doFormal _ _ =
                        (print "Err: Saw bad fun param.\n";
                         raise (Fail "FunParams"))
                  fun mdf [] (y::ys) = (doFormal NONE y)::(mdf [] ys)
                    | mdf xs [] = xs
                    | mdf (x::xs) (y::ys) = (doFormal (SOME x) y)::(mdf xs ys)

                  val params' = killEllipsis params
                  val pInfo = case PTUtil.IdMap.find(!funMap, funName) of
                                   NONE => [] | SOME l => l
                  val newInfo = mdf pInfo params'
              in 
                  funMap := PTUtil.IdMap.insert(!funMap, funName, newInfo)
              end
        and getField ct id =
              let val pairsopt = PTUtil.IdMap.find(!structMap, id)
                  fun isCt (ct',_) = typeq (ct,ct')
                  val ft = CType.fieldType ct id
                  val name = ("(unnamed field "^(AST.idToString id)^")")
                  fun getNew () = freshVarCtx ft name
              in
                  case pairsopt of
                      NONE       =>
                        let val n = getNew ()
                            val m = [(ct,n)]
                        in
                            structMap := PTUtil.IdMap.insert(!structMap, id, m);
                            n
                        end
                    | SOME pairs =>
                        (case map #2 (List.filter isCt pairs) of
                             [x] => x
                           | []  =>
                                  let val n = getNew ()
                                  in
                                      structMap := PTUtil.IdMap.insert(
                                          !structMap, id, (ct,n)::pairs );
                                      n
                                  end
                           | _   => (print "Bad getField().\n";
                                     raise (Fail "ambiguous getField")))
              end
        (* FIX: would be nice to detect strings which might not be
           nul-terminated, e.g. after a strncpy(). *)
        (* FIX: would be nice to detect char *'s which might hold NULL
           values that might be dereferenced, e.g. after a strchr(). *)
        (* FIX: add the memcopy(), bcopy(), type of calls.
           See e.g. man 3 bstring. *)
        (* FIX: To do:
            strxfrm(), strtok(), strsep(),
            stpncpy(), stpcpy(), swab(),
            scanf(), vsprintf(), syslog(), argv[], argc,
            fread(), sscanf(), read(), environ
	    getcwd(), getwd()
         *)
        (* FIX: double-check semantics.  Watch out for off-by-one errors;
           e.g. strlen(s) = #len s - 1. *)

        and checkStrOp _ "strcpy" [Str {siz=ds,len=dl},
                                 src as Str {siz=ss,len=sl}] =
               (sl subset' dl; SOME(src))
          | checkStrOp _ "strncpy"
                   [dst as Str {siz=ds,len=dl}, Str {siz=ss,len=sl},Int i] =
               ((min'(sl,i)) subset' dl; SOME(dst))
          | checkStrOp _ "strlen" [Str {siz,len}] =
               (* Subtract one for the '\0'. *)
               SOME(Int(len minus' (const' 1)))
          | checkStrOp _ ("strspn"|"strcspn") [Str {siz,len}] =
               let val t = Constraint.freshTerm NONE
               in  (const' 0) subset' t; (len minus' (const' 1)) subset' t;
                   SOME(Int(t))
               end
          | checkStrOp _ "strdup" [Str {siz,len}] =
               let val new as Str {siz=ns,len=nl} = freshAllocStr ()
               in
                   (* FIX: This is suboptimal, because if `len' is a
                      non-trivial range (e.g. not just a singleton set),
                      then the program will always find a (probably
                      spurious) vulnerability in the strdup().
                      Can we fix this somehow?? *)
                   (* FIX: Is unifying `ns' and `len' is a good idea?
                      Can you get get ``backwards information flow''? *)
                   (* ns eq' len; *)
                   len subset' ns;
                   len subset' nl;
                   SOME(new)
               end
          | checkStrOp _ "strcat"
                   [dst as Str {siz=ds,len=dl}, Str {siz=ss,len=sl}] =
               (* FIX: It's impossible to model strcat() correctly in a
                  flow-insensitive analysis, so we punt.  Oh well. *)
               (warnStrcat (); sl subset' dl; SOME(dst))
          | checkStrOp _ "strncat"
                   [dst as Str {siz=ds,len=dl}, Str {siz=ss,len=sl}, Int i] =
               (* Be warned: the semantics of strncat() are a little confusing.
                  Check the libc code yourself; don't trust anything you read
                  here. *)
               (* FIX: It's very hard to model strncat() accurately in a
                  flow-insensitive analysis.  I do something very pessimistic;
                  as a side effect, strncat() will almost always trigger false
                  alarms, unless you use e.g.
                      strncat(a, b, sizeof(a) - strlen(b) - 1);
                  It would be nice to do better.  Maybe we can assume
                  optimistically that any given strncat() will be
                  executed only "once"? *)
               (warnStrcat ();
                ((min'(sl minus' (const' 1), i)) plus' dl) subset' dl;
                SOME(dst))
          | checkStrOp _ ("strchr"|"strrchr"|"strstr"|"strpbrk"
                          |"index"|"rindex")
                   [s as Str {siz,len}, _] =
               let val n as Str {siz=ns,len=nl} = freshStr NONE
               in
                   (const' 1) subset' ns; siz subset' ns;
                   (const' 1) subset' nl; len subset' nl;
                   (* (ns minus' nl) ge' (siz minus' len) *)
                   SOME(n)
               end
               (* No longer as good.
               let val n = Constraint.freshTerm ()
                    (* ^-- represents the index into the string where the
                           named character first (last) appears, so that the
                           return value of the str*() function is s+i. *)
               in  (const' 0) subset' n; (len minus' (const' 1)) subset' n;
                   case advance (s,Int(n)) of
                       x as Str _ => SOME(x)
                     | None       => (print "Err: str([r]chr|str)().\n";
                                      raise (Fail "advance"))
               end
               *)
          | checkStrOp _ "fgets" [s as Str {siz,len}, Int i, _] =
               ((const' 1) subset' len; i subset' len; SOME(s))
          | checkStrOp _ "gets" [s as Str {siz,len}] =
               (warn "Dear god, a call to gets()!";
                forceLenTop len; SOME(s))
          (* sprintf(), snprintf() should have been caught earlier.
             (sprintf, snprintf don't fit in here, because
             we need to know the value of the string passed as the
             format (not the context associated with the format string).) *)
          | checkStrOp _ ("sprintf"|"snprintf") _ =
               (print "Err: unexpected s[n]printf().\n";
                raise (Fail "unexpected s[n]printf"))
          | checkStrOp _ "getenv" _ =
               let val t = Constraint.freshTerm (SOME "@getenv_return");
                   val s = Str {siz=t,len=t}
               in  forceLenTop t; SOME(s)
               end
          | checkStrOp _ ( "malloc" | "valloc" | "xmalloc" | "alloca"
                     | "__builtin_alloca" ) [Int i] =
                (* FIX: Should also add calloc, realloc support. *)
                let val new as Str {siz=ns,len=nl} = freshAllocStr ()
                in
                    (* i eq' ns; *)
                    i subset' ns;
                    SOME(new)
                end
          | checkStrOp _ ("fprintf"|"printf") _ =
                SOME(None) (* XXX: FIX: HACK for '...' *)
          | checkStrOp lookup (ghby as ("gethostbyname"|"gethostbyaddr"
                              |"gethostbyname2")) _ =
              let val s0 = AST.SyntacticType(NONE, AST.NamedAggregate(
                             AST.Struct, AST.mkIdentifier "hostent"), [], [])
                  val s = CType.resolveType (lookup,s0)
                  val Str {siz=ns,len=nl} =
                          getField s (AST.mkIdentifier "h_name") 
                  val Int l = getField s (AST.mkIdentifier "h_length") 
              in
                  debug ghby;
                  (top' ()) subset' l;
                  forceLenTop nl;
                  SOME(None)
              end
          | checkStrOp _ "gethostname" [Str {siz,len}, Int i] =
              (i subset' len; SOME(None))
	           (* FIX: should be 1..i subset len *)
          | checkStrOp _ "syserror" _ =
              (* FIX: just do this once, to the real @syserror_return term *)
              (* XXX: 48 is for Linux *)
              let val t = Constraint.freshTerm (SOME("@syserror_return"))
              in (const' 0) subset' t; (const' 48) subset' t; SOME(Int(t))
              end
          | checkStrOp _ "syssignal" _ =
              (* FIX: Same comments as for syserror(). *)
              let val t = Constraint.freshTerm (SOME("@syssignal_return"))
              in (const' 0) subset' t; (const' 25) subset' t; SOME(Int(t))
              end
          | checkStrOp _ _ _ = NONE






        and expr lookup (AST.Variable (v,_)) =
              let val ct = CType.getVarType lookup v
                  val t = case getVarInfo v of
                             NONE => newUnknownVar lookup v true
                           | SOME vInfo => varCtx vInfo
              in  (t, ct)
              end
          | expr _ (AST.Constant (AST.Integer(i),_)) =
              (Int(const' i), CType.intType)
          | expr _ (AST.Constant (AST.Character(c),_)) =
              let val t = (* if (AST.idToString c)="'\\0'"
                      then StringTerminator else *) None
              in  (t, CType.charType)
              end
          | expr _ (AST.Constant (AST.String(id),_)) =
              let val str = getStr id
                  (* Don't forget to add one for the trailing '\0'! *)
                  val slen = 1 + (size str)
	          val l = const' slen
              in
	          debug ("saw " ^ str ^ "; size(" ^ str
                          ^ ") = " ^ (Int.toString slen) ^ ".");
	          (Str({siz=l, len=l}), CType.stringType)
              end
          | expr _ (AST.Constant _) = (None, CType.nullType)
          | expr lookup (AST.Conditional(e1, e2, e3, _)) =
              (* Merge control flow *)
              let val (t1, ct1) = promoteToPtr (expr lookup e1)
                  val (t2, ct2) = promoteToPtr (expr lookup e2)
                  val (t3, ct3) = promoteToPtr (expr lookup e3)
              in
               case (t2, t3) of
                   (Int i,Int j) =>
                      let val t = Constraint.freshTerm NONE
                      in
                        i subset' t; j subset' t;
                        (Int(t), ct2)
                      end
                 | (Str {siz=s2,len=l2}, Str {siz=s3,len=l3}) =>
                      let val s as Str {siz=ns,len=nl} = freshStr NONE
                      in  
                        s2 subset' ns; s3 subset' ns;
                        l2 subset' nl; l3 subset' nl;
                        (s, ct2)
                      end
                 (*
                 | (StringTerminator,StringTerminator) =>
                      (StringTerminator, ct2)
                 *)
                 | (None,_) => (None, ct2)
                 | (_,None) => (None, ct2)
                 | (_,_) => (warn "Saw weird conditional; bad code?";
                             (None, ct2))
              end
          | expr lookup (AST.Application(e as AST.Variable(id,_), el, _, _)) =
              let val el' = map (#1 o promoteToPtr o (expr lookup)) el
                  val ct = CType.returnType (#2 (promoteToPtr (expr lookup e)))
                  val t = case AST.idToString id of
                              "sprintf" => modelSprintf el el'
                            | "snprintf" => modelSnprintf el el'
                            | _ => applyFun lookup id el'
              in  (t, ct)
              end
          | expr lookup (AST.Application (e, _, _, _)) =
              let val ct = CType.returnType (#2 (promoteToPtr (expr lookup e)))
              in
                  warn "function pointers; analysis is unsafe...";
                  (None, ct)
              end
          | expr lookup (node as AST.Operation(AST.Incr, [e], region)) =
                   expr lookup (AST.Assignment(e, SOME AST.Plus,
                            AST.Constant (AST.Integer 1, region), region))
          | expr lookup (node as AST.Operation(AST.Decr, [e], region)) =
                   expr lookup (AST.Assignment(e, SOME AST.Minus,
                            AST.Constant (AST.Integer 1, region), region))
          | expr lookup (node as AST.Operation(oper, el, region)) =
              (* XXX: must fix some of this for ctype *)
              let val l = map (promoteToPtr o (expr lookup)) el
                  val (ets,ects) = (map #1 l, map #2 l)
                  val ct = CType.operType oper ects
              in
              case (oper,ets) of
                  ((AST.Plus|AST.Minus|AST.Mul|AST.Div),[Int(i),Int(j)]) =>
                     let val t = 
                           case oper of
                                AST.Plus => Int(i plus' j)
                              | AST.Minus => Int(i minus' j)
                              | AST.Mul => Int(i mul' j)
                              | AST.Div => Int(i div' j)
                     in  (t, ct)
                     end
                | ((AST.Lt|AST.Le|AST.Gt|AST.Ge|AST.Eq|AST.Ne|AST.Mod|
                    AST.And|AST.Or|AST.BitAnd|AST.BitOr|AST.BitXor|AST.Rshift|
                    AST.Lshift),[Int _,Int _]) =>
                     let val i = Constraint.freshTerm NONE
                     in  (const' 0) subset' i; (const' 1) subset' i;
                         (*
                         debug ("ignoring "^(Unparse.expression node));
                         *)
                         (Int(i), ct)
                     end
                | (_,[Int _, Int _]) =>
                        (print "Err: Unknown binary op.\n";
                         raise (Fail "BadOp"))
                | (AST.UnaryMinus,[Int(i)]) =>
                     (Int(Constraint.negate i), ct)
                | (AST.Address,[Int _]) =>
                     (warn "int pointers; analysis is unsafe."; (None, ct))
                | (AST.Address,[Str _]) =>
                     (warn "string pointers; analysis is unsafe."; (None, ct))
                | (AST.Deref,[Int _]) =>
                     (print "Err: Dereferencing int constant!?\n";
                      raise (Fail "BadOper"))
                | (AST.Deref,[Str({siz,len})]) =>
                     (* redundant, but here for emphasis *)
                     let fun fencepost _ =
                               warn ("Possible deref after end in ``"^
                                      (Unparse.expression node)^"''.")
                     in  queryge (len, const' 1, fencepost); (None, ct)
                     end
                | (AST.Deref,[None]) =>
                     (case el of
                          ([AST.Variable(v,_)]|
                           [AST.Operation(AST.Plus, [AST.Variable(v,_), _], _)])
                          => (* FIX: this is a major kludge! *)
                             if (AST.idToString v)="argv@main()"
                             then (case getVarInfo starArgvId of
                                       NONE => (print "Err: argv[]\n";
                                                raise (Fail "argv"))
                                     | SOME info => (varCtx info, ct))
                             else (None, ct)
                        | _ => (None, ct))
                | (AST.Plus,[s as Str _,i as Int _]) => (advance (s,i), ct)
                | _ => (None, ct)
                (*
                | (_,[(Str(_)|Int(_)),Str _]) =>
                    (print "Err: Int/Str op Str!?  Probably bad code.\n";
                     raise (Fail "BadOper"))
                | (AST.Not,[Str _]) => (None, ct)
                | (_,([None]|[None,None]|[Int _]|[Int _,Int _])) => (None, ct)
                | _ => (print "Unknown operator.\n"; raise (Fail "BadOp"))
                *)
              end
              (* FIX: add case for + and - ops on strings when we have some
                 aliasing analysis for string pointers *)
          | expr lookup (AST.Assignment(e1, SOME oper, e2, region)) =
              expr lookup (AST.Assignment(e1, NONE,
                  AST.Operation(oper, [e1,e2], region), region))
          | expr lookup (AST.Assignment(dst, NONE, src, _)) =
              (* FIX: This is not right.  Should find out what dst points
                 to and set it to src'; but we'll need a points-to analysis
                 for that. *)
              let val (dst', _) = expr lookup dst
                  val (src', ct) = promoteToPtr (expr lookup src)
              in
                 assign (dst',src');
                 (src', ct)
              end
          | expr lookup (AST.Comma(e1, e2, _)) =
                (promoteToPtr (expr lookup e1); promoteToPtr (expr lookup e2))
          | expr lookup (AST.Cast(t, e, _)) =
                let val (e',_) = expr lookup e
                    val ct = CType.resolveType (lookup, t)
                in
                    case (tkind (#2 ct),e') of
                        (IntKind,Int _) => (e', ct)
                      | (IntKind,_)     => (Int(top' ()), ct)
                      | (StrKind,Str _) => (e', ct)
                      | (StrKind,_)     => (None, ct)
                      | (NoneKind,_)    => (None, ct)
                end
          | expr lookup (AST.SizeOfExpr(e, aI)) =
                   (case (expr lookup e) of
                        (Str({siz,len}),_) => (Int(siz), CType.intType)
                      | ((Int _|(*StringTerminator|*)None),(l,t)) =>
                            expr l (AST.SizeOfTypeExpr(t,aI))
                      | _ => (warn "Saw a Sizeof; implement later.\n";
                              (None, CType.intType)))
          | expr l (AST.SizeOfTypeExpr (t as AST.SyntacticType(_,sp,_,ds),aI)) =
                (case (sp,ds) of
                     (AST.NamedType _,_) =>
                        let val (l',t') = CType.resolveType (l,t)
                        in  expr l' (AST.SizeOfTypeExpr(t',aI))
                        end
                   | (((AST.Enum _|AST.Aggregate _|AST.NamedAggregate _
                       |AST.Void),_)|(_,_::_)) =>
                        (warn "SizeofType(Enum/Agg/Void/...); implement later.";
                         (None, CType.intType))
                   | ((AST.Char|AST.UnsignedChar),[])
                        => (Int(const' 1), CType.intType)
                   | ((AST.ShortInt|AST.UnsignedShortInt),[])
                        => (Int(const' 2), CType.intType)
                   | ((AST.Int|AST.UnsignedInt|AST.LongInt
                      |AST.UnsignedLongInt),[])
                        => (Int(const' 4), CType.intType)
                   | ((AST.LongLong|AST.UnsignedLongLong),[])
                        => (Int(const' 8), CType.intType)
                   | (AST.Float,[]) => (Int(const' 4), CType.intType)
                   | (AST.Double,[]) => (Int(const' 8), CType.intType)
                   | (AST.LongDouble,[]) => (Int(const' 12), CType.intType)
                   | (_,_) => (warn "SizeofType(???); got confused.";
                               (None, CType.intType))
                )
          | expr lookup (AST.FieldDirect(e, id, _)) =
                let val (e',ct) = promoteToPtr (expr lookup e)
                    val ft = CType.fieldType ct id
                in
                    (* debug ("FieldDirect `"^(AST.idToString id)^"'"); *)
                    (getField ct id,ft)
                end
          | expr lookup (AST.FieldIndirect(e, id, reg)) =
                expr lookup (AST.FieldDirect(
                               AST.Operation(AST.Deref, [e], reg), id, reg))
          | expr lookup (AST.Postfix(oper, e, region)) =
                expr lookup (AST.Operation(oper, [e], region))
          | expr lookup (AST.Index(a, i, region)) =
              expr lookup (AST.Operation (AST.Deref,
                [AST.Operation(AST.Plus, [a, i], region)], region))
          | expr _ (AST.Extension (stmt,_)) =
                 (case stmt of
                     AST.Compound {decls, body, lookup, ...} =>
                         (app (declaration One lookup) decls;
                          app (declaration Two lookup) decls;
                          app (statement lookup) body;
                          case List.last body of
                              AST.Expression (SOME e, _) =>
                                  expr lookup e
                            | _ => (print ("Err: Extension's last statement " ^
                                          "is not an expr\n");
                                    raise (Fail "Extension")))
                   | _ => (print "Err: Extns'n doesn't contain compound stmt\n";
                          raise (Fail "Extension")))
          | expr _ _ =
              (print "Err: Unimplemented expr!\n"; raise (Fail "expr"))

        and expression lookup e =
              (expr lookup e handle ex =>
                       (print (" "^(Unparse.expression e)^"\n"); raise ex);
               ())


        (* Generates constraints for expressions within the statement. *)
        and stmt lookup (AST.Return(SOME e, SOME f, astInfo)) =
              (* FIX: I suspect `fretvar' will have to go into some
                 symbol table somewhere, somehow. *)
              let val fretvar = AST.Variable(getFret f, PTUtil.nullASTInfo)
              in
                  expression lookup (AST.Assignment(fretvar, NONE, e, astInfo))
              end
          | stmt _ (AST.Return _) = ()
          | stmt lookup (AST.Expression(SOME e, _)) =
              expression lookup e
          | stmt _ (AST.Expression(NONE, _)) = ()
          | stmt lookup (AST.IfThen(e, s, _)) =
              (expression lookup e; stmt lookup s)
          | stmt lookup (AST.IfThenElse(e, s1, s2, _)) =
              (expression lookup e; stmt lookup s1; stmt lookup s2)
          | stmt lookup (AST.WhileDo(e, s, _)) =
              (expression lookup e; stmt lookup s)
          | stmt lookup (AST.DoWhile(s, e, _)) =
              (expression lookup e; stmt lookup s)
          | stmt _ (AST.GoTo _) = ()      (* Goto has no effect *)
          | stmt lookup (AST.For(eopt1, eopt2, eopt3, s, _)) =
            (applyOpt (expression lookup) eopt1;
             applyOpt (expression lookup) eopt2;
             applyOpt (expression lookup) eopt3;
             stmt lookup s)
          | stmt lookup (AST.Switch(e, s, _)) =
              (expression lookup e; stmt lookup s)
          | stmt lookup (AST.Case(_, NONE, _)) = ()
          | stmt lookup (AST.Case(_, SOME s, _)) = stmt lookup s
          | stmt lookup (AST.Labeled(_, NONE, _)) = ()
          | stmt lookup (AST.Labeled(_, SOME s, _)) = stmt lookup s
          | stmt _ (AST.Break _) = ()
          | stmt _ (AST.Continue _) = ()
            (* Break and continue have no effect *)
          | stmt _ (AST.Compound {decls, body, lookup, ...}) =
                (map (declaration One lookup) decls;
                 map (declaration Two lookup) decls;
                 map (stmt lookup) body; ())

        and statement lookup x =
              (* (debug ("statement: "^(Unparse.statement x)); *)
              (stmt lookup x)


        and declaration One lookup (AST.VarDecl{name, typ, ...}) =
            (newVar lookup name true typ; ())
          | declaration Two lookup (AST.VarDecl{name, init, region, typ}) =
            (* FIX: Need to handle the case `char a[] = "foo";' properly.
               Currently, it finds that 4 \in len(a), but it doesn't
               discover that siz(a) = 4. *)
            let val newCtx = newVar lookup name true typ
                val astInfo = {region=region, extraInfo=ref NONE, typeInfo=NONE}
                val varExpr = AST.Variable(name, astInfo)
                val expr' = #1 o (expr lookup)
                fun doInit (AST.SyntacticType(_, AST.Char, _,
                                                 [AST.Array(SOME _)] )) e =
                        assign' (expr' varExpr, expr' e)
                  | doInit (AST.SyntacticType(_, AST.Char, _,
                                                 [AST.Array(NONE)] ))
                           (e as AST.Constant(AST.String s, _)) =
                        let val v' as Str {siz=vs,...} = expr' varExpr
                            val Str {siz=es,...} = expr' e
                        in
                            vs eq' es;
                            assign' (v',expr' e)
                        end
                 | doInit _ e = assign (expr' varExpr, expr' e)
            in
               case init of
                   NONE => expression lookup varExpr
                 | SOME(AST.Initial e) => doInit typ e
                 | SOME(AST.InitialList _) =>
                      (warn "VarDecl InitialList.";
                       expression lookup varExpr)
                      (* FIX: handle array types in InitialList. *)
            end
          | declaration pass l (fd as AST.FunDef{name, decls, body, lookup,
                                              params, region, return, ...}) = (
	          newVar l (getFret name) true return;
                  rememberFunParams lookup name params false;
                  case pass of
                      One => ()
                    | Two => (map (declaration pass lookup) decls;
                              map (statement lookup) body; ())
	      )
          | declaration One lookup (AST.FunDecl{name,return,params=NONE,...}) =
                  (newVar lookup (getFret name) true return; ())
          | declaration One lookup (f as AST.FunDecl{name, return,
                                                 params=SOME(p), ...}) = (
                  newVar lookup (getFret name) true return;
                  (* debug ("FunDecl: "^(Unparse.declaration f)); *)
                  rememberFunParams lookup name p true;
                  ()
              )
          | declaration Two _ (AST.FunDecl _) = ()
          | declaration _ _ (AST.EllipsisDecl _) = ()
          | declaration _ _ (AST.TypeDef _) = ()

        and program pass (filename, AST.Program {decls, lookup}) =
              let val what = case pass of One => "protos" | Two => "code"
              in
                  showProgress ("Analyzing "^what^" in `"^filename^"'...");
                  map (declaration pass lookup) decls;
                  ()
              end

        fun errLookup _ _ = raise Match

        (* Do global stuff that's basically independent of the actual
           program source code. *)
        fun genericBounds () =
              let val t = Constraint.freshTerm (SOME("caller"))
                  val argc = Int(t)
                  val s as Str {siz,len} =
                        freshStr(SOME(AST.idToString starArgvId))
                  (* FIX: do environ[] too *)
              in
                  (const' 0) subset' t; (t plus' (const' 1)) subset' t;
                  varMap := PTUtil.IdMap.insert(!varMap, starArgvId,
                               VarInfo {varCtx=s, visible=true});
                  doFormals (AST.mkIdentifier "main") [argc,None]
              end


        fun parseFile filename =
            let val _ = showProgress ("Parsing `"^filename^"'...")
                val ind = TextIO.openIn filename
                val {sourceMap, tree=result} = (CParser.parse (ind,filename)
                                                handle (Fail s) =>
                                                    (TextIO.closeIn ind;
                                                     raise Fail s))
                val _ = TextIO.closeIn ind
            in
                result
            end

        fun convert filename =
            let val parsed = parseFile filename
                val _ = showProgress ("Alpha-converting `"^filename^"'...")
                val converted = Convert.convert filename parsed
            in  converted
            end

        fun walklist fileList = (
                (* Reset the system *)
                reset();
                (* Bane.Flags.set NONE "print_dontCompress"; *)
                (* Bane.reset(); *)
                ExtraInfo.reset();

                (* Analyze the program ... *)
                let val trees = map (fn x => (x, convert x)) fileList
                in
                    (* ... in two passes. *)
                    showProgress ("Memorizing prototypes...");
                    app (program One) trees;
                    genericBounds ();
                    showProgress ("Generating constraints...");
                    app (program Two) trees;
                    ()
                end;

                showProgress ("Solving constraints...");
                Constraint.solve ();
                flushOutput ();
                ()
              ) handle e as (Fail s) =>
                (print "FAILED!\n"; flushOutput (); print "FAILED!\n"; raise e)
        fun showexn e =
              let val msg = exnMessage e
                  val histl = SMLofNJ.exnHistory e
                  val histl' = map (fn s => (s^"\n")) histl
                  val hist = concat histl'
              in
                  print ("Error: uncaught exception: "
                         ^ "Could be a bug in my code, or something else\n"
                         ^ msg ^ "\n" ^ hist)
              end
        fun walkmany files =
              let val flattened =
                map (PTUtil.flatten_dir o PTUtil.translate_filename) files
              in  walklist (List.concat flattened)
              end handle e => (showexn e; raise e)
        fun walk filename = walkmany [filename]
        fun main (_,[]) =
              (print "Usage: boon [--nodebug|--debug] foo.i [...]\n";
               OS.Process.failure)
          | main (_,files) =
              let fun f "--nodebug" = (doDebug := false; false)
                    | f "--debug"   = (doDebug := true; false)
                    | f _ = true
                  val files' = List.filter f files
              in  walkmany files'; OS.Process.success
              end
        fun export () = SMLofNJ.exportFn ("boonheap", main)

    end
