
/*  ###############################################################
    ##
    ##     C Tree Builder
    ##
    ##     File:         tree.c
    ##
    ##     Revision 0.0.0.2  1996/04/03  10:26:06  kurt
    ##     Added line,column to some printouts;
    ##     addresses of nodes printed out, too.
    ##     -- Kurt Cockrum <kurt@grogatch.seaslug.org>
    ##
    ##     Revision 0.0.0.1  1996/04/03  10:08:46  kurt
    ##     o    missing return in make_for
    ##     o misc. cleanups
    ##     -- Kurt Cockrum <kurt@grogatch.seaslug.org>
    ##
    ##     Revision 0.0.0.0  1996/03/30  02:49:41  kurt
    ##     o    added right curly brackets inside comments for benefit of vi
    ##     o    Addresses of tree nodes are printed in messages;
    ##         minor content changes to some messages
    ##     o    Threw away ntype[]
    ##     o   Rev. 0.0 from
    ##         -rw-r--r--  1 1047 1047 44198 Aug  7 19:32
    ##             ftp://ftp.cs.wisc.edu/coral/tmp/spf/ctree.0.01.tar.gz
    ##     -- Kurt Cockrum <kurt@grogatch.seaslug.org>
    ##
    ##     Programmer:   Shaun Flisakowski
    ##     Date:         Jan 21, 1995
    ##
    ##
    ###############################################################  */

#include    <iostream.h>
#include    <string.h>
#include    <stdlib.h>
#include    <assert.h>

#include    "tree.h"
#include    "gram.h"
#include    "token.h"
#include    "project.h"

/*  ###############################################################  */
char*
print_ptr(void *ptr)
{
    static char buf[32];

#if defined(__STDC__) || defined(__cplusplus)
    sprintf( buf, "%p", ptr );
#else
    sprintf( buf, "0x%0x", (unsigned int) ptr );
#endif

    return buf;
}

/*  ###############################################################  */
baseNode::baseNode()
{
    static NameTbl fileTbl;
    filename = fileTbl.insert(gProject->Parse_TOS->filename,
                              strlen(gProject->Parse_TOS->filename))->_string;

    which = NONE_T;
    line = col = 0;
    tok = 0;
    type = TN_EMPTY;
}

/*  ###############################################################  */
treeNode::treeNode(tnType t)
         :baseNode()
{
    gProject->Parse_TOS->node_heap->insert(this);

    which = NODE_T;
    type  = t;

    lnode = (treeNode *) NULL;
    rnode = (treeNode *) NULL;
}

/*  ###############################################################  */
ifNode::ifNode(tnType t)
       :baseNode()
{
    gProject->Parse_TOS->node_heap->insert(this);

    which = IF_T;
    type  = t;
    cond  = (treeNode*) NULL;
    then_n = (treeNode*) NULL;
    else_n = (treeNode*) NULL;
}

/*  ###############################################################  */
forNode::forNode(tnType t)
        :baseNode()
{
    gProject->Parse_TOS->node_heap->insert(this);

    which = FOR_T;
    type = t;
    init = (treeNode *) NULL;
    test = (treeNode *) NULL;
    incr = (treeNode *) NULL;
    stemnt = (treeNode *) NULL;
}

/*  ###############################################################  */
leafNode::leafNode(tnType t)
         :baseNode()
{
    gProject->Parse_TOS->node_heap->insert(this);

    which = LEAF_T;
    type = t;
    syment = NULL;
}

/*  ###############################################################  */
treeNode *copy_tree(treeNode *root)
{
    leafNode *leaf, *newleaf;
    ifNode  *ifn, *newif;
    forNode *forn, *newfor;
    treeNode *newnode;

    if (!root)
        return(NULL);

    switch (root->which)
    {
    case LEAF_T:
        leaf = (leafNode *) root;
        newleaf = new leafNode(leaf->type);
        newleaf->line = leaf->line;
        newleaf->col = leaf->col;
        newleaf->tok = leaf->tok;
        newleaf->data = leaf->data;
        return((treeNode *)newleaf);

    case NODE_T:
        newnode = new treeNode(root->type);
        newnode->line = root->line;
        newnode->col = root->col;
        newnode->tok = root->tok;
        newnode->lnode = copy_tree(root->lnode);
        newnode->rnode = copy_tree(root->rnode);
        return(newnode);

    case IF_T:
        ifn = (ifNode *) root;
        newif = new ifNode(ifn->type);
        newif->line = ifn->line;
        newif->col = ifn->col;
        newif->tok = ifn->tok;
        newif->cond = copy_tree(ifn->cond);
        newif->then_n = copy_tree(ifn->then_n);
        newif->then_n = copy_tree(ifn->else_n);
        return((treeNode *)newif);

    case FOR_T:
        forn = (forNode *) root;
        newfor = new forNode(forn->type);
        newfor->line = forn->line;
        newfor->col = forn->col;
        newfor->tok = forn->tok;
        newfor->init = copy_tree(forn->init);
        newfor->test = copy_tree(forn->test);
        newfor->incr = copy_tree(forn->incr);
        newfor->stemnt = copy_tree(forn->stemnt);
        return((treeNode *)newfor);
        break;

    default:
    case NONE_T:
        newnode = new treeNode(TN_EMPTY);
        newnode->which = NONE_T;
        return (newnode);
    }
}

/*  ###############################################################  */
void free_tree(treeNode *root)
{
    leafNode *leaf;
    ifNode  *ifn;
    forNode *forn;

    if (!root)
        return;

    switch (root->which)
    {
    case LEAF_T:
        leaf = (leafNode *) root;
        if (leaf->tok == STRING)
            delete [] leaf->data.str;
        break;

    case NODE_T:
        free_tree(root->lnode);
        free_tree(root->rnode);
        break;

    case IF_T:
        ifn = (ifNode *) root;
        free_tree(ifn->cond);
        free_tree(ifn->then_n);
        free_tree(ifn->else_n);
        break;

    case FOR_T:
        forn = (forNode *) root;
        free_tree(forn->init);
        free_tree(forn->test);
        free_tree(forn->incr);
        free_tree(forn->stemnt);
        break;

    default:
    case NONE_T:
        break;
    }

    gProject->Parse_TOS->node_heap->free(root);
}

/*  ###############################################################  */
leafNode *leftmost(treeNode *root)
{
    ifNode *ifn;
    forNode *forn;

    switch (root->which)
    {
    case LEAF_T:
        return((leafNode *) root);
        break;

    case NODE_T:
        if (root->lnode)
            return(leftmost(root->lnode));
        else if (root->rnode)
            return(leftmost(root->rnode));
        cerr << "Tree node " << print_ptr(root)
             << " with no children reached in leftmost.\n";
        break;

    case IF_T:
        ifn = (ifNode *) root;
        if (ifn->cond)
            return(leftmost(ifn->cond));
        else if (ifn->then_n)
            return(leftmost(ifn->then_n));
        else if (ifn->else_n)
            return(leftmost(ifn->else_n));
        cerr << "If node " << print_ptr(root)
             << " with no children reached in leftmost.\n";
        break;

    case FOR_T:
        forn = (forNode *) root;
        if (forn->init)
            return(leftmost(forn->init));
        else if (forn->test)
            return(leftmost(forn->test));
        else if (forn->incr)
            return(leftmost(forn->incr));
        else if (forn->stemnt)
            return(leftmost(forn->stemnt));
        cerr << "For node " << print_ptr(root)
             << " with no children reached in leftmost.\n";
        break;

    case NONE_T:
    default:
        cerr << "Internal error: unknown node "
             << print_ptr(root) << " in leftmost;\n"
             << "  type is: " << name_of_nodetype(root->which) << endl;
        break;
    }

    return((leafNode *) NULL);
}

/*  ###############################################################  */
leafNode *rightmost(treeNode *root)
{
    ifNode *ifn;
    forNode *forn;

    switch (root->which)
    {
    case LEAF_T:
        return((leafNode *) root);
        break;

    case NODE_T:
        if (root->rnode)
            return(rightmost(root->rnode));
        else if (root->lnode)
            return(rightmost(root->lnode));
        cerr << "\nTree node " << print_ptr(root)
             << " with no children reached in rightmost.\n";
        break;

    case IF_T:
        ifn = (ifNode *) root;
        if (ifn->else_n)
            return(rightmost(ifn->else_n));
        else if (ifn->then_n)
            return(rightmost(ifn->then_n));
        else if (ifn->cond)
            return(rightmost(ifn->cond));
        cerr << "\nIf node " << print_ptr(root)
             << " with no children reached in rightmost.\n";
        break;

    case FOR_T:
        forn = (forNode *) root;
        if (forn->stemnt)
            return(rightmost(forn->stemnt));
        else if (forn->incr)
            return(rightmost(forn->incr));
        else if (forn->test)
            return(rightmost(forn->test));
        else if (forn->init)
            return(rightmost(forn->init));
        cerr << "\nFor node " << print_ptr(root)
             << " with no children reached in rightmost.\n";
        break;

    case NONE_T:
    default:
        cerr << "Internal error: unknown node "
             << print_ptr(root) << " in rightmost;\n"
             << "  type is: " << name_of_nodetype(root->which) << endl;
        break;
    }

    return((leafNode *) NULL);
}

/*  ###############################################################  */
/* Scans the typedef declaration node N for the ident
   naming the new type. --jsh */
void
find_typedef_name(treeNode *n, treeNode *def, FindFunction find)
{
    if (!n)
        return;

    switch(n->which)
    {
        case LEAF_T:
            (find)((leafNode*) n, def);
            break;

        case NODE_T:
        {
            switch(n->type)
            {
            case TN_DECL:
                find_typedef_name(n->rnode,def,find);
                break;

            case TN_ARRAY_DECL:
                find_typedef_name(n->lnode,def,find);
                break;

            case TN_PNTR:
                find_typedef_name(n->lnode,def,find);
                break;

            case TN_DECL_LIST:
            case TN_DECLS:
                find_typedef_name(n->lnode,def,find);
                find_typedef_name(n->rnode,def,find);
                break;

            case TN_FUNC_DECL:
                find_typedef_name(n->lnode,def,find);
                break;

            default:
                cerr << "Internal error: "
                     << "unknown node type (" << name_of_node(n->type)
                     << ") in find_typedef_name\n";
                break;
            }
        }
        break;

        case IF_T:
        case FOR_T:
        case NONE_T:
        default:
            cerr << "Internal error: unknown node "
                 << print_ptr(n) << " in find_typedef_name;\n"
                 << "  type is: " << name_of_nodetype(n->which) << endl;
            break;
    }
}

/*  ###############################################################  */
void  find_ident_name(treeNode *n, treeNode *def, FindFunction find)
{
    if (!n)
        return;

    switch(n->which)
    {
    case LEAF_T:
        (find)((leafNode*) n,def);
        break;

    case NODE_T:
    {
        switch(n->type)
        {
        case TN_IDENT:
            (find)((leafNode*) n,def);
            break;

        case TN_DECL:
            find_ident_name(n->rnode,def,find);
            break;

        case TN_COMP_DECL:
            find_ident_name(n->rnode,def,find);
            break;

        case TN_ASSIGN:
        case TN_ARRAY_DECL:
        case TN_BIT_FIELD:
            find_ident_name(n->lnode,def,find);
            break;

        case TN_FIELD_LIST:
        case TN_DECL_LIST:
        case TN_DECLS:
        case TN_PARAM_LIST:
        case TN_IDENT_LIST:
        case TN_PNTR:
            find_ident_name(n->lnode,def,find);
            find_ident_name(n->rnode,def,find);
            break;

        case TN_ELLIPSIS:
            break;

        case TN_FUNC_DECL:
            find_ident_name(n->lnode,def,find);
            break;

        default:
            cerr << "Internal error: "
                 << "unknown node type (" << name_of_node(n->type)
                 << ") in find_ident_name\n";
            break;
        }
    }
    break;

    case IF_T:
    case FOR_T:
    case NONE_T:
    default:
        cerr << "Internal error: unknown node "
             << print_ptr(n) << " in find_ident_name;\n"
             << "  type is: " << name_of_nodetype(n->which) << endl;
        break;
    }
}

/*  ###############################################################  */
/* Scans the function declaration node N for the ident
   naming the new function. */
leafNode *find_func_name(treeNode *n)
{
  forNode *f;

  while((n != 0) && (n->which != LEAF_T))
  {
    switch(n->which)
    {
    case LEAF_T:
      return (leafNode *) n;

    case NODE_T:
      switch(n->type)
        {
        case TN_DECL:
          n = n->rnode;
          break;

        case TN_PNTR:
          n = n->lnode;
          break;

        case TN_FUNC_DECL:
          n = n->lnode;
          break;

        default:
          cerr << "Internal error: "
               << "unknown node type (" << name_of_node(n->type)
               << ") in find_func_name\n";
          n = NULL;
          break;
        }
      break;

    case FOR_T:
        f = (forNode *) n;
        switch(n->type)
          {
          case TN_FUNC_DEF:
            n = f->test;
            break;

          default:
            break;
          }
        break;

    case IF_T:
    case NONE_T:
    default:
        cerr << "Internal error: unknown node "
             << print_ptr(n) << " in find_func_name;\n"
             << "  type is: " << name_of_nodetype(n->which) << endl;
        n = NULL;
        break;
    }
  }

  return (leafNode*) n;
}

/*  ###############################################################  */
void  find_params(treeNode *decl, FindFunction find)
{
    if (!decl)
        return;

    switch(decl->which)
    {
    case NODE_T:
    {
        switch(decl->type)
        {
        case TN_DECL:
            find_ident_name(decl,decl,find);
            break;

        case TN_DECL_LIST:
        case TN_DECLS:
            find_params(decl->lnode,find);
            find_params(decl->rnode,find);
            break;

        case TN_IDENT_LIST:
        case TN_PARAM_LIST:
            find_ident_name(decl->lnode,decl->lnode,find);
            find_ident_name(decl->rnode,decl->rnode,find);
            break;

        case TN_FUNC_DECL:
            find_params(decl->rnode,find);
            break;

        case TN_PNTR:
            find_params(decl->lnode,find);
            find_params(decl->rnode,find);
            break;

        default:
            cerr << "Internal error: "
                 << "unknown node type (" << name_of_node(decl->type)
                 << ") in find_params\n";
            break;
        }
    }
    break;

    case LEAF_T:
        cerr << "Internal error: "
             << "unknown node type (" << name_of_node(decl->type)
             << ") in find_params\n";
    break;

    case IF_T:
    case FOR_T:
    case NONE_T:
    default:
        cerr << "Internal error: unknown node "
             << print_ptr(decl) << " in find_params;\n"
             << "  type is: " << name_of_nodetype(decl->which) << endl;
        break;
    }
}

/*  ###############################################################  */
void  find_components(treeNode *decl, treeNode *def, FindFunction find)
{
    if (!decl)
        return;

    switch(decl->which)
    {
        case NODE_T:
        {
            switch(decl->type)
            {
            case TN_COMP_DECL:
            case TN_DECL:
                find_ident_name(decl,decl,find);
                break;

            case TN_DECL_LIST:
            case TN_DECLS:
                find_components(decl->lnode,def,find);
                find_components(decl->rnode,def,find);
                break;

            case TN_FIELD_LIST:
                find_components(decl->lnode,def,find);
                find_components(decl->rnode,def,find);
                break;

            case TN_FUNC_DECL:
                find_components(decl->rnode,def,find);
                break;

            default:
                cerr << "Internal error: "
                     << "unknown node type (" << name_of_node(decl->type)
                     << ") in find_components\n";
                break;
            }
        }
    break;

    case LEAF_T:
    case IF_T:
    case FOR_T:
    case NONE_T:
    default:
        cerr << "Internal error: unknown node "
             << print_ptr(decl) << " in find_components;\n"
             << "  type is: " << name_of_nodetype(decl->which) << endl;
        break;
    }
}

/*  ###############################################################  */
#define    SHOW(X)    case X: return #X

char*
name_of_node(tnType val)
{
    switch (val)
    {
        SHOW(TN_EMPTY);

        SHOW(TN_FUNC_DEF);
        SHOW(TN_FUNC_DECL);
        SHOW(TN_FUNC_CALL);
        SHOW(TN_BLOCK);
        SHOW(TN_DECL);
        SHOW(TN_ARRAY_DECL);

        SHOW(TN_TRANS_LIST);
        SHOW(TN_DECL_LIST);
        SHOW(TN_DECLS);
        SHOW(TN_STEMNT_LIST);
        SHOW(TN_EXPR_LIST);
        SHOW(TN_ENUM_LIST);
        SHOW(TN_FIELD_LIST);
        SHOW(TN_PARAM_LIST);
        SHOW(TN_IDENT_LIST);

        SHOW(TN_COMP_DECL);
        SHOW(TN_BIT_FIELD);
        SHOW(TN_PNTR);

        SHOW(TN_TYPE_LIST);
        SHOW(TN_TYPE_NME);

        SHOW(TN_INIT_LIST);
        SHOW(TN_INIT_BLK);

        SHOW(TN_OBJ_DEF);
        SHOW(TN_OBJ_REF);

        SHOW(TN_CAST);
        SHOW(TN_IF);
        SHOW(TN_ASSIGN);
        SHOW(TN_JUMP);
        SHOW(TN_FOR);
        SHOW(TN_WHILE);
        SHOW(TN_DOWHILE);
        SHOW(TN_SWITCH);
        SHOW(TN_LABEL);
        SHOW(TN_STEMNT);

        SHOW(TN_INDEX);
        SHOW(TN_DEREF);
        SHOW(TN_SELECT);

        SHOW(TN_EXPR);
        SHOW(TN_COND_EXPR);

        SHOW(TN_ELLIPSIS);
        SHOW(TN_IDENT);
        SHOW(TN_TYPE);
        SHOW(TN_STRING);
        SHOW(TN_INT);
        SHOW(TN_REAL);

        SHOW(TN_LINE);
        SHOW(TN_INCL);

        default:
           return "<Unknown Node Name>";
    }
}

char*
name_of_nodetype(nodeType val)
{
    switch (val)
    {
        SHOW(NONE_T);
        SHOW(LEAF_T);
        SHOW(IF_T);
        SHOW(FOR_T);
        SHOW(NODE_T);

        default:
            return "<Unknown Node Type>";
    }
}

#undef SHOW

/*  ###############################################################  */
