
/*  ###############################################################
    ##
    ##     C Tree Builder
    ##
    ##     File:         tree.h
    ##
    ##     Programmer:   Shawn Flisakowski
    ##     Date:         Jan 21, 1995
    ##
    ##
    ###############################################################  */

#ifndef   TREE_H
#define   TREE_H

#include  "config.h"
#include  "nametbl.h"
#include  "heap.h"

class SymEntry;

/*  ###############################################################  */

typedef enum tnType
{
    TN_EMPTY=0,

    TN_FUNC_DEF,
    TN_FUNC_DECL,
    TN_FUNC_CALL,
    TN_BLOCK,
    TN_DECL,
    TN_ARRAY_DECL,

    TN_TRANS_LIST,
    TN_DECL_LIST,
    TN_STEMNT_LIST,
    TN_EXPR_LIST,
    TN_ENUM_LIST,
    TN_FIELD_LIST,
    TN_PARAM_LIST,
    TN_IDENT_LIST,

    TN_DECLS,

        /* Dumb, but I'm not sure what to do yet */
    TN_COMP_DECL,
    TN_BIT_FIELD,
    TN_PNTR,

        /* Stores a list of type specifiers/qualifers (int,const,etc) */
    TN_TYPE_LIST,
    TN_TYPE_NME,

        /* Stores initial values for arrays */
    TN_INIT_LIST,
    TN_INIT_BLK,

    TN_OBJ_DEF,    /* Definition of struct, union, or enum */
    TN_OBJ_REF,    /* Reference to struct, union, or enum */

        /* More vague */
    TN_CAST,
    TN_IF,
    TN_ASSIGN,
    TN_JUMP,
    TN_FOR,
    TN_WHILE,
    TN_DOWHILE,
    TN_SWITCH,
    TN_LABEL,
    TN_STEMNT,

    TN_INDEX,     /* Index with [] */
    TN_ADDROF,    /* Address of via & */
    TN_DEREF,     /* Dereference with * */
    TN_SIZEOF,    /* sizeof */
    TN_SELECT,    /* -> and . */

    TN_EXPR,
    TN_COND_EXPR,

    TN_ELLIPSIS,
    TN_IDENT,
    TN_TYPE,
    TN_STRING,
    TN_INT,
    TN_REAL,

    TN_LINE,       /* Preprocessor #line directive. */
    TN_INCL        /* An included file. */

} tnType;


typedef enum nodeType
{
    NONE_T,
    LEAF_T,
    IF_T,
    FOR_T,
    NODE_T

} nodeType;

/*  ###############################################################  */

class baseNode : public HeapItem
{
  public:
    baseNode();

    nodeType  which;
    char*     filename;
    int       line;
    int       col;
    int       tok;
    tnType    type;
};

class treeNode : public baseNode
{
  public:
    treeNode( tnType t );

    treeNode *lnode;
    treeNode *rnode;
};

/*  ###############################################################  */

class ifNode : public baseNode
{
  public:
    ifNode( tnType t );

    treeNode *cond;
    treeNode *then_n;
    treeNode *else_n;
};

/*  ###############################################################  */

class forNode : public baseNode
{
  public:
    forNode( tnType t );

    treeNode *init;
    treeNode *test;
    treeNode *incr;
    treeNode *stemnt;
};

#define FORNODE_SZE        (sizeof(forNode))

/*  ###############################################################  */

struct type_item;

class leafNode : public baseNode
{
  public:
    leafNode( tnType t);

    /* The node in the symbol table. */
    SymEntry           *syment;

    union {
      int               cval;
      StrItem          *sval;
      char             *str;
      int               ival;
      long int          l_ival;
      unsigned int      u_ival;
      unsigned long int ul_ival; 
      float             fval;
      double            dval;
      long double       l_dval;
    } data;

};

#define LEAFNODE_SZE        (sizeof(leafNode))

/*  ###############################################################  */

    /*  External Declarations */

typedef  void (*FindFunction)(leafNode*, treeNode*);

extern char *name_of_node(tnType);
extern char *name_of_nodetype    ARGS((nodeType));

void      free_tree       ARGS((treeNode*));

treeNode *copy_tree       ARGS((treeNode*));

leafNode *leftmost        ARGS((treeNode*));
leafNode *rightmost       ARGS((treeNode*));

void find_typedef_name    ARGS((treeNode*,treeNode*, FindFunction));
void find_ident_name      ARGS((treeNode*,treeNode*, FindFunction));
leafNode *find_func_name  ARGS((treeNode*));
void find_params          ARGS((treeNode*, FindFunction));
void find_components      ARGS((treeNode*,treeNode*,FindFunction));

/*  ###############################################################  */

#endif    /* TREE_H */
