(*
Copyright (c) 1997,1998 The Regents of the University of California.
All rights reserved.

Permission to use, copy, modify, and distribute this software for any
purpose, without fee, and without written agreement is hereby granted,
provided that the above copyright notice and the following two
paragraphs appear in all copies of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*)


    (* Copyright (C) 1989,1990 James A. Roskind, All rights reserved. 
    This lexer description was written by James A.  Roskind.  Copying 
    of  this  file, as a whole, is permitted providing this notice is 
    intact  and  applicable   in   all   complete   copies.    Direct 
    translations  as a whole to other lexer generator input languages 
    (or lexical description languages)  is  permitted  provided  that 
    this  notice  is  intact and applicable in all such copies, along 
    with a disclaimer that  the  contents  are  a  translation.   The 
    reproduction  of derived files or text, such as modified versions 
    of this file, or the output of scanner generators, is  permitted, 
    provided   the  resulting  work  includes  the  copyright  notice 
    "Portions Copyright (c) 1989, 1990 James  A.   Roskind".  Derived 
    products  must  also  provide  the notice "Portions Copyright (c) 
    1989, 1990 James A.  Roskind" in  a  manner  appropriate  to  the 
    utility,   and  in  keeping  with  copyright  law  (e.g.:  EITHER 
    displayed when first invoked/executed; OR displayed  continuously 
    on  display terminal; OR via placement in the object code in form 
    readable in a printout, with or near the title of the work, or at 
    the end of the file).  No royalties, licenses or  commissions  of 
    any  kind  are  required  to copy this file, its translations, or 
    derivative products, when the copies are made in compliance  with 
    this  notice.  Persons  or  corporations  that  do make copies in 
    compliance  with  this  notice  may  charge  whatever  price   is 
    agreeable  to  a buyer, for such copies or derivative works. THIS 
    FILE IS PROVIDED ``AS IS'' AND WITHOUT  ANY  EXPRESS  OR  IMPLIED 
    WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES 
    OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.Lexeme.

    James A. Roskind
    Independent Consultant
    516 Latania Palm Drive
    Indialantic FL, 32903
    (407)729-4348
    jar@ileaf.com
    or ...!uunet!leafusa!jar

    ---end of copyright notice--- 


    Adapted from Guhan Viswanathan's (gviswana@cs.wisc.edu) Css ("C**")
    lexer. See ~/Guhan/css.lex

    The changes here are basically to remove any parallel features of C**,
    to make things known as "gen" rather than "css," and to rely on my own
    datatypes and modules. 

*)

structure N = Atom
structure SC = StringCvt
structure Tokens = Tokens
type pos = int
type svalue = Tokens.svalue
type ('a, 'b) token = ('a, 'b) Tokens.token
type lexresult = (svalue, pos) token
type lexarg = {sourceMap : SourceMap.sourcemap
              }
type arg = lexarg

structure SymbolTable = SymbolTable
structure AST = SymbolTable.AST

open Tokens

val eof = fn {sourceMap} => 
             let val pos = SourceMap.lastChange sourceMap
             in
                EOF(pos, pos)
             end

(* make old string conversions compatible with standard basis *)
exception Convert

fun atoi s = (case (SC.scanString (Int.scan SC.DEC) s) of 
	       SOME(i) => i
	     | NONE => raise Convert)

fun oatoi s = (case (SC.scanString (Int.scan SC.OCT) s) of 
	       SOME(i) => i
	     | NONE => raise Convert)

fun xatoi s = (case (SC.scanString (Int.scan SC.HEX) s) of 
	       SOME(i) => i
	     | NONE => raise Convert)

(* Take string of form # <line-no> <file-name> <flags> or
                       #line <line-no> <file-name> <flags>
   and resynchronize current position in sourcemap *)
fun resynch s sourceMap = 
    let fun delim c = (c = #" " orelse
		       c = #"\n" orelse
		       c = #"\012")
	val tokens = String.tokens delim s
	val _ = if length tokens < 3 then
	            raise Fail ("Unknown # directive " ^ s)
		else ()
	val (_::lineStr::fileStr::_) = tokens
	val line = case Int.fromString lineStr of
	    SOME i => i
	  | NONE => raise Fail ("Bad line number " ^ lineStr)
	val fileName = String.substring (fileStr, 1, size fileStr - 1)
    in
(*	print (concat ["Resynching to file ",
		       fileStr,
		       " line ",
		       Int.toString line,
		       "\n"]);*)
	SourceMap.resynch sourceMap (2, (* lexer initial position *)
				     {fileName = SOME fileName,
				      line = line,
				      column = NONE})
    end
	       

fun lookupId(id,pos) = 
      (case (KeywordTable.lookup (N.atom id)) of
	 SOME(kw) => kw(pos,pos + size id)
       | NONE =>
	   let
	     val id' = Lexeme.mkIdentifier id
	   in
	       if SymbolTable.idIsTypedefName id' then
		   TYPEDEF_NAME(id', pos, pos + size id)
	       else
		   IDENTIFIER(id', pos, pos + size id)
	   end)

(*
add these lines before the # rule below to get resynchronization
<INITIAL>"#"{white}+{decimal}{white}+{string}.*\n
                        =>      (resynch yytext sourceMap;
				 continue());
<INITIAL>"#line"{white}+{decimal}{white}+{string}.*\n
                        =>      (resynch yytext sourceMap;
				 continue());
*)

%%

%s C;
%header (functor GenLexFun(structure Tokens : Gen_TOKENS and
			       KeywordTable : KEYWORD_TABLE and
			       SymbolTable : SymbolTable
			   sharing Tokens = KeywordTable.Tokens));
%arg ({sourceMap});

identifier = [a-zA-Z_][0-9a-zA-Z_]* ;
qualifier  = \$[a-zA-Z_][0-9a-zA-Z_]* ;

digit = [0-9];
octDigit = [0-7];
hexDigit = [0-9a-fA-F];

exponent = [eE][-+]? {digit}+ ;
fraction = ({digit}* "." {digit}+)|({digit}+ ".") ;
floating = (({fraction} {exponent}?) | ({digit}+ {exponent})) [fF]? [lL]? ;


intSuffix = ([uU]? [lL]? [lL]?) | ([lL][uU]);
decimal	 = [1-9] {digit}* {intSuffix};
octal	 = 0 {octDigit}* {intSuffix};
hex	 = 0 [xX] {hexDigit}+ {intSuffix};

simpleEsc = [abfnrtv\\?'"] ;
octalEsc  = {octDigit}{octDigit}? {octDigit}?;
hexEsc	 = "x" {hexDigit}+ ;

escape	 = \\ ({simpleEsc} | {octalEsc} | {hexEsc});

string	 = \" ([^"\\\n] | {escape})* \" ;

wchar	 = "L"  ' ([^'\\\n] | {escape})+ ';
char	 = ' ([^'\\\n] | {escape})+ ';

comment  = "/*" ([^*] | \*+ [^*/])* \*+\/;
newline	 = \n;
white	 = [\ \t\012\010\004];

%%

<INITIAL>{newline}	=>	(SourceMap.newline sourceMap yypos; continue());
<INITIAL>{white}+	=>	(continue());

<INITIAL>{identifier}	=>	(lookupId(yytext,yypos));

<INITIAL>{qualifier}	=>	(QUALIFIER(Lexeme.mkIdentifier yytext,
					   yypos, yypos+size yytext));

<INITIAL>{char}		=>	(CONSTANT(Lexeme.Character(Lexeme.mkIdentifier yytext),
                                 yypos, yypos+size yytext));
<INITIAL>{wchar}	=>	(CONSTANT(Lexeme.Character(Lexeme.mkIdentifier yytext),
                                 yypos, yypos+size yytext));
<INITIAL>{decimal}	=>	(CONSTANT((Lexeme.Integer(atoi(yytext))
				   handle Overflow => 
				   Lexeme.BigInteger(Lexeme.mkIdentifier yytext)
					| Convert => Lexeme.BigInteger(Lexeme.mkIdentifier
								   yytext)),
				  yypos, yypos+size yytext));
<INITIAL>{octal}	=>	(CONSTANT((Lexeme.Integer(oatoi(yytext))
				   handle Overflow => 
				   Lexeme.BigInteger(Lexeme.mkIdentifier yytext)
					| Convert => Lexeme.BigInteger(Lexeme.mkIdentifier
								   yytext)),
				  yypos, yypos+size yytext));
<INITIAL>{hex}		=>	(CONSTANT((Lexeme.Integer(xatoi(yytext))
				   handle Overflow => 
				   Lexeme.BigInteger(Lexeme.mkIdentifier yytext)
					| Convert => Lexeme.BigInteger(Lexeme.mkIdentifier
								   yytext)),
				  yypos, yypos+size yytext));
				  
<INITIAL>{floating}	=>	(CONSTANT(Lexeme.Real(Lexeme.mkIdentifier yytext), 
                                 yypos, yypos+size yytext));
<INITIAL>{string}	=>	(STRINGCONSTANT(yytext,	yypos, yypos+size yytext));

<INITIAL>","		=>	(COMMA(yypos, yypos+1));
<INITIAL>"#".*\n	=>	(SourceMap.newline sourceMap (yypos+size yytext-1);
				 continue());
<INITIAL>"//".*\n	=>	(SourceMap.newline sourceMap (yypos+size yytext-1);
				 continue());


<INITIAL>"("		=>	(LP(yypos, yypos+1));
<INITIAL>")"		=>	(RP(yypos, yypos+1));
<INITIAL>"{"		=>	(LC(yypos, yypos+1));
<INITIAL>"}"		=>	(RC(yypos, yypos+1));
<INITIAL>"["		=>	(LB(yypos, yypos+1));
<INITIAL>"]"		=>	(RB(yypos, yypos+1));

<INITIAL>"+"		=>	(PLUS(yypos, yypos+1));
<INITIAL>"-"		=>	(MINUS(yypos, yypos+1));
<INITIAL>"!"		=>	(NOT(yypos, yypos+1));
<INITIAL>"*"		=>	(MUL(yypos, yypos+1));
<INITIAL>"/"		=>	(DIV(yypos, yypos+1));
<INITIAL>"%"		=>	(MOD(yypos, yypos+1));

<INITIAL>"&"		=>	(BIT_AND(yypos, yypos+1));
<INITIAL>"|"		=>	(BIT_OR(yypos, yypos+1));
<INITIAL>"~"		=>	(BIT_NOT(yypos, yypos+1));
<INITIAL>"^"		=>	(BIT_XOR(yypos, yypos+1));
<INITIAL>"<<"		=>	(LSHIFT(yypos, yypos+2));
<INITIAL>">>"		=>	(RSHIFT(yypos, yypos+2));

<INITIAL>"?"		=>	(QUESTION(yypos, yypos+1));
<INITIAL>":"		=>	(COLON(yypos, yypos+1));
<INITIAL>";"		=>	(SEMICOLON(yypos, yypos+1));


<INITIAL>"."		=>	(DOT(yypos, yypos+1));
<INITIAL>"->"		=>	(PTR(yypos, yypos+2));

<INITIAL>"++"		=>	(INCR(yypos, yypos+2));
<INITIAL>"--"		=>	(DECR(yypos, yypos+2));

<INITIAL>"<"		=>	(LT(yypos, yypos+1));
<INITIAL>">"		=>	(GT(yypos, yypos+1));
<INITIAL>"<="		=>	(LE(yypos, yypos+2));
<INITIAL>">="		=>	(GE(yypos, yypos+2));
<INITIAL>"=="		=>	(EQ(yypos, yypos+2));
<INITIAL>"!="		=>	(NE(yypos, yypos+2));

<INITIAL>"&&"		=>	(AND(yypos, yypos+2));
<INITIAL>"||"		=>	(OR(yypos, yypos+2));

<INITIAL>"="		=>	(ASSIGN(NONE, yypos, yypos+1));
<INITIAL>"+="		=>	(ASSIGN(SOME(Lexeme.Plus), yypos, yypos+2));
<INITIAL>"-="		=>	(ASSIGN(SOME(Lexeme.Minus), yypos, yypos+2));
<INITIAL>"*="		=>	(ASSIGN(SOME(Lexeme.Mul), yypos, yypos+2));
<INITIAL>"/="		=>	(ASSIGN(SOME(Lexeme.Div), yypos, yypos+2));
<INITIAL>"%="		=>	(ASSIGN(SOME(Lexeme.Mod), yypos, yypos+2));
<INITIAL>"<<="		=>	(ASSIGN(SOME(Lexeme.Lshift), yypos, yypos+3));
<INITIAL>">>="		=>	(ASSIGN(SOME(Lexeme.Rshift), yypos, yypos+3));
<INITIAL>"&="		=>	(ASSIGN(SOME(Lexeme.BitAnd), yypos, yypos+2));
<INITIAL>"^="		=>	(ASSIGN(SOME(Lexeme.BitXor), yypos, yypos+2));
<INITIAL>"|="		=>	(ASSIGN(SOME(Lexeme.BitOr), yypos, yypos+2));
<INITIAL>"<-="		=>	(ASSIGN(SOME(Lexeme.Flow), yypos, yypos+3));

<INITIAL>"..."		=>	(ELLIPSIS(yypos, yypos+3));

<INITIAL>.		=>	(TextIO.output(TextIO.stdErr, "Error: unknown character " ^ yytext ^ "\n"); continue());

<INITIAL>"/*"	=>	(YYBEGIN C; continue ());

<C>"*/"         =>      (YYBEGIN INITIAL; continue ());
<C>\n		=>      (SourceMap.newline sourceMap yypos; continue());
<C>.            =>      (continue());
