(*
Copyright (c) 1997,1998 The Regents of the University of California.
All rights reserved.

Permission to use, copy, modify, and distribute this software for any
purpose, without fee, and without written agreement is hereby granted,
provided that the above copyright notice and the following two
paragraphs appear in all copies of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*)

(*

keyword table for C, 
based on one by Guhan Viswanathan (gviswana@cs.wisc.edu)

*)

functor KeywordTable (structure Tokens : Gen_TOKENS) : KEYWORD_TABLE =
  struct

    structure N = Atom
    structure NT = AtomTable
    structure Tokens = Tokens
    type terminal = (int * int) -> (Tokens.svalue, int) Tokens.token

    exception NOTFOUND (* dummy exception *)

    val table = NT.mkTable (32, NOTFOUND) : terminal NT.hash_table
	
    local 
      fun insert (a, b) = NT.insert table (N.atom a, b)
      open Tokens
    in
      val _ = 
	(insert("__inline",INLINE);		   (* gcc special *)
	 insert("__inline__",INLINE);		   (* gcc special *)
	 insert("inline",INLINE);                  (* gcc special *)
	 insert("__attribute__",ATTRIBUTE);	   (* gcc special *)
	 insert("__extension__",EXTENSION);	   (* gcc special (used for varargs) *)
	 insert("typeof",TYPEOF);
	 insert("__typeof__", TYPEOF);
	 insert("__typeof", TYPEOF);
	 insert("__const",CONST);                (* gcc special *)
	 insert("__const__",CONST);              (* gcc special *)
	 insert("__FUNCTION__",FUNCTION);      (* gcc special -- current function *)
	 insert("auto", AUTO);
	 insert("break", BREAK);
	 insert("case", CASE);
	 insert("char", CHAR);
	 insert("const", CONST);
	 insert("continue", CONTINUE);
	 insert("default", DEFAULT);
	 insert("do", DO);
	 insert("double", DOUBLE);
	 insert("else", ELSE);
	 insert("enum", ENUM);
	 insert("extern", EXTERN);
	 insert("float", FLOAT);
	 insert("for", FOR);
	 insert("goto", GOTO);
	 insert("if", IF);
	 insert("int", INT);
	 insert("long", LONG);
	 insert("nonconst", NONCONST);
	 insert("register", REGISTER);
	 insert("return", RETURN);
	 insert("short", SHORT);
	 insert("signed", SIGNED);
	 insert("__signed__", SIGNED);  (* gcc special *)
	 insert("sizeof", SIZEOF);
	 insert("static", STATIC);
	 insert("struct", STRUCT);	 
	 insert("switch", SWITCH);
	 insert("typedef", TYPEDEF);
	 insert("union", UNION);
	 insert("unsigned", UNSIGNED);
	 insert("void", VOID);
	 insert("volatile", VOLATILE);
	 insert("while", WHILE);
	 insert("__alignof__", ALIGNOF);
	 insert("__label__", LABEL)
	)
    end

    fun lookup id = NT.find table id
	
  end;
