(*
Copyright (c) 1997,1998 The Regents of the University of California.
All rights reserved.

Permission to use, copy, modify, and distribute this software for any
purpose, without fee, and without written agreement is hereby granted,
provided that the above copyright notice and the following two
paragraphs appear in all copies of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*)

structure Lexeme : Lexeme =
    struct

	datatype identifier = Id of Atom.atom
	fun mkIdentifier s = Id(Atom.atom s)
	structure IdKey =
	    struct
		type ord_key = identifier
		fun compare (Id(id), Id(id')) = Atom.compare (id, id')
	    end
	val nullId = Id(Atom.atom "")
	fun idKey x = x
	fun keyId x = x
	val freshIdRef = ref 0
	fun mkFreshId () = 
            let val i = !freshIdRef
		val id = Atom.atom ("T" ^ (Int.toString i))
		val _ = (freshIdRef := i + 1)
	    in
		Id(id)
	    end
	fun mkAggrId (Id(nm)) = 
	    Id(Atom.atom ("$AGGREGATE#" ^ (Atom.toString nm)))
	fun idToString (Id(nm)) = Atom.toString nm

	fun reset () = (freshIdRef := 0)

	structure IdSet = BinarySetFn (IdKey)

	datatype constant =
	    Integer of int
	  | BigInteger of identifier
	  | String of identifier
	  | Character of identifier
	  | Real of identifier
	    
	datatype operator =
	    Plus
	  | Minus
	  | Mul
	  | Div
	  | Mod
	  | Incr
	  | Decr
	  | And
	  | Or
	  | Not
	  | BitAnd
	  | BitOr
	  | BitNot
	  | BitXor
	  | Rshift
	  | Lshift
	  | Lt
	  | Le
	  | Gt
	  | Ge
	  | Eq
	  | Ne
	  | Address
	  | LabelAddress
	  | Deref
	  | UnaryMinus
	  | Flow

    end
