(*
Copyright (c) 1997,1998 The Regents of the University of California.
All rights reserved.

Permission to use, copy, modify, and distribute this software for any
purpose, without fee, and without written agreement is hereby granted,
provided that the above copyright notice and the following two
paragraphs appear in all copies of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*)

functor GenFront(ExtraInfo : ExtraInfo) = 
  struct

      structure AST = ASTFn(structure ExtraInfo = ExtraInfo)
      structure SymbolTable = SymbolTableFn(structure AST = AST)
      structure Unparse = UnparseFn(structure AST = AST)

      (* Create the Lexer and Parser *)
      structure GenLrVals = GenLrValsFun(structure Token = LrParser.Token
					 structure SymbolTable = SymbolTable)
    
      (* Keyword table *)
      structure KeywordTable = KeywordTable(structure Tokens =
						GenLrVals.Tokens)
    
      structure GenLex = GenLexFun(structure Tokens = GenLrVals.Tokens
				   structure KeywordTable = KeywordTable
				   structure SymbolTable = SymbolTable)

      structure GenParser = JoinWithArg(structure ParserData =
					    GenLrVals.ParserData
					structure Lex = GenLex
					structure LrParser = LrParser)

      fun parse (dev,filename) =
	  let
	      val _ = SymbolTable.reset ()
(*	      val _ = Lexeme.reset () -- breaks multiple files *)
	      val sourceMap = SourceMap.newmap (2 (* lexer initial position *),
						{fileName=filename,
						 line=1,
						 column=1})
	      val lexarg = {sourceMap=sourceMap}

	      fun printError (s, p1, p2) =
		  let val loc = SourceMap.location_string sourceMap (p1,p2)
		  in
		      TextIO.output (TextIO.stdErr,
				     loc^" Error: " ^ s ^ "\n")
		  end

	      val anyErrors = ref false

	      fun error (p1,p2) s = 
		  (anyErrors:=true;
		   printError (s,p1,p2))

	      val stream = GenParser.makeLexer
		  (fn i => TextIO.inputN (dev, i)) lexarg

	      val (tree, stream) = 
		  let val res = GenParser.parse(15, stream, printError, error)
		      handle (Fail(s)) => 
			  (printError(s, 0, 0);
			   raise Fail s)
		  in
		      if !anyErrors then
			  (print "Abort due to parse errors\n";
			   raise (Fail "Parse"))
		      else
			  res
		  end
	  in
	      {tree=tree,sourceMap=sourceMap}
	  end


      type errorinfo = {pos:{filename:string, posstring:string,region:int*int}option,
			message:string}
      exception ParseErrors of errorinfo list
      val errors : errorinfo list ref = ref nil

      (* parse' may raise exception ParseErrors *)
      fun parse' (dev,filename) : {tree:AST.program ,sourceMap:SourceMap.sourcemap} =
	  let val _ = SymbolTable.reset ()
	      val _ = errors := nil
(*	      val _ = Lexeme.reset () -- breaks multiple files *)
	      val sourceMap = SourceMap.newmap (2 (* lexer initial position *),
						{fileName=filename,
						 line=1, column=1})
	      fun saveError (s, p1, p2) =
		let val loc = SourceMap.location_string sourceMap (p1,p2)
		in errors := {pos=SOME{filename=filename,posstring=loc,region=(p1,p2)},
			      message = s} :: !errors
		end

	      fun saveError' (p1,p2) s = saveError (s,p1,p2)

	      val stream = GenParser.makeLexer
		  (fn i => TextIO.inputN (dev, i)) {sourceMap=sourceMap}
	  
	      val (tree, stream) = GenParser.parse(15, stream, saveError, saveError')
		handle Fail s => (saveError(s, 0, 0); raise Fail s)
                     | X => (saveError(exnName X, 0, 0); 
			     raise Fail "Internal parse error")
	  in case !errors
	       of nil => {tree=tree,sourceMap=sourceMap}
		| errors => raise ParseErrors errors
	  end handle _ => raise ParseErrors (!errors)
                   
  end

