(*
Copyright (c) 1997,1998 The Regents of the University of California.
All rights reserved.

Permission to use, copy, modify, and distribute this software for any
purpose, without fee, and without written agreement is hereby granted,
provided that the above copyright notice and the following two
paragraphs appear in all copies of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*)

(* some top level helpers *)

structure NullExtraInfo =
    struct
	type scopeInfo = unit
	type symInfo = unit
	type astInfo = unit
	val nameAnonAggregates = false
	fun enterScope () = ()
	fun rootScopeInfo () = ()
	fun reset () = ()
    end

structure CParser = GenFront(NullExtraInfo)

fun applyToFile f fileName = 
  let
    val ind = TextIO.openIn fileName
    val result = f (ind,fileName)
    val _ = TextIO.closeIn ind
  in
    result
  end

fun parse fileName = applyToFile CParser.parse fileName
val unparse = CParser.Unparse.program

fun isSuffix suf s =
    let open List String
	val revStr = implode o rev o explode
	val sufRev = revStr suf
	val sRev = revStr s
    in
	isPrefix sufRev sRev
    end

fun parseIDir filename =
    let val allFiles = FileUtil.flattenDir filename
	val allIFiles = List.filter (isSuffix ".i") allFiles
	    fun doParse file =
		let val _ = ()
		in
		    print (concat ["Parsing ",
				   file,
				   "..."]);
		    parse file;
		    print "OK\n"
		end
	    handle _ => print "FAILED!\n"
    in
	app doParse allIFiles
    end

fun parserRegress () =
    let val sources = "/home/cs/jfoster/test.programs"
	val allFiles = FileUtil.flattenDir sources
	val allCFiles = List.filter (isSuffix ".c") allFiles
	fun doParse file =
	    let val _ = ()
	    in
		print (concat ["Parsing ",
			       file,
			       "..."]);
		parse file;
		print "OK\n"
	    end
    in
	(app doParse allCFiles;
	 print "Passed\n")
	handle x => (print "Failed regression test!\n";
		     raise x)
    end

fun export_parse_fn (arg0, arg1::rest) =
   let
      val ast = applyToFile typeCheckCParser.parse arg1
      val ast2 = pam_traverser.traverse_ast_and_view_results ast [arg1]
   in
      OS.Process.success
   end
| export_parse_fn _ = (print "Strange arguments given\n"; OS.Process.failure)

fun export () = SMLofNJ.exportFn ("parse", export_parse_fn)
