(*
Copyright (c) 1997,1998 The Regents of the University of California.
All rights reserved.

Permission to use, copy, modify, and distribute this software for any
purpose, without fee, and without written agreement is hereby granted,
provided that the above copyright notice and the following two
paragraphs appear in all copies of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*)

structure FileUtil :
    sig

	(* Given a directory, return all files in that directory,
	   recursively. *)
	val flattenDir : string -> string list

	(* Given a list of files, return the first on the list
	   that exists. *)
	val findFile : string list -> string

    end =
    struct

	open OS OS.FileSys OS.Path

	fun buildFileList dir getFile =
	    let val filename = getFile ()
		val invisible = String.isPrefix "."
	    in
		if size filename = 0 then []
		else
		    let val fullName = joinDirFile {dir=dir,
						    file=filename}
			val rest = buildFileList dir getFile
		    in
			if invisible filename then
			    rest
			else if isDir fullName then
			    (flattenDir fullName) @ rest
			     else
				 fullName :: rest
		    end
	    end

	(* Expand a file into a list of files *)
	and flattenDir filename =
	    if isDir filename then
		(let val dir = openDir filename
		 in
		     let val fileList = buildFileList filename
			 (fn () => readDir dir)
		     in
			 closeDir dir;
			 fileList
		     end
		     handle ex => (closeDir dir; raise ex)
		 end)
	    else
		[filename]

	val noent = Posix.Error.noent;

	(* Given a list of files, return the first on the list
	 that exists *)
	fun findFile [] = raise SysErr ("No such file or directory",
					SOME noent)
	  | findFile (file::files) =
	    let val _ = fileSize file
	    in
		file
	    end
	handle SysErr(_, noent) => findFile files
	     | ex => raise ex

    end
