
/*  ###############################################################
    ##
    ##     C Tree Builder
    ##
    ##     File:         ctdemo.cpp
    ##
    ##     Revision 0.0.0.0  1996/03/30  01:22:21  kurt
    ##     o   Allow compile-time override of LIB_CPP define.
    ##     o   deleting stack item popped off of DoneStack causes
    ##         segmentation violation in HeapFree() because of 0 value
    ##         in Parse_TOS; Setting Parse_TOS to value popped seems to
    ##         cure problem.
    ##     o   Rev. 0.0 from
    ##         -rw-r--r--  1 1047 1047 44198 Aug  7 19:32
    ##             ftp://ftp.cs.wisc.edu/coral/tmp/spf/ctree.0.01.tar.gz
    ##     -- Kurt Cockrum <kurt@grogatch.seaslug.org>
    ##
    ##     Programmer:   Shaun Flisakowski
    ##     Date:         Dec 16, 1994
    ##
    ##
    ###############################################################  */

#include    <stdio.h>
#include    <iostream.h>
#include    <string.h>
#include    <stdlib.h>
#include    <assert.h>

#include    "ctdemo.h"
#include    "ctree.h"

/*  ###############################################################  */

char  *cur_file;
int    show_as_code = 1;
int    show_prev_way = 0;
int    show_syms = 0;

const int MAX_FILES    = 256;

/* To hold the files/directories requested at startup */
char *file_list[MAX_FILES];

/*  ###############################################################  */

void Usage(char *prog)
{
    cerr << "Usage: " << prog << " [options] [filename(s)]\n\n";

    cerr << "This program parses C code, and builds a tree.\n\n";

    cerr << "Options:\n";
    cerr << "\t-code:\t\t Show tree as code. (-c, default).\n";
    cerr << "\t-tree:\t\t Show tree as a tree. (-t).\n";
    cerr << "\t-syms:\t\t Show symbol table. (-s).\n";
    cerr << "\t-version:\t Show ctree version (-V).\n";
    cerr << "\t-help:\t\t Show usage (-h).\n";
    
    exit(0);
}

/*  ###############################################################  */

int processArgs( int argc, char **argv )
{
    int num_files = 0;
    char *arg;
    char *prog = argv[0];

    argc--; 
    argv++;

    while(argc-- > 0)
    {
        arg = *argv++;

        if ((strcmp(arg,"-tree") == 0)
              || (strcmp(arg,"-t") == 0))
          show_as_code = 0;

        else if ((strcmp(arg,"-code") == 0)
              || (strcmp(arg,"-c") == 0))
          show_as_code = 1;

        else if ((strcmp(arg,"-prev") == 0)
              || (strcmp(arg,"-p") == 0))
          show_prev_way = 1;

        else if ((strcmp(arg,"-syms") == 0)
              || (strcmp(arg,"-s") == 0))
          show_syms = 1;

        else if ((strcmp(arg,"-version") == 0)
              || (strcmp(arg,"-V") == 0))
          cout << VERSION << endl;

        else if ((strcmp(arg,"-help") == 0)
            || (strcmp(arg,"-h") == 0))
          Usage(prog);

        else
          file_list[num_files++] = arg;    
    }

    file_list[num_files] = NULL;
    return num_files;
}

/*  ###############################################################  */

int main( int argc, char **argv )
{
    int   i, nf;

    if ((nf = processArgs(argc,argv)) < 0)
        return -1;

    if (nf == 0)
        Usage(argv[0]);

    InitStdPath();

    Project  *prj = new Project();

    for (i=0; i < nf; i++)
    {
        TransUnit *unit = prj->parse(file_list[i]);

        if (unit) 
        {
            /* Show the parse tree. */
            cout << *unit << endl;
        } else {
            cout << " translation unit for " << file_list[i] << " is NULL.\n";
        }

        if (show_syms)
        {
          cout << "Symbols:";
          unit->contxt->syms->Show(cout);
          cout << "Tags:";
          unit->contxt->tags->Show(cout);
          cout << "Labels:";
          unit->contxt->labels->Show(cout);
        }
    }

    delete prj;

    return 0;
}

/*  ###############################################################  */
