/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
    o+
    o+     File:         express.h
    o+
    o+     Programmer:   Shaun Flisakowski
    o+     Date:         Aug 9, 1998
    o+
    o+     A high-level view of expressions.
    o+
    o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */

#ifndef    EXPRESS_H
#define    EXPRESS_H

#include <stdlib.h>
#include <iostream.h>

#include "symbol.h"
#include "tree.h"
#include "callback.h"
#include "location.h"


class Decl;
class Type;

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

typedef void (*fnExprCallback)(Expression *expr);

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
enum ExpressionType
{
    ET_VoidExpr=0,           // No value

    ET_Constant,
    ET_Variable,
    ET_FunctionCall,

    ET_AssignExpr,
    ET_RelExpr,

    ET_UnaryExpr,
    ET_BinaryExpr,
    ET_TrinaryExpr,          // x ? y : z

    ET_CastExpr,             // (type) x
    ET_SizeofExpr,           // sizeof(type) -or- sizeof(expr)
    ET_IndexExpr             // x[y]...[]
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
enum ConstantType
{
    CT_Char,
    CT_Int,
    CT_Float,

    CT_String,
    CT_Array            // For an array initializer
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
enum AssignOp
{
    AO_Equal,          //  =

    AO_PlusEql,        // +=
    AO_MinusEql,       // -=
    AO_MultEql,        // *=
    AO_DivEql,         // /=
    AO_ModEql,         // %=

    AO_ShlEql,         // <<=
    AO_ShrEql,         // >>=

    AO_BitNotEql,      // ~=
    AO_BitAndEql,      // &=
    AO_BitXorEql,      // ^=
    AO_BitOrEql        // |=
};

enum RelOp
{
    RO_Equal,          // ==
    RO_NotEqual,       // !=

    RO_Less,           // < 
    RO_LessEql,        // <=
    RO_Grtr,           // > 
    RO_GrtrEql         // >=
};

enum UnaryOp
{
    UO_Plus=0,         // +
    UO_Minus,          // -
    UO_BitNot,         // ~
    UO_Not,            // !

    UO_PreInc,         // ++x
    UO_PreDec,         // --x
    UO_PostInc,        // x++
    UO_PostDec,        // x--

    UO_AddrOf,         // &
    UO_Deref           // *
};

enum BinaryOp
{
    BO_Plus,        // +
    BO_Minus,       // -
    BO_Mult,        // *
    BO_Div,         // /
    BO_Mod,         // %

    BO_Shl,         // <<
    BO_Shr,         // >>
    BO_BitAnd,      // &
    BO_BitXor,      // ^
    BO_BitOr,       // |

    BO_And,         // &&
    BO_Or,          // ||

    BO_Comma,       // x,y

    BO_Member,      // x.y
    BO_PtrMember,   // x->y

    BO_Assign,      // An AssignExpr
    BO_Rel          // A RelExpr
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class Expression
{
  public:
    Expression(ExpressionType et, const Location& l);
    virtual ~Expression();

    virtual int precedence() const { return 16; }

    virtual Expression *dup() const;

    virtual void print(ostream& out) const;

    virtual void findExpr( fnExprCallback cb ) { (cb)(this); }

    ExpressionType    etype;
    Location location;

  friend ostream& operator<< ( ostream& out, const Expression& expr );
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class Constant : public Expression
{
  public:
    Constant(ConstantType ct, const Location& l);
    virtual ~Constant();
 
    virtual Expression *dup() const = 0;
    virtual void print(ostream& out) const = 0;

    ConstantType    ctype;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class IntConstant : public Constant
{
  public:
    IntConstant(long val, const Location& l);
    virtual ~IntConstant();
 
    Expression *dup() const;
    void print(ostream& out) const;

    long    lng;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class FloatConstant : public Constant
{
  public:
    FloatConstant(double val, const Location& l);
    virtual ~FloatConstant();
 
    Expression *dup() const;
    void print(ostream& out) const;

    double  doub;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class CharConstant : public Constant
{
  public:
    CharConstant(char chr, const Location& l, bool isWide=false);
   ~CharConstant();

    Expression *dup() const;
    void print(ostream& out) const;

    char    ch;       // the character
    bool    wide;     // wide char?
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class StringConstant : public Constant
{
  public:
    StringConstant(char *str, const Location& l, bool isWide=false);
   ~StringConstant();

    int length() const;

    Expression *dup() const;
    void print(ostream& out) const;

    char    *buff;    // dynamic string storage
    bool    wide;     // wide string?
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class ArrayConstant : public Constant
{
  public:
    ArrayConstant(const Location& l);
   ~ArrayConstant();

    void addElement( Expression *expr);

    Expression *dup() const;
    void print(ostream& out) const;

    void findExpr( fnExprCallback cb );

    int           size;
    int           nItems;
    Expression  **items;        // Array of whatever, including another array.
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class Variable : public Expression
{
  public:
    Variable(Symbol *varname, const Location& l);
   ~Variable();

    Expression *dup() const;
    void print(ostream& out) const;

    Symbol    *name;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class FunctionCall : public Expression
{
  public:
    FunctionCall(Expression *func, const Location& l);
   ~FunctionCall();
    
    int  nArgs() const { return _nArgs; }
    void addArg( Expression *arg );

    void addArgs( treeNode *args );

    Expression *dup() const;
    void print(ostream& out) const;

    void findExpr( fnExprCallback cb );

    Expression   *function;
    int           size;
    int           _nArgs;
    Expression  **args;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class UnaryExpr : public Expression
{
  public:
    UnaryExpr( UnaryOp op, Expression *expr, const Location& l);
   ~UnaryExpr();

    int precedence() const;

    Expression *operand() const { return _operand; }

    UnaryOp op() const { return uOp; }

    Expression *dup() const;
    void print(ostream& out) const;

    void findExpr( fnExprCallback cb );

    UnaryOp    uOp;
    Expression *_operand;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class BinaryExpr : public Expression
{
  public:
    BinaryExpr( BinaryOp op, Expression *lExpr, Expression *rExpr,
		const Location& l);
   ~BinaryExpr();

    Expression *leftExpr() const { return _leftExpr; }
    Expression *rightExpr() const { return _rightExpr; }

    virtual int precedence() const;

    BinaryOp op() const { return bOp; }

    virtual Expression *dup() const;
    virtual void print(ostream& out) const;

    void findExpr( fnExprCallback cb );

    BinaryOp    bOp;

    Expression *_leftExpr;
    Expression *_rightExpr;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class TrinaryExpr : public Expression
{
  public:
    TrinaryExpr( Expression *cExpr, Expression *tExpr, Expression *fExpr,
		 const Location& l);
   ~TrinaryExpr();

    Expression *condExpr() const { return _condExpr; }
    Expression *trueExpr() const { return _trueExpr; }
    Expression *falseExpr() const { return _falseExpr; }

    int precedence() const { return 3; }

    Expression *dup() const;
    void print(ostream& out) const;

    void findExpr( fnExprCallback cb );

    Expression *_condExpr;
    Expression *_trueExpr;
    Expression *_falseExpr;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class AssignExpr : public BinaryExpr
{
  public:
    AssignExpr( AssignOp op, Expression *lExpr, Expression *rExpr,
		const Location& l);
   ~AssignExpr();

    Expression *lValue() const { return leftExpr(); }
    Expression *rValue() const { return rightExpr(); }

    AssignOp op() const { return aOp; }

    int precedence() const { return 2; }

    Expression *dup() const;
    void print(ostream& out) const;

    AssignOp    aOp;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class RelExpr : public BinaryExpr
{
  public:
    RelExpr( RelOp op, Expression *lExpr, Expression *rExpr,
	     const Location& l);
   ~RelExpr();

    int precedence() const;

    RelOp op() const { return rOp; }

    Expression *dup() const;
    void print(ostream& out) const;

    RelOp    rOp;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class CastExpr : public Expression
{
  public:
    CastExpr( Decl *typeExpr, Expression *operand,
	      const Location& l);
   ~CastExpr();

    Decl *type() { return castTo; }

    int precedence() const { return 14; }

    Expression *dup() const;
    void print(ostream& out) const;

    void findExpr( fnExprCallback cb );

    Decl        *castTo;
    Expression  *expr;    // The expression being cast.
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class SizeofExpr : public Expression
{
  public:
    SizeofExpr( Expression *operand, const Location& l);
    SizeofExpr( Decl *operand, const Location& l);
   ~SizeofExpr();

    int precedence() const { return 15; }

    Expression *dup() const;
    void print(ostream& out) const;

    void findExpr( fnExprCallback cb );

    Decl        *type;    // want the size of this type.
    Expression  *expr;    // want the size of this expression.
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class IndexExpr : public Expression
{
  // I should extend this to be able to handle
  // multi-dimensional arrays directly.

  public:
    IndexExpr( Expression *_array, Expression *sub, const Location& l);
   ~IndexExpr();
    
    //addSubscript( Expression *sub );
    Expression *subscript(int i);

    Expression *dup() const;
    void print(ostream& out) const;

    void findExpr( fnExprCallback cb );

    Expression   *array;
    Expression   *_subscript;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

Expression    *buildExpr(treeNode *node);
Expression    *buildIndexExpr(treeNode *node);

ArrayConstant *buildArrayInit(ArrayConstant *array, treeNode *node);

// For debugging, show the enumerated name.
char *nameOfExpressionType(ExpressionType type);
char *nameOfConstantType(ConstantType type);
char *nameOfAssignOp(AssignOp op);
char *nameOfRelOp(RelOp op);
char *nameOfUnaryOp(UnaryOp op);
char *nameOfBinaryOp(BinaryOp op);

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

#endif  /* EXPRESS_H */

