
/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
    o+
    o+     File:         nametbl.h
    o+
    o+     Programmer:   Shaun Flisakowski
    o+     Date:         Sep 16, 1998
    o+
    o+     A name-table, for storing strings.
    o+
    o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */

#include    <iostream.h>
#include    <string.h>
#include    <stdlib.h>
#include    <assert.h>

//#include    "lexer.h"
#include    "nametbl.h"

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
uint
calcHash( char *str )
{
    uint hsh = 0, c;

    while(*str)
    {
        c = *str++;
        hsh = (hsh << 1) ^ (hsh >> 20) ^ c;
    }

    return hsh;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
StrItem::StrItem(char *str, int len /* =0 */)
{
    _string = NULL;
    _next = NULL;
    setString(str,len);
}

StrItem::StrItem(char *str, int len, uint hsh)
{
    _string = new char [len+1];
    memcpy(_string,str,len+1);
    _hash = hsh;
    _next = NULL;
}

StrItem::~StrItem()
{
    delete [] _string;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
StrItem::setString(char *str, int len /* =0 */)
{
    delete [] _string;

    len = (len != 0) ? len : strlen(str);
    _string = new char [len+1];
    memcpy(_string,str,len+1);
    _hash = calcHash(_string);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
bool
StrItem::equal(StrItem *that) const
{
    if (!this || !that)
      return false;

    return (_string == that->_string);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
bool
StrItem::equal(char *that) const
{
    return (strcmp(_string,that) == 0);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
NameTbl::NameTbl()
{
    for (int j=0; j < kMaxHashBckts; j++)
        tbl[j] = NULL;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
NameTbl::~NameTbl()
{
    StrItem *item, *prev=NULL;

    for (int j=0; j < kMaxHashBckts; j++)
    {
        prev = NULL;

        for (item=tbl[j]; item; item=item->next())
        {
            delete prev;
            prev = item;
        }

        delete prev;
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
StrItem*
NameTbl::insert( char *sym, int len )
{
    uint hsh;
    int bckt;
    StrItem *symptr;

    hsh  = calcHash(sym);
    bckt = hsh % kMaxHashBckts;

    for (symptr=tbl[bckt]; symptr; symptr=symptr->next())
    {
      if ((symptr->hash() == hsh) && (strcmp(sym,symptr->string()) == 0))
          return symptr;
    } 

    if (len == 0)
        symptr = new StrItem(sym,strlen(sym),hsh);
    else
        symptr = new StrItem(sym,len,hsh);

    symptr->setNext(tbl[bckt]);
    tbl[bckt] = symptr;

    return symptr;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
/*
void nmeshow()
{
int  j;
HashItem *hptr;

    cout << "        Name Table\n";
    cout << "       ------------\n\n";

    for (j=0; j < MAX_HASH_BCKTS; j++)
    {
        cout << "Examining bucket: " << j << endl;
        for (hptr=NmeTab[j]; hptr; hptr=hptr->next)
        {
            cout << "Name: " << hptr->sym.str << endl;
        } 
    } 

    cout << "\n\n";
}
*/
// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

