/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
    o+
    o+     File:         stemnt.cpp
    o+
    o+     Programmer:   Shaun Flisakowski
    o+     Date:         Aug 9, 1998
    o+
    o+     A high-level view of statements.
    o+
    o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */

#include <string.h>
#include <assert.h>

#include "tree.h"
#include "stemnt.h"
#include "symbol.h"
#include "decl.h"
#include "project.h"

#include "gram.h"


// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
//  indent - 2 spaces per level.
void
indent(ostream& out, int level)
{
    if (level > 0)
    {
        for (int j=level; j > 0; j--)
          out << "  ";
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
printNull(ostream& out, int level)
{
    indent(out,level);
    out << ";\n";
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
printBlock(ostream& out, int level, Statement *block)
{
    if (block == NULL)
        printNull(out,level+1);
    else if (block->isBlock())
        block->print(out,level);
    else
    {
        block->print(out,level+1);
        if (block->needSemicolon())
            out << ";";
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Label::Label( LabelType labelType )
{
    type = labelType;
    expr = NULL;
    next = NULL;
}   

Label::Label( Expression *_expr )
{
    type = LT_Case;
    expr = _expr;
    next = NULL;
}   

Label::Label( Symbol *sym )
{
    type = LT_Goto;
    name = sym;
    next = NULL;
}   

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Label::~Label()
{
    if (type == LT_Case)
        delete expr;
    else if (type == LT_Goto)
        delete name;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Label*
Label::dup() const
{
    Label *ret = NULL;

    if (this)
    {
        ret = new Label(type);

        switch (type)
        {
          default:
          case LT_None:
          case LT_Default:
            break;

          case LT_Case:
            ret->expr = expr->dup();
            break;

          case LT_Goto:
            ret->name = name->dup();
            break;
        }

        // Don't duplicate the whole list.
        ret->next = NULL;
    }
    
    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
Label::print(ostream& out, int level) const
{
    indent(out,level-1);

    switch (type)
    {
        case LT_None:
            assert(0);
            break;

        case LT_Default:
            out << "default";
            break;

        case LT_Case:
            assert(expr);
            out << "case " << *expr;
            break;

        case LT_Goto:
            assert(name);
            out << *name;
            break;
    }

    out << ":\n";
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
Label::findExpr( fnExprCallback cb )
{
    if (type == LT_Case)
        expr->findExpr(cb);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Statement::Statement(StatementType stemntType, const Location& l):
  location(l)
{
    type  = stemntType;

    lblHead = lblTail = NULL;
    next = NULL;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Statement::~Statement()
{
    Label    *lbl, *prevLbl=NULL;

    for (lbl=lblHead; lbl; lbl=lbl->next)
    {
        delete prevLbl;
        prevLbl = lbl;
    }
    delete prevLbl;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
Statement::addLabel( Label *lbl )
{
    lbl->next = NULL;

    if (lblTail)
        lblTail->next = lbl;
    else
        lblHead = lbl;

    lblTail = lbl;

    // Hook the label's symtable entry back to this statement.
    if (lbl->type == LT_Goto)
    {
        if (lbl->name->entry != NULL)
            lbl->name->entry->uLabelPosition = this;
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Statement*
Statement::dup() const
{
    Statement *ret = NULL;

    if (this)
    {
        ret = new Statement(type, location);

        for (Label *lbl=lblHead; lbl; lbl=lbl->next)
            ret->addLabel(lbl->dup());
    }

    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
Statement::print(ostream& out, int level) const
{
    for (Label *lbl = lblHead; lbl; lbl=lbl->next)
        lbl->print(out,level);

    indent(out,level);

    switch (type)
    {
      default:
        out << __PRETTY_FUNCTION__ << endl;
        out << nameOfStatementType(type) << endl;
        break;

      case ST_NullStemnt:          // The null statement.
        out << ";";
        break;

      case ST_ContinueStemnt:
        out << "continue";
        break;

      case ST_BreakStemnt:
        out << "break";
        break;
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
Statement::findExpr( fnExprCallback cb )
{
    for (Label *lbl = lblHead; lbl; lbl=lbl->next)
        lbl->findExpr(cb);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
InclStemnt::InclStemnt(char *incl, const Location& l):
  Statement( ST_InclStemnt, l )
{
static const char  *NrmPath[] = { "./", NULL };

    int    j;

    isStandard = false;

    for (j=0; StdPath[j]; j++)
    {
        if (strncmp(incl,StdPath[j],strlen(StdPath[j])) == 0)
        {
            isStandard = true;
            incl = &incl[strlen(StdPath[j])];
            goto done;
        }
    }

    for (j=0; NrmPath[j]; j++)
    {
        if (strncmp(incl,NrmPath[j],strlen(NrmPath[j])) == 0)
        {
            incl = &incl[strlen(NrmPath[j])];
            goto done;
        }
    }

  done:
    filename = new char [strlen(incl)+1];
    strcpy(filename,incl);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
InclStemnt::~InclStemnt()
{
    delete [] filename;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Statement*
InclStemnt::dup() const
{
    InclStemnt *ret = NULL;

    if (this)
    {
        ret = new InclStemnt(filename, location);
        ret->isStandard = isStandard;

        for (Label *lbl=lblHead; lbl; lbl=lbl->next)
            ret->addLabel(lbl->dup());
    }

    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
InclStemnt::print(ostream& out, int level) const
{
    out << "#include " << (isStandard ? '<' : '"');
    out << filename << (isStandard ? '>' : '"') << endl;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
EndInclStemnt::EndInclStemnt(const Location& l)
           : Statement( ST_EndInclStemnt, l )
{
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
EndInclStemnt::~EndInclStemnt()
{
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Statement*
EndInclStemnt::dup() const
{
    EndInclStemnt *ret = NULL;

    if (this)
    {
        ret = new EndInclStemnt(location);

        for (Label *lbl=lblHead; lbl; lbl=lbl->next)
            ret->addLabel(lbl->dup());
    }

    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
EndInclStemnt::print(ostream& out, int level) const
{
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
ExpressionStemnt::ExpressionStemnt( Expression *expr, const Location& l)
                 : Statement( ST_ExpressionStemnt, l )
{
    expression = expr;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
ExpressionStemnt::~ExpressionStemnt()
{
    delete expression;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Statement*
ExpressionStemnt::dup() const
{
    ExpressionStemnt *ret = NULL;

    if (this)
    {
        if (expression)
            ret = new ExpressionStemnt(expression->dup(), location);
        else
            ret = new ExpressionStemnt(NULL, location);

        for (Label *lbl=lblHead; lbl; lbl=lbl->next)
            ret->addLabel(lbl->dup());
    }

    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
ExpressionStemnt::print(ostream& out, int level) const
{
    for (Label *lbl = lblHead; lbl; lbl=lbl->next)
        lbl->print(out,level);

    indent(out,level);
    assert(expression);
    out << *expression;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
ExpressionStemnt::findExpr( fnExprCallback cb )
{
    Statement::findExpr(cb);
    expression->findExpr(cb);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
IfStemnt::IfStemnt( Expression *_cond,
                    Statement *_thenBlk, const Location& l,
                    Statement *_elseBlk /* =NULL */)
         : Statement( ST_IfStemnt, l )
{
    cond = _cond;
    thenBlk = _thenBlk;
    elseBlk = _elseBlk;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
IfStemnt::~IfStemnt()
{
    delete cond;
    delete thenBlk;
    delete elseBlk;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Statement*
IfStemnt::dup() const
{
    IfStemnt *ret = NULL;

    if (this)
    {
        Expression  *_cond = (cond ? cond->dup() : (Expression*) NULL);
        Statement   *_thenBlk = (thenBlk ? thenBlk->dup() : (Statement*) NULL);
        Statement   *_elseBlk = (elseBlk ? elseBlk->dup() : (Statement*) NULL);

        ret = new IfStemnt(_cond,_thenBlk,location,_elseBlk);

        for (Label *lbl=lblHead; lbl; lbl=lbl->next)
            ret->addLabel(lbl->dup());
    }

    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
IfStemnt::print(ostream& out, int level) const
{
    for (Label *lbl = lblHead; lbl; lbl=lbl->next)
        lbl->print(out,level);

    indent(out,level);

    out << "if (" << *cond << ")\n";
    printBlock(out,level,thenBlk);

    if (elseBlk)
    {
        out << endl;
        indent(out,level);
        out << "else\n";
        printBlock(out,level,elseBlk);
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
IfStemnt::findExpr( fnExprCallback cb )
{
    Statement::findExpr(cb);

    cond->findExpr(cb);

    thenBlk->findExpr(cb);

    if (elseBlk)
        elseBlk->findExpr(cb);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
IfStemnt::findStemnt( fnStemntCallback cb )
{
    (cb)(this);

    thenBlk->findStemnt(cb);

    if (elseBlk)
        elseBlk->findStemnt(cb);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
SwitchStemnt::SwitchStemnt( Expression *_cond, Statement *_block,
			    const Location& l )
             : Statement( ST_SwitchStemnt, l )
{
    cond = _cond;
    block = _block;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
SwitchStemnt::~SwitchStemnt()
{
    delete cond;
    delete block;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Statement*
SwitchStemnt::dup() const
{
    SwitchStemnt *ret = NULL;

    if (this)
    {
        Expression  *_cond  = (cond ? cond->dup() : (Expression*) NULL);
        Statement   *_block = (block ? block->dup() : (Statement*) NULL);

        ret = new SwitchStemnt(_cond,_block, location);

        for (Label *lbl=lblHead; lbl; lbl=lbl->next)
            ret->addLabel(lbl->dup());
    }

    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
SwitchStemnt::print(ostream& out, int level) const
{
    for (Label *lbl = lblHead; lbl; lbl=lbl->next)
        lbl->print(out,level);

    indent(out,level);

    out << "switch (" << *cond << ")\n";

    printBlock(out,level,block);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
SwitchStemnt::findExpr( fnExprCallback cb )
{
    Statement::findExpr(cb);
    cond->findExpr(cb);
    block->findExpr(cb);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
SwitchStemnt::findStemnt( fnStemntCallback cb )
{
    (cb)(this);

    block->findStemnt(cb);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
ForStemnt::ForStemnt( Expression *_init, Expression *_cond,
                      Expression *_incr, const Location& l,
		      Statement *_block /* =NULL */)
          : Statement( ST_ForStemnt, l )
{
    init = _init;
    cond = _cond;
    incr = _incr;

    block = _block;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
ForStemnt::~ForStemnt()
{
    delete init;
    delete cond;
    delete incr;

    delete block;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Statement*
ForStemnt::dup() const
{
    ForStemnt *ret = NULL;

    if (this)
    {
        Expression  *_init  = (init ? init->dup() : (Expression*) NULL);
        Expression  *_cond  = (cond ? cond->dup() : (Expression*) NULL);
        Expression  *_incr  = (incr ? incr->dup() : (Expression*) NULL);
        Statement   *_block = (block ? block->dup() : (Statement*) NULL);

        ret = new ForStemnt(_init,_cond,_incr,location,_block);

        for (Label *lbl=lblHead; lbl; lbl=lbl->next)
            ret->addLabel(lbl->dup());
    }

    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
ForStemnt::print(ostream& out, int level) const
{
    for (Label *lbl = lblHead; lbl; lbl=lbl->next)
        lbl->print(out,level);

    indent(out,level);

    out << "for (";

    if (init)
        out << *init;
    out << "; ";

    if (cond)
        out << *cond;
    out << "; ";

    if (incr)
        out << *incr;
    out << ")\n";

    printBlock(out,level,block);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
ForStemnt::findExpr( fnExprCallback cb )
{
    Statement::findExpr(cb);

    if (init != NULL)
        init->findExpr(cb);

    if (cond != NULL)
        cond->findExpr(cb);

    if (incr != NULL)
        incr->findExpr(cb);

    if (block != NULL)
        block->findExpr(cb);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
ForStemnt::findStemnt( fnStemntCallback cb )
{
    (cb)(this);

    if (block)
        block->findStemnt(cb);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
WhileStemnt::WhileStemnt( Expression *_cond, Statement *_block,
			  const Location& l)
            : Statement( ST_WhileStemnt, l )
{
    cond = _cond;
    block = _block;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
WhileStemnt::~WhileStemnt()
{
    delete cond;
    delete block;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Statement*
WhileStemnt::dup() const
{
    WhileStemnt *ret = NULL;

    if (this)
    {
        Expression  *_cond  = (cond ? cond->dup() : (Expression*) NULL);
        Statement   *_block = (block ? block->dup() : (Statement*) NULL);

        ret = new WhileStemnt(_cond,_block, location);

        for (Label *lbl=lblHead; lbl; lbl=lbl->next)
            ret->addLabel(lbl->dup());
    }

    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
WhileStemnt::print(ostream& out, int level) const
{
    for (Label *lbl = lblHead; lbl; lbl=lbl->next)
        lbl->print(out,level);

    indent(out,level);

    out << "while (" << *cond << ")\n";

    printBlock(out,level,block);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
WhileStemnt::findExpr( fnExprCallback cb )
{
    Statement::findExpr(cb);

    cond->findExpr(cb);

    if (block)
        block->findExpr(cb);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
WhileStemnt::findStemnt( fnStemntCallback cb )
{
    (cb)(this);

    if (block)
        block->findStemnt(cb);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
DoWhileStemnt::DoWhileStemnt( Expression *_cond, Statement *_block,
			     const Location& l)
              : Statement( ST_DoWhileStemnt, l )
{
    cond = _cond;
    block = _block;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
DoWhileStemnt::~DoWhileStemnt()
{
    delete cond;
    delete block;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Statement*
DoWhileStemnt::dup() const
{
    DoWhileStemnt *ret = NULL;

    if (this)
    {
        Expression  *_cond  = (cond ? cond->dup() : (Expression*) NULL);
        Statement   *_block = (block ? block->dup() : (Statement*) NULL);

        ret = new DoWhileStemnt(_cond,_block, location);

        for (Label *lbl=lblHead; lbl; lbl=lbl->next)
            ret->addLabel(lbl->dup());
    }

    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
DoWhileStemnt::print(ostream& out, int level) const
{
    for (Label *lbl = lblHead; lbl; lbl=lbl->next)
        lbl->print(out,level);

    indent(out,level);
    out << "do ";

    if (!block->isBlock())
        out << endl;

    printBlock(out,level,block);

    if (!block->isBlock())
        out << endl;

    indent(out,level);
    out << "while (" << *cond << ")";
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
DoWhileStemnt::findExpr( fnExprCallback cb )
{
    Statement::findExpr(cb);

    cond->findExpr(cb);

    if (block)
        block->findExpr(cb);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
DoWhileStemnt::findStemnt( fnStemntCallback cb )
{
    (cb)(this);

    if (block)
        block->findStemnt(cb);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
GotoStemnt::GotoStemnt( Symbol *_dest, const Location& l)
           : Statement( ST_GotoStemnt, l )
{
    dest = _dest;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
GotoStemnt::~GotoStemnt()
{
    delete dest;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Statement*
GotoStemnt::dup() const
{
    GotoStemnt *ret = NULL;

    if (this)
    {
        ret = new GotoStemnt(dest->dup(), location);

        for (Label *lbl=lblHead; lbl; lbl=lbl->next)
            ret->addLabel(lbl->dup());
    }

    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
GotoStemnt::print(ostream& out, int level) const
{
    for (Label *lbl = lblHead; lbl; lbl=lbl->next)
        lbl->print(out,level);

    indent(out,level);
    out << "goto " << *dest;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
ReturnStemnt::ReturnStemnt( Expression *_result, const Location& l)
             : Statement( ST_ReturnStemnt, l )
{
    result = _result;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
ReturnStemnt::~ReturnStemnt()
{
    delete result;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Statement*
ReturnStemnt::dup() const
{
    ReturnStemnt *ret = NULL;

    if (this)
    {
        Expression  *_result  = (result ? result->dup() : (Expression*) NULL);

        ret = new ReturnStemnt(_result, location);

        for (Label *lbl=lblHead; lbl; lbl=lbl->next)
            ret->addLabel(lbl->dup());
    }

    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
ReturnStemnt::print(ostream& out, int level) const
{
    for (Label *lbl = lblHead; lbl; lbl=lbl->next)
        lbl->print(out,level);

    indent(out,level);
    out << "return";

    if (result)
        out << " " << *result;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
ReturnStemnt::findExpr( fnExprCallback cb )
{
    Statement::findExpr(cb);

    if (result)
        result->findExpr(cb);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
DeclStemnt::DeclStemnt(const Location& l,
		       StatementType stype /* =ST_DeclStemnt */)
      : Statement( stype, l )
{
    size = nDecls = 0;
    decls = NULL;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
DeclStemnt::~DeclStemnt()
{
    for (int j=0; j < nDecls; j++)
        delete decls[j];

    delete [] decls;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
DeclStemnt::addDecl(Decl *decl)
{
    if (size == nDecls)
    {
        if (size == 0)
            size = 4;
        else
            size += size;

        Decl   **oldDecls = decls;

        decls = new Decl* [size];

        for (int j=0; j < nDecls; j++)
        {
            decls[j] = oldDecls[j];
        }

        delete [] oldDecls;
    }

    decls[nDecls] = decl;
    nDecls++;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Statement*
DeclStemnt::dup() const
{
    DeclStemnt *ret = NULL;

    if (this)
    {
        if (type == ST_TypedefStemnt)
            ret = new TypedefStemnt(location);
        else
            ret = new DeclStemnt(location);

        ret->base = base;
        ret->base.name = base.name->dup();

        if (ret->base.form)
            ret->base.form = base.form->dup();

        if (ret->base.initializer)
            ret->base.initializer = base.initializer->dup();

        ret->size = size;
        ret->decls = new Decl* [size];

        for (int j=0; j < nDecls; j++)
            ret->addDecl(decls[j]->dup());

        for (Label *lbl=lblHead; lbl; lbl=lbl->next)
            ret->addLabel(lbl->dup());
    }

    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
DeclStemnt*
DeclStemnt::convertToTypedef()
{
    // Nothing to do?
    if (isTypedef() || !base.isTypedef())
        return this;

    TypedefStemnt *ret = new TypedefStemnt(location);

    // Since we are really the same thing,
    // let's just steal the insides.
    ret->lblHead = lblHead;
    ret->lblTail = lblTail;
    lblHead = lblTail = NULL;

    ret->next    = next;
    next = NULL;

    ret->base.copy(base);
    base.clear();

    ret->size   = size;
    ret->nDecls = nDecls;
    ret->decls  = decls;
    decls = NULL;

    delete this;        // Danger - Will Robinson!

    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
DeclStemnt::print(ostream& out, int level) const
{
    for (Label *lbl = lblHead; lbl; lbl=lbl->next)
        lbl->print(out,level);

    indent(out,level);

    // The base type.
    base.print(out,true,level);

    if (base.form == NULL)
        out << " ";

    if (nDecls > 0)
    {
        decls[0]->print(out,false,level);
        for (int j=1; j < nDecls; j++)
        {
            out << ", ";
            decls[j]->print(out,false,level);
        }
    }

    out << ";";
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
DeclStemnt::findExpr( fnExprCallback cb )
{
    Statement::findExpr(cb);

    base.findExpr(cb);

    for (int j=0; j < nDecls; j++)
    {
        decls[j]->findExpr(cb);
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
TypedefStemnt::TypedefStemnt(const Location& l)
              : DeclStemnt( l, ST_TypedefStemnt )
{
}

TypedefStemnt::~TypedefStemnt()
{
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
TypedefStemnt::print(ostream& out, int level) const
{
    for (Label *lbl = lblHead; lbl; lbl=lbl->next)
        lbl->print(out,level);

    indent(out,level);

    out << "typedef ";

    // The base type.
    base.print(out,true,level);

    if (base.form == NULL)
        out << " ";

    if (nDecls > 0)
    {
        decls[0]->print(out,false,level);
        for (int j=1; j < nDecls; j++)
        {
            out << ", ";
            decls[j]->print(out,false,level);
        }
    }

    out << ";";
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Block::Block(const Location& l)
      : Statement( ST_Block , l )
{
    head = tail = NULL;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Block::~Block()
{
    Statement *stemnt, *prevStemnt = NULL;

    for (stemnt=head; stemnt; stemnt=stemnt->next)
    {
        delete prevStemnt;
        prevStemnt = stemnt;
    }
    delete prevStemnt;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
Block::add(Statement *stemnt)
{
    if (stemnt)
    {
        stemnt->next = NULL;
    
        if (tail)
            tail->next = stemnt;
        else
            head = stemnt;
    
        tail = stemnt;
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
Block::addStatements(treeNode *node)
{
    Statement    *stemnt;
    Block        *blk;

    if (node)
    {
        switch (node->which)
        {
            default:
            case NONE_T:
            case LEAF_T:
            case IF_T:
            case FOR_T:
                break;
    
            case NODE_T:
                switch (node->type)
                {
                  default:
                      break;

                  case TN_DECL_LIST:
                  case TN_STEMNT_LIST:
                      addStatements(node->lnode);
                      addStatements(node->rnode);
                      break;

                  case TN_BLOCK:
                      blk = new Block(node);
                      blk->addStatements(node->lnode);
                      blk->addStatements(node->rnode);
                      add(blk);
                      break;

                  case TN_DECL:
                  case TN_DECLS:
                  case TN_STEMNT:
                      stemnt = buildStatement(node);
                      add(stemnt);
                      break;
                }
                break;
        }
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Statement*
Block::dup() const
{
    Block *ret = NULL;

    if (this)
    {
        Statement *stemnt;

        ret = new Block(location);

        for (stemnt=head; stemnt; stemnt=stemnt->next)
            ret->add(stemnt->dup());

        for (Label *lbl=lblHead; lbl; lbl=lbl->next)
            ret->addLabel(lbl->dup());
    }

    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
Block::print(ostream& out, int level) const
{
    bool       isDecl;
    Statement *stemnt;

    for (Label *lbl = lblHead; lbl; lbl=lbl->next)
        lbl->print(out,level);

    indent(out,level);
    out << "{\n";

    isDecl = (head != NULL) ? head->isDeclaration() : false;
    for (stemnt=head; stemnt; stemnt=stemnt->next)
    {
        if (isDecl && !stemnt->isDeclaration())
        {
            isDecl = false;
            out << endl;
        }

        stemnt->print(out,level+1);

        if (stemnt->needSemicolon())
            out << ";";
        out << endl;
    }

    indent(out,level);
    out << "}\n";
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
Block::findExpr( fnExprCallback cb )
{
    Statement *stemnt;

    Statement::findExpr(cb);

    for (stemnt=head; stemnt; stemnt=stemnt->next)
    {
        stemnt->findExpr(cb);
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
Block::findStemnt( fnStemntCallback cb )
{
    Statement *stemnt;

    (cb)(this);

    for (stemnt=head; stemnt; stemnt=stemnt->next)
    {
        stemnt->findStemnt(cb);
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
Block::insert(Statement *stemnt, Statement *after /* =NULL */)
{
    if (stemnt)
    {
        stemnt->next = NULL;
    
        if (tail)
        {
            if (after)
            {
                stemnt->next = after->next;
                after->next = stemnt;
            }
            else
            {
                stemnt->next = head;
                head = stemnt;
            }

            if (stemnt->next == NULL)
                tail = stemnt;
        }
        else
            head = tail = stemnt;
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
FunctionDef::FunctionDef(const Location& l)
            : Block(l)
{
    decl = NULL;
}

FunctionDef::~FunctionDef()
{
    delete decl;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Statement*
FunctionDef::dup() const
{
    FunctionDef *ret = NULL;

    if (this)
    {
        Statement *stemnt;

        ret = new FunctionDef(location);

        ret->decl = decl->dup();

        for (stemnt=head; stemnt; stemnt=stemnt->next)
            ret->add(stemnt->dup());

        for (Label *lbl=lblHead; lbl; lbl=lbl->next)
            ret->addLabel(lbl->dup());
    }

    return ret;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
FunctionDef::print(ostream& out, int) const
{
    decl->print(out,true); 
    out << endl;

    Block::print(out,0);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
FunctionDef::findExpr( fnExprCallback cb )
{
    decl->findExpr(cb);

    Block::findExpr(cb);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Symbol*
FunctionDef::FunctionName() const
{
    return decl->name;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Label*
buildLabel(treeNode *node)
{
    Label       *result = NULL;
    Expression  *expr;
    Symbol      *sym;
    leafNode    *leaf;

    if (node)
    {
        switch (node->which)
        {
            default:
            case NONE_T:
            case IF_T:
            case FOR_T:
                break;

            case LEAF_T:
                leaf = (leafNode*) node;
                switch (leaf->tok)
                {
                    default:
                        break;

                    case IDENT:       // goto label
                        sym = buildSymbol((leafNode*) node);
                        result = new Label(sym);
                        break;

                    case DEFLT:       // default
                        result = new Label(LT_Default);
                        break;
                }
                break;
    
            case NODE_T:
                switch (node->tok)
                {
                    default:
                        break;

                    case CASE:        // case label
                        expr = buildExpr(node->rnode);
                        result = new Label(expr);
                        break;
                }
                break;
        }
    }

    return result;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Statement*
buildStatement(treeNode *node)
{
    Statement    *result = NULL;

    if (node)
    {
        switch (node->which)
        {
            default:
            case NONE_T:
                break;

            case LEAF_T:
                switch (node->type)
                {
                  default:
                    assert(0);
                    break;

                  case TN_LINE:
                    result = new EndInclStemnt(node);
                    break;

                  case TN_INCL:
                    {
                    leafNode *leaf = (leafNode*) node;
                    result = new InclStemnt(leaf->data.sval->string(), node);
                    }
                    break;
                }
                break;
    
            case NODE_T:
                switch (node->type)
                {
                  default:
                      cerr << "Unhandled statement "
                           << name_of_node(node->type) << endl;
                      assert(0);
                      break;

                  case TN_DECL:
                  case TN_DECLS:
                      {
                      DeclStemnt *ds =  new DeclStemnt(node);
                      result = buildDeclStemnt(ds,node);
                      }
                      break;

                  case TN_BLOCK:
                      if ((node->lnode == NULL) && (node->rnode == NULL))
                        result = new Statement(ST_NullStemnt, node);
                      else
                      {
                        Block *blk = new Block(node);
                        blk->addStatements(node->lnode);
                        blk->addStatements(node->rnode);
                        result = blk;
                      }
                      break;

                  case TN_STEMNT:
                      if ((node->lnode == NULL) && (node->rnode == NULL))
                        result = new Statement(ST_NullStemnt, node);
                      else
                        result = buildStatement(node->rnode);
                      break;

                  case TN_LABEL:
                      {
                      result = buildStatement(node->rnode);
                      Label *lbl = buildLabel(node->lnode); 
                      result->addLabel(lbl);
                      }
                      break;

                  case TN_EXPR:
                  case TN_ASSIGN:
                  case TN_FUNC_CALL:
                  case TN_CAST:
                      {
                      Expression *expr = buildExpr(node);  
                      result = new ExpressionStemnt(expr, node);
                      }
                      break;

                  case TN_SWITCH:
                      {
                      Expression *cond = buildExpr(node->lnode);
                      Statement  *body = buildStatement(node->rnode);
                      result = new SwitchStemnt(cond,body,node);
                      }
                      break;

                  case TN_JUMP:
                      switch (node->tok)
                      {
                        case BREAK:
                          result = new Statement(ST_BreakStemnt, node);
                          break;

                        case CONT:
                          result = new Statement(ST_ContinueStemnt, node);
                          break;

                        case RETURN:
                          {
                          Expression *ret = buildExpr(node->lnode);
                          result = new ReturnStemnt(ret,node);
                          }
                          break;

                        case GOTO:
                          {
                          Symbol *dest = buildSymbol((leafNode*) node->lnode);
                          result = new GotoStemnt(dest, node);
                          }
                          break;
                      }
                      break;

                  case TN_DOWHILE:
                      {
                      Expression *cond = buildExpr(node->lnode);
                      Statement  *body = buildStatement(node->rnode);
                      assert(body);
                      result = new DoWhileStemnt(cond,body, node);
                      }
                      break;

                  case TN_WHILE:
                      {
                      Expression *cond = buildExpr(node->lnode);
                      Statement  *body = buildStatement(node->rnode);
                      assert(body);
                      result = new WhileStemnt(cond,body,node);
                      }
                      break;
                }
                break;

            case IF_T:
                {
                ifNode *ifn = (ifNode*) node;

                Expression *cond = buildExpr(ifn->cond);
                Statement  *thenStemnt = buildStatement(ifn->then_n); 
                Statement  *elseStemnt = NULL;

                if (ifn->else_n)
                    elseStemnt = buildStatement(ifn->else_n); 

                result = new IfStemnt(cond,thenStemnt,node,elseStemnt);
                }
                break;

            case FOR_T:
                switch (node->type)
                {
                  default:
                    cerr << "Unexpected token: " << node->tok << endl;
                    break;

                  case TN_FUNC_DEF:
                    {
                      FunctionDef  *fDef = new FunctionDef(node);
                      forNode *forn = (forNode*) node;

                      fDef->decl = new Decl();
                      buildDecl(fDef->decl,node);                      

                      // Hook the function's symtable entry back to here.
                      if (fDef->FunctionName()->entry != NULL)
                      {
                        fDef->FunctionName()->entry->uFuncDef = fDef;
                      }

                      if (forn->stemnt)
                      {
                        fDef->addStatements(forn->stemnt->lnode);
                        fDef->addStatements(forn->stemnt->rnode);
                      }

                      // Old-style parameter decl.
                      if (forn->incr != NULL)
                      {
                        FunctionType *ft = (FunctionType*) fDef->decl->form;

                        Block *k_and_r = new Block(node);
                        k_and_r->addStatements(forn->incr->lnode);
                        k_and_r->addStatements(forn->incr->rnode);

                        Statement *stemnt;
                        for (stemnt=k_and_r->head; stemnt; stemnt=stemnt->next)
                        {
                            DeclStemnt *decl = (DeclStemnt*) stemnt;

                            for (int j=0; j < decl->nDecls; j++)
                            {
                                ft->fixParam(decl->decls[j]);
                            }
                        }

                        delete k_and_r;
                      }

                      result = fDef; 
                    }
                    break;

                  case TN_FOR:
                    {
                      forNode *forn = (forNode*) node;
      
                      Expression *init = buildExpr(forn->init);
                      Expression *cond = buildExpr(forn->test);
                      Expression *incr = buildExpr(forn->incr);
                      Statement  *body = buildStatement(forn->stemnt); 
      
                      result = new ForStemnt(init,cond,incr,node,body);
                    }
                    break;
                }
        }
    }

    return result;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
DeclStemnt*
buildDeclStemnt(DeclStemnt *stemnt, treeNode *node)
{
    if (node)
    {
      switch (node->which)
      {
        default:
            assert(0);
            break;

        case LEAF_T:
            switch (node->type)
            {
                case TN_TYPE:
                    assert(0);
                    break;

                case TN_IDENT:
                    {
                        Decl *decl = stemnt->base.halfdup();
                        buildDecl(decl,node);

                        // Hook the decl's symtable entry back to here.
                        if (decl->name->entry != NULL)
                        {
                          decl->name->entry->uVarDecl = decl;
                        }

                        stemnt->addDecl(decl);
                    }
                    break;

                default:
                    assert(0);
                    break;
            }
            break;

        case NODE_T:
            switch (node->type)
            {
                case TN_TYPE_LIST:
                    buildBaseType(&(stemnt->base),node);
                    break;

                case TN_DECL:
                case TN_COMP_DECL:
                    // Top-level
                    assert(node->lnode->type == TN_TYPE_LIST);
                    stemnt = buildDeclStemnt(stemnt,node->lnode);
                    stemnt = buildDeclStemnt(stemnt,node->rnode);
                    break;

                case TN_DECLS:
                    // This is actually multiple types/declarations...
                    stemnt = buildDeclStemnt(stemnt,node->lnode);
                    stemnt = buildDeclStemnt(stemnt,node->rnode);
                    break; 

                case TN_PNTR:
                case TN_ARRAY_DECL:
                case TN_FUNC_DECL:
                case TN_ASSIGN:    // an initialized declaration
                    {
                        Decl *decl = stemnt->base.halfdup();
                        buildDecl(decl,node);

                        // Hook the decl's symtable entry back to here.
                        if (decl->name->entry != NULL)
                        {
                          decl->name->entry->uVarDecl = decl;
                        }

                        stemnt->addDecl(decl);
                    }
                    break;

                default:
                    cerr << "Bad Type: "
                         << name_of_node(node->type) << endl;
                    assert(0);
                    break;
            }
            break;
      }
    }

    // Convert to a TypedefStemnt if it's a typedef.
    stemnt = stemnt->convertToTypedef();

    return stemnt;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
ostream&
operator<< (ostream& out, const Statement& stemnt)
{
    stemnt.print(out,0);
    return out;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
#define  SHOW(X)  case X: return #X

char*
nameOfStatementType( StatementType type )
{
    switch (type)
    {
        default:
            return "Unknown StatementType";

        SHOW(ST_NullStemnt);
        SHOW(ST_DeclStemnt);
        SHOW(ST_ExpressionStemnt);
        SHOW(ST_IfStemnt);
        SHOW(ST_SwitchStemnt);
        SHOW(ST_ForStemnt);
        SHOW(ST_WhileStemnt);
        SHOW(ST_DoWhileStemnt);
        SHOW(ST_ContinueStemnt);
        SHOW(ST_BreakStemnt);
        SHOW(ST_GotoStemnt);
        SHOW(ST_ReturnStemnt);
        SHOW(ST_Block);
        SHOW(ST_InclStemnt);
        SHOW(ST_EndInclStemnt);
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
char*
nameOfLabelType( LabelType type )
{
    switch (type)
    {
        default:
            return "Unknown LabelType";

        SHOW(LT_None);        // No label - invalid.
        SHOW(LT_Default);     // default:
        SHOW(LT_Case);        // case <expr>:
        SHOW(LT_Goto);        // A named label (goto destination).
    }
}

#undef SHOW
// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

