/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
    o+
    o+     File:         symbol.cpp
    o+
    o+     Programmer:   Shaun Flisakowski
    o+     Date:         Aug 19, 1998
    o+
    o+     A symbol class.
    o+
    o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "symbol.h"
#include "stemnt.h"

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
SymEntry::SymEntry(SymEntryType _type)
{
    type = _type;
    name = NULL;
    def  = NULL;
    uVarDecl = NULL;
    Container = NULL;
    next = NULL;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
SymEntry::SymEntry(SymEntryType _type, StrItem *sym, treeNode *tn)
{
    type = _type;
    name = sym;
    def  = tn;
    uVarDecl = NULL;
    Container = NULL;
    next = NULL;
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
SymEntry::~SymEntry()
{
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
void
SymEntry::Show(ostream& out) const
{
    out << "Name: " << name->string() << ":\n";

    out << "Definition: ";

    switch (type)
    {
        default:
            out << "Unknown entry type.\n";
            break;

        case FuncdefEntry:
            if (uFuncDef != NULL)
                uFuncDef->print(out,0);
            else
                out << "NO definition for FuncdefEntry\n";
            break;

        case TypedefEntry:
        case VarDeclEntry:
            if (uVarDecl != NULL)
                uVarDecl->print(out,true,0);
            else
                out << "NO definition for VarDeclEntry\n";
            break;

        case EnumConstEntry:
            if (uEnumValue != NULL)
                uEnumValue->print(out);
            else
                out << "NO definition for EnumConstEntry\n";
            break;

        case LabelEntry:
            if (uLabelPosition != NULL)
                uLabelPosition->print(out,0);
            else
                out << "NO definition for LabelEntry\n";
            break;

        case TagEntry:
            if (uStructDef != NULL)
                uStructDef->print(out,true,0);
            else
                out << "NO definition for TagEntry\n";
            break;

        case ComponentEntry:
            if (uComponent != NULL)
                uComponent->print(out,true,0);
            else
                out << "NO definition for ComponentEntry\n";
            break;
    }

    out << endl;

    if (Container != NULL)
    {
        out << "Container:\n";
        Container->print(out,NULL,0);
    }

    out << "\n-------------\n";
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
SymEntry*
mk_typedef(StrItem *sym, treeNode *tn)
{
    SymEntry *ret = new SymEntry(TypedefEntry,sym,tn);
    return ret;
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
SymEntry*
mk_funcdef(StrItem *sym, treeNode *tn)
{
    SymEntry *ret = new SymEntry(FuncdefEntry,sym,tn);
    return ret;
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
SymEntry*
mk_vardecl(StrItem *sym, treeNode *tn)
{
    SymEntry *ret = new SymEntry(VarDeclEntry,sym,tn);
    return ret;
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
SymEntry*
mk_enum_const(StrItem *sym, treeNode *tn)
{
    SymEntry *ret = new SymEntry(EnumConstEntry,sym,tn);
    return ret;
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
SymEntry*
mk_label(StrItem *sym, treeNode *tn)
{
    SymEntry *ret = new SymEntry(LabelEntry,sym,tn);
    return ret;
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
SymEntry*
mk_tag(StrItem *sym, treeNode *tn)
{
    SymEntry *ret = new SymEntry(TagEntry,sym,tn);
    return ret;
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
SymEntry*
mk_component(StrItem *sym, treeNode *tn)
{
    SymEntry *ret = new SymEntry(ComponentEntry,sym,tn);
    return ret;
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
HashTbl::HashTbl(int initSize /* = INIT_HASHTAB_SIZE */)
{
    nent = 0;
    tsize = initSize;
  
    tab = new SymEntry* [tsize];

    for (int j=0; j < tsize; j++)
        tab[j] = NULL;
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
HashTbl::~HashTbl()
{
    for (int j=0; j < tsize; j++)
    {
        SymEntry *curr, *prev = NULL;

        for (curr=tab[j]; curr; curr = curr->next)
        {
            delete prev;
            prev = curr;
        }

        delete prev;
    }

    delete [] tab;
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
SymEntry*
HashTbl::Lookup( StrItem *sym )
{
    int j = sym->hash() % tsize;
  
    for (SymEntry *curr = tab[j]; curr; curr = curr->next)
        if (curr->name->equal(sym)) 
            return curr;

    return NULL;
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
SymEntry*
HashTbl::Insert( SymEntry *entry )
{
    SymEntry *ret;

    if ((ret=Lookup(entry->name)))
        return ret;

    int j = entry->name->hash() % tsize;
    entry->next = tab[j];
    tab[j] = entry;

    nent++;
    return entry;
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
void
HashTbl::Show(ostream& out) const
{
    out << "HashTab:  #entries: " << nent
        << "  size: " << tsize << endl;
  
    for (int j=0; j < tsize; j++)
    {
        SymEntry *list;

        list = tab[j];
        if (list)
        {
            out << "[" << j << "]: ";
            for (; list; list=list->next)
                list->Show(out);
            out << endl;
        }
    }
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
ScopeTbl::ScopeTbl( ScopeTbl *mom /* =NULL */,
                    int initSize /* = INIT_CHILD_SIZE */ )
{
    htab = NULL;
    nchild = 0;
    size = initSize;
    children = new ScopeTbl* [size];

    for (int j=0; j < size; j++)
        children[j] = NULL;

    nsyms = 0;
    parent = mom;

    if (mom)
        level = mom->level + 1;
    else
        level = EXTERN_SCOPE;
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
ScopeTbl::~ScopeTbl()
{
    delete htab;

    /*
    for (int j=0; j < size; j++)
        delete children[j];
    */

    delete [] children;
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
void
ScopeTbl::PostOrderTblDelete()
{
    if (this)
    {
        for (int j=0; j < nchild; j++)
            children[j]->PostOrderTblDelete();

        delete this;
    }
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
SymEntry*
ScopeTbl::Lookup( StrItem *sym )
{
    SymEntry *ret = NULL;

    if (this && htab)
        ret = htab->Lookup(sym);

    return ret;
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
SymEntry*
ScopeTbl::Find( StrItem *sym )
{
    SymEntry *ret = NULL;
  
    if (this)
    {
#ifdef  PRINT_LEVEL
        cout << "Looking for '" << sym->string() << "' in level "
             << level << " (<entry>)\n";
#endif
  
        if (htab)
            ret = htab->Lookup(sym);
  
        if (!ret)
            return parent->Find(sym);
  
#ifdef  PRINT_LEVEL
        cout << "Looking for '" << sym->string() << "' in level "
             << level << " (" << (ret ? "found" : "NOT found") << ")\n";
#endif
    }
  
    return ret;
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
SymEntry*
ScopeTbl::Insert(SymEntry *entry)
{
#ifdef  PRINT_LEVEL
    cout << "Inserting '" << entry->name->string()
         << "' in level " << level << endl;
#endif

    if (!htab)
        htab = new HashTbl();

    nsyms++;

    return htab->Insert(entry);
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
bool
ScopeTbl::ChildInsert(ScopeTbl *kid)
{
    if (nchild >= size)
    {
        ScopeTbl **oldkids = children;
        size += size;
        children = new ScopeTbl* [size];
    
        if (!children)
            return false;
    
        for (int j=0; j < nchild; j++)
            children[j] = oldkids[j];
    }
  
    children[nchild++] = kid;
  
    return true;
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
void
ScopeTbl::Show(ostream& out) const
{
    out << "\nScopeTab (" << (void*) this
        << "): (parent: " << (void*) parent << ")\n";
  
    out << "level: " << level << "  nsyms: " << nsyms
        << "  nchild: " << nchild << endl;
  
    if (htab)
        htab->Show(out);
    else
        out << "HashTab: NULL\n";
  
    for (int j=0; j < nchild; j++)
        children[j]->Show(out);
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
SymTbl::SymTbl()
{
    root = new ScopeTbl();
    clevel = EXTERN_SCOPE;
    current = root;
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
SymTbl::~SymTbl()
{
    /* delete all ScopeTbl's (post-order traversal). */ 
    root->PostOrderTblDelete();
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
SymEntry*
SymTbl::Lookup(StrItem *nme)
{
    SymEntry *ret = NULL;

    if (current)
        ret = current->Find(nme);

    return ret;
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
SymEntry*
SymTbl::LookupAt(StrItem *nme, int level)
{
    ScopeTbl *scp = current;

    while (scp && (scp->level < level))
        scp = scp->parent;

    return scp->Find(nme);
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
SymEntry*
SymTbl::Insert(SymEntry *entry)
{
    while (clevel > current->level)
    {
        ScopeTbl *child = new ScopeTbl(current);
  
        if (!child || !current->ChildInsert(child))
            return NULL;
  
        current = child;
    }
  
    return current->Insert(entry);
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
SymEntry*
SymTbl::InsertUp(SymEntry *entry)
{
    InsertAt( entry, clevel - 1);
  
    return entry;
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
bool
SymTbl::InsertAt(SymEntry *entry, int level)
{
    ScopeTbl *scp;

#ifdef  PRINT_LEVEL
    cerr << "scope level " << current->level << endl;
    cerr << "Current level " << clevel << endl;
    cerr << "request level " << level << endl;
#endif

    while ((clevel > current->level) && (clevel >= level))
    {
        ScopeTbl *child = new ScopeTbl(current);

        if (!child || !current->ChildInsert(child))
            return false;

        current = child;
    }

    scp = current;
    while (scp && (scp->level > level))
        scp = scp->parent;
  
    if (scp)
        return (scp->Insert(entry) != NULL);
    else
        return false;
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
int
SymTbl::EnterScope()
{
    return ++clevel;
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
void
SymTbl::ExitScope()
{
    clevel--;
    if (current->level > clevel)
        current = current->parent;
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
void
SymTbl::Show(ostream& out) const
{
    out << "\n-------------\nSymbol Table:\n";

    out << "current level: " << clevel
        << "  (" << (void*) current
        << ")  parent: (" << (void*) current->parent << ")\n";

    root->Show(out);

    out << "\n-------------\n";
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
Context::Context()
{
    labels = new SymTbl();
    tags   = new SymTbl();
    syms   = new SymTbl();
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
Context::~Context()
{
    delete labels;
    delete tags;
    delete syms;
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
int
Context::EnterScope()
{
#ifdef  PRINT_LEVEL
    cout << "Entering Scope: " << (syms->clevel + 1) << endl;
#endif
    labels->EnterScope();
    tags->EnterScope();
    return syms->EnterScope();
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
void
Context::ExitScope()
{
#ifdef  PRINT_LEVEL
    /*  cout << "Exiting Scope: " << (syms->clevel) << endl;  */
#endif
    labels->ExitScope();
    tags->ExitScope();
    syms->ExitScope();
}

/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */
void
Context::ExitScopes(int newlev)
{
    if (newlev < EXTERN_SCOPE)
        newlev = EXTERN_SCOPE;

    while (newlev < syms->current->level)
        ExitScope();
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Symbol::Symbol()
{
    name = NULL;
    entry = NULL;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Symbol::~Symbol()
{
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Symbol*
Symbol::dup() const
{
    Symbol *sym = NULL;

    if (this)
    {
        sym = new Symbol();

        sym->name  = name;
        sym->entry = entry;
    }

    return sym;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
ostream&
operator<< ( ostream& out, const Symbol& sym )
{
    //static int inSymShow = 0;

    if (sym.name)
    {
        out << sym.name->string();

        /*
        if (sym.entry && !inSymShow)
        {
            inSymShow = 1;
            out << "[";
            sym.entry->Show(out);
            out << "]";
            inSymShow = 0;
        }
        */
    }
    else
        out << "<bad symbol>";

    return out;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Symbol*
buildSymbol(leafNode *leaf)
{
    Symbol *result = NULL;

    if (leaf)
    {
        assert((leaf->type == TN_IDENT) || (leaf->type == TN_TYPE));

        result = new Symbol();

        result->name  = leaf->data.sval;
        result->entry = leaf->syment;
    }

    return result;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

