
functor GenLrValsFun(structure Token : TOKEN
				  structure SymbolTable : SymbolTable)
              : sig structure ParserData : PARSER_DATA
                    structure Tokens : Gen_TOKENS
                end = 
struct
structure ParserData=
struct
structure Header = 
struct
(* Copyright (c) 1997,1998 The Regents of the University of California.
   All rights reserved.

Permission to use, copy, modify, and distribute this software for any
purpose, without fee, and without written agreement is hereby granted,
provided that the above copyright notice and the following two
paragraphs appear in all copies of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

   ML-Yacc grammar for C. Based on grammar by Roskind
   Adapted from Guhan Viswanathan's (gviswana@cs.wisc.edu) Css ("C**")
   parser. See ~/Guhan/css.grm

*)
        structure N = Atom

	structure SymbolTable = SymbolTable
	structure AST = SymbolTable.AST

        type identifier = N.atom
        val cmpId = N.compare
        datatype longShort = Long | Short | LongLong
        datatype signed = Unsigned | Signed


        fun formTySpec err (lsO, suO, tsO) =
            let
      	val ts' = 
      	      (case (lsO, tsO) of
      		(NONE, NONE)		       => AST.Int
      	      | (NONE, SOME ts)		       => ts
      	      | (SOME(Short), NONE) 	       => AST.ShortInt
      	      |	(SOME(Short), SOME(AST.Int))   => AST.ShortInt
      	      | (SOME(LongLong), NONE) 	       => AST.LongLong
      	      | (SOME(LongLong), SOME(AST.Int))=> AST.LongLong
      	      | (SOME(Long), NONE) 	       => AST.LongInt
      	      |	(SOME(Long), SOME(AST.Int))    => AST.LongInt
      	      |	(SOME(Long), SOME(AST.Double)) => AST.LongDouble
      	      | _ => (err "ill-formed type specifier";
		      AST.Int))
      	val ts'' = 
      	      (case (suO, ts') of
      		(NONE, _)	             => ts'
      	      |	(SOME(Signed), AST.Int)      => ts'
      	      |	(SOME(Signed), AST.LongInt)  => ts'
      	      |	(SOME(Signed), AST.LongLong) => ts'
      	      |	(SOME(Signed), AST.ShortInt) => ts'
      	      |	(SOME(Signed), AST.Char)     => ts'
      	      |	(SOME(Unsigned), AST.Int)    => AST.UnsignedInt
      	      |	(SOME(Unsigned), AST.LongInt)  => AST.UnsignedLongInt
      	      |	(SOME(Unsigned), AST.LongLong)  => AST.UnsignedLongLong
      	      |	(SOME(Unsigned), AST.ShortInt) => AST.UnsignedShortInt
      	      |	(SOME(Unsigned), AST.Char)     => AST.UnsignedChar
      	      | _ => (err "ill-formed type specifier";
		      AST.Int))
            in
      	ts''
            end

        fun getParams err (AST.Function(params)::returnType) = SOME(params,returnType)
	  | getParams err (AST.OldFunction([],NONE)::returnType) = SOME([],returnType)
	  | getParams err (AST.OldFunction(_)::_) = 
	       (err "Invalid old-style function declaration\n"; NONE)
          | getParams err _ = NONE
      	

        fun buildFunctionDefinition (region,error) 
	    ((sc, ts, tq), (idregion,id, dL), oldParamDecls,
	     (decls, body, scop)) =
      	let
      	  val return = AST.SyntacticType(sc, ts, tq, List.tl dL) 
      	  val typ = AST.SyntacticType(sc, ts, tq, dL)
      	in
	    case (dL, oldParamDecls) of
		(AST.Function params :: _, []) =>
		    (SymbolTable.insertId id typ;
		     AST.FunDef({name   = id,
				 typ    = typ,
				 region = idregion,
				 return = return,
				 params = params,
				 decls  = decls,
				 body   = body,
				 lookup = SymbolTable.scopeLookup scop}))
	      | (AST.Function params :: _, _) =>
		    (error region "bad combination of new and old-style function definition";
		     raise (Fail "bad function definition"))
	      | (AST.OldFunction (params,ellipsis) :: rest, paramDecls) => 
		    let open AST
			fun paramType (idregion,id) =
			    let fun find [] = (* default int *)
				    VarDecl {name=id,
					     region=idregion,
					     typ=SyntacticType (NONE,Int,[],[]),
					     init=NONE}
				  | find ((vd as VarDecl {name,...})::rest) =
				    if IdKey.compare (name,id) = EQUAL then
					vd
				    else find rest
				  | find ((fd as FunDecl {name,...})::rest) =
			            if IdKey.compare (name,id) = EQUAL then
					fd
				    else find rest
				  | find (_::rest) =
					(error region "Invalid declaration in paramter list";
					 find rest)

					
			    in
				find paramDecls
			    end

			val params'  = List.map paramType params
			val params'' = case ellipsis of
			                 SOME region => params' @ [EllipsisDecl {region=region}]
				       | NONE => params'
			val newType =
			    AST.SyntacticType(sc, ts, tq,
					      (AST.Function params'')::rest)
		    in
			SymbolTable.insertId id newType;
			AST.FunDef({name   = id,
				    region = idregion,
				    return = return,
				    typ    = newType,
				    params = params'',
				    decls  = decls,
				    body   = body,
				    lookup = SymbolTable.scopeLookup scop})
		    end

	      | _ =>
		    (error region "function definition w/o parameter list";
		     raise (Fail "bad function definition"))
      	end


	fun mergeStorageClasses err (sc0, sc1) =
	    case (sc0, sc1) of
		(SOME(sc), NONE) => SOME(sc)
	      | (NONE, SOME(sc)) => SOME(sc)
	      | (NONE, NONE)     => NONE
	      | _	         => 
		    (err "multiple storage classes";
		     NONE)



	fun mergeTypeQualifiers err (tqnew, tql) =
	    (if List.exists (fn tq => AST.typeQualEq (tq, tqnew)) tql then
		 (err "redundant type qualifier";
		  tql)
	     else
		 tqnew::tql)

	fun mergeBaseTypeSpecifiers err (ts0, ts1) =
	    let val (sl, su, tsO) = ts0
		val (sl', su', tsO') = ts1
		val sl'' = (case (sl, sl') of
				(x, NONE) => x
			      | (NONE, x) => x
			      | (SOME(Short), SOME(Short))  => 
				    (err "duplicate short";
				     sl)
			      | (SOME(Long), SOME(Long))  => SOME LongLong
			      | _ => 
				    (err "bad short/long";
				     sl))
		val su'' = (case (su, su') of
				(x, NONE) => x
			      | (NONE, x) => x
			      | (SOME(Signed), SOME(Signed))  => 
				    (err "duplicate signed";
				     su)
			      | (SOME(Unsigned), SOME(Unsigned))  =>
				    (err "duplicate unsigned";
				     su)
			      | _ =>
				    (err "bad signed/unsigned";
				     su))
		val tsO'' = (case (tsO, tsO') of
				 (NONE, x) => x
			       | (x, NONE) => x
			       | _ => 
      				     (err "multiple type specifiers";
				      tsO))
	    in
		(sl'', su'', tsO'')
	    end

	    (* Given an enumerator, add all the enumerated names
	       to the symbol table *)
	    fun	addEnum (e as AST.Enum(idOpt, enumeratorList)) =
		let val enumType = AST.SyntacticType (NONE, e, [], [])
		    fun addElt (id, _) = SymbolTable.insertEnum id enumType
		in
		    app addElt enumeratorList
		end

  
end
structure LrTable = Token.LrTable
structure Token = Token
local open LrTable in 
val table=let val actionRows =
"\
\\001\000\001\000\238\001\002\000\238\001\003\000\238\001\006\000\238\001\
\\007\000\238\001\008\000\238\001\010\000\238\001\011\000\238\001\
\\013\000\238\001\029\000\238\001\033\000\238\001\034\000\238\001\
\\035\000\238\001\036\000\238\001\040\000\238\001\041\000\238\001\
\\042\000\238\001\043\000\238\001\045\000\238\001\046\000\238\001\
\\047\000\238\001\048\000\238\001\049\000\238\001\050\000\238\001\
\\051\000\238\001\052\000\238\001\053\000\238\001\054\000\238\001\
\\055\000\238\001\056\000\238\001\057\000\238\001\058\000\238\001\
\\059\000\238\001\060\000\238\001\061\000\238\001\062\000\238\001\
\\063\000\238\001\064\000\238\001\065\000\238\001\066\000\238\001\
\\067\000\238\001\069\000\238\001\070\000\238\001\071\000\238\001\
\\072\000\238\001\073\000\238\001\074\000\238\001\075\000\238\001\
\\076\000\238\001\077\000\238\001\078\000\238\001\079\000\238\001\
\\083\000\238\001\084\000\238\001\085\000\238\001\086\000\238\001\000\000\
\\001\000\001\000\239\001\002\000\239\001\003\000\239\001\006\000\239\001\
\\007\000\239\001\008\000\239\001\010\000\239\001\011\000\239\001\
\\013\000\239\001\029\000\239\001\033\000\239\001\034\000\239\001\
\\035\000\239\001\036\000\239\001\040\000\239\001\041\000\239\001\
\\042\000\239\001\043\000\239\001\045\000\239\001\046\000\239\001\
\\047\000\239\001\048\000\239\001\049\000\239\001\050\000\239\001\
\\051\000\239\001\052\000\239\001\053\000\239\001\054\000\239\001\
\\055\000\239\001\056\000\239\001\057\000\239\001\058\000\239\001\
\\059\000\239\001\060\000\239\001\061\000\239\001\062\000\239\001\
\\063\000\239\001\064\000\239\001\065\000\239\001\066\000\239\001\
\\067\000\239\001\069\000\239\001\070\000\239\001\071\000\239\001\
\\072\000\239\001\073\000\239\001\074\000\239\001\075\000\239\001\
\\076\000\239\001\077\000\239\001\078\000\239\001\079\000\239\001\
\\083\000\239\001\084\000\239\001\085\000\239\001\086\000\239\001\000\000\
\\001\000\001\000\076\002\002\000\076\002\003\000\076\002\006\000\076\002\
\\007\000\076\002\008\000\076\002\010\000\076\002\011\000\076\002\
\\013\000\076\002\029\000\076\002\033\000\076\002\034\000\076\002\
\\035\000\076\002\040\000\076\002\041\000\076\002\042\000\076\002\
\\043\000\076\002\067\000\076\002\069\000\076\002\070\000\076\002\
\\071\000\076\002\072\000\076\002\073\000\076\002\074\000\076\002\
\\075\000\076\002\076\000\076\002\077\000\076\002\078\000\076\002\
\\079\000\076\002\083\000\076\002\086\000\076\002\000\000\
\\001\000\001\000\077\002\002\000\077\002\003\000\077\002\006\000\077\002\
\\007\000\077\002\008\000\077\002\010\000\077\002\011\000\077\002\
\\013\000\077\002\029\000\077\002\033\000\077\002\034\000\077\002\
\\035\000\077\002\040\000\077\002\041\000\077\002\042\000\077\002\
\\043\000\077\002\067\000\077\002\069\000\077\002\070\000\077\002\
\\071\000\077\002\072\000\077\002\073\000\077\002\074\000\077\002\
\\075\000\077\002\076\000\077\002\077\000\077\002\078\000\077\002\
\\079\000\077\002\083\000\077\002\086\000\077\002\000\000\
\\001\000\001\000\078\002\002\000\078\002\003\000\078\002\006\000\078\002\
\\007\000\078\002\008\000\078\002\010\000\078\002\011\000\078\002\
\\013\000\078\002\029\000\078\002\030\000\078\002\033\000\078\002\
\\034\000\078\002\035\000\078\002\036\000\078\002\039\000\078\002\
\\040\000\078\002\041\000\078\002\042\000\078\002\043\000\078\002\
\\045\000\078\002\046\000\078\002\047\000\078\002\048\000\078\002\
\\049\000\078\002\050\000\078\002\051\000\078\002\052\000\078\002\
\\053\000\078\002\054\000\078\002\055\000\078\002\056\000\078\002\
\\057\000\078\002\058\000\078\002\059\000\078\002\060\000\078\002\
\\061\000\078\002\062\000\078\002\063\000\078\002\064\000\078\002\
\\065\000\078\002\066\000\078\002\067\000\078\002\068\000\078\002\
\\069\000\078\002\070\000\078\002\071\000\078\002\072\000\078\002\
\\073\000\078\002\074\000\078\002\075\000\078\002\076\000\078\002\
\\077\000\078\002\078\000\078\002\079\000\078\002\083\000\078\002\
\\084\000\078\002\085\000\078\002\086\000\078\002\087\000\078\002\000\000\
\\001\000\001\000\079\002\002\000\079\002\003\000\079\002\006\000\079\002\
\\007\000\079\002\008\000\079\002\010\000\079\002\011\000\079\002\
\\013\000\079\002\029\000\079\002\030\000\079\002\033\000\079\002\
\\034\000\079\002\035\000\079\002\036\000\079\002\039\000\079\002\
\\040\000\079\002\041\000\079\002\042\000\079\002\043\000\079\002\
\\045\000\079\002\046\000\079\002\047\000\079\002\048\000\079\002\
\\049\000\079\002\050\000\079\002\051\000\079\002\052\000\079\002\
\\053\000\079\002\054\000\079\002\055\000\079\002\056\000\079\002\
\\057\000\079\002\058\000\079\002\059\000\079\002\060\000\079\002\
\\061\000\079\002\062\000\079\002\063\000\079\002\064\000\079\002\
\\065\000\079\002\066\000\079\002\067\000\079\002\068\000\079\002\
\\069\000\079\002\070\000\079\002\071\000\079\002\072\000\079\002\
\\073\000\079\002\074\000\079\002\075\000\079\002\076\000\079\002\
\\077\000\079\002\078\000\079\002\079\000\079\002\083\000\079\002\
\\084\000\079\002\085\000\079\002\086\000\079\002\087\000\079\002\000\000\
\\001\000\001\000\080\002\002\000\080\002\003\000\080\002\006\000\080\002\
\\007\000\080\002\008\000\080\002\010\000\080\002\011\000\080\002\
\\013\000\080\002\029\000\080\002\030\000\080\002\033\000\080\002\
\\034\000\080\002\035\000\080\002\036\000\080\002\039\000\080\002\
\\040\000\080\002\041\000\080\002\042\000\080\002\043\000\080\002\
\\045\000\080\002\046\000\080\002\047\000\080\002\048\000\080\002\
\\049\000\080\002\050\000\080\002\051\000\080\002\052\000\080\002\
\\053\000\080\002\054\000\080\002\055\000\080\002\056\000\080\002\
\\057\000\080\002\058\000\080\002\059\000\080\002\060\000\080\002\
\\061\000\080\002\062\000\080\002\063\000\080\002\064\000\080\002\
\\065\000\080\002\066\000\080\002\067\000\080\002\068\000\080\002\
\\069\000\080\002\070\000\080\002\071\000\080\002\072\000\080\002\
\\073\000\080\002\074\000\080\002\075\000\080\002\076\000\080\002\
\\077\000\080\002\078\000\080\002\079\000\080\002\083\000\080\002\
\\084\000\080\002\085\000\080\002\086\000\080\002\087\000\080\002\000\000\
\\001\000\001\000\081\002\002\000\081\002\003\000\081\002\006\000\081\002\
\\007\000\081\002\008\000\081\002\010\000\081\002\011\000\081\002\
\\013\000\081\002\029\000\081\002\030\000\081\002\033\000\081\002\
\\034\000\081\002\035\000\081\002\036\000\081\002\039\000\081\002\
\\040\000\081\002\041\000\081\002\042\000\081\002\043\000\081\002\
\\045\000\081\002\046\000\081\002\047\000\081\002\048\000\081\002\
\\049\000\081\002\050\000\081\002\051\000\081\002\052\000\081\002\
\\053\000\081\002\054\000\081\002\055\000\081\002\056\000\081\002\
\\057\000\081\002\058\000\081\002\059\000\081\002\060\000\081\002\
\\061\000\081\002\062\000\081\002\063\000\081\002\064\000\081\002\
\\065\000\081\002\066\000\081\002\067\000\081\002\068\000\081\002\
\\069\000\081\002\070\000\081\002\071\000\081\002\072\000\081\002\
\\073\000\081\002\074\000\081\002\075\000\081\002\076\000\081\002\
\\077\000\081\002\078\000\081\002\079\000\081\002\083\000\081\002\
\\084\000\081\002\085\000\081\002\086\000\081\002\087\000\081\002\000\000\
\\001\000\001\000\082\002\002\000\082\002\003\000\082\002\006\000\082\002\
\\007\000\082\002\008\000\082\002\010\000\082\002\011\000\082\002\
\\013\000\082\002\029\000\082\002\033\000\082\002\034\000\082\002\
\\035\000\082\002\040\000\082\002\041\000\082\002\042\000\082\002\
\\043\000\082\002\067\000\082\002\068\000\082\002\069\000\082\002\
\\070\000\082\002\071\000\082\002\072\000\082\002\073\000\082\002\
\\074\000\082\002\075\000\082\002\076\000\082\002\077\000\082\002\
\\078\000\082\002\079\000\082\002\083\000\082\002\086\000\082\002\000\000\
\\001\000\001\000\083\002\002\000\083\002\003\000\083\002\006\000\083\002\
\\007\000\083\002\008\000\083\002\010\000\083\002\011\000\083\002\
\\013\000\083\002\029\000\083\002\033\000\083\002\034\000\083\002\
\\035\000\083\002\040\000\083\002\041\000\083\002\042\000\083\002\
\\043\000\083\002\067\000\083\002\068\000\083\002\069\000\083\002\
\\070\000\083\002\071\000\083\002\072\000\083\002\073\000\083\002\
\\074\000\083\002\075\000\083\002\076\000\083\002\077\000\083\002\
\\078\000\083\002\079\000\083\002\083\000\083\002\086\000\083\002\000\000\
\\001\000\001\000\084\002\002\000\084\002\003\000\084\002\006\000\084\002\
\\007\000\084\002\008\000\084\002\010\000\084\002\011\000\084\002\
\\013\000\084\002\029\000\084\002\033\000\084\002\034\000\084\002\
\\035\000\084\002\040\000\084\002\041\000\084\002\042\000\084\002\
\\043\000\084\002\067\000\084\002\068\000\084\002\069\000\084\002\
\\070\000\084\002\071\000\084\002\072\000\084\002\073\000\084\002\
\\074\000\084\002\075\000\084\002\076\000\084\002\077\000\084\002\
\\078\000\084\002\079\000\084\002\083\000\084\002\086\000\084\002\000\000\
\\001\000\001\000\085\002\002\000\085\002\003\000\085\002\006\000\085\002\
\\007\000\085\002\008\000\085\002\010\000\085\002\011\000\085\002\
\\013\000\085\002\029\000\085\002\033\000\085\002\034\000\085\002\
\\035\000\085\002\040\000\085\002\041\000\085\002\042\000\085\002\
\\043\000\085\002\067\000\085\002\068\000\085\002\069\000\085\002\
\\070\000\085\002\071\000\085\002\072\000\085\002\073\000\085\002\
\\074\000\085\002\075\000\085\002\076\000\085\002\077\000\085\002\
\\078\000\085\002\079\000\085\002\083\000\085\002\086\000\085\002\000\000\
\\001\000\001\000\086\002\002\000\086\002\003\000\086\002\006\000\086\002\
\\007\000\086\002\008\000\086\002\010\000\086\002\011\000\086\002\
\\013\000\086\002\029\000\086\002\033\000\086\002\034\000\086\002\
\\035\000\086\002\040\000\086\002\041\000\086\002\042\000\086\002\
\\043\000\086\002\067\000\086\002\068\000\086\002\069\000\086\002\
\\070\000\086\002\071\000\086\002\072\000\086\002\073\000\086\002\
\\074\000\086\002\075\000\086\002\076\000\086\002\077\000\086\002\
\\078\000\086\002\079\000\086\002\083\000\086\002\086\000\086\002\000\000\
\\001\000\001\000\087\002\002\000\087\002\003\000\087\002\006\000\087\002\
\\007\000\087\002\008\000\087\002\010\000\087\002\011\000\087\002\
\\013\000\087\002\029\000\087\002\033\000\087\002\034\000\087\002\
\\035\000\087\002\040\000\087\002\041\000\087\002\042\000\087\002\
\\043\000\087\002\067\000\087\002\068\000\087\002\069\000\087\002\
\\070\000\087\002\071\000\087\002\072\000\087\002\073\000\087\002\
\\074\000\087\002\075\000\087\002\076\000\087\002\077\000\087\002\
\\078\000\087\002\079\000\087\002\083\000\087\002\086\000\087\002\000\000\
\\001\000\001\000\088\002\002\000\088\002\003\000\088\002\006\000\088\002\
\\007\000\088\002\008\000\088\002\010\000\088\002\011\000\088\002\
\\013\000\088\002\029\000\088\002\033\000\088\002\034\000\088\002\
\\035\000\088\002\040\000\088\002\041\000\088\002\042\000\088\002\
\\043\000\088\002\067\000\088\002\068\000\088\002\069\000\088\002\
\\070\000\088\002\071\000\088\002\072\000\088\002\073\000\088\002\
\\074\000\088\002\075\000\088\002\076\000\088\002\077\000\088\002\
\\078\000\088\002\079\000\088\002\083\000\088\002\086\000\088\002\000\000\
\\001\000\001\000\089\002\002\000\089\002\003\000\089\002\006\000\089\002\
\\007\000\089\002\008\000\089\002\010\000\089\002\011\000\089\002\
\\013\000\089\002\029\000\089\002\033\000\089\002\034\000\089\002\
\\035\000\089\002\040\000\089\002\041\000\089\002\042\000\089\002\
\\043\000\089\002\067\000\089\002\068\000\089\002\069\000\089\002\
\\070\000\089\002\071\000\089\002\072\000\089\002\073\000\089\002\
\\074\000\089\002\075\000\089\002\076\000\089\002\077\000\089\002\
\\078\000\089\002\079\000\089\002\083\000\089\002\086\000\089\002\000\000\
\\001\000\001\000\090\002\002\000\090\002\003\000\090\002\006\000\090\002\
\\007\000\090\002\008\000\090\002\010\000\090\002\011\000\090\002\
\\013\000\090\002\029\000\090\002\033\000\090\002\034\000\090\002\
\\035\000\090\002\040\000\090\002\041\000\090\002\042\000\090\002\
\\043\000\090\002\067\000\090\002\068\000\172\001\069\000\090\002\
\\070\000\090\002\071\000\090\002\072\000\090\002\073\000\090\002\
\\074\000\090\002\075\000\090\002\076\000\090\002\077\000\090\002\
\\078\000\090\002\079\000\090\002\083\000\090\002\086\000\090\002\000\000\
\\001\000\001\000\091\002\002\000\091\002\003\000\091\002\006\000\091\002\
\\007\000\091\002\008\000\091\002\010\000\091\002\011\000\091\002\
\\013\000\091\002\029\000\091\002\033\000\091\002\034\000\091\002\
\\035\000\091\002\040\000\091\002\041\000\091\002\042\000\091\002\
\\043\000\091\002\067\000\091\002\068\000\091\002\069\000\091\002\
\\070\000\091\002\071\000\091\002\072\000\091\002\073\000\091\002\
\\074\000\091\002\075\000\091\002\076\000\091\002\077\000\091\002\
\\078\000\091\002\079\000\091\002\083\000\091\002\086\000\091\002\000\000\
\\001\000\001\000\092\002\002\000\092\002\003\000\092\002\006\000\092\002\
\\007\000\092\002\008\000\092\002\010\000\092\002\011\000\092\002\
\\013\000\092\002\029\000\092\002\033\000\092\002\034\000\092\002\
\\035\000\092\002\040\000\092\002\041\000\092\002\042\000\092\002\
\\043\000\092\002\067\000\092\002\068\000\092\002\069\000\092\002\
\\070\000\092\002\071\000\092\002\072\000\092\002\073\000\092\002\
\\074\000\092\002\075\000\092\002\076\000\092\002\077\000\092\002\
\\078\000\092\002\079\000\092\002\083\000\092\002\086\000\092\002\000\000\
\\001\000\001\000\093\002\002\000\093\002\003\000\093\002\006\000\093\002\
\\007\000\093\002\008\000\093\002\010\000\093\002\011\000\093\002\
\\013\000\093\002\029\000\093\002\033\000\093\002\034\000\093\002\
\\035\000\093\002\040\000\093\002\041\000\093\002\042\000\093\002\
\\043\000\093\002\067\000\093\002\068\000\093\002\069\000\093\002\
\\070\000\093\002\071\000\093\002\072\000\093\002\073\000\093\002\
\\074\000\093\002\075\000\093\002\076\000\093\002\077\000\093\002\
\\078\000\093\002\079\000\093\002\083\000\093\002\086\000\093\002\000\000\
\\001\000\001\000\094\002\002\000\094\002\003\000\094\002\006\000\094\002\
\\007\000\094\002\008\000\094\002\010\000\094\002\011\000\094\002\
\\013\000\094\002\029\000\094\002\033\000\094\002\034\000\094\002\
\\035\000\094\002\040\000\094\002\041\000\094\002\042\000\094\002\
\\043\000\094\002\067\000\094\002\068\000\094\002\069\000\094\002\
\\070\000\094\002\071\000\094\002\072\000\094\002\073\000\094\002\
\\074\000\094\002\075\000\094\002\076\000\094\002\077\000\094\002\
\\078\000\094\002\079\000\094\002\083\000\094\002\086\000\094\002\000\000\
\\001\000\001\000\095\002\002\000\095\002\003\000\095\002\006\000\095\002\
\\007\000\095\002\008\000\095\002\010\000\095\002\011\000\095\002\
\\013\000\095\002\029\000\095\002\033\000\095\002\034\000\095\002\
\\035\000\095\002\040\000\095\002\041\000\095\002\042\000\095\002\
\\043\000\095\002\067\000\095\002\068\000\095\002\069\000\095\002\
\\070\000\095\002\071\000\095\002\072\000\095\002\073\000\095\002\
\\074\000\095\002\075\000\095\002\076\000\095\002\077\000\095\002\
\\078\000\095\002\079\000\095\002\083\000\095\002\086\000\095\002\000\000\
\\001\000\001\000\096\002\002\000\096\002\003\000\096\002\006\000\096\002\
\\007\000\096\002\008\000\096\002\010\000\096\002\011\000\096\002\
\\013\000\096\002\029\000\096\002\033\000\096\002\034\000\096\002\
\\035\000\096\002\040\000\096\002\041\000\096\002\042\000\096\002\
\\043\000\096\002\067\000\096\002\068\000\096\002\069\000\096\002\
\\070\000\096\002\071\000\096\002\072\000\096\002\073\000\096\002\
\\074\000\096\002\075\000\096\002\076\000\096\002\077\000\096\002\
\\078\000\096\002\079\000\096\002\083\000\096\002\086\000\096\002\000\000\
\\001\000\001\000\097\002\002\000\097\002\003\000\097\002\006\000\097\002\
\\007\000\097\002\008\000\097\002\010\000\097\002\011\000\097\002\
\\013\000\097\002\029\000\097\002\033\000\097\002\034\000\097\002\
\\035\000\097\002\040\000\097\002\041\000\097\002\042\000\097\002\
\\043\000\097\002\067\000\097\002\068\000\097\002\069\000\097\002\
\\070\000\097\002\071\000\097\002\072\000\097\002\073\000\097\002\
\\074\000\097\002\075\000\097\002\076\000\097\002\077\000\097\002\
\\078\000\097\002\079\000\097\002\083\000\097\002\086\000\097\002\000\000\
\\001\000\001\000\098\002\002\000\098\002\003\000\098\002\006\000\098\002\
\\007\000\098\002\008\000\098\002\010\000\098\002\011\000\098\002\
\\013\000\098\002\029\000\098\002\033\000\098\002\034\000\098\002\
\\035\000\098\002\040\000\098\002\041\000\098\002\042\000\098\002\
\\043\000\098\002\067\000\098\002\068\000\098\002\069\000\098\002\
\\070\000\098\002\071\000\098\002\072\000\098\002\073\000\098\002\
\\074\000\098\002\075\000\098\002\076\000\098\002\077\000\098\002\
\\078\000\098\002\079\000\098\002\083\000\098\002\086\000\098\002\000\000\
\\001\000\001\000\099\002\002\000\099\002\003\000\099\002\006\000\099\002\
\\007\000\099\002\008\000\099\002\010\000\099\002\011\000\099\002\
\\013\000\099\002\029\000\099\002\033\000\099\002\034\000\099\002\
\\035\000\099\002\040\000\099\002\041\000\099\002\042\000\099\002\
\\043\000\099\002\067\000\099\002\068\000\099\002\069\000\099\002\
\\070\000\099\002\071\000\099\002\072\000\099\002\073\000\099\002\
\\074\000\099\002\075\000\099\002\076\000\099\002\077\000\099\002\
\\078\000\099\002\079\000\099\002\083\000\099\002\086\000\099\002\000\000\
\\001\000\001\000\100\002\002\000\100\002\003\000\100\002\006\000\100\002\
\\007\000\100\002\008\000\100\002\010\000\100\002\011\000\100\002\
\\013\000\100\002\029\000\100\002\033\000\100\002\034\000\100\002\
\\035\000\100\002\040\000\100\002\041\000\100\002\042\000\100\002\
\\043\000\100\002\067\000\100\002\068\000\100\002\069\000\100\002\
\\070\000\100\002\071\000\100\002\072\000\100\002\073\000\100\002\
\\074\000\100\002\075\000\100\002\076\000\100\002\077\000\100\002\
\\078\000\100\002\079\000\100\002\083\000\100\002\086\000\100\002\000\000\
\\001\000\001\000\101\002\002\000\101\002\003\000\101\002\006\000\101\002\
\\007\000\101\002\008\000\101\002\010\000\101\002\011\000\101\002\
\\013\000\101\002\029\000\101\002\033\000\101\002\035\000\101\002\
\\040\000\101\002\041\000\101\002\042\000\101\002\043\000\101\002\
\\067\000\101\002\069\000\101\002\070\000\101\002\071\000\101\002\
\\072\000\101\002\073\000\101\002\074\000\101\002\075\000\101\002\
\\076\000\101\002\077\000\101\002\078\000\101\002\079\000\101\002\
\\083\000\101\002\086\000\101\002\000\000\
\\001\000\001\000\102\002\002\000\102\002\003\000\102\002\006\000\102\002\
\\007\000\102\002\008\000\102\002\010\000\102\002\011\000\102\002\
\\013\000\102\002\029\000\102\002\033\000\102\002\035\000\102\002\
\\040\000\102\002\041\000\102\002\042\000\102\002\043\000\102\002\
\\067\000\102\002\069\000\102\002\070\000\102\002\071\000\102\002\
\\072\000\102\002\073\000\102\002\074\000\102\002\075\000\102\002\
\\076\000\102\002\077\000\102\002\078\000\102\002\079\000\102\002\
\\083\000\102\002\086\000\102\002\000\000\
\\001\000\001\000\103\002\002\000\103\002\003\000\103\002\006\000\103\002\
\\007\000\103\002\008\000\103\002\010\000\103\002\011\000\103\002\
\\013\000\103\002\029\000\103\002\033\000\103\002\035\000\103\002\
\\040\000\103\002\041\000\103\002\042\000\103\002\043\000\103\002\
\\067\000\103\002\069\000\103\002\070\000\103\002\071\000\103\002\
\\072\000\103\002\073\000\103\002\074\000\103\002\075\000\103\002\
\\076\000\103\002\077\000\103\002\078\000\103\002\079\000\103\002\
\\083\000\103\002\086\000\103\002\000\000\
\\001\000\001\000\104\002\002\000\104\002\003\000\104\002\006\000\104\002\
\\007\000\104\002\008\000\104\002\010\000\104\002\011\000\104\002\
\\013\000\104\002\029\000\104\002\033\000\104\002\035\000\104\002\
\\040\000\104\002\041\000\104\002\042\000\104\002\043\000\104\002\
\\067\000\104\002\069\000\104\002\070\000\104\002\071\000\104\002\
\\072\000\104\002\073\000\104\002\074\000\104\002\075\000\104\002\
\\076\000\104\002\077\000\104\002\078\000\104\002\079\000\104\002\
\\083\000\104\002\086\000\104\002\000\000\
\\001\000\001\000\115\002\002\000\115\002\003\000\115\002\004\000\115\002\
\\005\000\115\002\008\000\115\002\009\000\115\002\011\000\115\002\
\\012\000\115\002\014\000\115\002\015\000\115\002\016\000\115\002\
\\017\000\115\002\018\000\115\002\019\000\115\002\020\000\115\002\
\\021\000\115\002\022\000\115\002\023\000\014\001\026\000\115\002\
\\027\000\115\002\028\000\115\002\030\000\115\002\032\000\115\002\
\\039\000\115\002\079\000\115\002\000\000\
\\001\000\001\000\115\002\002\000\115\002\003\000\115\002\004\000\115\002\
\\005\000\115\002\008\000\115\002\009\000\115\002\011\000\115\002\
\\012\000\115\002\014\000\115\002\015\000\115\002\016\000\115\002\
\\017\000\115\002\018\000\115\002\019\000\115\002\020\000\115\002\
\\021\000\115\002\022\000\115\002\026\000\115\002\027\000\115\002\
\\028\000\115\002\030\000\115\002\032\000\115\002\034\000\115\002\
\\039\000\115\002\079\000\115\002\000\000\
\\001\000\001\000\129\002\002\000\129\002\003\000\179\000\004\000\178\000\
\\005\000\177\000\008\000\129\002\009\000\129\002\011\000\129\002\
\\012\000\129\002\014\000\129\002\015\000\129\002\016\000\129\002\
\\017\000\129\002\018\000\129\002\019\000\129\002\020\000\129\002\
\\021\000\129\002\022\000\129\002\026\000\129\002\027\000\129\002\
\\028\000\129\002\030\000\129\002\032\000\129\002\034\000\129\002\
\\039\000\129\002\079\000\129\002\000\000\
\\001\000\001\000\130\002\002\000\130\002\003\000\179\000\004\000\178\000\
\\005\000\177\000\008\000\130\002\009\000\130\002\011\000\130\002\
\\012\000\130\002\014\000\130\002\015\000\130\002\016\000\130\002\
\\017\000\130\002\018\000\130\002\019\000\130\002\020\000\130\002\
\\021\000\130\002\022\000\130\002\026\000\130\002\027\000\130\002\
\\028\000\130\002\030\000\130\002\032\000\130\002\034\000\130\002\
\\039\000\130\002\079\000\130\002\000\000\
\\001\000\001\000\131\002\002\000\131\002\003\000\131\002\004\000\131\002\
\\005\000\131\002\008\000\131\002\009\000\131\002\011\000\131\002\
\\012\000\131\002\014\000\131\002\015\000\131\002\016\000\131\002\
\\017\000\131\002\018\000\131\002\019\000\131\002\020\000\131\002\
\\021\000\131\002\022\000\131\002\026\000\131\002\027\000\131\002\
\\028\000\131\002\030\000\131\002\032\000\131\002\034\000\131\002\
\\039\000\131\002\079\000\131\002\000\000\
\\001\000\001\000\132\002\002\000\132\002\003\000\132\002\004\000\132\002\
\\005\000\132\002\008\000\132\002\009\000\132\002\011\000\132\002\
\\012\000\132\002\014\000\132\002\015\000\132\002\016\000\132\002\
\\017\000\132\002\018\000\132\002\019\000\132\002\020\000\132\002\
\\021\000\132\002\022\000\132\002\026\000\132\002\027\000\132\002\
\\028\000\132\002\030\000\132\002\032\000\132\002\034\000\132\002\
\\039\000\132\002\079\000\132\002\000\000\
\\001\000\001\000\133\002\002\000\133\002\003\000\133\002\004\000\133\002\
\\005\000\133\002\008\000\133\002\009\000\133\002\011\000\133\002\
\\012\000\133\002\014\000\133\002\015\000\133\002\016\000\133\002\
\\017\000\133\002\018\000\133\002\019\000\133\002\020\000\133\002\
\\021\000\133\002\022\000\133\002\026\000\133\002\027\000\133\002\
\\028\000\133\002\030\000\133\002\032\000\133\002\034\000\133\002\
\\039\000\133\002\079\000\133\002\000\000\
\\001\000\001\000\134\002\002\000\134\002\003\000\134\002\004\000\134\002\
\\005\000\134\002\008\000\134\002\009\000\134\002\011\000\134\002\
\\012\000\134\002\014\000\134\002\015\000\134\002\016\000\134\002\
\\017\000\134\002\018\000\134\002\019\000\134\002\020\000\134\002\
\\021\000\134\002\022\000\134\002\023\000\134\002\026\000\134\002\
\\027\000\134\002\028\000\134\002\030\000\134\002\032\000\134\002\
\\034\000\134\002\039\000\134\002\079\000\134\002\000\000\
\\001\000\001\000\135\002\002\000\135\002\003\000\135\002\004\000\135\002\
\\005\000\135\002\008\000\135\002\009\000\135\002\011\000\135\002\
\\012\000\135\002\014\000\135\002\015\000\135\002\016\000\135\002\
\\017\000\135\002\018\000\135\002\019\000\135\002\020\000\135\002\
\\021\000\135\002\022\000\135\002\023\000\135\002\026\000\135\002\
\\027\000\135\002\028\000\135\002\030\000\135\002\032\000\135\002\
\\034\000\135\002\039\000\135\002\079\000\135\002\000\000\
\\001\000\001\000\136\002\002\000\136\002\003\000\136\002\004\000\136\002\
\\005\000\136\002\008\000\136\002\009\000\136\002\011\000\136\002\
\\012\000\136\002\014\000\136\002\015\000\136\002\016\000\136\002\
\\017\000\136\002\018\000\136\002\019\000\136\002\020\000\136\002\
\\021\000\136\002\022\000\136\002\023\000\136\002\026\000\136\002\
\\027\000\136\002\028\000\136\002\030\000\136\002\032\000\136\002\
\\034\000\136\002\039\000\136\002\079\000\136\002\000\000\
\\001\000\001\000\137\002\002\000\137\002\003\000\137\002\004\000\137\002\
\\005\000\137\002\008\000\137\002\009\000\137\002\011\000\137\002\
\\012\000\137\002\014\000\137\002\015\000\137\002\016\000\137\002\
\\017\000\137\002\018\000\137\002\019\000\137\002\020\000\137\002\
\\021\000\137\002\022\000\137\002\023\000\137\002\026\000\137\002\
\\027\000\137\002\028\000\137\002\030\000\137\002\032\000\137\002\
\\034\000\137\002\039\000\137\002\079\000\137\002\000\000\
\\001\000\001\000\138\002\002\000\138\002\003\000\138\002\004\000\138\002\
\\005\000\138\002\006\000\161\000\007\000\160\000\008\000\138\002\
\\009\000\138\002\011\000\138\002\012\000\138\002\014\000\138\002\
\\015\000\138\002\016\000\138\002\017\000\138\002\018\000\138\002\
\\019\000\138\002\020\000\138\002\021\000\138\002\022\000\138\002\
\\023\000\138\002\024\000\159\000\025\000\158\000\026\000\138\002\
\\027\000\138\002\028\000\138\002\029\000\157\000\030\000\138\002\
\\031\000\156\000\032\000\138\002\034\000\138\002\039\000\138\002\
\\079\000\138\002\000\000\
\\001\000\001\000\139\002\002\000\139\002\003\000\139\002\004\000\139\002\
\\005\000\139\002\008\000\139\002\009\000\139\002\011\000\139\002\
\\012\000\139\002\014\000\139\002\015\000\139\002\016\000\139\002\
\\017\000\139\002\018\000\139\002\019\000\139\002\020\000\139\002\
\\021\000\139\002\022\000\139\002\023\000\139\002\026\000\139\002\
\\027\000\139\002\028\000\139\002\030\000\139\002\032\000\139\002\
\\034\000\139\002\039\000\139\002\079\000\139\002\000\000\
\\001\000\001\000\140\002\002\000\140\002\003\000\140\002\004\000\140\002\
\\005\000\140\002\008\000\140\002\009\000\140\002\011\000\140\002\
\\012\000\140\002\014\000\140\002\015\000\140\002\016\000\140\002\
\\017\000\140\002\018\000\140\002\019\000\140\002\020\000\140\002\
\\021\000\140\002\022\000\140\002\023\000\140\002\026\000\140\002\
\\027\000\140\002\028\000\140\002\030\000\140\002\032\000\140\002\
\\034\000\140\002\039\000\140\002\079\000\140\002\000\000\
\\001\000\001\000\141\002\002\000\141\002\003\000\141\002\004\000\141\002\
\\005\000\141\002\008\000\141\002\009\000\141\002\011\000\141\002\
\\012\000\141\002\014\000\141\002\015\000\141\002\016\000\141\002\
\\017\000\141\002\018\000\141\002\019\000\141\002\020\000\141\002\
\\021\000\141\002\022\000\141\002\023\000\141\002\026\000\141\002\
\\027\000\141\002\028\000\141\002\030\000\141\002\032\000\141\002\
\\034\000\141\002\039\000\141\002\079\000\141\002\000\000\
\\001\000\001\000\142\002\002\000\142\002\003\000\142\002\004\000\142\002\
\\005\000\142\002\008\000\142\002\009\000\142\002\011\000\142\002\
\\012\000\142\002\014\000\142\002\015\000\142\002\016\000\142\002\
\\017\000\142\002\018\000\142\002\019\000\142\002\020\000\142\002\
\\021\000\142\002\022\000\142\002\023\000\142\002\026\000\142\002\
\\027\000\142\002\028\000\142\002\030\000\142\002\032\000\142\002\
\\034\000\142\002\039\000\142\002\079\000\142\002\000\000\
\\001\000\001\000\143\002\002\000\143\002\003\000\143\002\004\000\143\002\
\\005\000\143\002\008\000\143\002\009\000\143\002\011\000\143\002\
\\012\000\143\002\014\000\143\002\015\000\143\002\016\000\143\002\
\\017\000\143\002\018\000\143\002\019\000\143\002\020\000\143\002\
\\021\000\143\002\022\000\143\002\023\000\143\002\026\000\143\002\
\\027\000\143\002\028\000\143\002\030\000\143\002\032\000\143\002\
\\034\000\143\002\039\000\143\002\079\000\143\002\000\000\
\\001\000\001\000\144\002\002\000\144\002\003\000\144\002\004\000\144\002\
\\005\000\144\002\008\000\144\002\009\000\144\002\011\000\144\002\
\\012\000\144\002\014\000\144\002\015\000\144\002\016\000\144\002\
\\017\000\144\002\018\000\144\002\019\000\144\002\020\000\144\002\
\\021\000\144\002\022\000\144\002\023\000\144\002\026\000\144\002\
\\027\000\144\002\028\000\144\002\030\000\144\002\032\000\144\002\
\\034\000\144\002\039\000\144\002\079\000\144\002\000\000\
\\001\000\001\000\145\002\002\000\145\002\003\000\145\002\004\000\145\002\
\\005\000\145\002\008\000\145\002\009\000\145\002\011\000\145\002\
\\012\000\145\002\014\000\145\002\015\000\145\002\016\000\145\002\
\\017\000\145\002\018\000\145\002\019\000\145\002\020\000\145\002\
\\021\000\145\002\022\000\145\002\023\000\145\002\026\000\145\002\
\\027\000\145\002\028\000\145\002\030\000\145\002\032\000\145\002\
\\034\000\145\002\039\000\145\002\079\000\145\002\000\000\
\\001\000\001\000\146\002\002\000\146\002\003\000\146\002\006\000\146\002\
\\007\000\146\002\008\000\146\002\010\000\146\002\011\000\146\002\
\\013\000\146\002\029\000\146\002\035\000\146\002\041\000\146\002\
\\042\000\146\002\043\000\146\002\083\000\146\002\086\000\146\002\000\000\
\\001\000\001\000\147\002\002\000\147\002\003\000\147\002\006\000\147\002\
\\007\000\147\002\008\000\147\002\010\000\147\002\011\000\147\002\
\\013\000\147\002\029\000\147\002\035\000\147\002\041\000\147\002\
\\042\000\147\002\043\000\147\002\083\000\147\002\086\000\147\002\000\000\
\\001\000\001\000\148\002\002\000\148\002\003\000\148\002\006\000\148\002\
\\007\000\148\002\008\000\148\002\010\000\148\002\011\000\148\002\
\\013\000\148\002\029\000\148\002\035\000\148\002\041\000\148\002\
\\042\000\148\002\043\000\148\002\083\000\148\002\086\000\148\002\000\000\
\\001\000\001\000\149\002\002\000\149\002\003\000\149\002\006\000\149\002\
\\007\000\149\002\008\000\149\002\010\000\149\002\011\000\149\002\
\\013\000\149\002\029\000\149\002\035\000\149\002\041\000\149\002\
\\042\000\149\002\043\000\149\002\083\000\149\002\086\000\149\002\000\000\
\\001\000\001\000\150\002\002\000\150\002\003\000\150\002\006\000\150\002\
\\007\000\150\002\008\000\150\002\010\000\150\002\011\000\150\002\
\\013\000\150\002\029\000\150\002\035\000\150\002\041\000\150\002\
\\042\000\150\002\043\000\150\002\083\000\150\002\086\000\150\002\000\000\
\\001\000\001\000\151\002\002\000\151\002\003\000\151\002\006\000\151\002\
\\007\000\151\002\008\000\151\002\010\000\151\002\011\000\151\002\
\\013\000\151\002\029\000\151\002\035\000\151\002\041\000\151\002\
\\042\000\151\002\043\000\151\002\083\000\151\002\086\000\151\002\000\000\
\\001\000\001\000\152\002\002\000\152\002\003\000\152\002\006\000\152\002\
\\007\000\152\002\008\000\152\002\010\000\152\002\011\000\152\002\
\\013\000\152\002\029\000\152\002\035\000\152\002\041\000\152\002\
\\042\000\152\002\043\000\152\002\083\000\152\002\086\000\152\002\000\000\
\\001\000\001\000\153\002\002\000\153\002\003\000\153\002\004\000\153\002\
\\005\000\153\002\006\000\153\002\007\000\153\002\008\000\153\002\
\\009\000\153\002\011\000\153\002\012\000\153\002\014\000\153\002\
\\015\000\153\002\016\000\153\002\017\000\153\002\018\000\153\002\
\\019\000\153\002\020\000\153\002\021\000\153\002\022\000\153\002\
\\023\000\153\002\024\000\153\002\025\000\153\002\026\000\153\002\
\\027\000\153\002\028\000\153\002\029\000\153\002\030\000\153\002\
\\031\000\153\002\032\000\153\002\034\000\153\002\039\000\153\002\
\\079\000\153\002\000\000\
\\001\000\001\000\154\002\002\000\154\002\003\000\154\002\004\000\154\002\
\\005\000\154\002\006\000\154\002\007\000\154\002\008\000\154\002\
\\009\000\154\002\011\000\154\002\012\000\154\002\014\000\154\002\
\\015\000\154\002\016\000\154\002\017\000\154\002\018\000\154\002\
\\019\000\154\002\020\000\154\002\021\000\154\002\022\000\154\002\
\\023\000\154\002\024\000\154\002\025\000\154\002\026\000\154\002\
\\027\000\154\002\028\000\154\002\029\000\154\002\030\000\154\002\
\\031\000\154\002\032\000\154\002\034\000\154\002\039\000\154\002\
\\079\000\154\002\000\000\
\\001\000\001\000\155\002\002\000\155\002\003\000\155\002\004\000\155\002\
\\005\000\155\002\006\000\155\002\007\000\155\002\008\000\155\002\
\\009\000\155\002\011\000\155\002\012\000\155\002\014\000\155\002\
\\015\000\155\002\016\000\155\002\017\000\155\002\018\000\155\002\
\\019\000\155\002\020\000\155\002\021\000\155\002\022\000\155\002\
\\023\000\155\002\024\000\155\002\025\000\155\002\026\000\155\002\
\\027\000\155\002\028\000\155\002\029\000\155\002\030\000\155\002\
\\031\000\155\002\032\000\155\002\034\000\155\002\039\000\155\002\
\\079\000\155\002\000\000\
\\001\000\001\000\156\002\002\000\156\002\003\000\156\002\004\000\156\002\
\\005\000\156\002\006\000\156\002\007\000\156\002\008\000\156\002\
\\009\000\156\002\011\000\156\002\012\000\156\002\014\000\156\002\
\\015\000\156\002\016\000\156\002\017\000\156\002\018\000\156\002\
\\019\000\156\002\020\000\156\002\021\000\156\002\022\000\156\002\
\\023\000\156\002\024\000\156\002\025\000\156\002\026\000\156\002\
\\027\000\156\002\028\000\156\002\029\000\156\002\030\000\156\002\
\\031\000\156\002\032\000\156\002\034\000\156\002\039\000\156\002\
\\079\000\156\002\000\000\
\\001\000\001\000\157\002\002\000\157\002\003\000\157\002\004\000\157\002\
\\005\000\157\002\006\000\157\002\007\000\157\002\008\000\157\002\
\\009\000\157\002\011\000\157\002\012\000\157\002\014\000\157\002\
\\015\000\157\002\016\000\157\002\017\000\157\002\018\000\157\002\
\\019\000\157\002\020\000\157\002\021\000\157\002\022\000\157\002\
\\023\000\157\002\024\000\157\002\025\000\157\002\026\000\157\002\
\\027\000\157\002\028\000\157\002\029\000\157\002\030\000\157\002\
\\031\000\157\002\032\000\157\002\034\000\157\002\039\000\157\002\
\\079\000\157\002\000\000\
\\001\000\001\000\158\002\002\000\158\002\003\000\158\002\004\000\158\002\
\\005\000\158\002\006\000\158\002\007\000\158\002\008\000\158\002\
\\009\000\158\002\011\000\158\002\012\000\158\002\014\000\158\002\
\\015\000\158\002\016\000\158\002\017\000\158\002\018\000\158\002\
\\019\000\158\002\020\000\158\002\021\000\158\002\022\000\158\002\
\\023\000\158\002\024\000\158\002\025\000\158\002\026\000\158\002\
\\027\000\158\002\028\000\158\002\029\000\158\002\030\000\158\002\
\\031\000\158\002\032\000\158\002\034\000\158\002\039\000\158\002\
\\079\000\158\002\000\000\
\\001\000\001\000\159\002\002\000\159\002\003\000\159\002\004\000\159\002\
\\005\000\159\002\006\000\159\002\007\000\159\002\008\000\159\002\
\\009\000\159\002\011\000\159\002\012\000\159\002\014\000\159\002\
\\015\000\159\002\016\000\159\002\017\000\159\002\018\000\159\002\
\\019\000\159\002\020\000\159\002\021\000\159\002\022\000\159\002\
\\023\000\159\002\024\000\159\002\025\000\159\002\026\000\159\002\
\\027\000\159\002\028\000\159\002\029\000\159\002\030\000\159\002\
\\031\000\159\002\032\000\159\002\034\000\159\002\039\000\159\002\
\\079\000\159\002\000\000\
\\001\000\001\000\160\002\002\000\160\002\003\000\160\002\004\000\160\002\
\\005\000\160\002\006\000\160\002\007\000\160\002\008\000\160\002\
\\009\000\160\002\011\000\160\002\012\000\160\002\014\000\160\002\
\\015\000\160\002\016\000\160\002\017\000\160\002\018\000\160\002\
\\019\000\160\002\020\000\160\002\021\000\160\002\022\000\160\002\
\\023\000\160\002\024\000\160\002\025\000\160\002\026\000\160\002\
\\027\000\160\002\028\000\160\002\029\000\160\002\030\000\160\002\
\\031\000\160\002\032\000\160\002\034\000\160\002\039\000\160\002\
\\079\000\160\002\000\000\
\\001\000\001\000\161\002\002\000\161\002\003\000\161\002\004\000\161\002\
\\005\000\161\002\006\000\161\002\007\000\161\002\008\000\161\002\
\\009\000\161\002\011\000\161\002\012\000\161\002\014\000\161\002\
\\015\000\161\002\016\000\161\002\017\000\161\002\018\000\161\002\
\\019\000\161\002\020\000\161\002\021\000\161\002\022\000\161\002\
\\023\000\161\002\024\000\161\002\025\000\161\002\026\000\161\002\
\\027\000\161\002\028\000\161\002\029\000\161\002\030\000\161\002\
\\031\000\161\002\032\000\161\002\034\000\161\002\039\000\161\002\
\\079\000\161\002\000\000\
\\001\000\001\000\162\002\002\000\162\002\003\000\162\002\004\000\162\002\
\\005\000\162\002\006\000\162\002\007\000\162\002\008\000\162\002\
\\009\000\162\002\011\000\162\002\012\000\162\002\014\000\162\002\
\\015\000\162\002\016\000\162\002\017\000\162\002\018\000\162\002\
\\019\000\162\002\020\000\162\002\021\000\162\002\022\000\162\002\
\\023\000\162\002\024\000\162\002\025\000\162\002\026\000\162\002\
\\027\000\162\002\028\000\162\002\029\000\162\002\030\000\162\002\
\\031\000\162\002\032\000\162\002\034\000\162\002\039\000\162\002\
\\079\000\162\002\000\000\
\\001\000\001\000\162\002\002\000\162\002\003\000\162\002\004\000\162\002\
\\005\000\162\002\006\000\162\002\007\000\162\002\008\000\162\002\
\\009\000\162\002\011\000\162\002\012\000\162\002\014\000\162\002\
\\015\000\162\002\016\000\162\002\017\000\162\002\018\000\162\002\
\\019\000\162\002\020\000\162\002\021\000\162\002\022\000\162\002\
\\023\000\162\002\024\000\162\002\025\000\162\002\026\000\162\002\
\\027\000\162\002\028\000\126\001\029\000\162\002\031\000\162\002\
\\079\000\162\002\000\000\
\\001\000\001\000\163\002\002\000\163\002\003\000\163\002\004\000\163\002\
\\005\000\163\002\006\000\163\002\007\000\163\002\008\000\163\002\
\\009\000\163\002\011\000\163\002\012\000\163\002\014\000\163\002\
\\015\000\163\002\016\000\163\002\017\000\163\002\018\000\163\002\
\\019\000\163\002\020\000\163\002\021\000\163\002\022\000\163\002\
\\023\000\163\002\024\000\163\002\025\000\163\002\026\000\163\002\
\\027\000\163\002\028\000\163\002\029\000\163\002\030\000\163\002\
\\031\000\163\002\032\000\163\002\034\000\163\002\039\000\163\002\
\\079\000\163\002\000\000\
\\001\000\001\000\164\002\002\000\164\002\003\000\164\002\004\000\164\002\
\\005\000\164\002\006\000\164\002\007\000\164\002\008\000\164\002\
\\009\000\164\002\011\000\164\002\012\000\164\002\014\000\164\002\
\\015\000\164\002\016\000\164\002\017\000\164\002\018\000\164\002\
\\019\000\164\002\020\000\164\002\021\000\164\002\022\000\164\002\
\\023\000\164\002\024\000\164\002\025\000\164\002\026\000\164\002\
\\027\000\164\002\028\000\164\002\029\000\164\002\030\000\164\002\
\\031\000\164\002\032\000\164\002\034\000\164\002\039\000\164\002\
\\079\000\164\002\000\000\
\\001\000\001\000\165\002\002\000\165\002\003\000\165\002\004\000\165\002\
\\005\000\165\002\006\000\165\002\007\000\165\002\008\000\165\002\
\\009\000\165\002\011\000\165\002\012\000\165\002\014\000\165\002\
\\015\000\165\002\016\000\165\002\017\000\165\002\018\000\165\002\
\\019\000\165\002\020\000\165\002\021\000\165\002\022\000\165\002\
\\023\000\165\002\024\000\165\002\025\000\165\002\026\000\165\002\
\\027\000\165\002\028\000\165\002\029\000\165\002\030\000\165\002\
\\031\000\165\002\032\000\165\002\034\000\165\002\039\000\165\002\
\\079\000\165\002\000\000\
\\001\000\001\000\166\002\002\000\166\002\003\000\166\002\004\000\166\002\
\\005\000\166\002\006\000\166\002\007\000\166\002\008\000\166\002\
\\009\000\166\002\011\000\166\002\012\000\166\002\014\000\166\002\
\\015\000\166\002\016\000\166\002\017\000\166\002\018\000\166\002\
\\019\000\166\002\020\000\166\002\021\000\166\002\022\000\166\002\
\\023\000\166\002\024\000\166\002\025\000\166\002\026\000\166\002\
\\027\000\166\002\028\000\166\002\029\000\166\002\030\000\166\002\
\\031\000\166\002\032\000\166\002\034\000\166\002\039\000\166\002\
\\079\000\166\002\000\000\
\\001\000\001\000\167\002\002\000\167\002\003\000\167\002\004\000\167\002\
\\005\000\167\002\006\000\167\002\007\000\167\002\008\000\167\002\
\\009\000\167\002\011\000\167\002\012\000\167\002\014\000\167\002\
\\015\000\167\002\016\000\167\002\017\000\167\002\018\000\167\002\
\\019\000\167\002\020\000\167\002\021\000\167\002\022\000\167\002\
\\023\000\167\002\024\000\167\002\025\000\167\002\026\000\167\002\
\\027\000\167\002\028\000\167\002\029\000\167\002\030\000\167\002\
\\031\000\167\002\032\000\167\002\034\000\167\002\039\000\167\002\
\\079\000\167\002\000\000\
\\001\000\001\000\168\002\002\000\168\002\003\000\168\002\004\000\168\002\
\\005\000\168\002\006\000\168\002\007\000\168\002\008\000\168\002\
\\009\000\168\002\011\000\168\002\012\000\168\002\014\000\168\002\
\\015\000\168\002\016\000\168\002\017\000\168\002\018\000\168\002\
\\019\000\168\002\020\000\168\002\021\000\168\002\022\000\168\002\
\\023\000\168\002\024\000\168\002\025\000\168\002\026\000\168\002\
\\027\000\168\002\028\000\168\002\029\000\168\002\030\000\168\002\
\\031\000\168\002\032\000\168\002\034\000\168\002\039\000\168\002\
\\041\000\107\000\043\000\105\000\079\000\168\002\000\000\
\\001\000\001\000\169\002\002\000\169\002\003\000\169\002\004\000\169\002\
\\005\000\169\002\006\000\169\002\007\000\169\002\008\000\169\002\
\\009\000\169\002\011\000\169\002\012\000\169\002\014\000\169\002\
\\015\000\169\002\016\000\169\002\017\000\169\002\018\000\169\002\
\\019\000\169\002\020\000\169\002\021\000\169\002\022\000\169\002\
\\023\000\169\002\024\000\169\002\025\000\169\002\026\000\169\002\
\\027\000\169\002\028\000\169\002\029\000\169\002\030\000\169\002\
\\031\000\169\002\032\000\169\002\034\000\169\002\039\000\169\002\
\\079\000\169\002\000\000\
\\001\000\001\000\170\002\002\000\170\002\003\000\170\002\004\000\170\002\
\\005\000\170\002\006\000\170\002\007\000\170\002\008\000\170\002\
\\009\000\170\002\011\000\170\002\012\000\170\002\014\000\170\002\
\\015\000\170\002\016\000\170\002\017\000\170\002\018\000\170\002\
\\019\000\170\002\020\000\170\002\021\000\170\002\022\000\170\002\
\\023\000\170\002\024\000\170\002\025\000\170\002\026\000\170\002\
\\027\000\170\002\028\000\170\002\029\000\170\002\030\000\170\002\
\\031\000\170\002\032\000\170\002\034\000\170\002\039\000\170\002\
\\041\000\170\002\043\000\170\002\079\000\170\002\000\000\
\\001\000\001\000\171\002\002\000\171\002\003\000\171\002\004\000\171\002\
\\005\000\171\002\006\000\171\002\007\000\171\002\008\000\171\002\
\\009\000\171\002\011\000\171\002\012\000\171\002\014\000\171\002\
\\015\000\171\002\016\000\171\002\017\000\171\002\018\000\171\002\
\\019\000\171\002\020\000\171\002\021\000\171\002\022\000\171\002\
\\023\000\171\002\024\000\171\002\025\000\171\002\026\000\171\002\
\\027\000\171\002\028\000\171\002\029\000\171\002\030\000\171\002\
\\031\000\171\002\032\000\171\002\034\000\171\002\039\000\171\002\
\\041\000\171\002\043\000\171\002\079\000\171\002\000\000\
\\001\000\001\000\172\002\002\000\172\002\003\000\172\002\004\000\172\002\
\\005\000\172\002\006\000\172\002\007\000\172\002\008\000\172\002\
\\009\000\172\002\011\000\172\002\012\000\172\002\014\000\172\002\
\\015\000\172\002\016\000\172\002\017\000\172\002\018\000\172\002\
\\019\000\172\002\020\000\172\002\021\000\172\002\022\000\172\002\
\\023\000\172\002\024\000\172\002\025\000\172\002\026\000\172\002\
\\027\000\172\002\028\000\172\002\029\000\172\002\030\000\172\002\
\\031\000\172\002\032\000\172\002\034\000\172\002\039\000\172\002\
\\079\000\172\002\000\000\
\\001\000\001\000\173\002\002\000\173\002\003\000\173\002\004\000\173\002\
\\005\000\173\002\006\000\173\002\007\000\173\002\008\000\173\002\
\\009\000\173\002\011\000\173\002\012\000\173\002\014\000\173\002\
\\015\000\173\002\016\000\173\002\017\000\173\002\018\000\173\002\
\\019\000\173\002\020\000\173\002\021\000\173\002\022\000\173\002\
\\023\000\173\002\024\000\173\002\025\000\173\002\026\000\173\002\
\\027\000\173\002\028\000\173\002\029\000\173\002\030\000\173\002\
\\031\000\173\002\032\000\173\002\034\000\173\002\039\000\173\002\
\\079\000\173\002\000\000\
\\001\000\001\000\176\002\002\000\176\002\003\000\176\002\006\000\176\002\
\\007\000\176\002\008\000\176\002\010\000\176\002\011\000\176\002\
\\013\000\176\002\029\000\176\002\033\000\176\002\034\000\176\002\
\\035\000\176\002\036\000\176\002\040\000\176\002\041\000\176\002\
\\042\000\176\002\043\000\176\002\045\000\176\002\046\000\176\002\
\\047\000\176\002\048\000\176\002\049\000\176\002\050\000\176\002\
\\051\000\176\002\052\000\176\002\053\000\176\002\054\000\176\002\
\\055\000\176\002\056\000\176\002\057\000\176\002\058\000\176\002\
\\059\000\176\002\060\000\176\002\061\000\176\002\062\000\176\002\
\\063\000\176\002\064\000\176\002\065\000\176\002\066\000\176\002\
\\067\000\176\002\069\000\176\002\070\000\176\002\071\000\176\002\
\\072\000\176\002\073\000\176\002\074\000\176\002\075\000\176\002\
\\076\000\176\002\077\000\176\002\078\000\176\002\079\000\176\002\
\\083\000\176\002\084\000\176\002\085\000\176\002\086\000\176\002\000\000\
\\001\000\001\000\119\000\002\000\118\000\003\000\117\000\006\000\116\000\
\\007\000\115\000\008\000\114\000\010\000\113\000\011\000\112\000\
\\013\000\111\000\026\000\020\002\029\000\110\000\033\000\037\001\
\\034\000\020\002\035\000\108\000\041\000\107\000\042\000\106\000\
\\043\000\105\000\083\000\104\000\086\000\103\000\000\000\
\\001\000\001\000\119\000\002\000\118\000\003\000\117\000\006\000\116\000\
\\007\000\115\000\008\000\114\000\010\000\113\000\011\000\112\000\
\\013\000\111\000\029\000\110\000\030\000\105\002\035\000\108\000\
\\040\000\194\000\041\000\107\000\042\000\106\000\043\000\105\000\
\\083\000\104\000\086\000\103\000\000\000\
\\001\000\001\000\119\000\002\000\118\000\003\000\117\000\006\000\116\000\
\\007\000\115\000\008\000\114\000\010\000\113\000\011\000\112\000\
\\013\000\111\000\029\000\110\000\030\000\244\000\035\000\108\000\
\\040\000\194\000\041\000\107\000\042\000\106\000\043\000\105\000\
\\083\000\104\000\086\000\103\000\000\000\
\\001\000\001\000\119\000\002\000\118\000\003\000\117\000\006\000\116\000\
\\007\000\115\000\008\000\114\000\010\000\113\000\011\000\112\000\
\\013\000\111\000\029\000\110\000\032\000\109\000\035\000\108\000\
\\041\000\107\000\042\000\106\000\043\000\105\000\083\000\104\000\
\\086\000\103\000\000\000\
\\001\000\001\000\119\000\002\000\118\000\003\000\117\000\006\000\116\000\
\\007\000\115\000\008\000\114\000\010\000\113\000\011\000\112\000\
\\013\000\111\000\029\000\110\000\033\000\195\000\034\000\180\002\
\\035\000\083\001\036\000\078\000\040\000\194\000\041\000\107\000\
\\042\000\106\000\043\000\105\000\045\000\048\000\046\000\047\000\
\\047\000\046\000\048\000\045\000\049\000\044\000\050\000\043\000\
\\051\000\042\000\052\000\041\000\053\000\040\000\054\000\039\000\
\\055\000\038\000\056\000\037\000\057\000\036\000\058\000\035\000\
\\059\000\034\000\060\000\033\000\061\000\032\000\062\000\031\000\
\\063\000\030\000\064\000\029\000\065\000\028\000\066\000\027\000\
\\067\000\082\001\069\000\081\001\070\000\080\001\071\000\079\001\
\\072\000\078\001\073\000\077\001\074\000\076\001\075\000\075\001\
\\076\000\074\001\077\000\073\001\078\000\072\001\079\000\105\002\
\\083\000\104\000\084\000\026\000\085\000\025\000\086\000\103\000\000\000\
\\001\000\001\000\119\000\002\000\118\000\003\000\117\000\006\000\116\000\
\\007\000\115\000\008\000\114\000\010\000\113\000\011\000\112\000\
\\013\000\111\000\029\000\110\000\033\000\195\000\034\000\180\002\
\\035\000\083\001\040\000\194\000\041\000\107\000\042\000\106\000\
\\043\000\105\000\067\000\082\001\069\000\081\001\070\000\080\001\
\\071\000\079\001\072\000\078\001\073\000\077\001\074\000\076\001\
\\075\000\075\001\076\000\074\001\077\000\073\001\078\000\072\001\
\\079\000\105\002\083\000\104\000\086\000\103\000\000\000\
\\001\000\001\000\119\000\002\000\118\000\003\000\117\000\006\000\116\000\
\\007\000\115\000\008\000\114\000\010\000\113\000\011\000\112\000\
\\013\000\111\000\029\000\110\000\033\000\195\000\035\000\108\000\
\\036\000\078\000\040\000\194\000\041\000\107\000\042\000\106\000\
\\043\000\105\000\045\000\048\000\046\000\047\000\047\000\046\000\
\\048\000\045\000\049\000\044\000\050\000\043\000\051\000\042\000\
\\052\000\041\000\053\000\040\000\054\000\039\000\055\000\038\000\
\\056\000\037\000\057\000\036\000\058\000\035\000\059\000\034\000\
\\060\000\033\000\061\000\032\000\062\000\031\000\063\000\030\000\
\\064\000\029\000\065\000\028\000\066\000\027\000\083\000\104\000\
\\084\000\026\000\085\000\025\000\086\000\103\000\000\000\
\\001\000\001\000\119\000\002\000\118\000\003\000\117\000\006\000\116\000\
\\007\000\115\000\008\000\114\000\010\000\113\000\011\000\112\000\
\\013\000\111\000\029\000\110\000\033\000\195\000\035\000\108\000\
\\040\000\194\000\041\000\107\000\042\000\106\000\043\000\105\000\
\\083\000\104\000\086\000\103\000\000\000\
\\001\000\001\000\119\000\002\000\118\000\003\000\117\000\006\000\116\000\
\\007\000\115\000\008\000\114\000\010\000\113\000\011\000\112\000\
\\013\000\111\000\029\000\110\000\033\000\195\000\035\000\083\001\
\\040\000\194\000\041\000\107\000\042\000\106\000\043\000\105\000\
\\067\000\082\001\069\000\081\001\070\000\080\001\071\000\079\001\
\\072\000\078\001\073\000\077\001\074\000\076\001\075\000\075\001\
\\076\000\074\001\077\000\073\001\078\000\072\001\079\000\105\002\
\\083\000\104\000\086\000\103\000\000\000\
\\001\000\001\000\119\000\002\000\118\000\003\000\117\000\006\000\116\000\
\\007\000\115\000\008\000\114\000\010\000\113\000\011\000\112\000\
\\013\000\111\000\029\000\110\000\033\000\215\000\035\000\108\000\
\\040\000\194\000\041\000\107\000\042\000\106\000\043\000\105\000\
\\083\000\104\000\086\000\103\000\000\000\
\\001\000\001\000\119\000\002\000\118\000\003\000\117\000\006\000\116\000\
\\007\000\115\000\008\000\114\000\010\000\113\000\011\000\112\000\
\\013\000\111\000\029\000\110\000\033\000\037\001\034\000\130\001\
\\035\000\108\000\041\000\107\000\042\000\106\000\043\000\105\000\
\\083\000\104\000\086\000\103\000\000\000\
\\001\000\001\000\119\000\002\000\118\000\003\000\117\000\006\000\116\000\
\\007\000\115\000\008\000\114\000\010\000\113\000\011\000\112\000\
\\013\000\111\000\029\000\110\000\033\000\037\001\034\000\156\001\
\\035\000\108\000\041\000\107\000\042\000\106\000\043\000\105\000\
\\083\000\104\000\086\000\103\000\000\000\
\\001\000\001\000\119\000\002\000\118\000\003\000\117\000\006\000\116\000\
\\007\000\115\000\008\000\114\000\010\000\113\000\011\000\112\000\
\\013\000\111\000\029\000\110\000\033\000\037\001\034\000\157\001\
\\035\000\108\000\041\000\107\000\042\000\106\000\043\000\105\000\
\\083\000\104\000\086\000\103\000\000\000\
\\001\000\001\000\119\000\002\000\118\000\003\000\117\000\006\000\116\000\
\\007\000\115\000\008\000\114\000\010\000\113\000\011\000\112\000\
\\013\000\111\000\029\000\110\000\033\000\063\001\035\000\108\000\
\\041\000\107\000\042\000\106\000\043\000\105\000\083\000\104\000\
\\086\000\103\000\000\000\
\\001\000\001\000\119\000\002\000\118\000\003\000\117\000\006\000\116\000\
\\007\000\115\000\008\000\114\000\010\000\113\000\011\000\112\000\
\\013\000\111\000\029\000\110\000\035\000\108\000\040\000\194\000\
\\041\000\107\000\042\000\106\000\043\000\105\000\079\000\105\002\
\\083\000\104\000\086\000\103\000\000\000\
\\001\000\001\000\119\000\002\000\118\000\003\000\117\000\006\000\116\000\
\\007\000\115\000\008\000\114\000\010\000\113\000\011\000\112\000\
\\013\000\111\000\029\000\110\000\035\000\108\000\040\000\194\000\
\\041\000\107\000\042\000\106\000\043\000\105\000\083\000\104\000\
\\086\000\103\000\000\000\
\\001\000\001\000\119\000\002\000\118\000\003\000\117\000\006\000\116\000\
\\007\000\115\000\008\000\114\000\010\000\113\000\011\000\112\000\
\\013\000\111\000\029\000\110\000\035\000\108\000\041\000\107\000\
\\042\000\106\000\043\000\105\000\083\000\104\000\086\000\103\000\000\000\
\\001\000\001\000\119\000\002\000\118\000\003\000\117\000\006\000\116\000\
\\007\000\115\000\008\000\114\000\010\000\113\000\011\000\112\000\
\\013\000\111\000\029\000\147\000\035\000\108\000\036\000\078\000\
\\041\000\107\000\042\000\106\000\043\000\105\000\045\000\048\000\
\\046\000\047\000\047\000\046\000\048\000\045\000\049\000\044\000\
\\050\000\043\000\051\000\042\000\052\000\041\000\053\000\040\000\
\\054\000\039\000\055\000\038\000\056\000\037\000\057\000\036\000\
\\058\000\035\000\059\000\034\000\060\000\033\000\061\000\032\000\
\\062\000\031\000\063\000\030\000\064\000\029\000\065\000\028\000\
\\066\000\027\000\083\000\104\000\084\000\026\000\085\000\025\000\
\\086\000\103\000\000\000\
\\001\000\001\000\119\000\002\000\118\000\003\000\117\000\006\000\116\000\
\\007\000\115\000\008\000\114\000\010\000\113\000\011\000\112\000\
\\013\000\111\000\029\000\147\000\035\000\108\000\041\000\107\000\
\\042\000\106\000\043\000\105\000\083\000\104\000\086\000\103\000\000\000\
\\001\000\001\000\119\000\002\000\118\000\003\000\117\000\006\000\116\000\
\\007\000\115\000\008\000\114\000\010\000\113\000\011\000\112\000\
\\013\000\111\000\029\000\185\000\035\000\108\000\041\000\107\000\
\\042\000\106\000\043\000\105\000\083\000\104\000\086\000\103\000\000\000\
\\001\000\001\000\119\000\002\000\118\000\003\000\117\000\006\000\116\000\
\\007\000\115\000\008\000\114\000\010\000\113\000\011\000\112\000\
\\013\000\111\000\029\000\187\000\035\000\108\000\041\000\107\000\
\\042\000\106\000\043\000\105\000\083\000\104\000\086\000\103\000\000\000\
\\001\000\001\000\181\000\002\000\180\000\003\000\179\000\004\000\178\000\
\\005\000\177\000\008\000\117\002\009\000\117\002\011\000\174\000\
\\012\000\173\000\014\000\172\000\015\000\171\000\016\000\170\000\
\\017\000\169\000\018\000\168\000\019\000\167\000\020\000\166\000\
\\021\000\165\000\022\000\164\000\026\000\117\002\027\000\117\002\
\\028\000\117\002\030\000\117\002\032\000\117\002\034\000\117\002\
\\039\000\117\002\079\000\117\002\000\000\
\\001\000\001\000\181\000\002\000\180\000\003\000\179\000\004\000\178\000\
\\005\000\177\000\008\000\118\002\009\000\118\002\011\000\174\000\
\\012\000\118\002\014\000\172\000\015\000\171\000\016\000\170\000\
\\017\000\169\000\018\000\168\000\019\000\167\000\020\000\166\000\
\\021\000\165\000\022\000\164\000\026\000\118\002\027\000\118\002\
\\028\000\118\002\030\000\118\002\032\000\118\002\034\000\118\002\
\\039\000\118\002\079\000\118\002\000\000\
\\001\000\001\000\181\000\002\000\180\000\003\000\179\000\004\000\178\000\
\\005\000\177\000\008\000\119\002\009\000\119\002\011\000\174\000\
\\012\000\119\002\014\000\119\002\015\000\171\000\016\000\170\000\
\\017\000\169\000\018\000\168\000\019\000\167\000\020\000\166\000\
\\021\000\165\000\022\000\164\000\026\000\119\002\027\000\119\002\
\\028\000\119\002\030\000\119\002\032\000\119\002\034\000\119\002\
\\039\000\119\002\079\000\119\002\000\000\
\\001\000\001\000\181\000\002\000\180\000\003\000\179\000\004\000\178\000\
\\005\000\177\000\008\000\120\002\009\000\120\002\011\000\120\002\
\\012\000\120\002\014\000\120\002\015\000\171\000\016\000\170\000\
\\017\000\169\000\018\000\168\000\019\000\167\000\020\000\166\000\
\\021\000\165\000\022\000\164\000\026\000\120\002\027\000\120\002\
\\028\000\120\002\030\000\120\002\032\000\120\002\034\000\120\002\
\\039\000\120\002\079\000\120\002\000\000\
\\001\000\001\000\181\000\002\000\180\000\003\000\179\000\004\000\178\000\
\\005\000\177\000\008\000\121\002\009\000\121\002\011\000\121\002\
\\012\000\121\002\014\000\121\002\015\000\171\000\016\000\170\000\
\\017\000\169\000\018\000\168\000\019\000\167\000\020\000\166\000\
\\021\000\121\002\022\000\121\002\026\000\121\002\027\000\121\002\
\\028\000\121\002\030\000\121\002\032\000\121\002\034\000\121\002\
\\039\000\121\002\079\000\121\002\000\000\
\\001\000\001\000\181\000\002\000\180\000\003\000\179\000\004\000\178\000\
\\005\000\177\000\008\000\122\002\009\000\122\002\011\000\122\002\
\\012\000\122\002\014\000\122\002\015\000\171\000\016\000\170\000\
\\017\000\169\000\018\000\168\000\019\000\167\000\020\000\166\000\
\\021\000\122\002\022\000\122\002\026\000\122\002\027\000\122\002\
\\028\000\122\002\030\000\122\002\032\000\122\002\034\000\122\002\
\\039\000\122\002\079\000\122\002\000\000\
\\001\000\001\000\181\000\002\000\180\000\003\000\179\000\004\000\178\000\
\\005\000\177\000\008\000\123\002\009\000\123\002\011\000\123\002\
\\012\000\123\002\014\000\123\002\015\000\171\000\016\000\170\000\
\\017\000\123\002\018\000\123\002\019\000\123\002\020\000\123\002\
\\021\000\123\002\022\000\123\002\026\000\123\002\027\000\123\002\
\\028\000\123\002\030\000\123\002\032\000\123\002\034\000\123\002\
\\039\000\123\002\079\000\123\002\000\000\
\\001\000\001\000\181\000\002\000\180\000\003\000\179\000\004\000\178\000\
\\005\000\177\000\008\000\124\002\009\000\124\002\011\000\124\002\
\\012\000\124\002\014\000\124\002\015\000\171\000\016\000\170\000\
\\017\000\124\002\018\000\124\002\019\000\124\002\020\000\124\002\
\\021\000\124\002\022\000\124\002\026\000\124\002\027\000\124\002\
\\028\000\124\002\030\000\124\002\032\000\124\002\034\000\124\002\
\\039\000\124\002\079\000\124\002\000\000\
\\001\000\001\000\181\000\002\000\180\000\003\000\179\000\004\000\178\000\
\\005\000\177\000\008\000\125\002\009\000\125\002\011\000\125\002\
\\012\000\125\002\014\000\125\002\015\000\171\000\016\000\170\000\
\\017\000\125\002\018\000\125\002\019\000\125\002\020\000\125\002\
\\021\000\125\002\022\000\125\002\026\000\125\002\027\000\125\002\
\\028\000\125\002\030\000\125\002\032\000\125\002\034\000\125\002\
\\039\000\125\002\079\000\125\002\000\000\
\\001\000\001\000\181\000\002\000\180\000\003\000\179\000\004\000\178\000\
\\005\000\177\000\008\000\126\002\009\000\126\002\011\000\126\002\
\\012\000\126\002\014\000\126\002\015\000\171\000\016\000\170\000\
\\017\000\126\002\018\000\126\002\019\000\126\002\020\000\126\002\
\\021\000\126\002\022\000\126\002\026\000\126\002\027\000\126\002\
\\028\000\126\002\030\000\126\002\032\000\126\002\034\000\126\002\
\\039\000\126\002\079\000\126\002\000\000\
\\001\000\001\000\181\000\002\000\180\000\003\000\179\000\004\000\178\000\
\\005\000\177\000\008\000\127\002\009\000\127\002\011\000\127\002\
\\012\000\127\002\014\000\127\002\015\000\127\002\016\000\127\002\
\\017\000\127\002\018\000\127\002\019\000\127\002\020\000\127\002\
\\021\000\127\002\022\000\127\002\026\000\127\002\027\000\127\002\
\\028\000\127\002\030\000\127\002\032\000\127\002\034\000\127\002\
\\039\000\127\002\079\000\127\002\000\000\
\\001\000\001\000\181\000\002\000\180\000\003\000\179\000\004\000\178\000\
\\005\000\177\000\008\000\128\002\009\000\128\002\011\000\128\002\
\\012\000\128\002\014\000\128\002\015\000\128\002\016\000\128\002\
\\017\000\128\002\018\000\128\002\019\000\128\002\020\000\128\002\
\\021\000\128\002\022\000\128\002\026\000\128\002\027\000\128\002\
\\028\000\128\002\030\000\128\002\032\000\128\002\034\000\128\002\
\\039\000\128\002\079\000\128\002\000\000\
\\001\000\001\000\181\000\002\000\180\000\003\000\179\000\004\000\178\000\
\\005\000\177\000\008\000\176\000\009\000\116\002\011\000\174\000\
\\012\000\173\000\014\000\172\000\015\000\171\000\016\000\170\000\
\\017\000\169\000\018\000\168\000\019\000\167\000\020\000\166\000\
\\021\000\165\000\022\000\164\000\026\000\116\002\027\000\116\002\
\\028\000\116\002\030\000\116\002\032\000\116\002\034\000\116\002\
\\039\000\116\002\079\000\116\002\000\000\
\\001\000\001\000\181\000\002\000\180\000\003\000\179\000\004\000\178\000\
\\005\000\177\000\008\000\176\000\009\000\175\000\011\000\174\000\
\\012\000\173\000\014\000\172\000\015\000\171\000\016\000\170\000\
\\017\000\169\000\018\000\168\000\019\000\167\000\020\000\166\000\
\\021\000\165\000\022\000\164\000\026\000\112\002\027\000\163\000\
\\028\000\112\002\030\000\112\002\032\000\112\002\034\000\112\002\
\\039\000\112\002\079\000\112\002\000\000\
\\001\000\003\000\180\001\029\000\180\001\035\000\180\001\036\000\180\001\
\\039\000\180\001\040\000\180\001\045\000\180\001\046\000\180\001\
\\047\000\180\001\048\000\180\001\049\000\180\001\050\000\180\001\
\\051\000\180\001\052\000\180\001\053\000\180\001\054\000\180\001\
\\055\000\180\001\056\000\180\001\057\000\180\001\058\000\180\001\
\\059\000\180\001\060\000\180\001\061\000\180\001\062\000\180\001\
\\063\000\180\001\064\000\180\001\065\000\180\001\066\000\180\001\
\\084\000\180\001\085\000\180\001\087\000\180\001\000\000\
\\001\000\003\000\181\001\029\000\181\001\035\000\181\001\036\000\181\001\
\\039\000\181\001\040\000\181\001\045\000\181\001\046\000\181\001\
\\047\000\181\001\048\000\181\001\049\000\181\001\050\000\181\001\
\\051\000\181\001\052\000\181\001\053\000\181\001\054\000\181\001\
\\055\000\181\001\056\000\181\001\057\000\181\001\058\000\181\001\
\\059\000\181\001\060\000\181\001\061\000\181\001\062\000\181\001\
\\063\000\181\001\064\000\181\001\065\000\181\001\066\000\181\001\
\\084\000\181\001\085\000\181\001\087\000\181\001\000\000\
\\001\000\003\000\182\001\029\000\182\001\035\000\182\001\036\000\182\001\
\\039\000\182\001\040\000\182\001\045\000\182\001\046\000\182\001\
\\047\000\182\001\048\000\182\001\049\000\182\001\050\000\182\001\
\\051\000\182\001\052\000\182\001\053\000\182\001\054\000\182\001\
\\055\000\182\001\056\000\182\001\057\000\182\001\058\000\182\001\
\\059\000\182\001\060\000\182\001\061\000\182\001\062\000\182\001\
\\063\000\182\001\064\000\182\001\065\000\182\001\066\000\182\001\
\\084\000\182\001\085\000\182\001\087\000\182\001\000\000\
\\001\000\003\000\183\001\029\000\183\001\035\000\183\001\036\000\183\001\
\\039\000\183\001\040\000\183\001\045\000\183\001\046\000\183\001\
\\047\000\183\001\048\000\183\001\049\000\183\001\050\000\183\001\
\\051\000\183\001\052\000\183\001\053\000\183\001\054\000\183\001\
\\055\000\183\001\056\000\183\001\057\000\183\001\058\000\183\001\
\\059\000\183\001\060\000\183\001\061\000\183\001\062\000\183\001\
\\063\000\183\001\064\000\183\001\065\000\183\001\066\000\183\001\
\\084\000\183\001\085\000\183\001\087\000\183\001\000\000\
\\001\000\003\000\184\001\029\000\184\001\035\000\184\001\036\000\184\001\
\\039\000\184\001\040\000\006\000\045\000\184\001\046\000\184\001\
\\047\000\184\001\048\000\184\001\049\000\184\001\050\000\184\001\
\\051\000\184\001\052\000\184\001\053\000\184\001\054\000\184\001\
\\055\000\184\001\056\000\184\001\057\000\184\001\058\000\184\001\
\\059\000\184\001\060\000\184\001\061\000\184\001\062\000\184\001\
\\063\000\184\001\064\000\184\001\065\000\184\001\066\000\184\001\
\\084\000\184\001\085\000\184\001\087\000\178\001\000\000\
\\001\000\003\000\184\001\029\000\184\001\035\000\184\001\036\000\184\001\
\\039\000\184\001\040\000\006\000\045\000\184\001\046\000\184\001\
\\047\000\184\001\048\000\184\001\049\000\184\001\050\000\184\001\
\\051\000\184\001\052\000\184\001\053\000\184\001\054\000\184\001\
\\055\000\184\001\056\000\184\001\057\000\184\001\058\000\184\001\
\\059\000\184\001\060\000\184\001\061\000\184\001\062\000\184\001\
\\063\000\184\001\064\000\184\001\065\000\184\001\066\000\184\001\
\\084\000\184\001\085\000\184\001\087\000\179\001\000\000\
\\001\000\003\000\185\001\029\000\185\001\035\000\185\001\036\000\185\001\
\\039\000\185\001\045\000\185\001\046\000\185\001\047\000\185\001\
\\048\000\185\001\049\000\185\001\050\000\185\001\051\000\185\001\
\\052\000\185\001\053\000\185\001\054\000\185\001\055\000\185\001\
\\056\000\185\001\057\000\185\001\058\000\185\001\059\000\185\001\
\\060\000\185\001\061\000\185\001\062\000\185\001\063\000\185\001\
\\064\000\185\001\065\000\185\001\066\000\185\001\084\000\185\001\
\\085\000\185\001\000\000\
\\001\000\003\000\186\001\029\000\186\001\035\000\186\001\036\000\186\001\
\\039\000\186\001\040\000\186\001\045\000\186\001\046\000\186\001\
\\047\000\186\001\048\000\186\001\049\000\186\001\050\000\186\001\
\\051\000\186\001\052\000\186\001\053\000\186\001\054\000\186\001\
\\055\000\186\001\056\000\186\001\057\000\186\001\058\000\186\001\
\\059\000\186\001\060\000\186\001\061\000\186\001\062\000\186\001\
\\063\000\186\001\064\000\186\001\065\000\186\001\066\000\186\001\
\\079\000\092\001\084\000\186\001\085\000\186\001\087\000\186\001\000\000\
\\001\000\003\000\187\001\029\000\187\001\035\000\187\001\036\000\187\001\
\\039\000\187\001\040\000\187\001\045\000\187\001\046\000\187\001\
\\047\000\187\001\048\000\187\001\049\000\187\001\050\000\187\001\
\\051\000\187\001\052\000\187\001\053\000\187\001\054\000\187\001\
\\055\000\187\001\056\000\187\001\057\000\187\001\058\000\187\001\
\\059\000\187\001\060\000\187\001\061\000\187\001\062\000\187\001\
\\063\000\187\001\064\000\187\001\065\000\187\001\066\000\187\001\
\\079\000\032\001\084\000\187\001\085\000\187\001\087\000\187\001\000\000\
\\001\000\003\000\188\001\029\000\188\001\035\000\188\001\036\000\188\001\
\\039\000\188\001\040\000\188\001\045\000\188\001\046\000\188\001\
\\047\000\188\001\048\000\188\001\049\000\188\001\050\000\188\001\
\\051\000\188\001\052\000\188\001\053\000\188\001\054\000\188\001\
\\055\000\188\001\056\000\188\001\057\000\188\001\058\000\188\001\
\\059\000\188\001\060\000\188\001\061\000\188\001\062\000\188\001\
\\063\000\188\001\064\000\188\001\065\000\188\001\066\000\188\001\
\\084\000\188\001\085\000\188\001\087\000\188\001\000\000\
\\001\000\003\000\189\001\029\000\189\001\035\000\189\001\036\000\189\001\
\\039\000\189\001\040\000\189\001\045\000\189\001\046\000\189\001\
\\047\000\189\001\048\000\189\001\049\000\189\001\050\000\189\001\
\\051\000\189\001\052\000\189\001\053\000\189\001\054\000\189\001\
\\055\000\189\001\056\000\189\001\057\000\189\001\058\000\189\001\
\\059\000\189\001\060\000\189\001\061\000\189\001\062\000\189\001\
\\063\000\189\001\064\000\189\001\065\000\189\001\066\000\189\001\
\\084\000\189\001\085\000\189\001\087\000\189\001\000\000\
\\001\000\003\000\190\001\026\000\190\001\028\000\190\001\029\000\190\001\
\\030\000\190\001\031\000\190\001\035\000\190\001\036\000\190\001\
\\039\000\190\001\045\000\190\001\046\000\190\001\047\000\190\001\
\\048\000\190\001\049\000\190\001\050\000\190\001\051\000\190\001\
\\052\000\190\001\053\000\190\001\054\000\190\001\055\000\190\001\
\\056\000\190\001\057\000\190\001\058\000\190\001\059\000\190\001\
\\060\000\190\001\061\000\190\001\062\000\190\001\063\000\190\001\
\\064\000\190\001\065\000\190\001\066\000\190\001\079\000\190\001\
\\084\000\190\001\085\000\190\001\000\000\
\\001\000\003\000\191\001\026\000\191\001\028\000\191\001\029\000\191\001\
\\030\000\191\001\031\000\191\001\035\000\191\001\036\000\191\001\
\\039\000\191\001\045\000\191\001\046\000\191\001\047\000\191\001\
\\048\000\191\001\049\000\191\001\050\000\191\001\051\000\191\001\
\\052\000\191\001\053\000\191\001\054\000\191\001\055\000\191\001\
\\056\000\191\001\057\000\191\001\058\000\191\001\059\000\191\001\
\\060\000\191\001\061\000\191\001\062\000\191\001\063\000\191\001\
\\064\000\191\001\065\000\191\001\066\000\191\001\079\000\191\001\
\\084\000\191\001\085\000\191\001\000\000\
\\001\000\003\000\192\001\026\000\192\001\028\000\192\001\029\000\192\001\
\\030\000\192\001\031\000\192\001\035\000\192\001\036\000\192\001\
\\039\000\192\001\045\000\192\001\046\000\192\001\047\000\192\001\
\\048\000\192\001\049\000\192\001\050\000\192\001\051\000\192\001\
\\052\000\192\001\053\000\192\001\054\000\192\001\055\000\192\001\
\\056\000\192\001\057\000\192\001\058\000\192\001\059\000\192\001\
\\060\000\192\001\061\000\192\001\062\000\192\001\063\000\192\001\
\\064\000\192\001\065\000\192\001\066\000\192\001\079\000\192\001\
\\084\000\192\001\085\000\192\001\000\000\
\\001\000\003\000\193\001\026\000\193\001\028\000\193\001\029\000\193\001\
\\030\000\193\001\031\000\193\001\035\000\193\001\036\000\193\001\
\\039\000\193\001\045\000\193\001\046\000\193\001\047\000\193\001\
\\048\000\193\001\049\000\193\001\050\000\193\001\051\000\193\001\
\\052\000\193\001\053\000\193\001\054\000\193\001\055\000\193\001\
\\056\000\193\001\057\000\193\001\058\000\193\001\059\000\193\001\
\\060\000\193\001\061\000\193\001\062\000\193\001\063\000\193\001\
\\064\000\193\001\065\000\193\001\066\000\193\001\079\000\193\001\
\\084\000\193\001\085\000\193\001\000\000\
\\001\000\003\000\194\001\026\000\194\001\028\000\194\001\029\000\194\001\
\\030\000\194\001\031\000\194\001\035\000\194\001\036\000\194\001\
\\039\000\194\001\045\000\194\001\046\000\194\001\047\000\194\001\
\\048\000\194\001\049\000\194\001\050\000\194\001\051\000\194\001\
\\052\000\194\001\053\000\194\001\054\000\194\001\055\000\194\001\
\\056\000\194\001\057\000\194\001\058\000\194\001\059\000\194\001\
\\060\000\194\001\061\000\194\001\062\000\194\001\063\000\194\001\
\\064\000\194\001\065\000\194\001\066\000\194\001\079\000\194\001\
\\084\000\194\001\085\000\194\001\000\000\
\\001\000\003\000\195\001\026\000\195\001\028\000\195\001\029\000\195\001\
\\030\000\195\001\031\000\195\001\035\000\195\001\036\000\195\001\
\\039\000\195\001\045\000\195\001\046\000\195\001\047\000\195\001\
\\048\000\195\001\049\000\195\001\050\000\195\001\051\000\195\001\
\\052\000\195\001\053\000\195\001\054\000\195\001\055\000\195\001\
\\056\000\195\001\057\000\195\001\058\000\195\001\059\000\195\001\
\\060\000\195\001\061\000\195\001\062\000\195\001\066\000\195\001\
\\079\000\195\001\085\000\195\001\000\000\
\\001\000\003\000\195\001\026\000\195\001\029\000\195\001\030\000\029\001\
\\031\000\195\001\035\000\195\001\036\000\195\001\039\000\195\001\
\\045\000\195\001\046\000\195\001\047\000\195\001\048\000\195\001\
\\049\000\195\001\050\000\195\001\051\000\195\001\052\000\195\001\
\\053\000\195\001\054\000\195\001\055\000\195\001\056\000\195\001\
\\057\000\195\001\058\000\195\001\059\000\195\001\060\000\195\001\
\\061\000\195\001\062\000\195\001\066\000\195\001\085\000\195\001\000\000\
\\001\000\003\000\196\001\026\000\196\001\028\000\196\001\029\000\196\001\
\\030\000\196\001\031\000\196\001\035\000\196\001\036\000\196\001\
\\039\000\196\001\045\000\196\001\046\000\196\001\047\000\196\001\
\\048\000\196\001\049\000\196\001\050\000\196\001\051\000\196\001\
\\052\000\196\001\053\000\196\001\054\000\196\001\055\000\196\001\
\\056\000\196\001\057\000\196\001\058\000\196\001\059\000\196\001\
\\060\000\196\001\061\000\196\001\062\000\196\001\066\000\196\001\
\\079\000\196\001\085\000\196\001\000\000\
\\001\000\003\000\197\001\026\000\197\001\028\000\197\001\029\000\197\001\
\\030\000\197\001\031\000\197\001\035\000\197\001\036\000\197\001\
\\039\000\197\001\045\000\197\001\046\000\197\001\047\000\197\001\
\\048\000\197\001\049\000\197\001\050\000\197\001\051\000\197\001\
\\052\000\197\001\053\000\197\001\054\000\197\001\055\000\197\001\
\\056\000\197\001\057\000\197\001\058\000\197\001\059\000\197\001\
\\060\000\197\001\061\000\197\001\062\000\197\001\066\000\197\001\
\\079\000\197\001\085\000\197\001\000\000\
\\001\000\003\000\198\001\026\000\198\001\028\000\198\001\029\000\198\001\
\\030\000\198\001\031\000\198\001\035\000\198\001\036\000\198\001\
\\039\000\198\001\045\000\198\001\046\000\198\001\047\000\198\001\
\\048\000\198\001\049\000\198\001\050\000\198\001\051\000\198\001\
\\052\000\198\001\053\000\198\001\054\000\198\001\055\000\198\001\
\\056\000\198\001\057\000\198\001\058\000\198\001\059\000\198\001\
\\060\000\198\001\061\000\198\001\062\000\198\001\066\000\198\001\
\\079\000\198\001\085\000\198\001\000\000\
\\001\000\003\000\199\001\026\000\199\001\028\000\199\001\029\000\199\001\
\\030\000\199\001\031\000\199\001\035\000\199\001\036\000\199\001\
\\039\000\199\001\045\000\199\001\046\000\199\001\047\000\199\001\
\\048\000\199\001\049\000\199\001\050\000\199\001\051\000\199\001\
\\052\000\199\001\053\000\199\001\054\000\199\001\055\000\199\001\
\\056\000\199\001\057\000\199\001\058\000\199\001\059\000\199\001\
\\060\000\199\001\061\000\199\001\062\000\199\001\066\000\199\001\
\\079\000\199\001\085\000\199\001\000\000\
\\001\000\003\000\200\001\026\000\200\001\028\000\200\001\029\000\200\001\
\\030\000\200\001\031\000\200\001\035\000\200\001\036\000\200\001\
\\039\000\200\001\045\000\200\001\046\000\200\001\047\000\200\001\
\\048\000\200\001\049\000\200\001\050\000\200\001\051\000\200\001\
\\052\000\200\001\053\000\200\001\054\000\200\001\055\000\200\001\
\\056\000\200\001\057\000\200\001\058\000\200\001\059\000\200\001\
\\060\000\200\001\061\000\200\001\062\000\200\001\066\000\200\001\
\\079\000\200\001\085\000\200\001\000\000\
\\001\000\003\000\201\001\026\000\201\001\028\000\201\001\029\000\201\001\
\\030\000\201\001\031\000\201\001\035\000\201\001\036\000\201\001\
\\039\000\201\001\045\000\201\001\046\000\201\001\047\000\201\001\
\\048\000\201\001\049\000\201\001\050\000\201\001\051\000\201\001\
\\052\000\201\001\053\000\201\001\054\000\201\001\055\000\201\001\
\\056\000\201\001\057\000\201\001\058\000\201\001\059\000\201\001\
\\060\000\201\001\061\000\201\001\062\000\201\001\066\000\201\001\
\\079\000\201\001\085\000\201\001\000\000\
\\001\000\003\000\202\001\026\000\202\001\028\000\202\001\029\000\202\001\
\\030\000\202\001\031\000\202\001\035\000\202\001\036\000\202\001\
\\039\000\202\001\045\000\202\001\046\000\202\001\047\000\202\001\
\\048\000\202\001\049\000\202\001\050\000\202\001\051\000\202\001\
\\052\000\202\001\053\000\202\001\054\000\202\001\055\000\202\001\
\\056\000\202\001\057\000\202\001\058\000\202\001\059\000\202\001\
\\060\000\202\001\061\000\202\001\062\000\202\001\066\000\202\001\
\\079\000\202\001\085\000\202\001\000\000\
\\001\000\003\000\203\001\026\000\203\001\028\000\203\001\029\000\203\001\
\\030\000\203\001\031\000\203\001\035\000\203\001\036\000\203\001\
\\039\000\203\001\045\000\203\001\046\000\203\001\047\000\203\001\
\\048\000\203\001\049\000\203\001\050\000\203\001\051\000\203\001\
\\052\000\203\001\053\000\203\001\054\000\203\001\055\000\203\001\
\\056\000\203\001\057\000\203\001\058\000\203\001\059\000\203\001\
\\060\000\203\001\061\000\203\001\062\000\203\001\066\000\203\001\
\\079\000\203\001\085\000\203\001\000\000\
\\001\000\003\000\204\001\026\000\204\001\028\000\204\001\029\000\204\001\
\\030\000\204\001\031\000\204\001\035\000\204\001\036\000\204\001\
\\039\000\204\001\045\000\204\001\046\000\204\001\047\000\204\001\
\\048\000\204\001\049\000\204\001\050\000\204\001\051\000\204\001\
\\052\000\204\001\053\000\204\001\054\000\204\001\055\000\204\001\
\\056\000\204\001\057\000\204\001\058\000\204\001\059\000\204\001\
\\060\000\204\001\061\000\204\001\062\000\204\001\066\000\204\001\
\\079\000\204\001\085\000\204\001\000\000\
\\001\000\003\000\205\001\026\000\205\001\028\000\205\001\029\000\205\001\
\\030\000\205\001\031\000\205\001\035\000\205\001\036\000\205\001\
\\039\000\205\001\045\000\205\001\046\000\205\001\047\000\205\001\
\\048\000\205\001\049\000\205\001\050\000\205\001\051\000\205\001\
\\052\000\205\001\053\000\205\001\066\000\205\001\079\000\205\001\000\000\
\\001\000\003\000\205\001\028\000\205\001\029\000\205\001\031\000\032\002\
\\033\000\032\002\035\000\205\001\036\000\205\001\039\000\205\001\
\\045\000\205\001\046\000\205\001\047\000\205\001\048\000\205\001\
\\049\000\205\001\050\000\205\001\051\000\205\001\052\000\205\001\
\\053\000\205\001\054\000\032\002\055\000\032\002\056\000\032\002\
\\057\000\032\002\058\000\032\002\059\000\032\002\060\000\032\002\
\\061\000\032\002\062\000\032\002\063\000\032\002\064\000\032\002\
\\065\000\032\002\066\000\205\001\079\000\205\001\084\000\032\002\
\\085\000\032\002\000\000\
\\001\000\003\000\206\001\026\000\206\001\028\000\206\001\029\000\206\001\
\\030\000\206\001\031\000\206\001\035\000\206\001\036\000\206\001\
\\039\000\206\001\045\000\206\001\046\000\206\001\047\000\206\001\
\\048\000\206\001\049\000\206\001\050\000\206\001\051\000\206\001\
\\052\000\206\001\053\000\206\001\066\000\206\001\079\000\206\001\000\000\
\\001\000\003\000\207\001\026\000\207\001\028\000\207\001\029\000\207\001\
\\030\000\207\001\031\000\207\001\035\000\207\001\036\000\207\001\
\\039\000\207\001\045\000\207\001\046\000\207\001\047\000\207\001\
\\048\000\207\001\049\000\207\001\050\000\207\001\051\000\207\001\
\\052\000\207\001\053\000\207\001\066\000\207\001\079\000\207\001\000\000\
\\001\000\003\000\208\001\026\000\208\001\028\000\208\001\029\000\208\001\
\\030\000\208\001\031\000\208\001\035\000\208\001\036\000\208\001\
\\039\000\208\001\045\000\208\001\046\000\208\001\047\000\208\001\
\\048\000\208\001\049\000\208\001\050\000\208\001\051\000\208\001\
\\052\000\208\001\053\000\208\001\066\000\208\001\079\000\208\001\000\000\
\\001\000\003\000\209\001\026\000\209\001\028\000\209\001\029\000\209\001\
\\030\000\209\001\031\000\209\001\035\000\209\001\036\000\209\001\
\\039\000\209\001\045\000\209\001\046\000\209\001\047\000\209\001\
\\048\000\209\001\049\000\209\001\050\000\209\001\051\000\209\001\
\\052\000\209\001\053\000\209\001\054\000\209\001\055\000\209\001\
\\056\000\209\001\057\000\209\001\058\000\209\001\059\000\209\001\
\\060\000\209\001\061\000\209\001\062\000\209\001\063\000\209\001\
\\064\000\209\001\065\000\209\001\066\000\209\001\079\000\209\001\
\\084\000\209\001\085\000\209\001\000\000\
\\001\000\003\000\210\001\026\000\210\001\028\000\210\001\029\000\210\001\
\\030\000\210\001\031\000\210\001\035\000\210\001\036\000\210\001\
\\039\000\210\001\045\000\210\001\046\000\210\001\047\000\210\001\
\\048\000\210\001\049\000\210\001\050\000\210\001\051\000\210\001\
\\052\000\210\001\053\000\210\001\054\000\210\001\055\000\210\001\
\\056\000\210\001\057\000\210\001\058\000\210\001\059\000\210\001\
\\060\000\210\001\061\000\210\001\062\000\210\001\063\000\210\001\
\\064\000\210\001\065\000\210\001\066\000\210\001\079\000\210\001\
\\084\000\210\001\085\000\210\001\000\000\
\\001\000\003\000\211\001\026\000\211\001\028\000\211\001\029\000\211\001\
\\030\000\211\001\031\000\211\001\035\000\211\001\036\000\211\001\
\\039\000\211\001\045\000\211\001\046\000\211\001\047\000\211\001\
\\048\000\211\001\049\000\211\001\050\000\211\001\051\000\211\001\
\\052\000\211\001\053\000\211\001\054\000\211\001\055\000\211\001\
\\056\000\211\001\057\000\211\001\058\000\211\001\059\000\211\001\
\\060\000\211\001\061\000\211\001\062\000\211\001\063\000\211\001\
\\064\000\211\001\065\000\211\001\066\000\211\001\079\000\211\001\
\\084\000\211\001\085\000\211\001\000\000\
\\001\000\003\000\212\001\026\000\212\001\028\000\212\001\029\000\212\001\
\\030\000\212\001\031\000\212\001\035\000\212\001\036\000\212\001\
\\039\000\212\001\045\000\212\001\046\000\212\001\047\000\212\001\
\\048\000\212\001\049\000\212\001\050\000\212\001\051\000\212\001\
\\052\000\212\001\053\000\212\001\054\000\212\001\055\000\212\001\
\\056\000\212\001\057\000\212\001\058\000\212\001\059\000\212\001\
\\060\000\212\001\061\000\212\001\062\000\212\001\063\000\212\001\
\\064\000\212\001\065\000\212\001\066\000\212\001\079\000\212\001\
\\084\000\212\001\085\000\212\001\000\000\
\\001\000\003\000\213\001\026\000\213\001\028\000\213\001\029\000\213\001\
\\030\000\213\001\031\000\213\001\035\000\213\001\036\000\213\001\
\\039\000\213\001\045\000\213\001\046\000\213\001\047\000\213\001\
\\048\000\213\001\049\000\213\001\050\000\213\001\051\000\213\001\
\\052\000\213\001\053\000\213\001\054\000\213\001\055\000\213\001\
\\056\000\213\001\057\000\213\001\058\000\213\001\059\000\213\001\
\\060\000\213\001\061\000\213\001\062\000\213\001\063\000\213\001\
\\064\000\213\001\065\000\213\001\066\000\213\001\079\000\213\001\
\\084\000\213\001\085\000\213\001\000\000\
\\001\000\003\000\214\001\026\000\214\001\029\000\214\001\030\000\214\001\
\\031\000\214\001\035\000\214\001\036\000\214\001\039\000\214\001\
\\045\000\048\000\050\000\043\000\051\000\042\000\052\000\041\000\
\\053\000\040\000\000\000\
\\001\000\003\000\215\001\026\000\215\001\029\000\215\001\030\000\215\001\
\\031\000\215\001\035\000\215\001\036\000\215\001\039\000\215\001\000\000\
\\001\000\003\000\218\001\026\000\218\001\028\000\218\001\029\000\218\001\
\\030\000\218\001\031\000\218\001\035\000\218\001\036\000\218\001\
\\039\000\218\001\045\000\218\001\046\000\218\001\047\000\218\001\
\\048\000\218\001\049\000\218\001\050\000\218\001\051\000\218\001\
\\052\000\218\001\053\000\218\001\066\000\218\001\079\000\218\001\000\000\
\\001\000\003\000\219\001\026\000\219\001\028\000\219\001\029\000\219\001\
\\030\000\219\001\031\000\219\001\035\000\219\001\036\000\219\001\
\\039\000\219\001\045\000\219\001\046\000\219\001\047\000\219\001\
\\048\000\219\001\049\000\219\001\050\000\219\001\051\000\219\001\
\\052\000\219\001\053\000\219\001\066\000\219\001\079\000\219\001\000\000\
\\001\000\003\000\220\001\026\000\220\001\028\000\220\001\029\000\220\001\
\\030\000\220\001\031\000\220\001\033\000\148\000\035\000\220\001\
\\036\000\220\001\039\000\220\001\045\000\220\001\046\000\220\001\
\\047\000\220\001\048\000\220\001\049\000\220\001\050\000\220\001\
\\051\000\220\001\052\000\220\001\053\000\220\001\066\000\220\001\
\\079\000\220\001\000\000\
\\001\000\003\000\227\001\026\000\227\001\028\000\227\001\029\000\227\001\
\\030\000\227\001\031\000\227\001\035\000\227\001\036\000\227\001\
\\039\000\227\001\045\000\227\001\046\000\227\001\047\000\227\001\
\\048\000\227\001\049\000\227\001\050\000\227\001\051\000\227\001\
\\052\000\227\001\053\000\227\001\066\000\227\001\079\000\227\001\000\000\
\\001\000\003\000\228\001\026\000\228\001\028\000\228\001\029\000\228\001\
\\030\000\228\001\031\000\228\001\035\000\228\001\036\000\228\001\
\\039\000\228\001\045\000\228\001\046\000\228\001\047\000\228\001\
\\048\000\228\001\049\000\228\001\050\000\228\001\051\000\228\001\
\\052\000\228\001\053\000\228\001\066\000\228\001\079\000\228\001\000\000\
\\001\000\003\000\229\001\026\000\229\001\028\000\229\001\029\000\229\001\
\\030\000\229\001\031\000\229\001\035\000\229\001\036\000\229\001\
\\039\000\229\001\045\000\229\001\046\000\229\001\047\000\229\001\
\\048\000\229\001\049\000\229\001\050\000\229\001\051\000\229\001\
\\052\000\229\001\053\000\229\001\066\000\229\001\079\000\229\001\000\000\
\\001\000\003\000\230\001\026\000\230\001\028\000\230\001\029\000\230\001\
\\030\000\230\001\031\000\230\001\033\000\133\000\035\000\230\001\
\\036\000\230\001\039\000\230\001\045\000\230\001\046\000\230\001\
\\047\000\230\001\048\000\230\001\049\000\230\001\050\000\230\001\
\\051\000\230\001\052\000\230\001\053\000\230\001\066\000\230\001\
\\079\000\230\001\000\000\
\\001\000\003\000\231\001\026\000\231\001\028\000\231\001\029\000\231\001\
\\030\000\231\001\031\000\231\001\033\000\132\000\035\000\231\001\
\\036\000\231\001\039\000\231\001\045\000\231\001\046\000\231\001\
\\047\000\231\001\048\000\231\001\049\000\231\001\050\000\231\001\
\\051\000\231\001\052\000\231\001\053\000\231\001\066\000\231\001\
\\079\000\231\001\000\000\
\\001\000\003\000\236\001\026\000\236\001\028\000\236\001\029\000\236\001\
\\030\000\236\001\031\000\236\001\035\000\236\001\036\000\236\001\
\\039\000\236\001\045\000\236\001\046\000\236\001\047\000\236\001\
\\048\000\236\001\049\000\236\001\050\000\236\001\051\000\236\001\
\\052\000\236\001\053\000\236\001\066\000\236\001\079\000\236\001\000\000\
\\001\000\003\000\237\001\026\000\237\001\028\000\237\001\029\000\237\001\
\\030\000\237\001\031\000\237\001\035\000\237\001\036\000\237\001\
\\039\000\237\001\045\000\237\001\046\000\237\001\047\000\237\001\
\\048\000\237\001\049\000\237\001\050\000\237\001\051\000\237\001\
\\052\000\237\001\053\000\237\001\066\000\237\001\079\000\237\001\000\000\
\\001\000\003\000\244\001\026\000\244\001\028\000\244\001\029\000\244\001\
\\030\000\244\001\031\000\244\001\035\000\244\001\036\000\244\001\
\\039\000\244\001\079\000\244\001\000\000\
\\001\000\003\000\245\001\026\000\245\001\028\000\245\001\029\000\245\001\
\\030\000\245\001\031\000\245\001\035\000\245\001\036\000\245\001\
\\039\000\245\001\079\000\245\001\000\000\
\\001\000\003\000\246\001\026\000\246\001\028\000\246\001\029\000\246\001\
\\030\000\246\001\031\000\246\001\035\000\246\001\036\000\246\001\
\\039\000\246\001\079\000\246\001\000\000\
\\001\000\003\000\247\001\026\000\247\001\028\000\247\001\029\000\247\001\
\\030\000\247\001\031\000\247\001\035\000\247\001\036\000\078\000\
\\039\000\247\001\045\000\048\000\046\000\047\000\047\000\046\000\
\\048\000\045\000\049\000\044\000\050\000\043\000\051\000\042\000\
\\052\000\041\000\053\000\040\000\054\000\039\000\055\000\038\000\
\\056\000\037\000\057\000\036\000\058\000\035\000\059\000\034\000\
\\060\000\033\000\061\000\032\000\062\000\031\000\063\000\030\000\
\\064\000\029\000\065\000\028\000\066\000\027\000\079\000\247\001\
\\084\000\026\000\085\000\025\000\000\000\
\\001\000\003\000\248\001\026\000\248\001\028\000\248\001\029\000\248\001\
\\030\000\248\001\031\000\248\001\035\000\248\001\036\000\248\001\
\\039\000\248\001\079\000\248\001\000\000\
\\001\000\003\000\249\001\026\000\249\001\028\000\249\001\029\000\249\001\
\\030\000\249\001\031\000\249\001\035\000\249\001\036\000\249\001\
\\039\000\249\001\079\000\249\001\000\000\
\\001\000\003\000\250\001\026\000\250\001\028\000\250\001\029\000\250\001\
\\030\000\250\001\031\000\250\001\035\000\250\001\036\000\250\001\
\\039\000\250\001\079\000\250\001\000\000\
\\001\000\003\000\251\001\026\000\251\001\028\000\251\001\029\000\251\001\
\\030\000\251\001\031\000\251\001\035\000\251\001\036\000\251\001\
\\039\000\251\001\079\000\251\001\000\000\
\\001\000\003\000\252\001\026\000\252\001\028\000\252\001\029\000\252\001\
\\030\000\252\001\031\000\252\001\035\000\252\001\036\000\252\001\
\\039\000\252\001\045\000\048\000\046\000\047\000\047\000\046\000\
\\048\000\045\000\049\000\044\000\050\000\043\000\051\000\042\000\
\\052\000\041\000\053\000\040\000\054\000\039\000\055\000\038\000\
\\056\000\037\000\057\000\036\000\058\000\035\000\059\000\034\000\
\\060\000\033\000\061\000\032\000\062\000\031\000\066\000\027\000\
\\079\000\252\001\085\000\025\000\000\000\
\\001\000\003\000\253\001\026\000\253\001\028\000\253\001\029\000\253\001\
\\030\000\253\001\031\000\253\001\035\000\253\001\036\000\253\001\
\\039\000\253\001\079\000\253\001\000\000\
\\001\000\003\000\254\001\026\000\254\001\028\000\254\001\029\000\254\001\
\\030\000\254\001\031\000\254\001\035\000\254\001\036\000\254\001\
\\039\000\254\001\079\000\254\001\000\000\
\\001\000\003\000\255\001\026\000\255\001\028\000\255\001\029\000\255\001\
\\030\000\255\001\031\000\255\001\035\000\255\001\036\000\255\001\
\\039\000\255\001\045\000\048\000\046\000\047\000\047\000\046\000\
\\048\000\045\000\049\000\044\000\050\000\043\000\051\000\042\000\
\\052\000\041\000\053\000\040\000\066\000\027\000\079\000\255\001\000\000\
\\001\000\003\000\000\002\026\000\000\002\028\000\000\002\029\000\000\002\
\\030\000\000\002\031\000\000\002\035\000\000\002\036\000\000\002\
\\039\000\000\002\045\000\000\002\046\000\000\002\047\000\000\002\
\\048\000\000\002\049\000\000\002\050\000\000\002\051\000\000\002\
\\052\000\000\002\053\000\000\002\054\000\000\002\055\000\000\002\
\\056\000\000\002\057\000\000\002\058\000\000\002\059\000\000\002\
\\060\000\000\002\061\000\000\002\062\000\000\002\066\000\000\002\
\\079\000\000\002\085\000\000\002\000\000\
\\001\000\003\000\001\002\026\000\001\002\028\000\001\002\029\000\001\002\
\\030\000\001\002\031\000\001\002\035\000\001\002\036\000\001\002\
\\039\000\001\002\045\000\001\002\046\000\001\002\047\000\001\002\
\\048\000\001\002\049\000\001\002\050\000\001\002\051\000\001\002\
\\052\000\001\002\053\000\001\002\054\000\001\002\055\000\001\002\
\\056\000\001\002\057\000\001\002\058\000\001\002\059\000\001\002\
\\060\000\001\002\061\000\001\002\062\000\001\002\066\000\001\002\
\\079\000\001\002\085\000\001\002\000\000\
\\001\000\003\000\002\002\026\000\002\002\028\000\002\002\029\000\002\002\
\\030\000\002\002\031\000\002\002\035\000\002\002\036\000\002\002\
\\039\000\002\002\045\000\002\002\046\000\002\002\047\000\002\002\
\\048\000\002\002\049\000\002\002\050\000\002\002\051\000\002\002\
\\052\000\002\002\053\000\002\002\054\000\002\002\055\000\002\002\
\\056\000\002\002\057\000\002\002\058\000\002\002\059\000\002\002\
\\060\000\002\002\061\000\002\002\062\000\002\002\066\000\002\002\
\\079\000\002\002\085\000\002\002\000\000\
\\001\000\003\000\003\002\026\000\003\002\028\000\003\002\029\000\003\002\
\\030\000\003\002\031\000\003\002\035\000\003\002\036\000\003\002\
\\039\000\003\002\045\000\003\002\046\000\003\002\047\000\003\002\
\\048\000\003\002\049\000\003\002\050\000\003\002\051\000\003\002\
\\052\000\003\002\053\000\003\002\054\000\003\002\055\000\003\002\
\\056\000\003\002\057\000\003\002\058\000\003\002\059\000\003\002\
\\060\000\003\002\061\000\003\002\062\000\003\002\066\000\003\002\
\\079\000\003\002\085\000\003\002\000\000\
\\001\000\003\000\004\002\026\000\004\002\028\000\004\002\029\000\004\002\
\\030\000\004\002\031\000\004\002\035\000\004\002\036\000\004\002\
\\039\000\004\002\045\000\004\002\046\000\004\002\047\000\004\002\
\\048\000\004\002\049\000\004\002\050\000\004\002\051\000\004\002\
\\052\000\004\002\053\000\004\002\054\000\004\002\055\000\004\002\
\\056\000\004\002\057\000\004\002\058\000\004\002\059\000\004\002\
\\060\000\004\002\061\000\004\002\062\000\004\002\066\000\004\002\
\\079\000\004\002\085\000\004\002\000\000\
\\001\000\003\000\052\000\026\000\025\002\029\000\025\002\030\000\025\002\
\\031\000\025\002\035\000\025\002\036\000\025\002\039\000\025\002\
\\045\000\048\000\050\000\043\000\051\000\042\000\052\000\041\000\
\\053\000\040\000\000\000\
\\001\000\003\000\052\000\026\000\026\002\029\000\026\002\030\000\026\002\
\\031\000\026\002\035\000\026\002\036\000\026\002\039\000\026\002\000\000\
\\001\000\003\000\052\000\026\000\074\002\029\000\028\001\030\000\074\002\
\\031\000\057\000\035\000\037\002\036\000\060\000\039\000\049\000\000\000\
\\001\000\003\000\052\000\028\000\069\000\029\000\051\000\035\000\037\002\
\\036\000\060\000\039\000\049\000\000\000\
\\001\000\003\000\052\000\028\000\069\000\029\000\051\000\035\000\037\002\
\\036\000\060\000\039\000\049\000\079\000\037\002\000\000\
\\001\000\003\000\052\000\029\000\051\000\035\000\037\002\036\000\050\000\
\\039\000\049\000\045\000\048\000\046\000\047\000\047\000\046\000\
\\048\000\045\000\049\000\044\000\050\000\043\000\051\000\042\000\
\\052\000\041\000\053\000\040\000\054\000\039\000\055\000\038\000\
\\056\000\037\000\057\000\036\000\058\000\035\000\059\000\034\000\
\\060\000\033\000\061\000\032\000\062\000\031\000\063\000\030\000\
\\064\000\029\000\065\000\028\000\066\000\027\000\084\000\026\000\
\\085\000\025\000\000\000\
\\001\000\003\000\052\000\029\000\051\000\035\000\037\002\036\000\060\000\
\\039\000\049\000\000\000\
\\001\000\003\000\052\000\029\000\227\000\030\000\216\001\031\000\057\000\000\000\
\\001\000\003\000\052\000\029\000\227\000\030\000\047\001\031\000\057\000\
\\036\000\176\002\045\000\176\002\046\000\176\002\047\000\176\002\
\\048\000\176\002\049\000\176\002\050\000\176\002\051\000\176\002\
\\052\000\176\002\053\000\176\002\054\000\176\002\055\000\176\002\
\\056\000\176\002\057\000\176\002\058\000\176\002\059\000\176\002\
\\060\000\176\002\061\000\176\002\062\000\176\002\063\000\176\002\
\\064\000\176\002\065\000\176\002\066\000\176\002\084\000\176\002\
\\085\000\176\002\000\000\
\\001\000\003\000\052\000\029\000\028\001\030\000\047\001\031\000\057\000\
\\035\000\037\002\036\000\176\002\039\000\049\000\045\000\176\002\
\\046\000\176\002\047\000\176\002\048\000\176\002\049\000\176\002\
\\050\000\176\002\051\000\176\002\052\000\176\002\053\000\176\002\
\\054\000\176\002\055\000\176\002\056\000\176\002\057\000\176\002\
\\058\000\176\002\059\000\176\002\060\000\176\002\061\000\176\002\
\\062\000\176\002\063\000\176\002\064\000\176\002\065\000\176\002\
\\066\000\176\002\084\000\176\002\085\000\176\002\000\000\
\\001\000\023\000\023\002\026\000\023\002\027\000\023\002\028\000\023\002\
\\029\000\058\000\030\000\023\002\031\000\057\000\033\000\023\002\
\\036\000\023\002\039\000\023\002\045\000\023\002\046\000\023\002\
\\047\000\023\002\048\000\023\002\049\000\023\002\050\000\023\002\
\\051\000\023\002\052\000\023\002\053\000\023\002\054\000\023\002\
\\055\000\023\002\056\000\023\002\057\000\023\002\058\000\023\002\
\\059\000\023\002\060\000\023\002\061\000\023\002\062\000\023\002\
\\063\000\023\002\064\000\023\002\065\000\023\002\066\000\023\002\
\\079\000\023\002\084\000\023\002\085\000\023\002\000\000\
\\001\000\023\000\024\002\026\000\024\002\027\000\024\002\028\000\024\002\
\\029\000\058\000\030\000\024\002\031\000\057\000\033\000\024\002\
\\036\000\024\002\039\000\024\002\045\000\024\002\046\000\024\002\
\\047\000\024\002\048\000\024\002\049\000\024\002\050\000\024\002\
\\051\000\024\002\052\000\024\002\053\000\024\002\054\000\024\002\
\\055\000\024\002\056\000\024\002\057\000\024\002\058\000\024\002\
\\059\000\024\002\060\000\024\002\061\000\024\002\062\000\024\002\
\\063\000\024\002\064\000\024\002\065\000\024\002\066\000\024\002\
\\079\000\024\002\084\000\024\002\085\000\024\002\000\000\
\\001\000\023\000\029\002\026\000\029\002\027\000\029\002\028\000\029\002\
\\029\000\029\002\030\000\029\002\031\000\029\002\033\000\029\002\
\\036\000\029\002\039\000\029\002\045\000\029\002\046\000\029\002\
\\047\000\029\002\048\000\029\002\049\000\029\002\050\000\029\002\
\\051\000\029\002\052\000\029\002\053\000\029\002\054\000\029\002\
\\055\000\029\002\056\000\029\002\057\000\029\002\058\000\029\002\
\\059\000\029\002\060\000\029\002\061\000\029\002\062\000\029\002\
\\063\000\029\002\064\000\029\002\065\000\029\002\066\000\029\002\
\\079\000\029\002\084\000\029\002\085\000\029\002\000\000\
\\001\000\023\000\030\002\026\000\030\002\027\000\030\002\028\000\030\002\
\\029\000\030\002\030\000\030\002\031\000\030\002\033\000\030\002\
\\036\000\030\002\039\000\030\002\045\000\030\002\046\000\030\002\
\\047\000\030\002\048\000\030\002\049\000\030\002\050\000\030\002\
\\051\000\030\002\052\000\030\002\053\000\030\002\054\000\030\002\
\\055\000\030\002\056\000\030\002\057\000\030\002\058\000\030\002\
\\059\000\030\002\060\000\030\002\061\000\030\002\062\000\030\002\
\\063\000\030\002\064\000\030\002\065\000\030\002\066\000\030\002\
\\079\000\030\002\084\000\030\002\085\000\030\002\000\000\
\\001\000\023\000\031\002\026\000\031\002\027\000\031\002\028\000\031\002\
\\029\000\031\002\030\000\031\002\031\000\031\002\033\000\031\002\
\\036\000\031\002\039\000\031\002\045\000\031\002\046\000\031\002\
\\047\000\031\002\048\000\031\002\049\000\031\002\050\000\031\002\
\\051\000\031\002\052\000\031\002\053\000\031\002\054\000\031\002\
\\055\000\031\002\056\000\031\002\057\000\031\002\058\000\031\002\
\\059\000\031\002\060\000\031\002\061\000\031\002\062\000\031\002\
\\063\000\031\002\064\000\031\002\065\000\031\002\066\000\031\002\
\\079\000\031\002\084\000\031\002\085\000\031\002\000\000\
\\001\000\023\000\032\002\026\000\032\002\027\000\032\002\028\000\032\002\
\\029\000\032\002\030\000\032\002\031\000\032\002\033\000\032\002\
\\036\000\032\002\039\000\032\002\045\000\032\002\046\000\032\002\
\\047\000\032\002\048\000\032\002\049\000\032\002\050\000\032\002\
\\051\000\032\002\052\000\032\002\053\000\032\002\054\000\032\002\
\\055\000\032\002\056\000\032\002\057\000\032\002\058\000\032\002\
\\059\000\032\002\060\000\032\002\061\000\032\002\062\000\032\002\
\\063\000\032\002\064\000\032\002\065\000\032\002\066\000\032\002\
\\079\000\032\002\084\000\032\002\085\000\032\002\000\000\
\\001\000\023\000\033\002\026\000\033\002\027\000\033\002\028\000\033\002\
\\029\000\033\002\030\000\033\002\031\000\033\002\033\000\033\002\
\\036\000\033\002\039\000\033\002\045\000\033\002\046\000\033\002\
\\047\000\033\002\048\000\033\002\049\000\033\002\050\000\033\002\
\\051\000\033\002\052\000\033\002\053\000\033\002\054\000\033\002\
\\055\000\033\002\056\000\033\002\057\000\033\002\058\000\033\002\
\\059\000\033\002\060\000\033\002\061\000\033\002\062\000\033\002\
\\063\000\033\002\064\000\033\002\065\000\033\002\066\000\033\002\
\\079\000\033\002\084\000\033\002\085\000\033\002\000\000\
\\001\000\023\000\034\002\026\000\034\002\027\000\034\002\028\000\034\002\
\\029\000\034\002\030\000\034\002\031\000\034\002\033\000\034\002\
\\036\000\034\002\039\000\034\002\045\000\034\002\046\000\034\002\
\\047\000\034\002\048\000\034\002\049\000\034\002\050\000\034\002\
\\051\000\034\002\052\000\034\002\053\000\034\002\054\000\034\002\
\\055\000\034\002\056\000\034\002\057\000\034\002\058\000\034\002\
\\059\000\034\002\060\000\034\002\061\000\034\002\062\000\034\002\
\\063\000\034\002\064\000\034\002\065\000\034\002\066\000\034\002\
\\079\000\034\002\084\000\034\002\085\000\034\002\000\000\
\\001\000\023\000\035\002\026\000\035\002\027\000\035\002\028\000\035\002\
\\029\000\035\002\030\000\035\002\031\000\035\002\033\000\035\002\
\\036\000\035\002\039\000\035\002\045\000\035\002\046\000\035\002\
\\047\000\035\002\048\000\035\002\049\000\035\002\050\000\035\002\
\\051\000\035\002\052\000\035\002\053\000\035\002\054\000\035\002\
\\055\000\035\002\056\000\035\002\057\000\035\002\058\000\035\002\
\\059\000\035\002\060\000\035\002\061\000\035\002\062\000\035\002\
\\063\000\035\002\064\000\035\002\065\000\035\002\066\000\035\002\
\\079\000\035\002\084\000\035\002\085\000\035\002\000\000\
\\001\000\023\000\036\002\026\000\036\002\027\000\036\002\028\000\036\002\
\\029\000\036\002\030\000\036\002\031\000\036\002\033\000\036\002\
\\036\000\036\002\039\000\036\002\045\000\036\002\046\000\036\002\
\\047\000\036\002\048\000\036\002\049\000\036\002\050\000\036\002\
\\051\000\036\002\052\000\036\002\053\000\036\002\054\000\036\002\
\\055\000\036\002\056\000\036\002\057\000\036\002\058\000\036\002\
\\059\000\036\002\060\000\036\002\061\000\036\002\062\000\036\002\
\\063\000\036\002\064\000\036\002\065\000\036\002\066\000\036\002\
\\079\000\036\002\084\000\036\002\085\000\036\002\000\000\
\\001\000\023\000\062\002\026\000\062\002\027\000\062\002\028\000\062\002\
\\029\000\062\002\030\000\062\002\031\000\062\002\033\000\062\002\
\\036\000\062\002\039\000\062\002\045\000\062\002\046\000\062\002\
\\047\000\062\002\048\000\062\002\049\000\062\002\050\000\062\002\
\\051\000\062\002\052\000\062\002\053\000\062\002\054\000\062\002\
\\055\000\062\002\056\000\062\002\057\000\062\002\058\000\062\002\
\\059\000\062\002\060\000\062\002\061\000\062\002\062\000\062\002\
\\063\000\062\002\064\000\062\002\065\000\062\002\066\000\062\002\
\\079\000\062\002\084\000\062\002\085\000\062\002\000\000\
\\001\000\023\000\063\002\026\000\063\002\027\000\063\002\028\000\063\002\
\\029\000\063\002\030\000\063\002\031\000\063\002\033\000\063\002\
\\036\000\063\002\039\000\063\002\045\000\063\002\046\000\063\002\
\\047\000\063\002\048\000\063\002\049\000\063\002\050\000\063\002\
\\051\000\063\002\052\000\063\002\053\000\063\002\054\000\063\002\
\\055\000\063\002\056\000\063\002\057\000\063\002\058\000\063\002\
\\059\000\063\002\060\000\063\002\061\000\063\002\062\000\063\002\
\\063\000\063\002\064\000\063\002\065\000\063\002\066\000\063\002\
\\079\000\063\002\084\000\063\002\085\000\063\002\000\000\
\\001\000\023\000\064\002\026\000\064\002\027\000\064\002\028\000\064\002\
\\029\000\064\002\030\000\064\002\031\000\064\002\033\000\064\002\
\\036\000\064\002\039\000\064\002\045\000\064\002\046\000\064\002\
\\047\000\064\002\048\000\064\002\049\000\064\002\050\000\064\002\
\\051\000\064\002\052\000\064\002\053\000\064\002\054\000\064\002\
\\055\000\064\002\056\000\064\002\057\000\064\002\058\000\064\002\
\\059\000\064\002\060\000\064\002\061\000\064\002\062\000\064\002\
\\063\000\064\002\064\000\064\002\065\000\064\002\066\000\064\002\
\\079\000\064\002\084\000\064\002\085\000\064\002\000\000\
\\001\000\023\000\065\002\026\000\065\002\027\000\065\002\028\000\065\002\
\\029\000\065\002\030\000\065\002\031\000\065\002\033\000\065\002\
\\036\000\065\002\039\000\065\002\045\000\065\002\046\000\065\002\
\\047\000\065\002\048\000\065\002\049\000\065\002\050\000\065\002\
\\051\000\065\002\052\000\065\002\053\000\065\002\054\000\065\002\
\\055\000\065\002\056\000\065\002\057\000\065\002\058\000\065\002\
\\059\000\065\002\060\000\065\002\061\000\065\002\062\000\065\002\
\\063\000\065\002\064\000\065\002\065\000\065\002\066\000\065\002\
\\079\000\065\002\084\000\065\002\085\000\065\002\000\000\
\\001\000\023\000\066\002\026\000\066\002\027\000\066\002\028\000\066\002\
\\029\000\066\002\030\000\066\002\031\000\066\002\033\000\066\002\
\\036\000\066\002\039\000\066\002\045\000\066\002\046\000\066\002\
\\047\000\066\002\048\000\066\002\049\000\066\002\050\000\066\002\
\\051\000\066\002\052\000\066\002\053\000\066\002\054\000\066\002\
\\055\000\066\002\056\000\066\002\057\000\066\002\058\000\066\002\
\\059\000\066\002\060\000\066\002\061\000\066\002\062\000\066\002\
\\063\000\066\002\064\000\066\002\065\000\066\002\066\000\066\002\
\\079\000\066\002\084\000\066\002\085\000\066\002\000\000\
\\001\000\023\000\067\002\026\000\067\002\027\000\067\002\028\000\067\002\
\\029\000\067\002\030\000\067\002\031\000\067\002\033\000\067\002\
\\036\000\067\002\039\000\067\002\045\000\067\002\046\000\067\002\
\\047\000\067\002\048\000\067\002\049\000\067\002\050\000\067\002\
\\051\000\067\002\052\000\067\002\053\000\067\002\054\000\067\002\
\\055\000\067\002\056\000\067\002\057\000\067\002\058\000\067\002\
\\059\000\067\002\060\000\067\002\061\000\067\002\062\000\067\002\
\\063\000\067\002\064\000\067\002\065\000\067\002\066\000\067\002\
\\079\000\067\002\084\000\067\002\085\000\067\002\000\000\
\\001\000\023\000\179\002\026\000\179\002\027\000\179\002\028\000\179\002\
\\029\000\179\002\030\000\179\002\031\000\179\002\033\000\179\002\
\\036\000\179\002\039\000\179\002\045\000\179\002\046\000\179\002\
\\047\000\179\002\048\000\179\002\049\000\179\002\050\000\179\002\
\\051\000\179\002\052\000\179\002\053\000\179\002\054\000\179\002\
\\055\000\179\002\056\000\179\002\057\000\179\002\058\000\179\002\
\\059\000\179\002\060\000\179\002\061\000\179\002\062\000\179\002\
\\063\000\179\002\064\000\179\002\065\000\179\002\066\000\179\002\
\\079\000\179\002\084\000\179\002\085\000\179\002\000\000\
\\001\000\023\000\128\000\026\000\007\002\027\000\127\000\028\000\126\000\
\\033\000\240\001\036\000\176\002\039\000\007\002\045\000\176\002\
\\046\000\176\002\047\000\176\002\048\000\176\002\049\000\176\002\
\\050\000\176\002\051\000\176\002\052\000\176\002\053\000\176\002\
\\054\000\176\002\055\000\176\002\056\000\176\002\057\000\176\002\
\\058\000\176\002\059\000\176\002\060\000\176\002\061\000\176\002\
\\062\000\176\002\063\000\176\002\064\000\176\002\065\000\176\002\
\\066\000\176\002\079\000\007\002\084\000\176\002\085\000\176\002\000\000\
\\001\000\023\000\128\000\026\000\007\002\027\000\127\000\028\000\126\000\
\\039\000\007\002\079\000\007\002\000\000\
\\001\000\023\000\050\001\026\000\225\001\034\000\225\001\000\000\
\\001\000\026\000\226\001\034\000\226\001\000\000\
\\001\000\026\000\005\002\039\000\005\002\079\000\005\002\000\000\
\\001\000\026\000\006\002\039\000\006\002\079\000\006\002\000\000\
\\001\000\026\000\008\002\039\000\008\002\079\000\008\002\000\000\
\\001\000\026\000\009\002\039\000\009\002\079\000\009\002\000\000\
\\001\000\026\000\010\002\039\000\010\002\079\000\010\002\000\000\
\\001\000\026\000\011\002\039\000\011\002\079\000\011\002\000\000\
\\001\000\026\000\012\002\039\000\012\002\079\000\012\002\000\000\
\\001\000\026\000\013\002\039\000\013\002\079\000\013\002\000\000\
\\001\000\026\000\014\002\039\000\014\002\079\000\014\002\000\000\
\\001\000\026\000\015\002\034\000\015\002\000\000\
\\001\000\026\000\016\002\034\000\016\002\000\000\
\\001\000\026\000\017\002\034\000\017\002\000\000\
\\001\000\026\000\021\002\034\000\021\002\000\000\
\\001\000\026\000\022\002\034\000\022\002\000\000\
\\001\000\026\000\027\002\029\000\027\002\030\000\027\002\031\000\027\002\
\\035\000\027\002\036\000\027\002\039\000\027\002\000\000\
\\001\000\026\000\028\002\029\000\028\002\030\000\028\002\031\000\028\002\
\\035\000\028\002\036\000\028\002\039\000\028\002\000\000\
\\001\000\026\000\042\002\030\000\042\002\000\000\
\\001\000\026\000\043\002\030\000\043\002\000\000\
\\001\000\026\000\044\002\030\000\044\002\035\000\240\000\036\000\239\000\
\\050\000\238\000\051\000\237\000\053\000\236\000\000\000\
\\001\000\026\000\045\002\029\000\051\001\030\000\045\002\000\000\
\\001\000\026\000\046\002\030\000\046\002\000\000\
\\001\000\026\000\047\002\029\000\047\002\030\000\047\002\000\000\
\\001\000\026\000\048\002\029\000\048\002\030\000\048\002\000\000\
\\001\000\026\000\049\002\029\000\049\002\030\000\049\002\000\000\
\\001\000\026\000\050\002\029\000\050\002\030\000\050\002\000\000\
\\001\000\026\000\051\002\029\000\051\002\030\000\051\002\000\000\
\\001\000\026\000\052\002\029\000\028\001\030\000\052\002\031\000\057\000\
\\035\000\037\002\036\000\060\000\039\000\049\000\000\000\
\\001\000\026\000\053\002\029\000\043\001\030\000\053\002\031\000\057\000\000\000\
\\001\000\026\000\054\002\029\000\043\001\030\000\054\002\031\000\057\000\000\000\
\\001\000\026\000\055\002\029\000\055\002\030\000\055\002\031\000\055\002\000\000\
\\001\000\026\000\056\002\029\000\056\002\030\000\056\002\031\000\056\002\000\000\
\\001\000\026\000\057\002\029\000\057\002\030\000\057\002\031\000\057\002\000\000\
\\001\000\026\000\058\002\029\000\058\002\030\000\058\002\031\000\058\002\000\000\
\\001\000\026\000\059\002\029\000\059\002\030\000\059\002\031\000\059\002\000\000\
\\001\000\026\000\060\002\029\000\060\002\030\000\060\002\031\000\060\002\000\000\
\\001\000\026\000\061\002\029\000\061\002\030\000\061\002\031\000\061\002\000\000\
\\001\000\026\000\068\002\030\000\068\002\000\000\
\\001\000\026\000\069\002\030\000\069\002\000\000\
\\001\000\026\000\070\002\030\000\070\002\000\000\
\\001\000\026\000\071\002\030\000\071\002\000\000\
\\001\000\026\000\072\002\030\000\072\002\000\000\
\\001\000\026\000\073\002\030\000\073\002\000\000\
\\001\000\026\000\075\002\030\000\075\002\000\000\
\\001\000\026\000\107\002\028\000\107\002\030\000\107\002\032\000\107\002\
\\079\000\107\002\000\000\
\\001\000\026\000\108\002\028\000\108\002\030\000\108\002\032\000\108\002\
\\079\000\108\002\000\000\
\\001\000\026\000\109\002\028\000\109\002\030\000\109\002\032\000\109\002\
\\039\000\109\002\079\000\109\002\000\000\
\\001\000\026\000\110\002\028\000\110\002\030\000\110\002\032\000\110\002\
\\039\000\110\002\079\000\110\002\000\000\
\\001\000\026\000\111\002\028\000\111\002\030\000\111\002\032\000\111\002\
\\039\000\111\002\079\000\111\002\000\000\
\\001\000\026\000\113\002\028\000\113\002\030\000\113\002\032\000\113\002\
\\034\000\113\002\039\000\113\002\079\000\113\002\000\000\
\\001\000\026\000\114\002\028\000\114\002\032\000\114\002\034\000\114\002\
\\039\000\114\002\079\000\114\002\000\000\
\\001\000\026\000\174\002\030\000\174\002\000\000\
\\001\000\026\000\175\002\030\000\175\002\000\000\
\\001\000\026\000\130\000\039\000\049\000\079\000\037\002\000\000\
\\001\000\026\000\016\001\028\000\057\001\000\000\
\\001\000\026\000\016\001\030\000\106\002\079\000\106\002\000\000\
\\001\000\026\000\016\001\030\000\015\001\000\000\
\\001\000\026\000\016\001\030\000\133\001\000\000\
\\001\000\026\000\016\001\030\000\162\001\000\000\
\\001\000\026\000\016\001\030\000\164\001\000\000\
\\001\000\026\000\016\001\030\000\165\001\000\000\
\\001\000\026\000\016\001\030\000\170\001\000\000\
\\001\000\026\000\016\001\032\000\054\001\000\000\
\\001\000\026\000\022\001\030\000\021\001\000\000\
\\001\000\026\000\024\001\030\000\023\001\000\000\
\\001\000\026\000\049\001\034\000\222\001\000\000\
\\001\000\026\000\053\001\030\000\052\001\000\000\
\\001\000\026\000\056\001\030\000\055\001\000\000\
\\001\000\026\000\094\001\034\000\093\001\000\000\
\\001\000\026\000\132\001\034\000\131\001\000\000\
\\001\000\026\000\135\001\034\000\134\001\000\000\
\\001\000\028\000\114\001\000\000\
\\001\000\028\000\144\001\000\000\
\\001\000\029\000\051\000\035\000\037\002\036\000\060\000\039\000\049\000\000\000\
\\001\000\029\000\084\000\000\000\
\\001\000\029\000\087\000\000\000\
\\001\000\029\000\151\000\000\000\
\\001\000\029\000\227\000\030\000\052\002\031\000\057\000\000\000\
\\001\000\029\000\118\001\000\000\
\\001\000\029\000\122\001\000\000\
\\001\000\029\000\124\001\000\000\
\\001\000\029\000\125\001\000\000\
\\001\000\029\000\158\001\000\000\
\\001\000\030\000\217\001\000\000\
\\001\000\030\000\176\002\035\000\176\002\036\000\176\002\045\000\176\002\
\\046\000\176\002\047\000\176\002\048\000\176\002\049\000\176\002\
\\050\000\176\002\051\000\176\002\052\000\176\002\053\000\176\002\
\\054\000\176\002\055\000\176\002\056\000\176\002\057\000\176\002\
\\058\000\176\002\059\000\176\002\060\000\176\002\061\000\176\002\
\\062\000\176\002\063\000\176\002\064\000\176\002\065\000\176\002\
\\066\000\176\002\084\000\176\002\085\000\176\002\000\000\
\\001\000\030\000\152\000\000\000\
\\001\000\030\000\204\000\035\000\203\000\036\000\078\000\045\000\048\000\
\\046\000\047\000\047\000\046\000\048\000\045\000\049\000\044\000\
\\050\000\043\000\051\000\042\000\052\000\041\000\053\000\040\000\
\\054\000\039\000\055\000\038\000\056\000\037\000\057\000\036\000\
\\058\000\202\000\059\000\034\000\060\000\033\000\061\000\032\000\
\\062\000\031\000\063\000\030\000\064\000\029\000\065\000\028\000\
\\066\000\027\000\084\000\026\000\085\000\025\000\000\000\
\\001\000\030\000\221\000\000\000\
\\001\000\030\000\228\000\000\000\
\\001\000\030\000\017\001\000\000\
\\001\000\030\000\018\001\000\000\
\\001\000\030\000\058\001\000\000\
\\001\000\030\000\059\001\000\000\
\\001\000\030\000\096\001\036\000\176\002\045\000\176\002\046\000\176\002\
\\047\000\176\002\048\000\176\002\049\000\176\002\050\000\176\002\
\\051\000\176\002\052\000\176\002\053\000\176\002\054\000\176\002\
\\055\000\176\002\056\000\176\002\057\000\176\002\058\000\176\002\
\\059\000\176\002\060\000\176\002\061\000\176\002\062\000\176\002\
\\063\000\176\002\064\000\176\002\065\000\176\002\066\000\176\002\
\\084\000\176\002\085\000\176\002\000\000\
\\001\000\030\000\097\001\000\000\
\\001\000\030\000\100\001\000\000\
\\001\000\030\000\127\001\000\000\
\\001\000\030\000\128\001\000\000\
\\001\000\030\000\176\001\000\000\
\\001\000\032\000\182\000\000\000\
\\001\000\033\000\232\001\035\000\232\001\036\000\232\001\000\000\
\\001\000\033\000\233\001\035\000\233\001\036\000\233\001\000\000\
\\001\000\033\000\240\001\036\000\176\002\045\000\176\002\046\000\176\002\
\\047\000\176\002\048\000\176\002\049\000\176\002\050\000\176\002\
\\051\000\176\002\052\000\176\002\053\000\176\002\054\000\176\002\
\\055\000\176\002\056\000\176\002\057\000\176\002\058\000\176\002\
\\059\000\176\002\060\000\176\002\061\000\176\002\062\000\176\002\
\\063\000\176\002\064\000\176\002\065\000\176\002\066\000\176\002\
\\084\000\176\002\085\000\176\002\000\000\
\\001\000\033\000\241\001\000\000\
\\001\000\033\000\177\002\000\000\
\\001\000\033\000\179\002\036\000\078\000\045\000\048\000\046\000\047\000\
\\047\000\046\000\048\000\045\000\049\000\044\000\050\000\043\000\
\\051\000\042\000\052\000\041\000\053\000\040\000\054\000\039\000\
\\055\000\038\000\056\000\037\000\057\000\036\000\058\000\035\000\
\\059\000\034\000\060\000\033\000\061\000\032\000\062\000\031\000\
\\063\000\030\000\064\000\029\000\065\000\028\000\066\000\027\000\
\\084\000\026\000\085\000\025\000\000\000\
\\001\000\033\000\073\000\035\000\072\000\036\000\071\000\000\000\
\\001\000\033\000\086\000\035\000\085\000\000\000\
\\001\000\033\000\195\000\000\000\
\\001\000\034\000\221\001\035\000\223\001\000\000\
\\001\000\034\000\234\001\036\000\178\002\045\000\178\002\046\000\178\002\
\\047\000\178\002\048\000\178\002\049\000\178\002\050\000\178\002\
\\051\000\178\002\052\000\178\002\053\000\178\002\054\000\178\002\
\\055\000\178\002\056\000\178\002\057\000\178\002\058\000\178\002\
\\059\000\178\002\060\000\178\002\061\000\178\002\062\000\178\002\
\\063\000\178\002\064\000\178\002\065\000\178\002\066\000\178\002\
\\084\000\178\002\085\000\178\002\000\000\
\\001\000\034\000\235\001\000\000\
\\001\000\034\000\179\002\000\000\
\\001\000\034\000\179\002\036\000\078\000\045\000\048\000\046\000\047\000\
\\047\000\046\000\048\000\045\000\049\000\044\000\050\000\043\000\
\\051\000\042\000\052\000\041\000\053\000\040\000\054\000\039\000\
\\055\000\038\000\056\000\037\000\057\000\036\000\058\000\035\000\
\\059\000\034\000\060\000\033\000\061\000\032\000\062\000\031\000\
\\063\000\030\000\064\000\029\000\065\000\028\000\066\000\027\000\
\\084\000\026\000\085\000\025\000\000\000\
\\001\000\034\000\220\000\000\000\
\\001\000\034\000\232\000\000\000\
\\001\000\034\000\038\001\000\000\
\\001\000\034\000\039\001\000\000\
\\001\000\034\000\048\001\000\000\
\\001\000\034\000\136\001\000\000\
\\001\000\034\000\159\001\000\000\
\\001\000\034\000\160\001\000\000\
\\001\000\034\000\167\001\000\000\
\\001\000\035\000\224\001\000\000\
\\001\000\035\000\038\002\039\000\049\000\079\000\038\002\000\000\
\\001\000\035\000\039\002\039\000\039\002\079\000\039\002\000\000\
\\001\000\035\000\040\002\039\000\040\002\079\000\040\002\000\000\
\\001\000\035\000\041\002\039\000\041\002\079\000\041\002\000\000\
\\001\000\035\000\062\000\000\000\
\\001\000\035\000\062\000\079\000\243\001\000\000\
\\001\000\035\000\231\000\000\000\
\\001\000\035\000\247\000\036\000\246\000\000\000\
\\001\000\035\000\090\001\036\000\089\001\080\000\088\001\000\000\
\\001\000\035\000\113\001\000\000\
\\001\000\036\000\078\000\045\000\048\000\046\000\047\000\047\000\046\000\
\\048\000\045\000\049\000\044\000\050\000\043\000\051\000\042\000\
\\052\000\041\000\053\000\040\000\054\000\039\000\055\000\038\000\
\\056\000\037\000\057\000\036\000\058\000\035\000\059\000\034\000\
\\060\000\033\000\061\000\032\000\062\000\031\000\063\000\030\000\
\\064\000\029\000\065\000\028\000\066\000\027\000\080\000\086\001\
\\084\000\026\000\085\000\025\000\000\000\
\\001\000\036\000\078\000\045\000\048\000\046\000\047\000\047\000\046\000\
\\048\000\045\000\049\000\044\000\050\000\043\000\051\000\042\000\
\\052\000\041\000\053\000\040\000\054\000\039\000\055\000\038\000\
\\056\000\037\000\057\000\036\000\058\000\035\000\059\000\034\000\
\\060\000\033\000\061\000\032\000\062\000\031\000\063\000\030\000\
\\064\000\029\000\065\000\028\000\066\000\027\000\084\000\026\000\
\\085\000\025\000\000\000\
\\001\000\036\000\078\000\045\000\048\000\046\000\047\000\047\000\046\000\
\\048\000\045\000\049\000\044\000\050\000\043\000\051\000\042\000\
\\052\000\041\000\053\000\040\000\054\000\039\000\055\000\038\000\
\\056\000\037\000\057\000\036\000\058\000\202\000\059\000\034\000\
\\060\000\033\000\061\000\032\000\062\000\031\000\063\000\030\000\
\\064\000\029\000\065\000\028\000\066\000\027\000\084\000\026\000\
\\085\000\025\000\000\000\
\\001\000\070\000\137\001\000\000\
\\001\000\079\000\242\001\000\000\
\\001\000\079\000\070\000\000\000\
\\001\000\079\000\206\000\000\000\
\\001\000\079\000\031\001\000\000\
\\001\000\079\000\106\001\000\000\
\\001\000\079\000\119\001\000\000\
\\001\000\079\000\120\001\000\000\
\\001\000\079\000\141\001\000\000\
\\001\000\079\000\142\001\000\000\
\\001\000\079\000\163\001\000\000\
\\001\000\079\000\171\001\000\000\
\\001\000\079\000\173\001\000\000\
\\001\000\087\000\000\000\000\000\
\"
val actionRowNumbers =
"\120\000\186\000\115\000\119\000\
\\121\000\192\000\028\001\080\001\
\\079\001\083\001\057\001\166\000\
\\185\000\094\001\146\000\061\001\
\\144\000\145\000\167\000\175\000\
\\176\000\167\000\117\000\137\000\
\\029\001\130\000\056\001\062\001\
\\055\001\136\000\135\000\134\000\
\\133\000\131\000\139\000\138\000\
\\141\000\140\000\150\000\148\000\
\\151\000\147\000\129\000\128\000\
\\127\000\126\000\149\000\030\001\
\\143\000\187\000\181\000\116\000\
\\200\000\199\000\198\000\083\000\
\\039\001\191\000\196\000\081\001\
\\195\000\090\001\059\001\084\001\
\\208\000\212\000\008\001\096\000\
\\118\000\161\000\160\000\065\001\
\\170\000\165\000\167\000\167\000\
\\142\000\171\000\175\000\175\000\
\\172\000\164\000\097\000\156\000\
\\078\001\031\001\040\001\226\000\
\\182\000\152\000\057\000\042\000\
\\096\000\038\000\032\000\114\000\
\\054\001\005\001\073\000\072\000\
\\068\000\099\000\100\000\075\000\
\\071\000\076\000\066\000\193\000\
\\086\000\055\000\051\000\056\000\
\\050\000\098\000\098\000\052\000\
\\054\000\053\000\041\001\185\000\
\\095\001\060\001\063\001\059\001\
\\096\000\096\000\089\000\093\001\
\\184\000\217\000\065\001\065\001\
\\090\001\069\001\169\000\168\000\
\\174\000\173\000\177\000\180\000\
\\179\000\178\000\042\001\188\000\
\\043\001\087\000\078\001\085\001\
\\070\001\230\000\197\000\227\000\
\\153\000\059\000\095\000\082\000\
\\086\001\086\001\063\000\062\000\
\\045\000\095\000\096\000\096\000\
\\096\000\096\000\096\000\096\000\
\\096\000\096\000\096\000\096\000\
\\096\000\096\000\096\000\096\000\
\\096\000\096\000\096\000\096\000\
\\194\000\074\000\048\000\086\000\
\\046\000\086\000\031\000\001\001\
\\255\000\011\001\044\001\045\001\
\\095\000\079\000\044\000\043\000\
\\248\000\018\001\019\001\183\000\
\\132\000\250\000\207\000\209\000\
\\000\000\058\001\096\001\123\000\
\\063\001\216\000\214\000\218\000\
\\215\000\080\000\213\000\071\001\
\\072\001\068\001\159\000\162\000\
\\245\000\239\000\038\001\243\000\
\\032\001\189\000\163\000\073\001\
\\020\001\210\000\155\000\231\000\
\\228\000\021\001\237\000\235\000\
\\236\000\234\000\233\000\017\001\
\\022\001\006\001\077\000\061\000\
\\065\000\064\000\060\000\009\001\
\\106\000\105\000\110\000\107\000\
\\109\000\108\000\112\000\111\000\
\\103\000\102\000\104\000\113\000\
\\101\000\037\000\036\000\035\000\
\\033\000\034\000\046\001\047\001\
\\095\000\069\000\095\000\070\000\
\\093\000\003\001\084\000\207\000\
\\089\001\207\000\087\001\254\000\
\\238\000\253\000\190\000\207\000\
\\205\000\001\000\125\000\122\000\
\\221\000\023\001\224\000\080\000\
\\158\000\157\000\066\001\244\000\
\\242\000\048\001\240\000\091\001\
\\049\001\247\000\154\000\064\001\
\\096\000\095\000\050\001\230\000\
\\058\000\078\000\095\000\096\000\
\\049\000\047\000\002\001\000\001\
\\039\000\080\000\067\001\010\001\
\\097\001\088\000\002\000\008\000\
\\085\000\084\000\094\000\088\001\
\\026\001\096\000\033\001\098\001\
\\099\001\034\001\035\001\030\000\
\\036\001\067\000\201\000\249\000\
\\051\001\204\000\052\001\252\000\
\\251\000\203\000\124\000\219\000\
\\090\000\024\001\246\000\241\000\
\\211\000\012\001\082\001\229\000\
\\007\001\004\001\025\001\074\001\
\\015\000\092\001\067\001\003\000\
\\067\001\085\000\100\001\101\001\
\\014\000\027\001\012\000\088\000\
\\095\000\023\000\024\000\088\000\
\\094\000\088\000\095\000\095\000\
\\010\000\207\000\207\000\225\000\
\\220\000\222\000\091\000\232\000\
\\040\000\092\000\007\000\037\001\
\\075\001\076\001\067\001\025\000\
\\022\000\013\000\026\000\011\000\
\\018\000\013\001\021\000\102\001\
\\019\000\014\001\015\001\009\000\
\\202\000\206\000\223\000\041\000\
\\095\000\006\000\005\000\077\001\
\\027\000\094\000\028\000\088\000\
\\016\001\004\000\103\001\016\000\
\\104\001\081\000\088\000\020\000\
\\053\001\017\000\029\000\105\001"
val gotoT =
"\
\\001\000\175\001\002\000\003\000\003\000\002\000\039\000\001\000\000\000\
\\004\000\022\000\005\000\021\000\006\000\020\000\007\000\019\000\
\\008\000\018\000\011\000\017\000\015\000\016\000\016\000\015\000\
\\018\000\014\000\021\000\013\000\022\000\012\000\024\000\011\000\
\\038\000\010\000\040\000\009\000\041\000\008\000\042\000\007\000\
\\046\000\006\000\048\000\005\000\000\000\
\\000\000\
\\003\000\051\000\039\000\001\000\000\000\
\\000\000\
\\047\000\054\000\051\000\053\000\052\000\052\000\000\000\
\\040\000\009\000\041\000\008\000\042\000\007\000\048\000\057\000\000\000\
\\000\000\
\\042\000\059\000\000\000\
\\000\000\
\\020\000\062\000\080\000\061\000\000\000\
\\000\000\
\\028\000\066\000\029\000\065\000\038\000\064\000\040\000\063\000\
\\041\000\008\000\042\000\007\000\046\000\006\000\048\000\005\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\005\000\075\000\006\000\020\000\007\000\019\000\008\000\074\000\
\\011\000\017\000\015\000\016\000\016\000\015\000\018\000\014\000\
\\023\000\073\000\024\000\072\000\000\000\
\\005\000\079\000\008\000\078\000\025\000\077\000\000\000\
\\027\000\080\000\000\000\
\\005\000\075\000\006\000\020\000\007\000\019\000\008\000\074\000\
\\011\000\017\000\015\000\016\000\016\000\015\000\018\000\014\000\
\\023\000\081\000\024\000\072\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\038\000\086\000\040\000\009\000\041\000\008\000\042\000\007\000\
\\046\000\006\000\048\000\005\000\000\000\
\\008\000\089\000\009\000\088\000\046\000\087\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\070\000\097\000\
\\071\000\096\000\072\000\095\000\073\000\094\000\074\000\093\000\
\\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\080\000\118\000\000\000\
\\047\000\054\000\051\000\053\000\052\000\052\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\005\000\021\000\006\000\020\000\007\000\019\000\008\000\018\000\
\\011\000\017\000\015\000\016\000\016\000\015\000\018\000\014\000\
\\019\000\121\000\021\000\120\000\022\000\119\000\024\000\011\000\000\000\
\\081\000\122\000\000\000\
\\000\000\
\\020\000\123\000\080\000\061\000\000\000\
\\000\000\
\\040\000\127\000\041\000\008\000\042\000\007\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\070\000\097\000\
\\071\000\129\000\072\000\095\000\073\000\094\000\074\000\093\000\
\\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\017\000\133\000\082\000\132\000\000\000\
\\000\000\
\\000\000\
\\005\000\075\000\006\000\020\000\007\000\019\000\008\000\074\000\
\\011\000\017\000\015\000\016\000\016\000\015\000\018\000\014\000\
\\023\000\134\000\024\000\072\000\000\000\
\\005\000\075\000\006\000\020\000\007\000\019\000\008\000\074\000\
\\011\000\017\000\015\000\016\000\016\000\015\000\018\000\014\000\
\\023\000\135\000\024\000\072\000\000\000\
\\000\000\
\\000\000\
\\005\000\079\000\008\000\078\000\025\000\136\000\000\000\
\\005\000\079\000\008\000\078\000\025\000\137\000\000\000\
\\005\000\141\000\006\000\140\000\008\000\139\000\026\000\138\000\000\000\
\\000\000\
\\005\000\021\000\006\000\020\000\007\000\019\000\008\000\018\000\
\\010\000\144\000\011\000\017\000\015\000\016\000\016\000\015\000\
\\018\000\014\000\022\000\143\000\024\000\011\000\035\000\100\000\
\\036\000\099\000\037\000\098\000\074\000\142\000\075\000\092\000\
\\076\000\091\000\077\000\090\000\000\000\
\\000\000\
\\012\000\148\000\013\000\147\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\046\000\151\000\000\000\
\\008\000\089\000\009\000\152\000\000\000\
\\000\000\
\\078\000\153\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\073\000\160\000\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\036\000\181\000\037\000\098\000\000\000\
\\000\000\
\\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\074\000\182\000\
\\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\074\000\184\000\
\\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\005\000\021\000\006\000\020\000\007\000\019\000\008\000\018\000\
\\010\000\191\000\011\000\017\000\015\000\016\000\016\000\015\000\
\\018\000\014\000\022\000\143\000\024\000\011\000\035\000\100\000\
\\036\000\099\000\037\000\098\000\058\000\190\000\068\000\189\000\
\\069\000\188\000\070\000\187\000\072\000\095\000\073\000\186\000\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\074\000\194\000\
\\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\074\000\195\000\
\\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\005\000\021\000\006\000\020\000\007\000\019\000\008\000\018\000\
\\011\000\017\000\015\000\016\000\016\000\015\000\018\000\014\000\
\\022\000\199\000\024\000\011\000\053\000\198\000\054\000\197\000\
\\055\000\196\000\000\000\
\\028\000\066\000\029\000\065\000\038\000\203\000\040\000\063\000\
\\041\000\008\000\042\000\007\000\046\000\006\000\048\000\005\000\000\000\
\\000\000\
\\005\000\021\000\006\000\020\000\007\000\019\000\008\000\018\000\
\\011\000\017\000\015\000\016\000\016\000\015\000\018\000\014\000\
\\021\000\206\000\022\000\119\000\024\000\011\000\083\000\205\000\000\000\
\\058\000\207\000\000\000\
\\081\000\208\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\070\000\097\000\
\\071\000\209\000\072\000\095\000\073\000\094\000\074\000\093\000\
\\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\070\000\097\000\
\\071\000\210\000\072\000\095\000\073\000\094\000\074\000\093\000\
\\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\030\000\212\000\035\000\100\000\036\000\099\000\037\000\098\000\
\\069\000\211\000\070\000\187\000\072\000\095\000\073\000\186\000\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\000\000\
\\029\000\214\000\038\000\203\000\040\000\009\000\041\000\008\000\
\\042\000\007\000\046\000\006\000\048\000\005\000\000\000\
\\000\000\
\\017\000\215\000\082\000\132\000\000\000\
\\017\000\216\000\082\000\132\000\000\000\
\\005\000\021\000\006\000\020\000\007\000\019\000\008\000\018\000\
\\011\000\017\000\015\000\016\000\016\000\015\000\018\000\014\000\
\\019\000\217\000\021\000\120\000\022\000\119\000\024\000\011\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\046\000\224\000\047\000\223\000\049\000\222\000\050\000\221\000\
\\051\000\220\000\000\000\
\\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\058\000\190\000\
\\068\000\189\000\069\000\188\000\070\000\187\000\072\000\095\000\
\\073\000\186\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\012\000\227\000\013\000\147\000\000\000\
\\014\000\228\000\000\000\
\\000\000\
\\043\000\233\000\044\000\232\000\045\000\231\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\068\000\239\000\
\\069\000\188\000\070\000\187\000\072\000\095\000\073\000\186\000\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\069\000\241\000\
\\070\000\187\000\072\000\095\000\073\000\186\000\074\000\093\000\
\\075\000\092\000\076\000\091\000\077\000\090\000\079\000\240\000\000\000\
\\034\000\243\000\000\000\
\\034\000\246\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\068\000\247\000\
\\069\000\188\000\070\000\187\000\072\000\095\000\073\000\186\000\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\072\000\248\000\
\\073\000\094\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\072\000\249\000\
\\073\000\094\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\072\000\250\000\
\\073\000\094\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\072\000\251\000\
\\073\000\094\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\072\000\252\000\
\\073\000\094\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\072\000\253\000\
\\073\000\094\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\072\000\254\000\
\\073\000\094\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\072\000\255\000\
\\073\000\094\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\072\000\000\001\
\\073\000\094\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\072\000\001\001\
\\073\000\094\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\072\000\002\001\
\\073\000\094\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\072\000\003\001\
\\073\000\094\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\072\000\004\001\
\\073\000\094\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\072\000\005\001\
\\073\000\094\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\072\000\006\001\
\\073\000\094\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\072\000\007\001\
\\073\000\094\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\072\000\008\001\
\\073\000\094\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\072\000\009\001\
\\073\000\094\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\005\000\021\000\006\000\020\000\007\000\019\000\008\000\018\000\
\\010\000\010\001\011\000\017\000\015\000\016\000\016\000\015\000\
\\018\000\014\000\022\000\143\000\024\000\011\000\035\000\100\000\
\\036\000\099\000\037\000\098\000\058\000\190\000\068\000\189\000\
\\069\000\188\000\070\000\187\000\072\000\095\000\073\000\186\000\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\000\000\
\\005\000\021\000\006\000\020\000\007\000\019\000\008\000\018\000\
\\010\000\011\001\011\000\017\000\015\000\016\000\016\000\015\000\
\\018\000\014\000\022\000\143\000\024\000\011\000\035\000\100\000\
\\036\000\099\000\037\000\098\000\058\000\190\000\068\000\189\000\
\\069\000\188\000\070\000\187\000\072\000\095\000\073\000\186\000\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\069\000\017\001\
\\070\000\187\000\072\000\095\000\073\000\186\000\074\000\093\000\
\\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\080\000\018\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\038\000\025\001\040\000\009\000\041\000\008\000\042\000\007\000\
\\046\000\024\001\047\000\223\000\048\000\005\000\049\000\023\001\
\\050\000\221\000\051\000\220\000\000\000\
\\000\000\
\\000\000\
\\083\000\028\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\058\000\031\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\031\000\034\001\032\000\033\001\035\000\100\000\036\000\099\000\
\\037\000\098\000\070\000\097\000\071\000\032\001\072\000\095\000\
\\073\000\094\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\005\000\021\000\006\000\020\000\007\000\019\000\008\000\018\000\
\\011\000\017\000\015\000\016\000\016\000\015\000\018\000\014\000\
\\021\000\206\000\022\000\119\000\024\000\011\000\083\000\038\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\047\000\040\001\051\000\039\001\000\000\
\\000\000\
\\000\000\
\\047\000\223\000\050\000\042\001\051\000\220\000\000\000\
\\046\000\224\000\047\000\223\000\049\000\044\001\050\000\221\000\
\\051\000\220\000\080\000\043\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\069\000\058\001\
\\070\000\187\000\072\000\095\000\073\000\186\000\074\000\093\000\
\\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\069\000\059\001\
\\070\000\187\000\072\000\095\000\073\000\186\000\074\000\093\000\
\\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\073\000\060\001\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\000\000\
\\005\000\021\000\006\000\020\000\007\000\019\000\008\000\018\000\
\\011\000\017\000\015\000\016\000\016\000\015\000\018\000\014\000\
\\019\000\069\001\021\000\120\000\022\000\119\000\024\000\011\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\057\000\068\001\
\\058\000\067\001\059\000\066\001\066\000\065\001\067\000\064\001\
\\068\000\063\001\069\000\188\000\070\000\187\000\072\000\095\000\
\\073\000\186\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\084\000\062\001\000\000\
\\083\000\082\001\000\000\
\\005\000\021\000\006\000\020\000\007\000\019\000\008\000\018\000\
\\011\000\017\000\015\000\016\000\016\000\015\000\018\000\014\000\
\\022\000\199\000\024\000\011\000\055\000\083\001\000\000\
\\083\000\085\001\000\000\
\\000\000\
\\000\000\
\\040\000\009\000\041\000\008\000\042\000\007\000\047\000\223\000\
\\048\000\057\000\050\000\042\001\051\000\220\000\000\000\
\\000\000\
\\038\000\086\000\040\000\009\000\041\000\008\000\042\000\007\000\
\\046\000\024\001\047\000\223\000\048\000\005\000\049\000\044\001\
\\050\000\221\000\051\000\220\000\080\000\043\001\000\000\
\\083\000\089\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\031\000\034\001\032\000\093\001\035\000\100\000\036\000\099\000\
\\037\000\098\000\070\000\097\000\071\000\032\001\072\000\095\000\
\\073\000\094\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\080\000\043\001\000\000\
\\047\000\040\001\051\000\039\001\000\000\
\\005\000\021\000\006\000\020\000\007\000\019\000\008\000\018\000\
\\011\000\017\000\015\000\016\000\016\000\015\000\018\000\014\000\
\\022\000\199\000\024\000\011\000\054\000\197\000\055\000\196\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\070\000\097\000\
\\071\000\096\001\072\000\095\000\073\000\094\000\074\000\093\000\
\\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\068\000\097\001\
\\069\000\188\000\070\000\187\000\072\000\095\000\073\000\186\000\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\000\000\
\\044\000\099\001\045\000\231\000\000\000\
\\000\000\
\\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\069\000\100\001\
\\070\000\187\000\072\000\095\000\073\000\186\000\074\000\093\000\
\\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\070\000\101\001\
\\072\000\095\000\073\000\094\000\074\000\093\000\075\000\092\000\
\\076\000\091\000\077\000\090\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\031\000\034\001\032\000\102\001\035\000\100\000\036\000\099\000\
\\037\000\098\000\070\000\097\000\071\000\032\001\072\000\095\000\
\\073\000\094\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\083\000\103\001\000\000\
\\000\000\
\\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\058\000\067\001\
\\059\000\105\001\066\000\065\001\067\000\064\001\068\000\063\001\
\\069\000\188\000\070\000\187\000\072\000\095\000\073\000\186\000\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\000\000\
\\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\058\000\067\001\
\\059\000\107\001\066\000\065\001\067\000\064\001\068\000\063\001\
\\069\000\188\000\070\000\187\000\072\000\095\000\073\000\186\000\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\
\\084\000\106\001\000\000\
\\005\000\021\000\006\000\020\000\007\000\019\000\008\000\018\000\
\\011\000\017\000\015\000\016\000\016\000\015\000\018\000\014\000\
\\021\000\206\000\022\000\119\000\024\000\011\000\035\000\100\000\
\\036\000\099\000\037\000\098\000\057\000\109\001\058\000\067\001\
\\059\000\066\001\066\000\065\001\067\000\064\001\068\000\063\001\
\\069\000\188\000\070\000\187\000\072\000\095\000\073\000\186\000\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\
\\084\000\108\001\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\067\000\110\001\
\\068\000\063\001\069\000\188\000\070\000\187\000\072\000\095\000\
\\073\000\186\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\000\000\
\\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\062\000\114\001\
\\070\000\097\000\071\000\113\001\072\000\095\000\073\000\094\000\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\061\000\115\001\000\000\
\\000\000\
\\000\000\
\\065\000\119\001\000\000\
\\064\000\121\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\031\000\127\001\035\000\100\000\036\000\099\000\037\000\098\000\
\\070\000\097\000\071\000\032\001\072\000\095\000\073\000\094\000\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\083\000\136\001\000\000\
\\000\000\
\\083\000\137\001\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\058\000\067\001\
\\059\000\107\001\066\000\065\001\067\000\064\001\068\000\063\001\
\\069\000\188\000\070\000\187\000\072\000\095\000\073\000\186\000\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\
\\084\000\138\001\000\000\
\\000\000\
\\000\000\
\\060\000\141\001\000\000\
\\000\000\
\\060\000\143\001\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\058\000\067\001\
\\059\000\144\001\066\000\065\001\067\000\064\001\068\000\063\001\
\\069\000\188\000\070\000\187\000\072\000\095\000\073\000\186\000\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\068\000\145\001\
\\069\000\188\000\070\000\187\000\072\000\095\000\073\000\186\000\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\000\000\
\\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\058\000\067\001\
\\059\000\146\001\066\000\065\001\067\000\064\001\068\000\063\001\
\\069\000\188\000\070\000\187\000\072\000\095\000\073\000\186\000\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\067\000\147\001\
\\068\000\063\001\069\000\188\000\070\000\187\000\072\000\095\000\
\\073\000\186\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\058\000\067\001\
\\059\000\148\001\066\000\065\001\067\000\064\001\068\000\063\001\
\\069\000\188\000\070\000\187\000\072\000\095\000\073\000\186\000\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\068\000\149\001\
\\069\000\188\000\070\000\187\000\072\000\095\000\073\000\186\000\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\068\000\150\001\
\\069\000\188\000\070\000\187\000\072\000\095\000\073\000\186\000\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\060\000\151\001\000\000\
\\083\000\152\001\000\000\
\\083\000\153\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\031\000\127\001\035\000\100\000\036\000\099\000\037\000\098\000\
\\070\000\097\000\071\000\032\001\072\000\095\000\073\000\094\000\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\000\000\
\\000\000\
\\031\000\127\001\035\000\100\000\036\000\099\000\037\000\098\000\
\\070\000\097\000\071\000\032\001\072\000\095\000\073\000\094\000\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\083\000\159\001\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\068\000\164\001\
\\069\000\188\000\070\000\187\000\072\000\095\000\073\000\186\000\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\067\000\166\001\
\\068\000\063\001\069\000\188\000\070\000\187\000\072\000\095\000\
\\073\000\186\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\058\000\067\001\
\\059\000\167\001\066\000\065\001\067\000\064\001\068\000\063\001\
\\069\000\188\000\070\000\187\000\072\000\095\000\073\000\186\000\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\067\000\172\001\
\\068\000\063\001\069\000\188\000\070\000\187\000\072\000\095\000\
\\073\000\186\000\074\000\093\000\075\000\092\000\076\000\091\000\
\\077\000\090\000\000\000\
\\035\000\100\000\036\000\099\000\037\000\098\000\058\000\067\001\
\\059\000\173\001\066\000\065\001\067\000\064\001\068\000\063\001\
\\069\000\188\000\070\000\187\000\072\000\095\000\073\000\186\000\
\\074\000\093\000\075\000\092\000\076\000\091\000\077\000\090\000\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\\000\000\
\"
val numstates = 432
val numrules = 259
val s = ref "" and index = ref 0
val string_to_int = fn () => 
let val i = !index
in index := i+2; Char.ord(String.sub(!s,i)) + Char.ord(String.sub(!s,i+1)) * 256
end
val string_to_list = fn s' =>
    let val len = String.size s'
        fun f () =
           if !index < len then string_to_int() :: f()
           else nil
   in index := 0; s := s'; f ()
   end
val string_to_pairlist = fn (conv_key,conv_entry) =>
     let fun f () =
         case string_to_int()
         of 0 => EMPTY
          | n => PAIR(conv_key (n-1),conv_entry (string_to_int()),f())
     in f
     end
val string_to_pairlist_default = fn (conv_key,conv_entry) =>
    let val conv_row = string_to_pairlist(conv_key,conv_entry)
    in fn () =>
       let val default = conv_entry(string_to_int())
           val row = conv_row()
       in (row,default)
       end
   end
val string_to_table = fn (convert_row,s') =>
    let val len = String.size s'
        fun f ()=
           if !index < len then convert_row() :: f()
           else nil
     in (s := s'; index := 0; f ())
     end
local
  val memo = Array.array(numstates+numrules,ERROR)
  val _ =let fun g i=(Array.update(memo,i,REDUCE(i-numstates)); g(i+1))
       fun f i =
            if i=numstates then g i
            else (Array.update(memo,i,SHIFT (STATE i)); f (i+1))
          in f 0 handle Subscript => ()
          end
in
val entry_to_action = fn 0 => ACCEPT | 1 => ERROR | j => Array.sub(memo,(j-2))
end
val gotoT=Array.fromList(string_to_table(string_to_pairlist(NT,STATE),gotoT))
val actionRows=string_to_table(string_to_pairlist_default(T,entry_to_action),actionRows)
val actionRowNumbers = string_to_list actionRowNumbers
val actionT = let val actionRowLookUp=
let val a=Array.fromList(actionRows) in fn i=>Array.sub(a,i) end
in Array.fromList(map actionRowLookUp actionRowNumbers)
end
in LrTable.mkLrTable {actions=actionT,gotos=gotoT,numRules=numrules,
numStates=numstates,initialState=STATE 0}
end
end
local open Header in
type pos = AST.pos
type arg = pos*pos -> string -> unit
structure MlyValue = 
struct
datatype svalue = VOID' | ntVOID of unit
 | QUALIFIER of  (Lexeme.identifier) | POUND of  (string)
 | STRINGCONSTANT of  (string) | CONSTANT of  (Lexeme.constant)
 | TYPEDEF_NAME of  (Lexeme.identifier)
 | IDENTIFIER of  (Lexeme.identifier)
 | ASSIGN of  (Lexeme.operator option)
 | getScope of  (SymbolTable.scope) | popScope of  (unit)
 | pushStructScope of  (unit) | recoverScope of  (unit)
 | pushScope of  (unit) | argList of  (AST.expression list)
 | args of  (AST.expression list) | primary of  (AST.expression)
 | postfix of  (AST.expression) | unaryOp of  (AST.operator)
 | unaryExpr of  (AST.expression) | cast of  (AST.expression)
 | binaryExpr of  (AST.expression) | constantExpr of  (AST.expression)
 | conditionalExpr of  (AST.expression)
 | assignExpr of  (AST.expression) | expr of  (AST.expression)
 | exprOpt of  (AST.expression option) | startDo of  (unit)
 | startFor of  (AST.expression option*AST.expression option*AST.expression option)
 | startWhile of  (AST.expression) | startDefault of  (AST.statement)
 | startCase of  (AST.expression option)
 | startSwitch of  (AST.expression)
 | stmtmaybe of  (AST.statement option) | stmt of  (AST.statement)
 | compoundStmt of  (AST.declaration list*AST.statement list*SymbolTable.scope)
 | stmtList of  (AST.statement list) | startBlock of  (unit)
 | parameterDeclaration of  (AST.declaration)
 | parameterList of  (AST.declaration list)
 | identifierList of  ( ( AST.region * AST.identifier )  list)
 | oldParameterList of  ( ( AST.region * AST.identifier )  list*AST.region option)
 | parameterTypeList of  (AST.declaration list)
 | directAbstractDeclarator of  (AST.declarator list)
 | abstractDeclarator of  (AST.declarator list)
 | directDeclarator of  (AST.region*AST.identifier*AST.declarator list)
 | arrayDeclarator of  (AST.expression option)
 | pointer of  (AST.declarator list) | any_word of  (unit)
 | attrib of  (unit) | attribute_list of  (unit)
 | attribute of  (unit) | attributes of  (unit)
 | attributeOpt of  (unit) | extensionOpt of  (unit)
 | declarator of  (AST.region*AST.identifier*AST.declarator list)
 | string of  (string) | stringconstant of  (string)
 | constant of  (AST.constant) | field of  (AST.identifier)
 | initCastOpt of  (AST.syntacticType option)
 | braceInitializerList of  (AST.initializer list)
 | braceInitializer of  (AST.initializer)
 | initializer of  (AST.initializer)
 | initDeclarator of  ( ( AST.region * AST.identifier *
				     AST.declarator list ) *AST.initializer option)
 | initDeclaratorList of  ( ( (AST.region * AST.identifier * AST.declarator list) * 
					 AST.initializer option )  list)
 | baseDeclarationSpecifierList of  (AST.storageClass option* ( longShort option * signed option * AST.typeSpecifier option ) *AST.typeQualifier list)
 | baseDeclarationSpecifier of  (AST.storageClass option* ( longShort option * signed option * AST.typeSpecifier option ) *AST.typeQualifier option)
 | sqDeclarationSpecifierList of  (AST.storageClass option*AST.typeQualifier list)
 | definedDeclarationSpecifierList of  (AST.storageClass option*AST.typeSpecifier*AST.typeQualifier list)
 | declarationSpecifierListAux of  (AST.storageClass option*AST.typeSpecifier*AST.typeQualifier list)
 | declarationSpecifierList of  (AST.storageClass option*AST.typeSpecifier*AST.typeQualifier list)
 | declaration of  (AST.declaration list)
 | paramDeclarationList of  (AST.declaration list)
 | declarationList of  (AST.declaration list)
 | typeOfSpecifier of  (AST.typeSpecifier)
 | structDeclarationList of  (AST.declaration list)
 | structUnion of  (AST.aggregate)
 | structSpecifier of  (AST.typeSpecifier)
 | enumerator of  (AST.identifier*AST.expression option)
 | enumList2 of  ( ( AST.identifier * AST.expression option )  list)
 | enumeratorList of  ( ( AST.identifier * AST.expression option )  list)
 | enumSpecifier of  (AST.typeSpecifier)
 | typeName of  (AST.syntacticType)
 | typeQualifierList of  (AST.typeQualifier list)
 | typeQualifier of  (AST.typeQualifier)
 | typeSpecifier of  (AST.typeSpecifier)
 | baseTypeSpecifier of  (longShort option*signed option*AST.typeSpecifier option)
 | storageClassSpecifier of  (AST.storageClass)
 | functionDefinition of  (AST.declaration)
 | externalDeclaration of  (AST.declaration list)
 | program of  (AST.declaration list) | start of  (AST.program)
end
type svalue = MlyValue.svalue
type result = AST.program
end
structure EC=
struct
open LrTable
val is_keyword =
fn (T 45) => true | (T 46) => true | (T 47) => true | (T 48) => true
 | (T 53) => true | (T 54) => true | (T 55) => true | (T 56) => true
 | (T 49) => true | (T 51) => true | (T 57) => true | (T 58) => true
 | (T 59) => true | (T 60) => true | (T 61) => true | (T 62) => true
 | (T 63) => true | (T 64) => true | (T 65) => true | (T 66) => true
 | (T 67) => true | (T 68) => true | (T 69) => true | (T 70) => true
 | (T 71) => true | (T 72) => true | (T 73) => true | (T 74) => true
 | (T 75) => true | (T 76) => true | (T 77) => true | (T 82) => true
 | (T 83) => true | (T 84) => true | (T 85) => true | _ => false
val preferred_change = 
nil
val noShift = 
fn (T 86) => true | _ => false
val showTerminal =
fn (T 0) => "PLUS"
  | (T 1) => "MINUS"
  | (T 2) => "MUL"
  | (T 3) => "DIV"
  | (T 4) => "MOD"
  | (T 5) => "INCR"
  | (T 6) => "DECR"
  | (T 7) => "AND"
  | (T 8) => "OR"
  | (T 9) => "NOT"
  | (T 10) => "BIT_AND"
  | (T 11) => "BIT_OR"
  | (T 12) => "BIT_NOT"
  | (T 13) => "BIT_XOR"
  | (T 14) => "RSHIFT"
  | (T 15) => "LSHIFT"
  | (T 16) => "LT"
  | (T 17) => "LE"
  | (T 18) => "GT"
  | (T 19) => "GE"
  | (T 20) => "EQ"
  | (T 21) => "NE"
  | (T 22) => "ASSIGN"
  | (T 23) => "DOT"
  | (T 24) => "PTR"
  | (T 25) => "COMMA"
  | (T 26) => "QUESTION"
  | (T 27) => "COLON"
  | (T 28) => "LP"
  | (T 29) => "RP"
  | (T 30) => "LB"
  | (T 31) => "RB"
  | (T 32) => "LC"
  | (T 33) => "RC"
  | (T 34) => "IDENTIFIER"
  | (T 35) => "TYPEDEF_NAME"
  | (T 36) => "TYPE_DEC_SPEC"
  | (T 37) => "HIGHEST"
  | (T 38) => "ATTRIBUTE"
  | (T 39) => "EXTENSION"
  | (T 40) => "FUNCTION"
  | (T 41) => "CONSTANT"
  | (T 42) => "STRINGCONSTANT"
  | (T 43) => "POUND"
  | (T 44) => "INLINE"
  | (T 45) => "AUTO"
  | (T 46) => "REGISTER"
  | (T 47) => "STATIC"
  | (T 48) => "EXTERN"
  | (T 49) => "CONST"
  | (T 50) => "QUALIFIER"
  | (T 51) => "VOLATILE"
  | (T 52) => "NONCONST"
  | (T 53) => "SIGNED"
  | (T 54) => "UNSIGNED"
  | (T 55) => "SHORT"
  | (T 56) => "LONG"
  | (T 57) => "VOID"
  | (T 58) => "CHAR"
  | (T 59) => "INT"
  | (T 60) => "FLOAT"
  | (T 61) => "DOUBLE"
  | (T 62) => "STRUCT"
  | (T 63) => "ENUM"
  | (T 64) => "UNION"
  | (T 65) => "TYPEDEF"
  | (T 66) => "IF"
  | (T 67) => "ELSE"
  | (T 68) => "DO"
  | (T 69) => "WHILE"
  | (T 70) => "FOR"
  | (T 71) => "BREAK"
  | (T 72) => "CONTINUE"
  | (T 73) => "SWITCH"
  | (T 74) => "CASE"
  | (T 75) => "DEFAULT"
  | (T 76) => "GOTO"
  | (T 77) => "RETURN"
  | (T 78) => "SEMICOLON"
  | (T 79) => "ELLIPSIS"
  | (T 80) => "BEGIN"
  | (T 81) => "END"
  | (T 82) => "SIZEOF"
  | (T 83) => "TYPEOF"
  | (T 84) => "LABEL"
  | (T 85) => "ALIGNOF"
  | (T 86) => "EOF"
  | _ => "bogus-term"
local open Header in
val errtermvalue=
fn _ => MlyValue.VOID'
end
val terms = (T 0) :: (T 1) :: (T 2) :: (T 3) :: (T 4) :: (T 5) :: (T 6
) :: (T 7) :: (T 8) :: (T 9) :: (T 10) :: (T 11) :: (T 12) :: (T 13)
 :: (T 14) :: (T 15) :: (T 16) :: (T 17) :: (T 18) :: (T 19) :: (T 20)
 :: (T 21) :: (T 23) :: (T 24) :: (T 25) :: (T 26) :: (T 27) :: (T 28)
 :: (T 29) :: (T 30) :: (T 31) :: (T 32) :: (T 33) :: (T 36) :: (T 37)
 :: (T 38) :: (T 39) :: (T 40) :: (T 44) :: (T 45) :: (T 46) :: (T 47)
 :: (T 48) :: (T 49) :: (T 51) :: (T 52) :: (T 53) :: (T 54) :: (T 55)
 :: (T 56) :: (T 57) :: (T 58) :: (T 59) :: (T 60) :: (T 61) :: (T 62)
 :: (T 63) :: (T 64) :: (T 65) :: (T 66) :: (T 67) :: (T 68) :: (T 69)
 :: (T 70) :: (T 71) :: (T 72) :: (T 73) :: (T 74) :: (T 75) :: (T 76)
 :: (T 77) :: (T 78) :: (T 79) :: (T 80) :: (T 81) :: (T 82) :: (T 83)
 :: (T 84) :: (T 85) :: (T 86) :: nil
end
structure Actions =
struct 
exception mlyAction of int
local open Header in
val actions = 
fn (i392,defaultPos,stack,
    (error):arg) =>
case (i392,stack)
of (0,(_,(MlyValue.program program,program1left,program1right))::
rest671) => let val result=MlyValue.start((
AST.Program {decls = rev program,
				  lookup = SymbolTable.mkLookupFun ()}
))
 in (LrTable.NT 0,(result,program1left,program1right),rest671) end
| (1,rest671) => let val result=MlyValue.start((
AST.Program {decls = [],
			    lookup = fn _ => fn _ => raise Fail "Empty program"}
))
 in (LrTable.NT 0,(result,defaultPos,defaultPos),rest671) end
| (2,(_,(MlyValue.externalDeclaration externalDeclaration,
externalDeclaration1left,externalDeclaration1right))::rest671) => let 
val result=MlyValue.program((externalDeclaration))
 in (LrTable.NT 1,(result,externalDeclaration1left,
externalDeclaration1right),rest671) end
| (3,(_,(MlyValue.externalDeclaration externalDeclaration,_,
externalDeclaration1right))::(_,(MlyValue.program program,program1left
,_))::rest671) => let val result=MlyValue.program((
externalDeclaration @ program ))
 in (LrTable.NT 1,(result,program1left,externalDeclaration1right),
rest671) end
| (4,(_,(MlyValue.functionDefinition functionDefinition,_,
functionDefinition1right))::(_,(_,extensionOpt1left,_))::rest671) => 
let val result=MlyValue.externalDeclaration(([functionDefinition]))
 in (LrTable.NT 2,(result,extensionOpt1left,functionDefinition1right),
rest671) end
| (5,(_,(_,_,SEMICOLON1right))::(_,(MlyValue.declaration declaration,_
,_))::(_,(_,extensionOpt1left,_))::rest671) => let val result=
MlyValue.externalDeclaration((declaration))
 in (LrTable.NT 2,(result,extensionOpt1left,SEMICOLON1right),rest671)
 end
| (6,rest671) => let val result=MlyValue.extensionOpt(())
 in (LrTable.NT 38,(result,defaultPos,defaultPos),rest671) end
| (7,(_,(_,EXTENSION1left,EXTENSION1right))::rest671) => let val 
result=MlyValue.extensionOpt(())
 in (LrTable.NT 38,(result,EXTENSION1left,EXTENSION1right),rest671)
 end
| (8,(_,(MlyValue.compoundStmt compoundStmt,_,compoundStmt1right))::_
::(_,(MlyValue.paramDeclarationList paramDeclarationList,_,_))::(_,(
MlyValue.declarator declarator,_,declaratorright))::(_,(
MlyValue.declarationSpecifierList declarationSpecifierList,
declarationSpecifierListleft as declarationSpecifierList1left,_))::
rest671) => let val result=MlyValue.functionDefinition((
let val region = (declarationSpecifierListleft, declaratorright)
			 in
				buildFunctionDefinition (region, error) handle _ => (error region "Error in function definition.";
										     raise (Fail "bad function definition"))
			 end
			 (declarationSpecifierList,
			  declarator, paramDeclarationList, compoundStmt)
))
 in (LrTable.NT 3,(result,declarationSpecifierList1left,
compoundStmt1right),rest671) end
| (9,(_,(MlyValue.compoundStmt compoundStmt,_,compoundStmt1right))::_
::(_,(MlyValue.paramDeclarationList paramDeclarationList,_,_))::(_,(
MlyValue.declarator declarator,declaratorleft as declarator1left,
declaratorright))::rest671) => let val result=
MlyValue.functionDefinition((
let val region = (declaratorleft,declaratorright)
			 in
				buildFunctionDefinition (region, error) handle _ => (error region "Error in function definition.";
										     raise (Fail "bad function definition"))
			 end
			 ((NONE, AST.Int, []),
			  declarator, paramDeclarationList, compoundStmt)
))
 in (LrTable.NT 3,(result,declarator1left,compoundStmt1right),rest671)
 end
| (10,(_,(_,_,SEMICOLON1right))::(_,(MlyValue.compoundStmt 
compoundStmt,_,_))::_::(_,(MlyValue.paramDeclarationList 
paramDeclarationList,_,_))::(_,(MlyValue.declarator declarator,_,
declaratorright))::(_,(MlyValue.declarationSpecifierList 
declarationSpecifierList,declarationSpecifierListleft as 
declarationSpecifierList1left,_))::rest671) => let val result=
MlyValue.functionDefinition((
let val region = (declarationSpecifierListleft, declaratorright)
			 in
				buildFunctionDefinition (region, error) handle _ => (error region "Error in function definition.";
										     raise (Fail "bad function definition"))
			 end
			 (declarationSpecifierList,
			  declarator, paramDeclarationList, compoundStmt)
))
 in (LrTable.NT 3,(result,declarationSpecifierList1left,
SEMICOLON1right),rest671) end
| (11,(_,(_,_,SEMICOLON1right))::(_,(MlyValue.compoundStmt 
compoundStmt,_,_))::_::(_,(MlyValue.paramDeclarationList 
paramDeclarationList,_,_))::(_,(MlyValue.declarator declarator,
declaratorleft as declarator1left,declaratorright))::rest671) => let 
val result=MlyValue.functionDefinition((
let val region = (declaratorleft,declaratorright)
			 in
				buildFunctionDefinition (region, error) handle _ => (error region "Error in function definition.";
										     raise (Fail "bad function definition"))
			 end
			 ((NONE, AST.Int, []),
			  declarator, paramDeclarationList, compoundStmt)
))
 in (LrTable.NT 3,(result,declarator1left,SEMICOLON1right),rest671)
 end
| (12,(_,(_,AUTO1left,AUTO1right))::rest671) => let val result=
MlyValue.storageClassSpecifier((AST.Auto))
 in (LrTable.NT 4,(result,AUTO1left,AUTO1right),rest671) end
| (13,(_,(_,REGISTER1left,REGISTER1right))::rest671) => let val result
=MlyValue.storageClassSpecifier((AST.Register))
 in (LrTable.NT 4,(result,REGISTER1left,REGISTER1right),rest671) end
| (14,(_,(_,STATIC1left,STATIC1right))::rest671) => let val result=
MlyValue.storageClassSpecifier((AST.Static))
 in (LrTable.NT 4,(result,STATIC1left,STATIC1right),rest671) end
| (15,(_,(_,EXTERN1left,EXTERN1right))::rest671) => let val result=
MlyValue.storageClassSpecifier((AST.Extern))
 in (LrTable.NT 4,(result,EXTERN1left,EXTERN1right),rest671) end
| (16,(_,(_,TYPEDEF1left,TYPEDEF1right))::rest671) => let val result=
MlyValue.storageClassSpecifier((AST.Typedef))
 in (LrTable.NT 4,(result,TYPEDEF1left,TYPEDEF1right),rest671) end
| (17,(_,(_,VOID1left,VOID1right))::rest671) => let val result=
MlyValue.baseTypeSpecifier((NONE, NONE, SOME(AST.Void)))
 in (LrTable.NT 5,(result,VOID1left,VOID1right),rest671) end
| (18,(_,(_,CHAR1left,CHAR1right))::rest671) => let val result=
MlyValue.baseTypeSpecifier((NONE, NONE, SOME(AST.Char)))
 in (LrTable.NT 5,(result,CHAR1left,CHAR1right),rest671) end
| (19,(_,(_,INT1left,INT1right))::rest671) => let val result=
MlyValue.baseTypeSpecifier((NONE, NONE, SOME(AST.Int)))
 in (LrTable.NT 5,(result,INT1left,INT1right),rest671) end
| (20,(_,(_,FLOAT1left,FLOAT1right))::rest671) => let val result=
MlyValue.baseTypeSpecifier((NONE, NONE, SOME(AST.Float)))
 in (LrTable.NT 5,(result,FLOAT1left,FLOAT1right),rest671) end
| (21,(_,(_,DOUBLE1left,DOUBLE1right))::rest671) => let val result=
MlyValue.baseTypeSpecifier((NONE, NONE, SOME(AST.Double)))
 in (LrTable.NT 5,(result,DOUBLE1left,DOUBLE1right),rest671) end
| (22,(_,(_,LABEL1left,LABEL1right))::rest671) => let val result=
MlyValue.baseTypeSpecifier((NONE, NONE, SOME(AST.Label)))
 in (LrTable.NT 5,(result,LABEL1left,LABEL1right),rest671) end
| (23,(_,(_,SHORT1left,SHORT1right))::rest671) => let val result=
MlyValue.baseTypeSpecifier((SOME(Short), NONE, NONE))
 in (LrTable.NT 5,(result,SHORT1left,SHORT1right),rest671) end
| (24,(_,(_,LONG1left,LONG1right))::rest671) => let val result=
MlyValue.baseTypeSpecifier((SOME(Long), NONE, NONE))
 in (LrTable.NT 5,(result,LONG1left,LONG1right),rest671) end
| (25,(_,(_,SIGNED1left,SIGNED1right))::rest671) => let val result=
MlyValue.baseTypeSpecifier((NONE, SOME(Signed), NONE))
 in (LrTable.NT 5,(result,SIGNED1left,SIGNED1right),rest671) end
| (26,(_,(_,UNSIGNED1left,UNSIGNED1right))::rest671) => let val result
=MlyValue.baseTypeSpecifier((NONE, SOME(Unsigned), NONE))
 in (LrTable.NT 5,(result,UNSIGNED1left,UNSIGNED1right),rest671) end
| (27,(_,(MlyValue.TYPEDEF_NAME TYPEDEF_NAME,TYPEDEF_NAME1left,
TYPEDEF_NAME1right))::rest671) => let val result=
MlyValue.typeSpecifier((AST.NamedType(TYPEDEF_NAME)))
 in (LrTable.NT 6,(result,TYPEDEF_NAME1left,TYPEDEF_NAME1right),
rest671) end
| (28,(_,(MlyValue.structSpecifier structSpecifier,
structSpecifier1left,structSpecifier1right))::rest671) => let val 
result=MlyValue.typeSpecifier((structSpecifier))
 in (LrTable.NT 6,(result,structSpecifier1left,structSpecifier1right),
rest671) end
| (29,(_,(MlyValue.enumSpecifier enumSpecifier,enumSpecifier1left,
enumSpecifier1right))::rest671) => let val result=
MlyValue.typeSpecifier((enumSpecifier))
 in (LrTable.NT 6,(result,enumSpecifier1left,enumSpecifier1right),
rest671) end
| (30,(_,(MlyValue.typeOfSpecifier typeOfSpecifier,
typeOfSpecifier1left,typeOfSpecifier1right))::rest671) => let val 
result=MlyValue.typeSpecifier((typeOfSpecifier))
 in (LrTable.NT 6,(result,typeOfSpecifier1left,typeOfSpecifier1right),
rest671) end
| (31,(_,(_,CONST1left,CONST1right))::rest671) => let val result=
MlyValue.typeQualifier((AST.Const))
 in (LrTable.NT 7,(result,CONST1left,CONST1right),rest671) end
| (32,(_,(_,VOLATILE1left,VOLATILE1right))::rest671) => let val result
=MlyValue.typeQualifier((AST.Volatile))
 in (LrTable.NT 7,(result,VOLATILE1left,VOLATILE1right),rest671) end
| (33,(_,(_,INLINE1left,INLINE1right))::rest671) => let val result=
MlyValue.typeQualifier((AST.Inline))
 in (LrTable.NT 7,(result,INLINE1left,INLINE1right),rest671) end
| (34,(_,(_,NONCONST1left,NONCONST1right))::rest671) => let val result
=MlyValue.typeQualifier((AST.NonConst))
 in (LrTable.NT 7,(result,NONCONST1left,NONCONST1right),rest671) end
| (35,(_,(MlyValue.QUALIFIER QUALIFIER,QUALIFIER1left,QUALIFIER1right)
)::rest671) => let val result=MlyValue.typeQualifier((
AST.Qualifier (QUALIFIER)))
 in (LrTable.NT 7,(result,QUALIFIER1left,QUALIFIER1right),rest671) end
| (36,(_,(MlyValue.typeQualifier typeQualifier,typeQualifier1left,
typeQualifier1right))::rest671) => let val result=
MlyValue.typeQualifierList(([typeQualifier]))
 in (LrTable.NT 8,(result,typeQualifier1left,typeQualifier1right),
rest671) end
| (37,(_,(MlyValue.typeQualifierList typeQualifierList,_,
typeQualifierList1right))::(_,(MlyValue.typeQualifier typeQualifier,
typeQualifierleft as typeQualifier1left,typeQualifierright))::rest671)
 => let val result=MlyValue.typeQualifierList((
let val err = error (typeQualifierleft,typeQualifierright)
		       in
			   mergeTypeQualifiers err (typeQualifier, typeQualifierList)
				handle _ => (err "Error merging type qualifiers"; raise Fail "Bad type qualify merging")
		       end
))
 in (LrTable.NT 8,(result,typeQualifier1left,typeQualifierList1right),
rest671) end
| (38,(_,(MlyValue.declarationSpecifierList declarationSpecifierList,
declarationSpecifierList1left,declarationSpecifierList1right))::
rest671) => let val result=MlyValue.typeName((
let
			   val (sc, ts, tq) = declarationSpecifierList
			 in
			   AST.SyntacticType(sc, ts, tq, nil)
			 end
))
 in (LrTable.NT 9,(result,declarationSpecifierList1left,
declarationSpecifierList1right),rest671) end
| (39,(_,(MlyValue.abstractDeclarator abstractDeclarator,_,
abstractDeclarator1right))::(_,(MlyValue.declarationSpecifierList 
declarationSpecifierList,declarationSpecifierList1left,_))::rest671)
 => let val result=MlyValue.typeName((
let
			   val (sc, ts, tq) = declarationSpecifierList
			 in
			   AST.SyntacticType(sc, ts, tq, abstractDeclarator)
			 end
))
 in (LrTable.NT 9,(result,declarationSpecifierList1left,
abstractDeclarator1right),rest671) end
| (40,(_,(_,_,RCright as RC1right))::(_,(MlyValue.enumeratorList 
enumeratorList,_,_))::_::(_,(MlyValue.IDENTIFIER IDENTIFIER,_,_))::(_,
(_,ENUMleft as ENUM1left,_))::rest671) => let val result=
MlyValue.enumSpecifier((
let val enum = AST.Enum(SOME(IDENTIFIER),
						 rev enumeratorList)
			 in
			     addEnum enum
				handle _ => (error (ENUMleft, RCright) "Bad enum specifier"; raise Fail "Bad enum");
			     enum
			 end
))
 in (LrTable.NT 10,(result,ENUM1left,RC1right),rest671) end
| (41,(_,(_,_,RCright as RC1right))::(_,(MlyValue.enumeratorList 
enumeratorList,_,_))::_::(_,(_,ENUMleft as ENUM1left,_))::rest671) => 
let val result=MlyValue.enumSpecifier((
let val enum = AST.Enum(NONE, rev enumeratorList)
			 in
			     addEnum enum
				handle _ => (error (ENUMleft, RCright) "Bad enum specifier"; raise Fail "Bad enum");
			     enum
			 end
))
 in (LrTable.NT 10,(result,ENUM1left,RC1right),rest671) end
| (42,(_,(MlyValue.IDENTIFIER IDENTIFIER,_,IDENTIFIER1right))::(_,(_,
ENUM1left,_))::rest671) => let val result=MlyValue.enumSpecifier((
AST.Enum(SOME(IDENTIFIER), nil)))
 in (LrTable.NT 10,(result,ENUM1left,IDENTIFIER1right),rest671) end
| (43,(_,(_,_,COMMA1right))::(_,(MlyValue.enumerator enumerator,_,_))
::(_,(MlyValue.enumList2 enumList2,enumList21left,_))::rest671) => 
let val result=MlyValue.enumeratorList((rev (enumerator :: enumList2))
)
 in (LrTable.NT 11,(result,enumList21left,COMMA1right),rest671) end
| (44,(_,(MlyValue.enumerator enumerator,_,enumerator1right))::(_,(
MlyValue.enumList2 enumList2,enumList21left,_))::rest671) => let val 
result=MlyValue.enumeratorList((rev (enumerator :: enumList2)))
 in (LrTable.NT 11,(result,enumList21left,enumerator1right),rest671)
 end
| (45,(_,(_,_,COMMA1right))::(_,(MlyValue.enumerator enumerator,_,_))
::(_,(MlyValue.enumList2 enumList2,enumList21left,_))::rest671) => 
let val result=MlyValue.enumList2((enumerator :: enumList2))
 in (LrTable.NT 12,(result,enumList21left,COMMA1right),rest671) end
| (46,rest671) => let val result=MlyValue.enumList2(([]))
 in (LrTable.NT 12,(result,defaultPos,defaultPos),rest671) end
| (47,(_,(MlyValue.IDENTIFIER IDENTIFIER,IDENTIFIER1left,
IDENTIFIER1right))::rest671) => let val result=MlyValue.enumerator((
IDENTIFIER, NONE))
 in (LrTable.NT 13,(result,IDENTIFIER1left,IDENTIFIER1right),rest671)
 end
| (48,(_,(MlyValue.constantExpr constantExpr,_,constantExpr1right))::_
::(_,(MlyValue.IDENTIFIER IDENTIFIER,IDENTIFIER1left,_))::rest671) => 
let val result=MlyValue.enumerator((IDENTIFIER, SOME(constantExpr)))
 in (LrTable.NT 13,(result,IDENTIFIER1left,constantExpr1right),rest671
) end
| (49,(_,(_,_,RCright as RC1right))::(_,(
MlyValue.structDeclarationList structDeclarationList,_,_))::_::(_,(
MlyValue.IDENTIFIER IDENTIFIER,_,_))::(_,(MlyValue.structUnion 
structUnion,structUnionleft as structUnion1left,_))::rest671) => let 
val result=MlyValue.structSpecifier((
let val theAgg = AST.Aggregate (structUnion,
						  SOME IDENTIFIER,
						  structDeclarationList)
		       val agType = AST.SyntacticType(NONE, theAgg, [], [])
		   in
			SymbolTable.insertStruct IDENTIFIER agType
				handle _ => (error (structUnionleft, RCright) "Bad struct specification"; raise Fail "Bad struct");
			theAgg
		   end
))
 in (LrTable.NT 14,(result,structUnion1left,RC1right),rest671) end
| (50,(_,(_,_,RCright as RC1right))::(_,(
MlyValue.structDeclarationList structDeclarationList,_,_))::_::(_,(
MlyValue.TYPEDEF_NAME TYPEDEF_NAME,_,_))::(_,(MlyValue.structUnion 
structUnion,structUnionleft as structUnion1left,_))::rest671) => let 
val result=MlyValue.structSpecifier((
let val theAgg = AST.Aggregate (structUnion,
						   SOME TYPEDEF_NAME,
						   structDeclarationList)
		       val agType = AST.SyntacticType(NONE, theAgg, [], [])
		   in
		       SymbolTable.insertStruct TYPEDEF_NAME agType
				handle _ => (error (structUnionleft, RCright) "Bad struct specification"; raise Fail "Bad struct");
		       theAgg
		   end
))
 in (LrTable.NT 14,(result,structUnion1left,RC1right),rest671) end
| (51,(_,(_,_,RCright as RC1right))::(_,(
MlyValue.structDeclarationList structDeclarationList,_,_))::_::(_,(
MlyValue.structUnion structUnion,structUnionleft as structUnion1left,_
))::rest671) => let val result=MlyValue.structSpecifier((
let val structName =
			     if AST.ExtraInfo.nameAnonAggregates then
				 SOME (AST.mkAggrId (AST.mkFreshId ()))
			     else
				 NONE
			 in
			     AST.Aggregate(structUnion, structName,
					   structDeclarationList)
			 end handle _ => (error (structUnionleft, RCright) "Bad struct specification"; raise Fail "Bad struct")
))
 in (LrTable.NT 14,(result,structUnion1left,RC1right),rest671) end
| (52,(_,(MlyValue.IDENTIFIER IDENTIFIER,_,IDENTIFIER1right))::(_,(
MlyValue.structUnion structUnion,structUnion1left,_))::rest671) => 
let val result=MlyValue.structSpecifier((
AST.NamedAggregate(structUnion, IDENTIFIER)))
 in (LrTable.NT 14,(result,structUnion1left,IDENTIFIER1right),rest671)
 end
| (53,(_,(MlyValue.TYPEDEF_NAME TYPEDEF_NAME,_,TYPEDEF_NAME1right))::(
_,(MlyValue.structUnion structUnion,structUnion1left,_))::rest671) => 
let val result=MlyValue.structSpecifier((
AST.NamedAggregate(structUnion, TYPEDEF_NAME)))
 in (LrTable.NT 14,(result,structUnion1left,TYPEDEF_NAME1right),
rest671) end
| (54,(_,(_,STRUCT1left,STRUCT1right))::rest671) => let val result=
MlyValue.structUnion((AST.Struct))
 in (LrTable.NT 15,(result,STRUCT1left,STRUCT1right),rest671) end
| (55,(_,(_,UNION1left,UNION1right))::rest671) => let val result=
MlyValue.structUnion((AST.Union))
 in (LrTable.NT 15,(result,UNION1left,UNION1right),rest671) end
| (56,rest671) => let val result=MlyValue.structDeclarationList((nil))
 in (LrTable.NT 16,(result,defaultPos,defaultPos),rest671) end
| (57,(_,(_,_,popScope1right))::(_,(MlyValue.declarationList 
declarationList,_,_))::(_,(_,pushStructScope1left,_))::rest671) => 
let val result=MlyValue.structDeclarationList((rev declarationList))
 in (LrTable.NT 16,(result,pushStructScope1left,popScope1right),
rest671) end
| (58,(_,(_,_,RPright as RP1right))::(_,(MlyValue.unaryExpr unaryExpr,
_,_))::_::(_,(_,TYPEOFleft as TYPEOF1left,_))::rest671) => let val 
result=MlyValue.typeOfSpecifier((
AST.TypeOfExpr(unaryExpr, {region=(TYPEOFleft, RPright),
				  extraInfo=ref NONE,
				  typeInfo=NONE})
))
 in (LrTable.NT 17,(result,TYPEOF1left,RP1right),rest671) end
| (59,(_,(_,_,RPright as RP1right))::(_,(MlyValue.typeName typeName,_,
_))::_::(_,(_,TYPEOFleft as TYPEOF1left,_))::rest671) => let val 
result=MlyValue.typeOfSpecifier((
AST.TypeOfTypeExpr(typeName, {region=(TYPEOFleft, RPright),
				     extraInfo=ref NONE,
				     typeInfo=NONE})
))
 in (LrTable.NT 17,(result,TYPEOF1left,RP1right),rest671) end
| (60,(_,(_,_,SEMICOLON1right))::(_,(MlyValue.declaration declaration,
declaration1left,_))::rest671) => let val result=
MlyValue.declarationList((declaration))
 in (LrTable.NT 18,(result,declaration1left,SEMICOLON1right),rest671)
 end
| (61,(_,(_,_,SEMICOLON1right))::(_,(MlyValue.declaration declaration,
_,_))::(_,(MlyValue.declarationList declarationList,
declarationList1left,_))::rest671) => let val result=
MlyValue.declarationList((declaration @ declarationList))
 in (LrTable.NT 18,(result,declarationList1left,SEMICOLON1right),
rest671) end
| (62,rest671) => let val result=MlyValue.paramDeclarationList(([]))
 in (LrTable.NT 19,(result,defaultPos,defaultPos),rest671) end
| (63,(_,(_,_,popScope1right))::(_,(MlyValue.declarationList 
declarationList,_,_))::(_,(_,pushScope1left,_))::rest671) => let val 
result=MlyValue.paramDeclarationList((declarationList))
 in (LrTable.NT 19,(result,pushScope1left,popScope1right),rest671) end
| (64,(_,(_,_,attributeOptright as attributeOpt1right))::(_,(
MlyValue.initDeclaratorList initDeclaratorList,_,_))::(_,(
MlyValue.declarationSpecifierList declarationSpecifierList,
declarationSpecifierListleft as declarationSpecifierList1left,_))::
rest671) => let val result=MlyValue.declaration((
let
			   val (sc, ts, tq) = declarationSpecifierList
			   fun auxMap ((region, id, dL), init) = 
				 let
				   val name = id
				   val typ  = AST.SyntacticType(sc, ts, tq, dL)
				 in
				   (SymbolTable.insertId id typ;
				    case sc of
				     SOME(AST.Typedef) =>
					 AST.TypeDef{name = name,
						     region = region,
						     typ  = typ}
				   | _ =>
					 (case dL of
					      (AST.Function(params)::returnT) =>
						  (AST.FunDecl{name   = name,
							       region = region,
							       return = AST.SyntacticType(sc,ts,tq,returnT),
							       typ    = AST.SyntacticType(sc,ts,tq,dL),
							       params = SOME params})
					    | (AST.OldFunction([],NONE)::returnT) => 
						  (AST.FunDecl{name   = name,
							       region = region,
							       return = AST.SyntacticType(sc,ts,tq,returnT),
							       typ    = AST.SyntacticType(sc,ts,tq,dL),
							       params = NONE})
					  | (AST.OldFunction(_)::returnT) => 
						  (error region "Invalid old-style function declaration";
						   AST.FunDecl{name   = name,
							       region = region,
							       return = AST.SyntacticType(sc,ts,tq,returnT),
							       typ    = AST.SyntacticType(sc,ts,tq,dL),
							       params = NONE})
					  | _ =>
						AST.VarDecl{name = name,
							    region = region,
							    typ	 = typ,
							    init = init}))
				 end
			in
			  map auxMap initDeclaratorList
			end handle _ => (error (declarationSpecifierListleft, attributeOptright) "Bad declaration"; raise Fail "Bad decl.")
))
 in (LrTable.NT 20,(result,declarationSpecifierList1left,
attributeOpt1right),rest671) end
| (65,(_,(_,_,attributeOpt1right))::(_,(
MlyValue.declarationSpecifierList declarationSpecifierList,
declarationSpecifierListleft as declarationSpecifierList1left,
declarationSpecifierListright))::rest671) => let val result=
MlyValue.declaration((
let
			   val (sc, ts, tq) = declarationSpecifierList
			   val typ  = AST.SyntacticType(sc, ts, tq, nil)
			   val name = 
				(case ts of 
				     AST.Aggregate(_,SOME(id),_) => AST.mkAggrId id
				   | _ => AST.mkFreshId ())
			 in
			   [AST.TypeDef{name = name,
					region=(declarationSpecifierListleft,
						declarationSpecifierListright),
					typ  = typ}]
			 end handle _ =>
			(error (declarationSpecifierListleft, declarationSpecifierListright) "Bad declaration"; raise Fail "Bad decl.")
))
 in (LrTable.NT 20,(result,declarationSpecifierList1left,
attributeOpt1right),rest671) end
| (66,(_,(MlyValue.declarationSpecifierListAux 
declarationSpecifierListAux,_,declarationSpecifierListAuxright as 
declarationSpecifierListAux1right))::(_,(
MlyValue.storageClassSpecifier storageClassSpecifier,
storageClassSpecifierleft as storageClassSpecifier1left,_))::rest671)
 => let val result=MlyValue.declarationSpecifierList((
let val (sc,ts,tq) = declarationSpecifierListAux
			     val err = error (storageClassSpecifierleft,
					      declarationSpecifierListAuxright)
			 in
			     (mergeStorageClasses err (sc, SOME storageClassSpecifier), ts, tq)
			 end
		  handle _ => (error (storageClassSpecifierleft, declarationSpecifierListAuxright)
				     "Bad declaration list "; raise Fail "Bad decl.")
))
 in (LrTable.NT 21,(result,storageClassSpecifier1left,
declarationSpecifierListAux1right),rest671) end
| (67,(_,(MlyValue.declarationSpecifierListAux 
declarationSpecifierListAux,_,declarationSpecifierListAuxright as 
declarationSpecifierListAux1right))::(_,(MlyValue.typeQualifier 
typeQualifier,typeQualifierleft as typeQualifier1left,_))::rest671)
 => let val result=MlyValue.declarationSpecifierList((
let val (sc,ts,tql) = declarationSpecifierListAux
			     val err = error (typeQualifierleft,
					      declarationSpecifierListAuxright)
			 in
			     (sc, ts, mergeTypeQualifiers err (typeQualifier,tql))
			 end 
		  handle _ => (error (typeQualifierleft, declarationSpecifierListAuxright)
				     "Bad declaration list "; raise Fail "Bad decl.")
))
 in (LrTable.NT 21,(result,typeQualifier1left,
declarationSpecifierListAux1right),rest671) end
| (68,(_,(MlyValue.definedDeclarationSpecifierList 
definedDeclarationSpecifierList,definedDeclarationSpecifierList1left,
definedDeclarationSpecifierList1right))::rest671) => let val result=
MlyValue.declarationSpecifierList((definedDeclarationSpecifierList))
 in (LrTable.NT 21,(result,definedDeclarationSpecifierList1left,
definedDeclarationSpecifierList1right),rest671) end
| (69,rest671) => let val result=MlyValue.declarationSpecifierListAux(
(NONE, AST.Int, []))
 in (LrTable.NT 22,(result,defaultPos,defaultPos),rest671) end
| (70,(_,(MlyValue.declarationSpecifierListAux 
declarationSpecifierListAux,_,declarationSpecifierListAuxright as 
declarationSpecifierListAux1right))::(_,(
MlyValue.storageClassSpecifier storageClassSpecifier,
storageClassSpecifierleft as storageClassSpecifier1left,_))::rest671)
 => let val result=MlyValue.declarationSpecifierListAux((
let val (sc,ts,tq) = declarationSpecifierListAux
			     val err = error (storageClassSpecifierleft,
					      declarationSpecifierListAuxright)
			 in
			     (mergeStorageClasses err (sc, SOME storageClassSpecifier), ts, tq)
			 end
				 handle _ => (error (storageClassSpecifierleft, declarationSpecifierListAuxright)
					     "Bad declaration list "; raise Fail "Bad decl.")
))
 in (LrTable.NT 22,(result,storageClassSpecifier1left,
declarationSpecifierListAux1right),rest671) end
| (71,(_,(MlyValue.declarationSpecifierListAux 
declarationSpecifierListAux,_,declarationSpecifierListAuxright as 
declarationSpecifierListAux1right))::(_,(MlyValue.typeQualifier 
typeQualifier,typeQualifierleft as typeQualifier1left,_))::rest671)
 => let val result=MlyValue.declarationSpecifierListAux((
let val (sc,ts,tql) = declarationSpecifierListAux
			     val err = error (typeQualifierleft,
					      declarationSpecifierListAuxright)
			 in
			     (sc, ts, mergeTypeQualifiers err (typeQualifier,tql))
			 end
				 handle _ => (error (typeQualifierleft, declarationSpecifierListAuxright)
					     "Bad declaration list "; raise Fail "Bad decl.")
))
 in (LrTable.NT 22,(result,typeQualifier1left,
declarationSpecifierListAux1right),rest671) end
| (72,(_,(MlyValue.definedDeclarationSpecifierList 
definedDeclarationSpecifierList,definedDeclarationSpecifierList1left,
definedDeclarationSpecifierList1right))::rest671) => let val result=
MlyValue.declarationSpecifierListAux((definedDeclarationSpecifierList)
)
 in (LrTable.NT 22,(result,definedDeclarationSpecifierList1left,
definedDeclarationSpecifierList1right),rest671) end
| (73,(_,(MlyValue.sqDeclarationSpecifierList 
sqDeclarationSpecifierList,_,sqDeclarationSpecifierListright as 
sqDeclarationSpecifierList1right))::(_,(MlyValue.typeSpecifier 
typeSpecifier,typeSpecifierleft as typeSpecifier1left,_))::rest671)
 => let val result=MlyValue.definedDeclarationSpecifierList((
let val (sc,tq) = sqDeclarationSpecifierList
			 in
			     (sc, typeSpecifier, tq)
			 end
				 handle _ => (error (typeSpecifierleft, sqDeclarationSpecifierListright)
					     "Bad declaration list "; raise Fail "Bad decl.")
))
 in (LrTable.NT 23,(result,typeSpecifier1left,
sqDeclarationSpecifierList1right),rest671) end
| (74,(_,(MlyValue.baseDeclarationSpecifierList 
baseDeclarationSpecifierList,_,baseDeclarationSpecifierListright as 
baseDeclarationSpecifierList1right))::(_,(MlyValue.baseTypeSpecifier 
baseTypeSpecifier,baseTypeSpecifierleft as baseTypeSpecifier1left,_))
::rest671) => let val result=MlyValue.definedDeclarationSpecifierList(
(
let
			   val (scO, ts, tq) = baseDeclarationSpecifierList
			   val err = error (baseTypeSpecifierleft,
					    baseDeclarationSpecifierListright)

			   val ts' = mergeBaseTypeSpecifiers err (ts,baseTypeSpecifier)
			 in
			     (scO, formTySpec err ts', tq)
			 end
				 handle _ => (error (baseTypeSpecifierleft, baseDeclarationSpecifierListright)
					     "Bad declaration list "; raise Fail "Bad decl.")
))
 in (LrTable.NT 23,(result,baseTypeSpecifier1left,
baseDeclarationSpecifierList1right),rest671) end
| (75,(_,(MlyValue.sqDeclarationSpecifierList 
sqDeclarationSpecifierList,_,sqDeclarationSpecifierListright as 
sqDeclarationSpecifierList1right))::(_,(MlyValue.storageClassSpecifier
 storageClassSpecifier,storageClassSpecifierleft as 
storageClassSpecifier1left,_))::rest671) => let val result=
MlyValue.sqDeclarationSpecifierList((
let val (sc,tq) = sqDeclarationSpecifierList
			     val err = error (storageClassSpecifierleft,
					      sqDeclarationSpecifierListright)
			 in
			     (mergeStorageClasses err (sc, SOME storageClassSpecifier), tq)
			 end
				 handle _ => (error (storageClassSpecifierleft, sqDeclarationSpecifierListright)
					     "Bad declaration list "; raise Fail "Bad decl.")
))
 in (LrTable.NT 24,(result,storageClassSpecifier1left,
sqDeclarationSpecifierList1right),rest671) end
| (76,(_,(MlyValue.sqDeclarationSpecifierList 
sqDeclarationSpecifierList,_,sqDeclarationSpecifierListright as 
sqDeclarationSpecifierList1right))::(_,(MlyValue.typeQualifier 
typeQualifier,typeQualifierleft as typeQualifier1left,_))::rest671)
 => let val result=MlyValue.sqDeclarationSpecifierList((
let val (sc,tql) = sqDeclarationSpecifierList
			     val err = error (typeQualifierleft,sqDeclarationSpecifierListright)
			 in
			     (sc, mergeTypeQualifiers err (typeQualifier,tql))
			 end
				 handle _ => (error (typeQualifierleft,sqDeclarationSpecifierListright)
					     "Bad declaration list "; raise Fail "Bad decl.")
))
 in (LrTable.NT 24,(result,typeQualifier1left,
sqDeclarationSpecifierList1right),rest671) end
| (77,rest671) => let val result=MlyValue.sqDeclarationSpecifierList((
NONE,[]))
 in (LrTable.NT 24,(result,defaultPos,defaultPos),rest671) end
| (78,rest671) => let val result=MlyValue.baseDeclarationSpecifierList
((NONE, (NONE,NONE,NONE), []))
 in (LrTable.NT 26,(result,defaultPos,defaultPos),rest671) end
| (79,(_,(MlyValue.baseDeclarationSpecifier baseDeclarationSpecifier,_
,baseDeclarationSpecifierright as baseDeclarationSpecifier1right))::(_
,(MlyValue.baseDeclarationSpecifierList baseDeclarationSpecifierList,
baseDeclarationSpecifierListleft as baseDeclarationSpecifierList1left,
_))::rest671) => let val result=MlyValue.baseDeclarationSpecifierList(
(
let
			   val (scO, ts, tql) = baseDeclarationSpecifierList
			   val (scO', ts', tqo) = baseDeclarationSpecifier
			   val err = error (baseDeclarationSpecifierListleft,
					    baseDeclarationSpecifierright)

			   val scO'' = mergeStorageClasses err (scO,scO')
			   val ts'' = mergeBaseTypeSpecifiers err (ts,ts')
			   val tql' = case tqo of
					SOME tq => mergeTypeQualifiers err (tq,tql)
				      | NONE => tql
			 in
			     (scO'', ts'', tql')
			 end
			    handle _ => (error (baseDeclarationSpecifierListleft, baseDeclarationSpecifierright)
					 "Bad declaration list "; raise Fail "Bad decl.")
))
 in (LrTable.NT 26,(result,baseDeclarationSpecifierList1left,
baseDeclarationSpecifier1right),rest671) end
| (80,(_,(MlyValue.storageClassSpecifier storageClassSpecifier,
storageClassSpecifier1left,storageClassSpecifier1right))::rest671) => 
let val result=MlyValue.baseDeclarationSpecifier((
SOME(storageClassSpecifier), (NONE, NONE, NONE), NONE))
 in (LrTable.NT 25,(result,storageClassSpecifier1left,
storageClassSpecifier1right),rest671) end
| (81,(_,(MlyValue.baseTypeSpecifier baseTypeSpecifier,
baseTypeSpecifier1left,baseTypeSpecifier1right))::rest671) => let val 
result=MlyValue.baseDeclarationSpecifier((
NONE, baseTypeSpecifier, NONE))
 in (LrTable.NT 25,(result,baseTypeSpecifier1left,
baseTypeSpecifier1right),rest671) end
| (82,(_,(MlyValue.typeQualifier typeQualifier,typeQualifier1left,
typeQualifier1right))::rest671) => let val result=
MlyValue.baseDeclarationSpecifier((
NONE, (NONE, NONE, NONE), SOME typeQualifier))
 in (LrTable.NT 25,(result,typeQualifier1left,typeQualifier1right),
rest671) end
| (83,(_,(MlyValue.initDeclarator initDeclarator,initDeclarator1left,
initDeclarator1right))::rest671) => let val result=
MlyValue.initDeclaratorList(([initDeclarator]))
 in (LrTable.NT 27,(result,initDeclarator1left,initDeclarator1right),
rest671) end
| (84,(_,(MlyValue.initDeclarator initDeclarator,_,
initDeclarator1right))::_::(_,(MlyValue.initDeclaratorList 
initDeclaratorList,initDeclaratorList1left,_))::rest671) => let val 
result=MlyValue.initDeclaratorList((
initDeclarator :: initDeclaratorList))
 in (LrTable.NT 27,(result,initDeclaratorList1left,
initDeclarator1right),rest671) end
| (85,(_,(MlyValue.declarator declarator,declarator1left,
declarator1right))::rest671) => let val result=MlyValue.initDeclarator
((declarator, NONE))
 in (LrTable.NT 28,(result,declarator1left,declarator1right),rest671)
 end
| (86,(_,(MlyValue.constantExpr constantExpr,_,constantExpr1right))::_
::(_,(MlyValue.declarator declarator,declarator1left,_))::rest671) => 
let val result=MlyValue.initDeclarator((
declarator, SOME(AST.Initial(constantExpr))))
 in (LrTable.NT 28,(result,declarator1left,constantExpr1right),rest671
) end
| (87,(_,(MlyValue.initializer initializer,_,initializer1right))::_::(
_,(MlyValue.declarator declarator,declarator1left,_))::rest671) => 
let val result=MlyValue.initDeclarator((declarator, SOME(initializer))
)
 in (LrTable.NT 28,(result,declarator1left,initializer1right),rest671)
 end
| (88,(_,(MlyValue.constantExpr constantExpr,_,constantExpr1right))::_
::(_,(MlyValue.declarator declarator,declarator1left,_))::rest671) => 
let val result=MlyValue.initDeclarator((
let val (region,id, dL) = declarator in
			     ((region, id, (AST.BitField(constantExpr)) :: dL),
			      NONE)
			end
))
 in (LrTable.NT 28,(result,declarator1left,constantExpr1right),rest671
) end
| (89,(_,(MlyValue.constantExpr constantExpr,_,constantExpr1right))::(
_,(_,COLONleft as COLON1left,COLONright))::rest671) => let val result=
MlyValue.initDeclarator((
((COLONleft, COLONright), AST.mkFreshId(),
			  [AST.BitField(constantExpr)]),
			 NONE
))
 in (LrTable.NT 28,(result,COLON1left,constantExpr1right),rest671) end
| (90,(_,(MlyValue.assignExpr assignExpr,assignExpr1left,
assignExpr1right))::rest671) => let val result=MlyValue.initializer((
AST.Initial(assignExpr)))
 in (LrTable.NT 29,(result,assignExpr1left,assignExpr1right),rest671)
 end
| (91,(_,(_,_,RC1right))::(_,(MlyValue.braceInitializerList 
braceInitializerList,_,_))::(_,(_,LC1left,_))::rest671) => let val 
result=MlyValue.initializer((AST.InitialList(rev braceInitializerList)
))
 in (LrTable.NT 29,(result,LC1left,RC1right),rest671) end
| (92,(_,(_,_,RC1right))::_::(_,(MlyValue.braceInitializerList 
braceInitializerList,_,_))::(_,(_,LC1left,_))::rest671) => let val 
result=MlyValue.initializer((AST.InitialList(rev braceInitializerList)
))
 in (LrTable.NT 29,(result,LC1left,RC1right),rest671) end
| (93,(_,(MlyValue.constantExpr constantExpr,constantExpr1left,
constantExpr1right))::rest671) => let val result=
MlyValue.braceInitializer((AST.Initial(constantExpr)))
 in (LrTable.NT 30,(result,constantExpr1left,constantExpr1right),
rest671) end
| (94,(_,(_,_,RC1right))::(_,(MlyValue.braceInitializerList 
braceInitializerList,_,_))::(_,(_,LC1left,_))::rest671) => let val 
result=MlyValue.braceInitializer((
AST.InitialList(rev braceInitializerList)))
 in (LrTable.NT 30,(result,LC1left,RC1right),rest671) end
| (95,(_,(_,_,RC1right))::_::(_,(MlyValue.braceInitializerList 
braceInitializerList,_,_))::(_,(_,LC1left,_))::rest671) => let val 
result=MlyValue.braceInitializer((
AST.InitialList(rev braceInitializerList)))
 in (LrTable.NT 30,(result,LC1left,RC1right),rest671) end
| (96,rest671) => let val result=MlyValue.initCastOpt((NONE))
 in (LrTable.NT 32,(result,defaultPos,defaultPos),rest671) end
| (97,(_,(_,_,RP1right))::(_,(MlyValue.typeName typeName,_,_))::(_,(_,
LP1left,_))::rest671) => let val result=MlyValue.initCastOpt((
SOME typeName))
 in (LrTable.NT 32,(result,LP1left,RP1right),rest671) end
| (98,rest671) => let val result=MlyValue.braceInitializerList((nil))
 in (LrTable.NT 31,(result,defaultPos,defaultPos),rest671) end
| (99,(_,(MlyValue.braceInitializer braceInitializer,
braceInitializer1left,braceInitializer1right))::rest671) => let val 
result=MlyValue.braceInitializerList(([braceInitializer]))
 in (LrTable.NT 31,(result,braceInitializer1left,
braceInitializer1right),rest671) end
| (100,(_,(MlyValue.braceInitializer braceInitializer,_,
braceInitializer1right))::_::(_,(MlyValue.braceInitializerList 
braceInitializerList,braceInitializerList1left,_))::rest671) => let 
val result=MlyValue.braceInitializerList((
braceInitializer :: braceInitializerList))
 in (LrTable.NT 31,(result,braceInitializerList1left,
braceInitializer1right),rest671) end
| (101,(_,(MlyValue.directDeclarator directDeclarator,_,
directDeclarator1right))::(_,(MlyValue.pointer pointer,pointer1left,_)
)::rest671) => let val result=MlyValue.declarator((
let
			   val (region, id, dL) = directDeclarator
			 in
			   (region, id, List.rev dL @ (List.rev pointer))
			 end
))
 in (LrTable.NT 37,(result,pointer1left,directDeclarator1right),
rest671) end
| (102,(_,(MlyValue.directDeclarator directDeclarator,
directDeclarator1left,directDeclarator1right))::rest671) => let val 
result=MlyValue.declarator((
let val (region,id,dL) = directDeclarator
			 in
			     (region,id,List.rev dL)
			 end
))
 in (LrTable.NT 37,(result,directDeclarator1left,
directDeclarator1right),rest671) end
| (103,(_,(_,MUL1left,MUL1right))::rest671) => let val result=
MlyValue.pointer(([AST.Pointer([])]))
 in (LrTable.NT 45,(result,MUL1left,MUL1right),rest671) end
| (104,(_,(MlyValue.typeQualifierList typeQualifierList,_,
typeQualifierList1right))::(_,(_,MUL1left,_))::rest671) => let val 
result=MlyValue.pointer(([AST.Pointer(typeQualifierList)]))
 in (LrTable.NT 45,(result,MUL1left,typeQualifierList1right),rest671)
 end
| (105,(_,(MlyValue.pointer pointer,_,pointer1right))::(_,(_,MUL1left,
_))::rest671) => let val result=MlyValue.pointer((
(AST.Pointer([]))
						 :: pointer))
 in (LrTable.NT 45,(result,MUL1left,pointer1right),rest671) end
| (106,(_,(MlyValue.pointer pointer,_,pointer1right))::(_,(
MlyValue.typeQualifierList typeQualifierList,_,_))::(_,(_,MUL1left,_))
::rest671) => let val result=MlyValue.pointer((
(AST.Pointer(typeQualifierList)) :: pointer))
 in (LrTable.NT 45,(result,MUL1left,pointer1right),rest671) end
| (107,(_,(_,_,RB1right))::(_,(_,LB1left,_))::rest671) => let val 
result=MlyValue.arrayDeclarator((NONE))
 in (LrTable.NT 46,(result,LB1left,RB1right),rest671) end
| (108,(_,(_,_,RB1right))::(_,(MlyValue.constantExpr constantExpr,_,_)
)::(_,(_,LB1left,_))::rest671) => let val result=
MlyValue.arrayDeclarator((SOME(constantExpr)))
 in (LrTable.NT 46,(result,LB1left,RB1right),rest671) end
| (109,(_,(MlyValue.IDENTIFIER IDENTIFIER,IDENTIFIERleft,
IDENTIFIERright as IDENTIFIER1right))::(_,(_,attributeOpt1left,_))::
rest671) => let val result=MlyValue.directDeclarator((
(IDENTIFIERleft,
						  IDENTIFIERright),
						 IDENTIFIER, nil))
 in (LrTable.NT 47,(result,attributeOpt1left,IDENTIFIER1right),rest671
) end
| (110,(_,(MlyValue.TYPEDEF_NAME TYPEDEF_NAME,TYPEDEF_NAMEleft as 
TYPEDEF_NAME1left,TYPEDEF_NAMEright as TYPEDEF_NAME1right))::rest671)
 => let val result=MlyValue.directDeclarator((
(TYPEDEF_NAMEleft,
						  TYPEDEF_NAMEright),
						 TYPEDEF_NAME, nil
))
 in (LrTable.NT 47,(result,TYPEDEF_NAME1left,TYPEDEF_NAME1right),
rest671) end
| (111,(_,(_,_,RP1right))::(_,(MlyValue.declarator declarator,_,_))::(
_,(_,LP1left,_))::rest671) => let val result=MlyValue.directDeclarator
((
let val (region, id, dL) = declarator
			 in
			     (region, id, List.rev dL)
			 end
))
 in (LrTable.NT 47,(result,LP1left,RP1right),rest671) end
| (112,(_,(MlyValue.arrayDeclarator arrayDeclarator,_,
arrayDeclarator1right))::(_,(MlyValue.directDeclarator 
directDeclarator,directDeclarator1left,_))::rest671) => let val result
=MlyValue.directDeclarator((
let 
			   val (region, id, dL) = directDeclarator
			 in
			   (region, id, (AST.Array(arrayDeclarator)) :: dL)
			 end
))
 in (LrTable.NT 47,(result,directDeclarator1left,arrayDeclarator1right
),rest671) end
| (113,(_,(MlyValue.parameterTypeList parameterTypeList,_,
parameterTypeList1right))::(_,(MlyValue.directDeclarator 
directDeclarator,directDeclarator1left,_))::rest671) => let val result
=MlyValue.directDeclarator((
let
			   val (region, id, dL) = directDeclarator
			 in
			   (region, id,
			    (AST.Function(rev parameterTypeList)) :: dL)
			 end
))
 in (LrTable.NT 47,(result,directDeclarator1left,
parameterTypeList1right),rest671) end
| (114,(_,(MlyValue.oldParameterList oldParameterList,_,
oldParameterList1right))::(_,(MlyValue.directDeclarator 
directDeclarator,directDeclarator1left,_))::rest671) => let val result
=MlyValue.directDeclarator((
let
			   val (region, id, dL) = directDeclarator
			 in
			   (region, id,
			    (AST.OldFunction(oldParameterList)) :: dL)
			 end
))
 in (LrTable.NT 47,(result,directDeclarator1left,
oldParameterList1right),rest671) end
| (115,rest671) => let val result=MlyValue.attributeOpt(())
 in (LrTable.NT 39,(result,defaultPos,defaultPos),rest671) end
| (116,(_,(_,attributes1left,attributes1right))::rest671) => let val 
result=MlyValue.attributeOpt(())
 in (LrTable.NT 39,(result,attributes1left,attributes1right),rest671)
 end
| (117,(_,(_,attribute1left,attribute1right))::rest671) => let val 
result=MlyValue.attributes(())
 in (LrTable.NT 40,(result,attribute1left,attribute1right),rest671)
 end
| (118,(_,(_,_,attribute1right))::(_,(_,attributes1left,_))::rest671)
 => let val result=MlyValue.attributes(())
 in (LrTable.NT 40,(result,attributes1left,attribute1right),rest671)
 end
| (119,(_,(_,_,RP2right))::_::_::_::_::(_,(_,ATTRIBUTE1left,_))::
rest671) => let val result=MlyValue.attribute(())
 in (LrTable.NT 41,(result,ATTRIBUTE1left,RP2right),rest671) end
| (120,(_,(_,attrib1left,attrib1right))::rest671) => let val result=
MlyValue.attribute_list(())
 in (LrTable.NT 42,(result,attrib1left,attrib1right),rest671) end
| (121,(_,(_,_,attrib1right))::_::(_,(_,attribute_list1left,_))::
rest671) => let val result=MlyValue.attribute_list(())
 in (LrTable.NT 42,(result,attribute_list1left,attrib1right),rest671)
 end
| (122,rest671) => let val result=MlyValue.attrib(())
 in (LrTable.NT 43,(result,defaultPos,defaultPos),rest671) end
| (123,(_,(_,any_word1left,any_word1right))::rest671) => let val 
result=MlyValue.attrib(())
 in (LrTable.NT 43,(result,any_word1left,any_word1right),rest671) end
| (124,(_,(_,_,RP1right))::_::_::(_,(_,any_word1left,_))::rest671) => 
let val result=MlyValue.attrib(())
 in (LrTable.NT 43,(result,any_word1left,RP1right),rest671) end
| (125,(_,(_,IDENTIFIER1left,IDENTIFIER1right))::rest671) => let val 
result=MlyValue.any_word(())
 in (LrTable.NT 44,(result,IDENTIFIER1left,IDENTIFIER1right),rest671)
 end
| (126,(_,(_,TYPEDEF_NAME1left,TYPEDEF_NAME1right))::rest671) => let 
val result=MlyValue.any_word(())
 in (LrTable.NT 44,(result,TYPEDEF_NAME1left,TYPEDEF_NAME1right),
rest671) end
| (127,(_,(_,QUALIFIER1left,QUALIFIER1right))::rest671) => let val 
result=MlyValue.any_word(())
 in (LrTable.NT 44,(result,QUALIFIER1left,QUALIFIER1right),rest671)
 end
| (128,(_,(_,CONST1left,CONST1right))::rest671) => let val result=
MlyValue.any_word(())
 in (LrTable.NT 44,(result,CONST1left,CONST1right),rest671) end
| (129,(_,(_,NONCONST1left,NONCONST1right))::rest671) => let val 
result=MlyValue.any_word(())
 in (LrTable.NT 44,(result,NONCONST1left,NONCONST1right),rest671) end
| (130,(_,(MlyValue.pointer pointer,pointer1left,pointer1right))::
rest671) => let val result=MlyValue.abstractDeclarator((
List.rev pointer))
 in (LrTable.NT 48,(result,pointer1left,pointer1right),rest671) end
| (131,(_,(MlyValue.directAbstractDeclarator directAbstractDeclarator,
directAbstractDeclarator1left,directAbstractDeclarator1right))::
rest671) => let val result=MlyValue.abstractDeclarator((
List.rev directAbstractDeclarator))
 in (LrTable.NT 48,(result,directAbstractDeclarator1left,
directAbstractDeclarator1right),rest671) end
| (132,(_,(MlyValue.directAbstractDeclarator directAbstractDeclarator,
_,directAbstractDeclarator1right))::(_,(MlyValue.pointer pointer,
pointer1left,_))::rest671) => let val result=
MlyValue.abstractDeclarator((
List.rev directAbstractDeclarator @ List.rev pointer))
 in (LrTable.NT 48,(result,pointer1left,directAbstractDeclarator1right
),rest671) end
| (133,(_,(_,_,RP1right))::(_,(MlyValue.abstractDeclarator 
abstractDeclarator,_,_))::(_,(_,LP1left,_))::rest671) => let val 
result=MlyValue.directAbstractDeclarator((abstractDeclarator))
 in (LrTable.NT 49,(result,LP1left,RP1right),rest671) end
| (134,(_,(MlyValue.arrayDeclarator arrayDeclarator,_,
arrayDeclarator1right))::(_,(MlyValue.directAbstractDeclarator 
directAbstractDeclarator,directAbstractDeclarator1left,_))::rest671)
 => let val result=MlyValue.directAbstractDeclarator((
(AST.Array(arrayDeclarator)) :: directAbstractDeclarator))
 in (LrTable.NT 49,(result,directAbstractDeclarator1left,
arrayDeclarator1right),rest671) end
| (135,(_,(MlyValue.arrayDeclarator arrayDeclarator,
arrayDeclarator1left,arrayDeclarator1right))::rest671) => let val 
result=MlyValue.directAbstractDeclarator(([AST.Array(arrayDeclarator)]
))
 in (LrTable.NT 49,(result,arrayDeclarator1left,arrayDeclarator1right)
,rest671) end
| (136,(_,(MlyValue.parameterTypeList parameterTypeList,_,
parameterTypeList1right))::(_,(MlyValue.directAbstractDeclarator 
directAbstractDeclarator,directAbstractDeclarator1left,_))::rest671)
 => let val result=MlyValue.directAbstractDeclarator((
(AST.Function(rev  parameterTypeList)) :: 
			  directAbstractDeclarator
))
 in (LrTable.NT 49,(result,directAbstractDeclarator1left,
parameterTypeList1right),rest671) end
| (137,(_,(MlyValue.parameterTypeList parameterTypeList,
parameterTypeList1left,parameterTypeList1right))::rest671) => let val 
result=MlyValue.directAbstractDeclarator((
[AST.Function(rev parameterTypeList)]))
 in (LrTable.NT 49,(result,parameterTypeList1left,
parameterTypeList1right),rest671) end
| (138,(_,(_,_,RP1right))::_::(_,(MlyValue.directAbstractDeclarator 
directAbstractDeclarator,directAbstractDeclarator1left,_))::rest671)
 => let val result=MlyValue.directAbstractDeclarator((
AST.OldFunction([],NONE) ::
			 directAbstractDeclarator))
 in (LrTable.NT 49,(result,directAbstractDeclarator1left,RP1right),
rest671) end
| (139,(_,(_,_,RP1right))::(_,(_,LP1left,_))::rest671) => let val 
result=MlyValue.directAbstractDeclarator(([AST.OldFunction ([],NONE)])
)
 in (LrTable.NT 49,(result,LP1left,RP1right),rest671) end
| (140,(_,(_,_,popScope1right))::_::(_,(MlyValue.parameterList 
parameterList,_,_))::_::(_,(_,LP1left,_))::rest671) => let val result=
MlyValue.parameterTypeList((parameterList))
 in (LrTable.NT 50,(result,LP1left,popScope1right),rest671) end
| (141,(_,(_,_,popScope1right))::_::(_,(_,ELLIPSISleft,ELLIPSISright))
::_::(_,(MlyValue.parameterList parameterList,_,_))::_::(_,(_,LP1left,
_))::rest671) => let val result=MlyValue.parameterTypeList((
(AST.EllipsisDecl {region = (ELLIPSISleft,
						      ELLIPSISright)})
				  :: parameterList
))
 in (LrTable.NT 50,(result,LP1left,popScope1right),rest671) end
| (142,(_,(_,_,popScope1right))::_::_::_::(_,(_,LP1left,_))::rest671)
 => let val result=MlyValue.parameterTypeList(([]))
 in (LrTable.NT 50,(result,LP1left,popScope1right),rest671) end
| (143,(_,(_,_,popScope1right))::_::(_,(MlyValue.identifierList 
identifierList,_,_))::_::(_,(_,LP1left,_))::rest671) => let val result
=MlyValue.oldParameterList(((List.rev identifierList,NONE)))
 in (LrTable.NT 51,(result,LP1left,popScope1right),rest671) end
| (144,(_,(_,_,popScope1right))::_::_::(_,(_,LP1left,_))::rest671) => 
let val result=MlyValue.oldParameterList((([],NONE)))
 in (LrTable.NT 51,(result,LP1left,popScope1right),rest671) end
| (145,(_,(_,_,popScope1right))::_::(_,(_,ELLIPSISleft,ELLIPSISright))
::_::(_,(MlyValue.identifierList identifierList,_,_))::_::(_,(_,
LP1left,_))::rest671) => let val result=MlyValue.oldParameterList((
(List.rev identifierList, SOME (ELLIPSISleft,ELLIPSISright))))
 in (LrTable.NT 51,(result,LP1left,popScope1right),rest671) end
| (146,(_,(MlyValue.parameterDeclaration parameterDeclaration,
parameterDeclaration1left,parameterDeclaration1right))::rest671) => 
let val result=MlyValue.parameterList(([parameterDeclaration]))
 in (LrTable.NT 53,(result,parameterDeclaration1left,
parameterDeclaration1right),rest671) end
| (147,(_,(MlyValue.parameterDeclaration parameterDeclaration,_,
parameterDeclaration1right))::_::(_,(MlyValue.parameterList 
parameterList,parameterList1left,_))::rest671) => let val result=
MlyValue.parameterList((parameterDeclaration :: parameterList))
 in (LrTable.NT 53,(result,parameterList1left,
parameterDeclaration1right),rest671) end
| (148,(_,(MlyValue.IDENTIFIER IDENTIFIER,IDENTIFIERleft as 
IDENTIFIER1left,IDENTIFIERright as IDENTIFIER1right))::rest671) => 
let val result=MlyValue.identifierList((
[((IDENTIFIERleft,IDENTIFIERright),IDENTIFIER)]))
 in (LrTable.NT 52,(result,IDENTIFIER1left,IDENTIFIER1right),rest671)
 end
| (149,(_,(MlyValue.IDENTIFIER IDENTIFIER,IDENTIFIERleft,
IDENTIFIERright as IDENTIFIER1right))::_::(_,(MlyValue.identifierList 
identifierList,identifierList1left,_))::rest671) => let val result=
MlyValue.identifierList((
((IDENTIFIERleft,IDENTIFIERright),IDENTIFIER) :: 
								 identifierList
))
 in (LrTable.NT 52,(result,identifierList1left,IDENTIFIER1right),
rest671) end
| (150,(_,(MlyValue.TYPEDEF_NAME TYPEDEF_NAME,TYPEDEF_NAMEleft,
TYPEDEF_NAMEright as TYPEDEF_NAME1right))::_::(_,(
MlyValue.identifierList identifierList,identifierList1left,_))::
rest671) => let val result=MlyValue.identifierList((
((TYPEDEF_NAMEleft,TYPEDEF_NAMEright),
								  TYPEDEF_NAME) :: identifierList
))
 in (LrTable.NT 52,(result,identifierList1left,TYPEDEF_NAME1right),
rest671) end
| (151,(_,(MlyValue.declarator declarator,_,declaratorright as 
declarator1right))::(_,(MlyValue.declarationSpecifierList 
declarationSpecifierList,declarationSpecifierListleft as 
declarationSpecifierList1left,_))::rest671) => let val result=
MlyValue.parameterDeclaration((
let
			   val (sc, ts, tq) = declarationSpecifierList
			   val (region, id, dL) = declarator
			   val name = id
			   val typ = AST.SyntacticType (sc, ts, tq, dL)
			 in
			   SymbolTable.insertId name typ;
			   AST.VarDecl{name = name,
				       region = region,
				       typ  = typ,
				       init = NONE}
			 end
				handle _ => (error (declarationSpecifierListleft, declaratorright) "Bad parameter declaration";
					     raise Fail "Bad parameter")
))
 in (LrTable.NT 54,(result,declarationSpecifierList1left,
declarator1right),rest671) end
| (152,(_,(MlyValue.declarationSpecifierList declarationSpecifierList,
declarationSpecifierListleft as declarationSpecifierList1left,
declarationSpecifierListright as declarationSpecifierList1right))::
rest671) => let val result=MlyValue.parameterDeclaration((
let
			   val (sc, ts, tq) = declarationSpecifierList
			   val id = AST.mkFreshId ()
			   val name = id
			   val typ = AST.SyntacticType (sc, ts, tq, nil)
			 in
			   SymbolTable.insertId name typ;
			   AST.VarDecl{name = name,
				       region = (declarationSpecifierListleft,
						 declarationSpecifierListright),
				       typ  = typ,
				       init = NONE}
			 end
				handle _ => (error (declarationSpecifierListleft, declarationSpecifierListright) "Bad parameter declaration";
					     raise Fail "Bad parameter")
))
 in (LrTable.NT 54,(result,declarationSpecifierList1left,
declarationSpecifierList1right),rest671) end
| (153,(_,(MlyValue.abstractDeclarator abstractDeclarator,_,
abstractDeclaratorright as abstractDeclarator1right))::(_,(
MlyValue.declarationSpecifierList declarationSpecifierList,
declarationSpecifierListleft as declarationSpecifierList1left,_))::
rest671) => let val result=MlyValue.parameterDeclaration((
let
			   val (sc, ts, tq) = declarationSpecifierList
			   val typ = 
				AST.SyntacticType (sc, ts, tq, abstractDeclarator)
			   val id = AST.mkFreshId ()
			   val name = id
			 in
			   SymbolTable.insertId name typ;
			   AST.VarDecl{name = name,
				       region = (declarationSpecifierListleft,
						 abstractDeclaratorright),
				       typ  = typ,
				       init = NONE}
			 end
				handle _ => (error (declarationSpecifierListleft, abstractDeclaratorright) "Bad parameter declaration";
					     raise Fail "Bad parameter")
))
 in (LrTable.NT 54,(result,declarationSpecifierList1left,
abstractDeclarator1right),rest671) end
| (154,(_,(MlyValue.stmt stmt,stmt1left,stmt1right))::rest671) => let 
val result=MlyValue.stmtList(([stmt]))
 in (LrTable.NT 56,(result,stmt1left,stmt1right),rest671) end
| (155,(_,(MlyValue.stmt stmt,_,stmt1right))::(_,(MlyValue.stmtList 
stmtList,stmtList1left,_))::rest671) => let val result=
MlyValue.stmtList((stmt :: stmtList))
 in (LrTable.NT 56,(result,stmtList1left,stmt1right),rest671) end
| (156,(_,(_,_,RC1right))::_::(_,(MlyValue.getScope getScope,_,_))::(_
,(MlyValue.stmtList stmtList,_,_))::(_,(MlyValue.declarationList 
declarationList,_,_))::_::(_,(_,LC1left,_))::rest671) => let val 
result=MlyValue.compoundStmt((
rev declarationList, rev stmtList, getScope))
 in (LrTable.NT 57,(result,LC1left,RC1right),rest671) end
| (157,(_,(_,_,RC1right))::_::(_,(MlyValue.getScope getScope,_,_))::(_
,(MlyValue.declarationList declarationList,_,_))::_::(_,(_,LC1left,_))
::rest671) => let val result=MlyValue.compoundStmt((
rev declarationList, nil, getScope))
 in (LrTable.NT 57,(result,LC1left,RC1right),rest671) end
| (158,(_,(_,_,RC1right))::_::(_,(MlyValue.getScope getScope,_,_))::(_
,(MlyValue.stmtList stmtList,_,_))::_::(_,(_,LC1left,_))::rest671) => 
let val result=MlyValue.compoundStmt((nil, rev stmtList, getScope))
 in (LrTable.NT 57,(result,LC1left,RC1right),rest671) end
| (159,(_,(_,_,RC1right))::_::(_,(MlyValue.getScope getScope,_,_))::_
::(_,(_,LC1left,_))::rest671) => let val result=MlyValue.compoundStmt(
(nil, nil, getScope))
 in (LrTable.NT 57,(result,LC1left,RC1right),rest671) end
| (160,(_,(MlyValue.compoundStmt compoundStmt,compoundStmtleft as 
compoundStmt1left,compoundStmtright as compoundStmt1right))::rest671)
 => let val result=MlyValue.stmt((
let val (dl, sl, scop) = compoundStmt
       in
	   AST.Compound {decls = dl,
			 body = sl,
			 region = (compoundStmtleft, compoundStmtright),
			 extraInfo = ref NONE,
			 lookup = SymbolTable.scopeLookup scop}
       end
))
 in (LrTable.NT 58,(result,compoundStmt1left,compoundStmt1right),
rest671) end
| (161,(_,(MlyValue.stmtmaybe stmtmaybe,_,stmtmayberight as 
stmtmaybe1right))::_::(_,(MlyValue.IDENTIFIER IDENTIFIER,
IDENTIFIERleft as IDENTIFIER1left,_))::rest671) => let val result=
MlyValue.stmt((
AST.Labeled(IDENTIFIER, stmtmaybe, {region=(IDENTIFIERleft,
						   stmtmayberight),
					   extraInfo=ref NONE,
					   typeInfo=NONE})
))
 in (LrTable.NT 58,(result,IDENTIFIER1left,stmtmaybe1right),rest671)
 end
| (162,(_,(_,_,COLONright as COLON1right))::(_,(MlyValue.IDENTIFIER 
IDENTIFIER,IDENTIFIERleft as IDENTIFIER1left,_))::rest671) => let val 
result=MlyValue.stmt((
AST.Labeled(IDENTIFIER, NONE, {region=(IDENTIFIERleft,
					     COLONright),
					   extraInfo=ref NONE,
					   typeInfo=NONE})
))
 in (LrTable.NT 58,(result,IDENTIFIER1left,COLON1right),rest671) end
| (163,(_,(MlyValue.stmtmaybe stmtmaybe,_,stmtmayberight as 
stmtmaybe1right))::(_,(MlyValue.startCase startCase,_,_))::(_,(_,
CASEleft as CASE1left,_))::rest671) => let val result=MlyValue.stmt((
AST.Case(startCase, stmtmaybe, {region=(CASEleft, stmtmayberight),
				       extraInfo=ref NONE,
				       typeInfo=NONE})
))
 in (LrTable.NT 58,(result,CASE1left,stmtmaybe1right),rest671) end
| (164,(_,(MlyValue.startCase startCase,_,startCaseright as 
startCase1right))::(_,(_,CASEleft as CASE1left,_))::rest671) => let 
val result=MlyValue.stmt((
AST.Case(startCase, NONE, {region=(CASEleft, startCaseright),
				       extraInfo=ref NONE,
				       typeInfo=NONE})
))
 in (LrTable.NT 58,(result,CASE1left,startCase1right),rest671) end
| (165,(_,(MlyValue.stmtmaybe stmtmaybe,_,stmtmayberight as 
stmtmaybe1right))::_::(_,(_,DEFAULTleft as DEFAULT1left,_))::rest671)
 => let val result=MlyValue.stmt((
AST.Case(NONE, stmtmaybe, {region=(DEFAULTleft, stmtmayberight),
				  extraInfo=ref NONE,
				  typeInfo=NONE})
))
 in (LrTable.NT 58,(result,DEFAULT1left,stmtmaybe1right),rest671) end
| (166,(_,(_,_,COLONright as COLON1right))::(_,(_,DEFAULTleft as 
DEFAULT1left,_))::rest671) => let val result=MlyValue.stmt((
AST.Case(NONE, NONE, {region=(DEFAULTleft, COLONright),
				  extraInfo=ref NONE,
				  typeInfo=NONE})
))
 in (LrTable.NT 58,(result,DEFAULT1left,COLON1right),rest671) end
| (167,(_,(_,_,SEMICOLONright as SEMICOLON1right))::(_,(
MlyValue.exprOpt exprOpt,exprOptleft as exprOpt1left,_))::rest671) => 
let val result=MlyValue.stmt((
AST.Expression(exprOpt, {region=(exprOptleft, SEMICOLONright),
				     extraInfo=ref NONE,
				     typeInfo=NONE})
))
 in (LrTable.NT 58,(result,exprOpt1left,SEMICOLON1right),rest671) end
| (168,(_,(MlyValue.stmt stmt,_,stmtright as stmt1right))::_::(_,(
MlyValue.expr expr,_,_))::_::(_,(_,IFleft as IF1left,_))::rest671) => 
let val result=MlyValue.stmt((
AST.IfThen(expr, stmt, {region=(IFleft, stmtright),
				    extraInfo=ref NONE,
				    typeInfo=NONE})
))
 in (LrTable.NT 58,(result,IF1left,stmt1right),rest671) end
| (169,(_,(MlyValue.stmt stmt2,_,stmt2right))::_::(_,(MlyValue.stmt 
stmt1,_,stmtright))::_::(_,(MlyValue.expr expr,_,_))::_::(_,(_,IFleft
 as IF1left,_))::rest671) => let val result=MlyValue.stmt((
AST.IfThenElse(expr, stmt1, stmt2, {region=(IFleft, stmtright),
						extraInfo=ref NONE,
						typeInfo=NONE})
))
 in (LrTable.NT 58,(result,IF1left,stmt2right),rest671) end
| (170,(_,(MlyValue.stmt stmt,_,stmtright as stmt1right))::(_,(
MlyValue.startSwitch startSwitch,_,_))::(_,(_,SWITCHleft as 
SWITCH1left,_))::rest671) => let val result=MlyValue.stmt((
AST.Switch(startSwitch, stmt, {region=(SWITCHleft, stmtright),
					   extraInfo=ref NONE,
					   typeInfo=NONE})
))
 in (LrTable.NT 58,(result,SWITCH1left,stmt1right),rest671) end
| (171,(_,(MlyValue.stmt stmt,_,stmtright as stmt1right))::(_,(
MlyValue.startWhile startWhile,_,_))::(_,(_,WHILEleft as WHILE1left,_)
)::rest671) => let val result=MlyValue.stmt((
AST.WhileDo(startWhile, stmt, {region=(WHILEleft, stmtright),
					   extraInfo=ref NONE,
					   typeInfo=NONE})
))
 in (LrTable.NT 58,(result,WHILE1left,stmt1right),rest671) end
| (172,(_,(_,_,SEMICOLONright as SEMICOLON1right))::_::(_,(
MlyValue.expr expr,_,_))::_::_::(_,(MlyValue.stmt stmt,_,_))::(_,(_,
startDoleft as startDo1left,_))::rest671) => let val result=
MlyValue.stmt((
AST.DoWhile(stmt, expr, {region=(startDoleft, SEMICOLONright),
				     extraInfo=ref NONE,
				     typeInfo=NONE})
))
 in (LrTable.NT 58,(result,startDo1left,SEMICOLON1right),rest671) end
| (173,(_,(MlyValue.stmt stmt,_,stmtright as stmt1right))::(_,(
MlyValue.startFor startFor,_,_))::(_,(_,FORleft as FOR1left,_))::
rest671) => let val result=MlyValue.stmt((
let
		val (e, e', e'') = startFor
	    in	
		AST.For(e, e', e'', stmt, {region=(FORleft, stmtright),
					   extraInfo=ref NONE,
					   typeInfo=NONE})
	    end
))
 in (LrTable.NT 58,(result,FOR1left,stmt1right),rest671) end
| (174,(_,(_,_,SEMICOLONright as SEMICOLON1right))::(_,(
MlyValue.IDENTIFIER IDENTIFIER,_,_))::(_,(_,GOTOleft as GOTO1left,_))
::rest671) => let val result=MlyValue.stmt((
AST.GoTo(IDENTIFIER, {region=(GOTOleft, SEMICOLONright),
				       extraInfo=ref NONE,
				       typeInfo=NONE})
))
 in (LrTable.NT 58,(result,GOTO1left,SEMICOLON1right),rest671) end
| (175,(_,(_,_,SEMICOLONright as SEMICOLON1right))::(_,(_,CONTINUEleft
 as CONTINUE1left,_))::rest671) => let val result=MlyValue.stmt((
AST.Continue {region=(CONTINUEleft, SEMICOLONright),
			       extraInfo=ref NONE,
			       typeInfo=NONE}
))
 in (LrTable.NT 58,(result,CONTINUE1left,SEMICOLON1right),rest671) end
| (176,(_,(_,_,SEMICOLONright as SEMICOLON1right))::(_,(_,BREAKleft
 as BREAK1left,_))::rest671) => let val result=MlyValue.stmt((
AST.Break {region=(BREAKleft, SEMICOLONright),
			    extraInfo=ref NONE,
			    typeInfo=NONE}
))
 in (LrTable.NT 58,(result,BREAK1left,SEMICOLON1right),rest671) end
| (177,(_,(_,_,SEMICOLONright as SEMICOLON1right))::(_,(
MlyValue.exprOpt exprOpt,_,_))::(_,(_,RETURNleft as RETURN1left,_))::
rest671) => let val result=MlyValue.stmt((
AST.Return(exprOpt, NONE, {region=(RETURNleft,
						    SEMICOLONright),
					    extraInfo=ref NONE,
					    typeInfo=NONE})
))
 in (LrTable.NT 58,(result,RETURN1left,SEMICOLON1right),rest671) end
| (178,(_,(_,_,COLON1right))::(_,(MlyValue.constantExpr constantExpr,
constantExpr1left,_))::rest671) => let val result=MlyValue.startCase((
SOME(constantExpr)))
 in (LrTable.NT 61,(result,constantExpr1left,COLON1right),rest671) end
| (179,(_,(_,_,RP1right))::(_,(MlyValue.expr expr,_,_))::(_,(_,LP1left
,_))::rest671) => let val result=MlyValue.startSwitch((expr))
 in (LrTable.NT 60,(result,LP1left,RP1right),rest671) end
| (180,(_,(_,_,RP1right))::(_,(MlyValue.expr expr,_,_))::(_,(_,LP1left
,_))::rest671) => let val result=MlyValue.startWhile((expr))
 in (LrTable.NT 63,(result,LP1left,RP1right),rest671) end
| (181,(_,(_,_,RP1right))::(_,(MlyValue.exprOpt exprOpt3,_,_))::_::(_,
(MlyValue.exprOpt exprOpt2,_,_))::_::(_,(MlyValue.exprOpt exprOpt1,_,_
))::(_,(_,LP1left,_))::rest671) => let val result=MlyValue.startFor((
exprOpt1, exprOpt2, exprOpt3))
 in (LrTable.NT 64,(result,LP1left,RP1right),rest671) end
| (182,(_,(_,DO1left,DO1right))::rest671) => let val result=
MlyValue.startDo((()))
 in (LrTable.NT 65,(result,DO1left,DO1right),rest671) end
| (183,rest671) => let val result=MlyValue.exprOpt((NONE))
 in (LrTable.NT 66,(result,defaultPos,defaultPos),rest671) end
| (184,(_,(MlyValue.expr expr,expr1left,expr1right))::rest671) => let 
val result=MlyValue.exprOpt((SOME(expr)))
 in (LrTable.NT 66,(result,expr1left,expr1right),rest671) end
| (185,(_,(MlyValue.assignExpr assignExpr,assignExpr1left,
assignExpr1right))::rest671) => let val result=MlyValue.expr((
assignExpr))
 in (LrTable.NT 67,(result,assignExpr1left,assignExpr1right),rest671)
 end
| (186,(_,(MlyValue.assignExpr assignExpr,_,assignExprright as 
assignExpr1right))::_::(_,(MlyValue.expr expr,exprleft as expr1left,_)
)::rest671) => let val result=MlyValue.expr((
AST.Comma(expr, assignExpr, {region=(exprleft, assignExprright),
				    extraInfo=ref NONE,
				    typeInfo=NONE})
))
 in (LrTable.NT 67,(result,expr1left,assignExpr1right),rest671) end
| (187,(_,(MlyValue.conditionalExpr conditionalExpr,
conditionalExpr1left,conditionalExpr1right))::rest671) => let val 
result=MlyValue.assignExpr((conditionalExpr))
 in (LrTable.NT 68,(result,conditionalExpr1left,conditionalExpr1right)
,rest671) end
| (188,(_,(MlyValue.assignExpr assignExpr,_,assignExprright as 
assignExpr1right))::(_,(MlyValue.ASSIGN ASSIGN,_,_))::(_,(
MlyValue.cast cast,castleft as cast1left,_))::rest671) => let val 
result=MlyValue.assignExpr((
AST.Assignment(cast, ASSIGN, assignExpr,
		      {region=(castleft, assignExprright),
		       extraInfo=ref NONE,
		       typeInfo=NONE})
))
 in (LrTable.NT 68,(result,cast1left,assignExpr1right),rest671) end
| (189,(_,(MlyValue.assignExpr assignExpr,_,assignExpr1right))::(_,(_,
EXTENSION1left,_))::rest671) => let val result=MlyValue.assignExpr((
assignExpr))
 in (LrTable.NT 68,(result,EXTENSION1left,assignExpr1right),rest671)
 end
| (190,(_,(MlyValue.binaryExpr binaryExpr,binaryExpr1left,
binaryExpr1right))::rest671) => let val result=
MlyValue.conditionalExpr((binaryExpr))
 in (LrTable.NT 69,(result,binaryExpr1left,binaryExpr1right),rest671)
 end
| (191,(_,(MlyValue.conditionalExpr conditionalExpr,_,
conditionalExprright as conditionalExpr1right))::_::(_,(MlyValue.expr 
expr,_,_))::_::(_,(MlyValue.binaryExpr binaryExpr,binaryExprleft as 
binaryExpr1left,_))::rest671) => let val result=
MlyValue.conditionalExpr((
AST.Conditional(binaryExpr, expr, conditionalExpr,
		       {region=(binaryExprleft, conditionalExprright),
			extraInfo=ref NONE,
			typeInfo=NONE})
))
 in (LrTable.NT 69,(result,binaryExpr1left,conditionalExpr1right),
rest671) end
| (192,(_,(MlyValue.conditionalExpr conditionalExpr,
conditionalExpr1left,conditionalExpr1right))::rest671) => let val 
result=MlyValue.constantExpr((conditionalExpr))
 in (LrTable.NT 70,(result,conditionalExpr1left,conditionalExpr1right)
,rest671) end
| (193,(_,(MlyValue.cast cast,cast1left,cast1right))::rest671) => let 
val result=MlyValue.binaryExpr((cast))
 in (LrTable.NT 71,(result,cast1left,cast1right),rest671) end
| (194,(_,(MlyValue.binaryExpr binaryExpr2,_,binaryExpr2right))::_::(_
,(MlyValue.binaryExpr binaryExpr1,binaryExpr1left,_))::rest671) => 
let val result=MlyValue.binaryExpr((
AST.Operation(AST.Or,[binaryExpr1, binaryExpr2],
		     {region=(binaryExpr1left, binaryExpr2right),
		      extraInfo=ref NONE,
		      typeInfo=NONE})
))
 in (LrTable.NT 71,(result,binaryExpr1left,binaryExpr2right),rest671)
 end
| (195,(_,(MlyValue.binaryExpr binaryExpr2,_,binaryExpr2right))::_::(_
,(MlyValue.binaryExpr binaryExpr1,binaryExpr1left,_))::rest671) => 
let val result=MlyValue.binaryExpr((
AST.Operation(AST.And, [binaryExpr1, binaryExpr2],
		     {region=(binaryExpr1left, binaryExpr2right),
		      extraInfo=ref NONE,
		      typeInfo=NONE})
))
 in (LrTable.NT 71,(result,binaryExpr1left,binaryExpr2right),rest671)
 end
| (196,(_,(MlyValue.binaryExpr binaryExpr2,_,binaryExpr2right))::_::(_
,(MlyValue.binaryExpr binaryExpr1,binaryExpr1left,_))::rest671) => 
let val result=MlyValue.binaryExpr((
AST.Operation(AST.BitOr, [binaryExpr1, binaryExpr2],
		     {region=(binaryExpr1left, binaryExpr2right),
		      extraInfo=ref NONE,
		      typeInfo=NONE})
))
 in (LrTable.NT 71,(result,binaryExpr1left,binaryExpr2right),rest671)
 end
| (197,(_,(MlyValue.binaryExpr binaryExpr2,_,binaryExpr2right))::_::(_
,(MlyValue.binaryExpr binaryExpr1,binaryExpr1left,_))::rest671) => 
let val result=MlyValue.binaryExpr((
AST.Operation(AST.BitXor, [binaryExpr1, binaryExpr2],
		     {region=(binaryExpr1left, binaryExpr2right),
		      extraInfo=ref NONE,
		      typeInfo=NONE})
))
 in (LrTable.NT 71,(result,binaryExpr1left,binaryExpr2right),rest671)
 end
| (198,(_,(MlyValue.binaryExpr binaryExpr2,_,binaryExpr2right))::_::(_
,(MlyValue.binaryExpr binaryExpr1,binaryExpr1left,_))::rest671) => 
let val result=MlyValue.binaryExpr((
AST.Operation(AST.BitAnd, [binaryExpr1, binaryExpr2],
		     {region=(binaryExpr1left, binaryExpr2right),
		      extraInfo=ref NONE,
		      typeInfo=NONE})
))
 in (LrTable.NT 71,(result,binaryExpr1left,binaryExpr2right),rest671)
 end
| (199,(_,(MlyValue.binaryExpr binaryExpr2,_,binaryExpr2right))::_::(_
,(MlyValue.binaryExpr binaryExpr1,binaryExpr1left,_))::rest671) => 
let val result=MlyValue.binaryExpr((
AST.Operation(AST.Eq, [binaryExpr1, binaryExpr2],
		     {region=(binaryExpr1left, binaryExpr2right),
		      extraInfo=ref NONE,
		      typeInfo=NONE})
))
 in (LrTable.NT 71,(result,binaryExpr1left,binaryExpr2right),rest671)
 end
| (200,(_,(MlyValue.binaryExpr binaryExpr2,_,binaryExpr2right))::_::(_
,(MlyValue.binaryExpr binaryExpr1,binaryExpr1left,_))::rest671) => 
let val result=MlyValue.binaryExpr((
AST.Operation(AST.Ne, [binaryExpr1, binaryExpr2],
		     {region=(binaryExpr1left, binaryExpr2right),
		      extraInfo=ref NONE,
		      typeInfo=NONE})
))
 in (LrTable.NT 71,(result,binaryExpr1left,binaryExpr2right),rest671)
 end
| (201,(_,(MlyValue.binaryExpr binaryExpr2,_,binaryExpr2right))::_::(_
,(MlyValue.binaryExpr binaryExpr1,binaryExpr1left,_))::rest671) => 
let val result=MlyValue.binaryExpr((
AST.Operation(AST.Gt, [binaryExpr1, binaryExpr2],
		     {region=(binaryExpr1left, binaryExpr2right),
		      extraInfo=ref NONE,
		      typeInfo=NONE})
))
 in (LrTable.NT 71,(result,binaryExpr1left,binaryExpr2right),rest671)
 end
| (202,(_,(MlyValue.binaryExpr binaryExpr2,_,binaryExpr2right))::_::(_
,(MlyValue.binaryExpr binaryExpr1,binaryExpr1left,_))::rest671) => 
let val result=MlyValue.binaryExpr((
AST.Operation(AST.Lt, [binaryExpr1, binaryExpr2],
		     {region=(binaryExpr1left, binaryExpr2right),
		      extraInfo=ref NONE,
		      typeInfo=NONE})
))
 in (LrTable.NT 71,(result,binaryExpr1left,binaryExpr2right),rest671)
 end
| (203,(_,(MlyValue.binaryExpr binaryExpr2,_,binaryExpr2right))::_::(_
,(MlyValue.binaryExpr binaryExpr1,binaryExpr1left,_))::rest671) => 
let val result=MlyValue.binaryExpr((
AST.Operation(AST.Le, [binaryExpr1, binaryExpr2],
		     {region=(binaryExpr1left, binaryExpr2right),
		      extraInfo=ref NONE,
		      typeInfo=NONE})
))
 in (LrTable.NT 71,(result,binaryExpr1left,binaryExpr2right),rest671)
 end
| (204,(_,(MlyValue.binaryExpr binaryExpr2,_,binaryExpr2right))::_::(_
,(MlyValue.binaryExpr binaryExpr1,binaryExpr1left,_))::rest671) => 
let val result=MlyValue.binaryExpr((
AST.Operation(AST.Ge, [binaryExpr1, binaryExpr2],
		     {region=(binaryExpr1left, binaryExpr2right),
		      extraInfo=ref NONE,
		      typeInfo=NONE})
))
 in (LrTable.NT 71,(result,binaryExpr1left,binaryExpr2right),rest671)
 end
| (205,(_,(MlyValue.binaryExpr binaryExpr2,_,binaryExpr2right))::_::(_
,(MlyValue.binaryExpr binaryExpr1,binaryExpr1left,_))::rest671) => 
let val result=MlyValue.binaryExpr((
AST.Operation(AST.Rshift, [binaryExpr1, binaryExpr2],
		     {region=(binaryExpr1left, binaryExpr2right),
		      extraInfo=ref NONE,
		      typeInfo=NONE})
))
 in (LrTable.NT 71,(result,binaryExpr1left,binaryExpr2right),rest671)
 end
| (206,(_,(MlyValue.binaryExpr binaryExpr2,_,binaryExpr2right))::_::(_
,(MlyValue.binaryExpr binaryExpr1,binaryExpr1left,_))::rest671) => 
let val result=MlyValue.binaryExpr((
AST.Operation(AST.Lshift, [binaryExpr1, binaryExpr2],
		     {region=(binaryExpr1left, binaryExpr2right),
		      extraInfo=ref NONE,
		      typeInfo=NONE})
))
 in (LrTable.NT 71,(result,binaryExpr1left,binaryExpr2right),rest671)
 end
| (207,(_,(MlyValue.binaryExpr binaryExpr2,_,binaryExpr2right))::_::(_
,(MlyValue.binaryExpr binaryExpr1,binaryExpr1left,_))::rest671) => 
let val result=MlyValue.binaryExpr((
AST.Operation(AST.Minus, [binaryExpr1, binaryExpr2],
		     {region=(binaryExpr1left, binaryExpr2right),
		      extraInfo=ref NONE,
		      typeInfo=NONE})
))
 in (LrTable.NT 71,(result,binaryExpr1left,binaryExpr2right),rest671)
 end
| (208,(_,(MlyValue.binaryExpr binaryExpr2,_,binaryExpr2right))::_::(_
,(MlyValue.binaryExpr binaryExpr1,binaryExpr1left,_))::rest671) => 
let val result=MlyValue.binaryExpr((
AST.Operation(AST.Plus, [binaryExpr1, binaryExpr2],
		     {region=(binaryExpr1left, binaryExpr2right),
		      extraInfo=ref NONE,
		      typeInfo=NONE})
))
 in (LrTable.NT 71,(result,binaryExpr1left,binaryExpr2right),rest671)
 end
| (209,(_,(MlyValue.binaryExpr binaryExpr2,_,binaryExpr2right))::_::(_
,(MlyValue.binaryExpr binaryExpr1,binaryExpr1left,_))::rest671) => 
let val result=MlyValue.binaryExpr((
AST.Operation(AST.Mul, [binaryExpr1, binaryExpr2],
		     {region=(binaryExpr1left, binaryExpr2right),
		      extraInfo=ref NONE,
		      typeInfo=NONE})
))
 in (LrTable.NT 71,(result,binaryExpr1left,binaryExpr2right),rest671)
 end
| (210,(_,(MlyValue.binaryExpr binaryExpr2,_,binaryExpr2right))::_::(_
,(MlyValue.binaryExpr binaryExpr1,binaryExpr1left,_))::rest671) => 
let val result=MlyValue.binaryExpr((
AST.Operation(AST.Div, [binaryExpr1, binaryExpr2],
		     {region=(binaryExpr1left, binaryExpr2right),
		      extraInfo=ref NONE,
		      typeInfo=NONE})
))
 in (LrTable.NT 71,(result,binaryExpr1left,binaryExpr2right),rest671)
 end
| (211,(_,(MlyValue.binaryExpr binaryExpr2,_,binaryExpr2right))::_::(_
,(MlyValue.binaryExpr binaryExpr1,binaryExpr1left,_))::rest671) => 
let val result=MlyValue.binaryExpr((
AST.Operation(AST.Mod, [binaryExpr1, binaryExpr2],
		     {region=(binaryExpr1left, binaryExpr2right),
		      extraInfo=ref NONE,
		      typeInfo=NONE})
))
 in (LrTable.NT 71,(result,binaryExpr1left,binaryExpr2right),rest671)
 end
| (212,(_,(MlyValue.unaryExpr unaryExpr,unaryExpr1left,unaryExpr1right
))::rest671) => let val result=MlyValue.cast((unaryExpr))
 in (LrTable.NT 72,(result,unaryExpr1left,unaryExpr1right),rest671)
 end
| (213,(_,(MlyValue.cast cast,_,castright as cast1right))::_::(_,(
MlyValue.typeName typeName,_,_))::(_,(_,LPleft as LP1left,_))::rest671
) => let val result=MlyValue.cast((
AST.Cast(typeName, cast, {region=(LPleft, castright),
				 extraInfo=ref NONE,
				 typeInfo=NONE})
))
 in (LrTable.NT 72,(result,LP1left,cast1right),rest671) end
| (214,(_,(_,_,RCright as RC1right))::(_,(
MlyValue.braceInitializerList braceInitializerList,_,_))::_::_::(_,(
MlyValue.typeName typeName,_,_))::(_,(_,LPleft as LP1left,_))::rest671
) => let val result=MlyValue.cast((
AST.CastList(typeName, braceInitializerList, 
				{region=(LPleft, RCright),
				 extraInfo=ref NONE,
				 typeInfo=NONE})
))
 in (LrTable.NT 72,(result,LP1left,RC1right),rest671) end
| (215,(_,(_,_,RCright as RC1right))::_::(_,(
MlyValue.braceInitializerList braceInitializerList,_,_))::_::_::(_,(
MlyValue.typeName typeName,_,_))::(_,(_,LPleft as LP1left,_))::rest671
) => let val result=MlyValue.cast((
AST.CastList(typeName, braceInitializerList, 
				{region=(LPleft, RCright),
				 extraInfo=ref NONE,
				 typeInfo=NONE})
))
 in (LrTable.NT 72,(result,LP1left,RC1right),rest671) end
| (216,(_,(MlyValue.postfix postfix,postfix1left,postfix1right))::
rest671) => let val result=MlyValue.unaryExpr((postfix))
 in (LrTable.NT 73,(result,postfix1left,postfix1right),rest671) end
| (217,(_,(MlyValue.unaryExpr unaryExpr,_,unaryExprright as 
unaryExpr1right))::(_,(_,INCRleft as INCR1left,_))::rest671) => let 
val result=MlyValue.unaryExpr((
AST.Operation(AST.Incr, [unaryExpr],
		     {region=(INCRleft, unaryExprright),
		      extraInfo=ref NONE,
		      typeInfo=NONE})
))
 in (LrTable.NT 73,(result,INCR1left,unaryExpr1right),rest671) end
| (218,(_,(MlyValue.unaryExpr unaryExpr,_,unaryExprright as 
unaryExpr1right))::(_,(_,DECRleft as DECR1left,_))::rest671) => let 
val result=MlyValue.unaryExpr((
AST.Operation(AST.Decr, [unaryExpr],
		     {region=(DECRleft, unaryExprright),
		      extraInfo=ref NONE,
		      typeInfo=NONE})
))
 in (LrTable.NT 73,(result,DECR1left,unaryExpr1right),rest671) end
| (219,(_,(MlyValue.cast cast,_,castright as cast1right))::(_,(
MlyValue.unaryOp unaryOp,unaryOpleft as unaryOp1left,_))::rest671) => 
let val result=MlyValue.unaryExpr((
AST.Operation(unaryOp, [cast], {region=(unaryOpleft, castright),
				       extraInfo=ref NONE,
				       typeInfo=NONE})
))
 in (LrTable.NT 73,(result,unaryOp1left,cast1right),rest671) end
| (220,(_,(MlyValue.unaryExpr unaryExpr,_,unaryExprright as 
unaryExpr1right))::(_,(_,SIZEOFleft as SIZEOF1left,_))::rest671) => 
let val result=MlyValue.unaryExpr((
AST.SizeOfExpr(unaryExpr, {region=(SIZEOFleft, unaryExprright),
				  extraInfo=ref NONE,
				  typeInfo=NONE})
))
 in (LrTable.NT 73,(result,SIZEOF1left,unaryExpr1right),rest671) end
| (221,(_,(_,_,RPright as RP1right))::(_,(MlyValue.typeName typeName,_
,_))::_::(_,(_,SIZEOFleft as SIZEOF1left,_))::rest671) => let val 
result=MlyValue.unaryExpr((
AST.SizeOfTypeExpr(typeName, {region=(SIZEOFleft, RPright),
				     extraInfo=ref NONE,
				     typeInfo=NONE})
))
 in (LrTable.NT 73,(result,SIZEOF1left,RP1right),rest671) end
| (222,(_,(MlyValue.unaryExpr unaryExpr,_,unaryExprright as 
unaryExpr1right))::(_,(_,ALIGNOFleft as ALIGNOF1left,_))::rest671) => 
let val result=MlyValue.unaryExpr((
AST.AlignOfExpr(unaryExpr, {region=(ALIGNOFleft, unaryExprright),
				  extraInfo=ref NONE,
				  typeInfo=NONE})
))
 in (LrTable.NT 73,(result,ALIGNOF1left,unaryExpr1right),rest671) end
| (223,(_,(_,_,RPright as RP1right))::(_,(MlyValue.typeName typeName,_
,_))::_::(_,(_,ALIGNOFleft as ALIGNOF1left,_))::rest671) => let val 
result=MlyValue.unaryExpr((
AST.AlignOfTypeExpr(typeName, {region=(ALIGNOFleft, RPright),
				     extraInfo=ref NONE,
				     typeInfo=NONE})
))
 in (LrTable.NT 73,(result,ALIGNOF1left,RP1right),rest671) end
| (224,(_,(_,AND1left,AND1right))::rest671) => let val result=
MlyValue.unaryOp((AST.LabelAddress))
 in (LrTable.NT 74,(result,AND1left,AND1right),rest671) end
| (225,(_,(_,BIT_AND1left,BIT_AND1right))::rest671) => let val result=
MlyValue.unaryOp((AST.Address))
 in (LrTable.NT 74,(result,BIT_AND1left,BIT_AND1right),rest671) end
| (226,(_,(_,MUL1left,MUL1right))::rest671) => let val result=
MlyValue.unaryOp((AST.Deref))
 in (LrTable.NT 74,(result,MUL1left,MUL1right),rest671) end
| (227,(_,(_,PLUS1left,PLUS1right))::rest671) => let val result=
MlyValue.unaryOp((AST.Plus))
 in (LrTable.NT 74,(result,PLUS1left,PLUS1right),rest671) end
| (228,(_,(_,MINUS1left,MINUS1right))::rest671) => let val result=
MlyValue.unaryOp((AST.Minus))
 in (LrTable.NT 74,(result,MINUS1left,MINUS1right),rest671) end
| (229,(_,(_,BIT_NOT1left,BIT_NOT1right))::rest671) => let val result=
MlyValue.unaryOp((AST.BitNot))
 in (LrTable.NT 74,(result,BIT_NOT1left,BIT_NOT1right),rest671) end
| (230,(_,(_,NOT1left,NOT1right))::rest671) => let val result=
MlyValue.unaryOp((AST.Not))
 in (LrTable.NT 74,(result,NOT1left,NOT1right),rest671) end
| (231,(_,(MlyValue.primary primary,primary1left,primary1right))::
rest671) => let val result=MlyValue.postfix((primary))
 in (LrTable.NT 75,(result,primary1left,primary1right),rest671) end
| (232,(_,(_,_,RBright as RB1right))::(_,(MlyValue.expr expr,_,_))::_
::(_,(MlyValue.postfix postfix,postfixleft as postfix1left,_))::
rest671) => let val result=MlyValue.postfix((
AST.Index(postfix, expr, {region=(postfixleft, RBright),
				 extraInfo=ref NONE,
				 typeInfo=NONE})
))
 in (LrTable.NT 75,(result,postfix1left,RB1right),rest671) end
| (233,(_,(MlyValue.args args,_,args1right))::(_,(MlyValue.postfix 
postfix,postfixleft as postfix1left,postfixright))::rest671) => let 
val result=MlyValue.postfix((
AST.Application(postfix, args, {region=(postfixleft, postfixright),
				       extraInfo=ref NONE,
				       typeInfo=NONE},
		       AST.mkFreshLab())
))
 in (LrTable.NT 75,(result,postfix1left,args1right),rest671) end
| (234,(_,(MlyValue.field field,fieldleft,fieldright as field1right))
::_::(_,(MlyValue.postfix postfix,postfix1left,_))::rest671) => let 
val result=MlyValue.postfix((
AST.FieldDirect(postfix, field, {region=(fieldleft, fieldright),
					extraInfo=ref NONE,
					typeInfo=NONE})
))
 in (LrTable.NT 75,(result,postfix1left,field1right),rest671) end
| (235,(_,(MlyValue.field field,fieldleft,fieldright as field1right))
::_::(_,(MlyValue.postfix postfix,postfix1left,_))::rest671) => let 
val result=MlyValue.postfix((
AST.FieldIndirect(postfix, field, {region=(fieldleft, fieldright),
					  extraInfo=ref NONE,
					  typeInfo=NONE})
))
 in (LrTable.NT 75,(result,postfix1left,field1right),rest671) end
| (236,(_,(_,_,INCRright as INCR1right))::(_,(MlyValue.postfix postfix
,postfixleft as postfix1left,_))::rest671) => let val result=
MlyValue.postfix((
AST.Postfix(AST.Incr, postfix, {region=(postfixleft, INCRright),
				       extraInfo=ref NONE,
				       typeInfo=NONE})
))
 in (LrTable.NT 75,(result,postfix1left,INCR1right),rest671) end
| (237,(_,(_,_,DECRright as DECR1right))::(_,(MlyValue.postfix postfix
,postfixleft as postfix1left,_))::rest671) => let val result=
MlyValue.postfix((
AST.Postfix(AST.Decr, postfix, {region=(postfixleft, DECRright),
				       extraInfo=ref NONE,
				       typeInfo=NONE})
))
 in (LrTable.NT 75,(result,postfix1left,DECR1right),rest671) end
| (238,(_,(MlyValue.IDENTIFIER IDENTIFIER,IDENTIFIER1left,
IDENTIFIER1right))::rest671) => let val result=MlyValue.field((
IDENTIFIER))
 in (LrTable.NT 33,(result,IDENTIFIER1left,IDENTIFIER1right),rest671)
 end
| (239,(_,(MlyValue.TYPEDEF_NAME TYPEDEF_NAME,TYPEDEF_NAME1left,
TYPEDEF_NAME1right))::rest671) => let val result=MlyValue.field((
TYPEDEF_NAME))
 in (LrTable.NT 33,(result,TYPEDEF_NAME1left,TYPEDEF_NAME1right),
rest671) end
| (240,(_,(MlyValue.IDENTIFIER IDENTIFIER,IDENTIFIERleft as 
IDENTIFIER1left,IDENTIFIERright as IDENTIFIER1right))::rest671) => 
let val result=MlyValue.primary((
AST.Variable(IDENTIFIER, {region=(IDENTIFIERleft, IDENTIFIERright),
				 extraInfo=ref NONE,
				 typeInfo=NONE})
))
 in (LrTable.NT 76,(result,IDENTIFIER1left,IDENTIFIER1right),rest671)
 end
| (241,(_,(MlyValue.constant constant,constantleft as constant1left,
constantright as constant1right))::rest671) => let val result=
MlyValue.primary((
AST.Constant(constant, {region=(constantleft, constantright),
			       extraInfo=ref NONE,
			       typeInfo=NONE})
))
 in (LrTable.NT 76,(result,constant1left,constant1right),rest671) end
| (242,(_,(_,_,RP1right))::(_,(MlyValue.expr expr,_,_))::(_,(_,LP1left
,_))::rest671) => let val result=MlyValue.primary((expr))
 in (LrTable.NT 76,(result,LP1left,RP1right),rest671) end
| (243,(_,(_,_,RPright as RP1right))::(_,(MlyValue.compoundStmt 
compoundStmt,compoundStmtleft,compoundStmtright))::(_,(_,LPleft as 
LP1left,_))::rest671) => let val result=MlyValue.primary((
let val (dl, sl, scop) = compoundStmt
					 in
					     AST.Extension
					     (AST.Compound {decls = dl,
							    body = sl,
							    region = (compoundStmtleft, compoundStmtright),
							    extraInfo = ref NONE,
							    lookup = SymbolTable.scopeLookup scop},
					      {region=(LPleft, RPright),
					       extraInfo=ref NONE,
					       typeInfo=NONE})
					 end
				handle _ => (error (LPleft, RPright) "Bad constant declaration";
					     raise Fail "Bad constant")
))
 in (LrTable.NT 76,(result,LP1left,RP1right),rest671) end
| (244,(_,(MlyValue.CONSTANT CONSTANT,CONSTANT1left,CONSTANT1right))::
rest671) => let val result=MlyValue.constant((CONSTANT))
 in (LrTable.NT 34,(result,CONSTANT1left,CONSTANT1right),rest671) end
| (245,(_,(MlyValue.stringconstant stringconstant,stringconstant1left,
stringconstant1right))::rest671) => let val result=MlyValue.constant((
AST.String (AST.mkIdentifier stringconstant)))
 in (LrTable.NT 34,(result,stringconstant1left,stringconstant1right),
rest671) end
| (246,(_,(MlyValue.string string,string1left,string1right))::rest671)
 => let val result=MlyValue.stringconstant((string))
 in (LrTable.NT 35,(result,string1left,string1right),rest671) end
| (247,(_,(MlyValue.stringconstant stringconstant,_,
stringconstant1right))::(_,(MlyValue.string string,string1left,_))::
rest671) => let val result=MlyValue.stringconstant((
concat [string, stringconstant]))
 in (LrTable.NT 35,(result,string1left,stringconstant1right),rest671)
 end
| (248,(_,(MlyValue.STRINGCONSTANT STRINGCONSTANT,STRINGCONSTANT1left,
STRINGCONSTANT1right))::rest671) => let val result=MlyValue.string((
STRINGCONSTANT))
 in (LrTable.NT 36,(result,STRINGCONSTANT1left,STRINGCONSTANT1right),
rest671) end
| (249,(_,(_,FUNCTION1left,FUNCTION1right))::rest671) => let val 
result=MlyValue.string(("__FUNCTION__"))
 in (LrTable.NT 36,(result,FUNCTION1left,FUNCTION1right),rest671) end
| (250,(_,(_,_,RP1right))::(_,(_,LP1left,_))::rest671) => let val 
result=MlyValue.args((nil))
 in (LrTable.NT 77,(result,LP1left,RP1right),rest671) end
| (251,(_,(_,_,RP1right))::(_,(MlyValue.argList argList,_,_))::(_,(_,
LP1left,_))::rest671) => let val result=MlyValue.args((rev argList))
 in (LrTable.NT 77,(result,LP1left,RP1right),rest671) end
| (252,(_,(MlyValue.assignExpr assignExpr,assignExpr1left,
assignExpr1right))::rest671) => let val result=MlyValue.argList((
[assignExpr]))
 in (LrTable.NT 78,(result,assignExpr1left,assignExpr1right),rest671)
 end
| (253,(_,(MlyValue.assignExpr assignExpr,_,assignExpr1right))::_::(_,
(MlyValue.argList argList,argList1left,_))::rest671) => let val result
=MlyValue.argList((assignExpr :: argList))
 in (LrTable.NT 78,(result,argList1left,assignExpr1right),rest671) end
| (254,rest671) => let val result=MlyValue.pushScope((
SymbolTable.pushScope ()))
 in (LrTable.NT 79,(result,defaultPos,defaultPos),rest671) end
| (255,rest671) => let val result=MlyValue.recoverScope((
SymbolTable.recoverParameterScopeOnNextPush ()))
 in (LrTable.NT 80,(result,defaultPos,defaultPos),rest671) end
| (256,rest671) => let val result=MlyValue.pushStructScope((
SymbolTable.pushStructScope ()))
 in (LrTable.NT 81,(result,defaultPos,defaultPos),rest671) end
| (257,rest671) => let val result=MlyValue.popScope((
SymbolTable.popScope ()))
 in (LrTable.NT 82,(result,defaultPos,defaultPos),rest671) end
| (258,rest671) => let val result=MlyValue.getScope((
SymbolTable.getScope ()))
 in (LrTable.NT 83,(result,defaultPos,defaultPos),rest671) end
| _ => raise (mlyAction i392)
end
val void = MlyValue.VOID'
val extract = fn a => (fn MlyValue.start x => x
| _ => let exception ParseInternal
	in raise ParseInternal end) a 
end
end
structure Tokens : Gen_TOKENS =
struct
type svalue = ParserData.svalue
type ('a,'b) token = ('a,'b) Token.token
fun PLUS (p1,p2) = Token.TOKEN (ParserData.LrTable.T 0,(
ParserData.MlyValue.VOID',p1,p2))
fun MINUS (p1,p2) = Token.TOKEN (ParserData.LrTable.T 1,(
ParserData.MlyValue.VOID',p1,p2))
fun MUL (p1,p2) = Token.TOKEN (ParserData.LrTable.T 2,(
ParserData.MlyValue.VOID',p1,p2))
fun DIV (p1,p2) = Token.TOKEN (ParserData.LrTable.T 3,(
ParserData.MlyValue.VOID',p1,p2))
fun MOD (p1,p2) = Token.TOKEN (ParserData.LrTable.T 4,(
ParserData.MlyValue.VOID',p1,p2))
fun INCR (p1,p2) = Token.TOKEN (ParserData.LrTable.T 5,(
ParserData.MlyValue.VOID',p1,p2))
fun DECR (p1,p2) = Token.TOKEN (ParserData.LrTable.T 6,(
ParserData.MlyValue.VOID',p1,p2))
fun AND (p1,p2) = Token.TOKEN (ParserData.LrTable.T 7,(
ParserData.MlyValue.VOID',p1,p2))
fun OR (p1,p2) = Token.TOKEN (ParserData.LrTable.T 8,(
ParserData.MlyValue.VOID',p1,p2))
fun NOT (p1,p2) = Token.TOKEN (ParserData.LrTable.T 9,(
ParserData.MlyValue.VOID',p1,p2))
fun BIT_AND (p1,p2) = Token.TOKEN (ParserData.LrTable.T 10,(
ParserData.MlyValue.VOID',p1,p2))
fun BIT_OR (p1,p2) = Token.TOKEN (ParserData.LrTable.T 11,(
ParserData.MlyValue.VOID',p1,p2))
fun BIT_NOT (p1,p2) = Token.TOKEN (ParserData.LrTable.T 12,(
ParserData.MlyValue.VOID',p1,p2))
fun BIT_XOR (p1,p2) = Token.TOKEN (ParserData.LrTable.T 13,(
ParserData.MlyValue.VOID',p1,p2))
fun RSHIFT (p1,p2) = Token.TOKEN (ParserData.LrTable.T 14,(
ParserData.MlyValue.VOID',p1,p2))
fun LSHIFT (p1,p2) = Token.TOKEN (ParserData.LrTable.T 15,(
ParserData.MlyValue.VOID',p1,p2))
fun LT (p1,p2) = Token.TOKEN (ParserData.LrTable.T 16,(
ParserData.MlyValue.VOID',p1,p2))
fun LE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 17,(
ParserData.MlyValue.VOID',p1,p2))
fun GT (p1,p2) = Token.TOKEN (ParserData.LrTable.T 18,(
ParserData.MlyValue.VOID',p1,p2))
fun GE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 19,(
ParserData.MlyValue.VOID',p1,p2))
fun EQ (p1,p2) = Token.TOKEN (ParserData.LrTable.T 20,(
ParserData.MlyValue.VOID',p1,p2))
fun NE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 21,(
ParserData.MlyValue.VOID',p1,p2))
fun ASSIGN (i,p1,p2) = Token.TOKEN (ParserData.LrTable.T 22,(
ParserData.MlyValue.ASSIGN i,p1,p2))
fun DOT (p1,p2) = Token.TOKEN (ParserData.LrTable.T 23,(
ParserData.MlyValue.VOID',p1,p2))
fun PTR (p1,p2) = Token.TOKEN (ParserData.LrTable.T 24,(
ParserData.MlyValue.VOID',p1,p2))
fun COMMA (p1,p2) = Token.TOKEN (ParserData.LrTable.T 25,(
ParserData.MlyValue.VOID',p1,p2))
fun QUESTION (p1,p2) = Token.TOKEN (ParserData.LrTable.T 26,(
ParserData.MlyValue.VOID',p1,p2))
fun COLON (p1,p2) = Token.TOKEN (ParserData.LrTable.T 27,(
ParserData.MlyValue.VOID',p1,p2))
fun LP (p1,p2) = Token.TOKEN (ParserData.LrTable.T 28,(
ParserData.MlyValue.VOID',p1,p2))
fun RP (p1,p2) = Token.TOKEN (ParserData.LrTable.T 29,(
ParserData.MlyValue.VOID',p1,p2))
fun LB (p1,p2) = Token.TOKEN (ParserData.LrTable.T 30,(
ParserData.MlyValue.VOID',p1,p2))
fun RB (p1,p2) = Token.TOKEN (ParserData.LrTable.T 31,(
ParserData.MlyValue.VOID',p1,p2))
fun LC (p1,p2) = Token.TOKEN (ParserData.LrTable.T 32,(
ParserData.MlyValue.VOID',p1,p2))
fun RC (p1,p2) = Token.TOKEN (ParserData.LrTable.T 33,(
ParserData.MlyValue.VOID',p1,p2))
fun IDENTIFIER (i,p1,p2) = Token.TOKEN (ParserData.LrTable.T 34,(
ParserData.MlyValue.IDENTIFIER i,p1,p2))
fun TYPEDEF_NAME (i,p1,p2) = Token.TOKEN (ParserData.LrTable.T 35,(
ParserData.MlyValue.TYPEDEF_NAME i,p1,p2))
fun TYPE_DEC_SPEC (p1,p2) = Token.TOKEN (ParserData.LrTable.T 36,(
ParserData.MlyValue.VOID',p1,p2))
fun HIGHEST (p1,p2) = Token.TOKEN (ParserData.LrTable.T 37,(
ParserData.MlyValue.VOID',p1,p2))
fun ATTRIBUTE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 38,(
ParserData.MlyValue.VOID',p1,p2))
fun EXTENSION (p1,p2) = Token.TOKEN (ParserData.LrTable.T 39,(
ParserData.MlyValue.VOID',p1,p2))
fun FUNCTION (p1,p2) = Token.TOKEN (ParserData.LrTable.T 40,(
ParserData.MlyValue.VOID',p1,p2))
fun CONSTANT (i,p1,p2) = Token.TOKEN (ParserData.LrTable.T 41,(
ParserData.MlyValue.CONSTANT i,p1,p2))
fun STRINGCONSTANT (i,p1,p2) = Token.TOKEN (ParserData.LrTable.T 42,(
ParserData.MlyValue.STRINGCONSTANT i,p1,p2))
fun POUND (i,p1,p2) = Token.TOKEN (ParserData.LrTable.T 43,(
ParserData.MlyValue.POUND i,p1,p2))
fun INLINE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 44,(
ParserData.MlyValue.VOID',p1,p2))
fun AUTO (p1,p2) = Token.TOKEN (ParserData.LrTable.T 45,(
ParserData.MlyValue.VOID',p1,p2))
fun REGISTER (p1,p2) = Token.TOKEN (ParserData.LrTable.T 46,(
ParserData.MlyValue.VOID',p1,p2))
fun STATIC (p1,p2) = Token.TOKEN (ParserData.LrTable.T 47,(
ParserData.MlyValue.VOID',p1,p2))
fun EXTERN (p1,p2) = Token.TOKEN (ParserData.LrTable.T 48,(
ParserData.MlyValue.VOID',p1,p2))
fun CONST (p1,p2) = Token.TOKEN (ParserData.LrTable.T 49,(
ParserData.MlyValue.VOID',p1,p2))
fun QUALIFIER (i,p1,p2) = Token.TOKEN (ParserData.LrTable.T 50,(
ParserData.MlyValue.QUALIFIER i,p1,p2))
fun VOLATILE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 51,(
ParserData.MlyValue.VOID',p1,p2))
fun NONCONST (p1,p2) = Token.TOKEN (ParserData.LrTable.T 52,(
ParserData.MlyValue.VOID',p1,p2))
fun SIGNED (p1,p2) = Token.TOKEN (ParserData.LrTable.T 53,(
ParserData.MlyValue.VOID',p1,p2))
fun UNSIGNED (p1,p2) = Token.TOKEN (ParserData.LrTable.T 54,(
ParserData.MlyValue.VOID',p1,p2))
fun SHORT (p1,p2) = Token.TOKEN (ParserData.LrTable.T 55,(
ParserData.MlyValue.VOID',p1,p2))
fun LONG (p1,p2) = Token.TOKEN (ParserData.LrTable.T 56,(
ParserData.MlyValue.VOID',p1,p2))
fun VOID (p1,p2) = Token.TOKEN (ParserData.LrTable.T 57,(
ParserData.MlyValue.VOID',p1,p2))
fun CHAR (p1,p2) = Token.TOKEN (ParserData.LrTable.T 58,(
ParserData.MlyValue.VOID',p1,p2))
fun INT (p1,p2) = Token.TOKEN (ParserData.LrTable.T 59,(
ParserData.MlyValue.VOID',p1,p2))
fun FLOAT (p1,p2) = Token.TOKEN (ParserData.LrTable.T 60,(
ParserData.MlyValue.VOID',p1,p2))
fun DOUBLE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 61,(
ParserData.MlyValue.VOID',p1,p2))
fun STRUCT (p1,p2) = Token.TOKEN (ParserData.LrTable.T 62,(
ParserData.MlyValue.VOID',p1,p2))
fun ENUM (p1,p2) = Token.TOKEN (ParserData.LrTable.T 63,(
ParserData.MlyValue.VOID',p1,p2))
fun UNION (p1,p2) = Token.TOKEN (ParserData.LrTable.T 64,(
ParserData.MlyValue.VOID',p1,p2))
fun TYPEDEF (p1,p2) = Token.TOKEN (ParserData.LrTable.T 65,(
ParserData.MlyValue.VOID',p1,p2))
fun IF (p1,p2) = Token.TOKEN (ParserData.LrTable.T 66,(
ParserData.MlyValue.VOID',p1,p2))
fun ELSE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 67,(
ParserData.MlyValue.VOID',p1,p2))
fun DO (p1,p2) = Token.TOKEN (ParserData.LrTable.T 68,(
ParserData.MlyValue.VOID',p1,p2))
fun WHILE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 69,(
ParserData.MlyValue.VOID',p1,p2))
fun FOR (p1,p2) = Token.TOKEN (ParserData.LrTable.T 70,(
ParserData.MlyValue.VOID',p1,p2))
fun BREAK (p1,p2) = Token.TOKEN (ParserData.LrTable.T 71,(
ParserData.MlyValue.VOID',p1,p2))
fun CONTINUE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 72,(
ParserData.MlyValue.VOID',p1,p2))
fun SWITCH (p1,p2) = Token.TOKEN (ParserData.LrTable.T 73,(
ParserData.MlyValue.VOID',p1,p2))
fun CASE (p1,p2) = Token.TOKEN (ParserData.LrTable.T 74,(
ParserData.MlyValue.VOID',p1,p2))
fun DEFAULT (p1,p2) = Token.TOKEN (ParserData.LrTable.T 75,(
ParserData.MlyValue.VOID',p1,p2))
fun GOTO (p1,p2) = Token.TOKEN (ParserData.LrTable.T 76,(
ParserData.MlyValue.VOID',p1,p2))
fun RETURN (p1,p2) = Token.TOKEN (ParserData.LrTable.T 77,(
ParserData.MlyValue.VOID',p1,p2))
fun SEMICOLON (p1,p2) = Token.TOKEN (ParserData.LrTable.T 78,(
ParserData.MlyValue.VOID',p1,p2))
fun ELLIPSIS (p1,p2) = Token.TOKEN (ParserData.LrTable.T 79,(
ParserData.MlyValue.VOID',p1,p2))
fun BEGIN (p1,p2) = Token.TOKEN (ParserData.LrTable.T 80,(
ParserData.MlyValue.VOID',p1,p2))
fun END (p1,p2) = Token.TOKEN (ParserData.LrTable.T 81,(
ParserData.MlyValue.VOID',p1,p2))
fun SIZEOF (p1,p2) = Token.TOKEN (ParserData.LrTable.T 82,(
ParserData.MlyValue.VOID',p1,p2))
fun TYPEOF (p1,p2) = Token.TOKEN (ParserData.LrTable.T 83,(
ParserData.MlyValue.VOID',p1,p2))
fun LABEL (p1,p2) = Token.TOKEN (ParserData.LrTable.T 84,(
ParserData.MlyValue.VOID',p1,p2))
fun ALIGNOF (p1,p2) = Token.TOKEN (ParserData.LrTable.T 85,(
ParserData.MlyValue.VOID',p1,p2))
fun EOF (p1,p2) = Token.TOKEN (ParserData.LrTable.T 86,(
ParserData.MlyValue.VOID',p1,p2))
end
end
